!> Create the additional nodes for the DLP scheme \cite DroLeP_2011 .
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_commons_DLP

  use mod_numerics,    only: PR, eps, InitMin
  use type_IO, only: StdOut, StdErr
  use mod_physics_var, only: NbEq
  use type_mesh_DLP,   only: max_Card_S, NbM, IntInterfaces
  use type_mesh_DLP,   only: ratio_init, step, iter_M_max, nghb_level
  use mod_tools,       only: CrossProduct2D, BinomialCoeff_Rec
  use mod_numerics,    only: IsZero
  use mod_tools_array, only: IsPresent, GetIndex
  use mod_tools_mesh,  only: IsOut
  use mod_tools_mesh,  only: compute_bar_coord_triangle, compute_bar_coord_stared_polygon
  use mod_tools_mesh,  only: compute_tangent
  use type_mesh,       only: pmesh, BndInterfaces, dim_, Max_Nb_Nghb
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells, chunk_IntInterfaces
#endif /* _FC_PAR_OMP_ */
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
  use type_HO_MOOD, only: MaxDegree
  use mod_HO_MOOD_quad, only: quad_seg
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */
#if defined(_FC_PGI_)
  use mod_tools_array, only: NORM2
#endif /* defined(_FC_PGI_) */

  implicit none

  private
  public create_nodes_DLP
  public create_stencil_DLP, free_IntInterfaces


#define _STOP_ON_FAILURE_

!#define _IO_EXTENSIVE_PRINT_

!#define _DLP_SAVE_COORD_BAR_
!#define _DLP_SAVE_xKJ_

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Create the additional nodes for the DLP scheme
  !>
  !> For each interfaces \f$i\f$ we define two points \f$M_{K,i}\f$ and \f$M_{L,i}\f$ which are
  !> on the ray from \f$K\f$ and \f$L\f$ and direction \f$\vec{n}\f$ or
  !> \f$-\vec{n}\f$.
  !>
  !> And then we define those points M as a convex combination of their nghb.
  !>
  !> @remark In the original article \cite DroLeP_2011 \f$K\f$ is the point 1,
  !> \f$L\f$ is the point 2, and \f$M_2\f$ is associated with \f$M_{K,i}\f$,
  !> and \f$M_1\f$ is associated with \f$M_{L,i}\f$.
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine create_nodes_DLP()
      integer :: IdxIntInt
      integer :: Card_S
      integer :: CellK, CellL
      integer :: i_K, i_L
      integer, save :: NbInterAdm
      real(PR) :: dKL
      real(PR), dimension(dim_) :: xK, xL, xKL
      real(PR), dimension(dim_) :: n_K, n_L
#if !defined(_TP_FLUX_)
      integer :: iter_M_K, iter_M_L
      integer :: nb_points
      real(PR) :: dLM, dKM
      real(PR) :: ratio_M_K, ratio_M_L
      real(PR), dimension(dim_) :: xM_K, xM_L
      real(PR), dimension(dim_,nghb_level*Max_Nb_Nghb) :: x_points
      integer,  dimension(nghb_level*Max_Nb_Nghb) :: idx_points
      real(PR), dimension(_DLP_NB_PT_INTERP_) :: coord_bar_K, coord_bar_L
      integer,  dimension(_DLP_NB_PT_INTERP_) :: idx_K, idx_L
#endif /* .not. _TP_FLUX_ */
#if defined(_SRC_ENABLE_) && (_HO_MOOD_MAX_DEGREE_ >=1) && defined(_TP_FLUX_)
      integer :: IdxQuad
#endif /* _SRC_ENABLE_ && (_HO_MOOD_MAX_DEGREE_ >=1) && _TP_FLUX_ */

      !$OMP SINGLE
      write(StdOut,'(A)')'Creating the additional nodes for DLP...'
      !$OMP END SINGLE NOWAIT

      !$OMP SINGLE
      max_Card_S = 0
      NbINterAdm = 0
      !$OMP END SINGLE

      !$OMP DO &
      !$OMP REDUCTION(MAX:max_Card_S) &
      !$OMP REDUCTION(+:NbInterAdm) &
      !$OMP SCHEDULE(STATIC,chunk_IntInterfaces)
      loop_int_interfaces: do IdxIntInt = 1, pmesh%NbIntInterfaces

        CellK = IntInterfaces(IdxIntInt)%nghb(1)
        CellL = IntInterfaces(IdxIntInt)%nghb(2)

        ! Get the indexes of the interfaces
        i_K = GetIndex(CellL, pmesh%elt(CellK)%nghb)
        i_L = GetIndex(CellK, pmesh%elt(CellL)%nghb)

        ! Get the normal
        n_K = pmesh%elt(CellK)%normal(:,i_K)
        n_L = - n_K

        ! Connectivity betweeen interior interfaces and interfaces of the elements
        IntInterfaces(IdxIntInt)%i_Cell(1) = i_K
        IntInterfaces(IdxIntInt)%i_Cell(2) = i_L
        IntInterfaces(IdxIntInt)%le = pmesh%elt(CellK)%le(i_K)

        xK(:) = pmesh%elt(CellK)%x(:)
        xL(:) = pmesh%elt(CellL)%x(:)
        xKL(:) = xL(:) - xK(:)
        dKL = NORM2(xKL)

        !================================================================!
        !              Test if an interface is admissible or not
        !================================================================!
        !IntInterfaces(IdxIntInt)%admissible = IsZero(CrossProduct2D(xL-xK,n_K)) ! WARNING: not enough precise
        IntInterfaces(IdxIntInt)%admissible = IsZero(dot_product(xKL,n_K) - NORM2(xKL)*NORM2(n_K))
        if (IntInterfaces(IdxIntInt)%admissible) NbINterAdm = NbInterAdm + 1
        !================================================================!

#if defined(_TP_FLUX_)
        !=====================================================================!
        ! Create few arrays for the TP flux
        !=====================================================================!
        Card_S = 1
        call save_infos_interfaces_adm(IdxIntInt,CellK,xK,CellL,xL,xKL,dKL,Card_S)
        !=====================================================================!
#else /* _TP_FLUX_ */
        test_admissible : if (IntInterfaces(IdxIntInt)%admissible) then
          !================================================================!
          ! In the case the interface is admissible the defintions of S
          !               and of the barycentric are easy
          !================================================================!
          Card_S = 1
          call save_infos_interfaces_adm(IdxIntInt,CellK,xK,CellL,xL,xKL,dKL,Card_S)

        else
          !================================================================!
          !               Definition of M_K issued from K
          !================================================================!
          ratio_M_K = ratio_init
          dKM = ratio_M_K*dKL
          xM_K(:) = xK(:) + dKM*n_K(:)
          ! If M_K is outside the computational area, then we take a closer point.
          iter_M_K = 0
          do while( IsOut(xM_K) )
            ratio_M_K = ratio_M_K - step
            dKM = ratio_M_K*dKL
            xM_K(:) = xK(:) + dKM*n_K(:)
            iter_M_K = iter_M_K + 1
          end do
          !================================================================!

          !================================================================!
          !               Definition of M_L issued from L
          !================================================================!
          ratio_M_L = ratio_init
          dLM = ratio_M_L*dKL
          xM_L(:) = xL(:) + dLM*n_L(:)
          iter_M_L = 0
          do while( IsOut(xM_L) )
            ratio_M_L = ratio_M_L - step
            dLM = ratio_M_L*dKL
            xM_L(:) = xL(:) + dLM*n_L(:)
            iter_M_L =  iter_M_L + 1
          end do
          !================================================================!

          !================================================================!
          ! In a more generic case let's do some search and computations!
          !================================================================!
          Card_S = _DLP_NB_PT_INTERP_
          ! Get the points that could be used to reconstruct M_K and M_L
          ! namely it's the neighbors of K and L (or more)
          call get_canditates_points(CellK,CellL,nb_points,idx_points,x_points)

          ! For M_K we need to find some points such that M_K
          ! lie into the polygon constitued with L and those points
          call find_points(CellK,xK,CellL,xL,n_K,xM_K,nb_points,idx_points,x_points,&
                             dKM,dKL,ratio_M_K,iter_M_K,idx_K,coord_bar_K,Card_S)

          ! For M_L we need to find some points such that M_L
          ! lie into the polygon constitued with K and those points
          call find_points(CellL,xL,CellK,xK,n_L,xM_L,nb_points,idx_points,x_points,&
                             dLM,dKL,ratio_M_L,iter_M_L,idx_L,coord_bar_L,Card_S)

          ! Then we save the informations
          call save_infos_interfaces(IdxIntInt,CellK,xK,xM_K,dKM,idx_K,coord_bar_K, &
              CellL,xL,xM_L,dLM,idx_L,coord_bar_L, &
              Card_S)


        end if test_admissible

#endif /* _TP_FLUX_ */

        max_Card_S = max(max_Card_S,Card_S)

        IntInterfaces(IdxIntInt)%Card_S = Card_S

#if defined(_DEBUG_) && !defined(_TP_FLUX_)
        call check_normal(IdxIntInt, n_K, n_L)
#endif /* defined(_DEBUG_) && !defined(_TP_FLUX_) */

        allocate(IntInterfaces(IdxIntInt)%nu(NbEq,Card_S,NbM))
#if defined(_TP_FLUX_)
        IntInterfaces(IdxIntInt)%nu(:,:,:) = 1._PR
#endif /* _TP_FLUX_ */

#if (_HO_MOOD_MAX_DEGREE_ >= 1)
        allocate(IntInterfaces(IdxIntInt)%nu_tilde(NbEq,Card_S,quad_seg(MaxDegree)%nb_points,NbM))
#if defined(_TP_FLUX_)
        IntInterfaces(IdxIntInt)%nu_tilde(:,:,:,:) = 1._PR
#endif /* _TP_FLUX_ */
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */

#if defined(_SRC_ENABLE_)
        allocate(IntInterfaces(IdxIntInt)%nu_bar(Card_S,NbM))
#if defined(_TP_FLUX_)
        IntInterfaces(IdxIntInt)%nu_bar(:,:) = 1._PR/IntInterfaces(IdxIntInt)%dKJ(:,:)
#endif /* _TP_FLUX_ */

!#if (_HO_MOOD_MAX_DEGREE_ >= 1)
        !allocate(IntInterfaces(IdxIntInt)%nu_bar_tilde(Card_S,quad_seg(MaxDegree)%nb_points,NbM))
!#if defined(_TP_FLUX_)
        !do IdxQuad = 1, quad_seg(MaxDegree)%nb_points
          !IntInterfaces(IdxIntInt)%nu_bar_tilde(:,IdxQuad,:) = 1._PR/IntInterfaces(IdxIntInt)%dKJ(:,:)
        !end do
!#endif /* _TP_FLUX_ */
!#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */
#endif /* _SRC_ENABLE_ */

      end do loop_int_interfaces
      !$OMP END DO

      !$OMP SINGLE
      write(StdOut,'(A,F6.2,A)')'Nb admissible interfaces : ', real(NbInterAdm,PR)/real(pmesh%NbIntInterfaces,PR)*100._PR, '%'
      !$OMP END SINGLE

      !STOP

  end subroutine create_nodes_DLP

  !--------------------------------------------------------------------!
  !> @brief
  !> Looks for points that could be used to reconstruct M_K and M_L
  !>
  !> @param[in] CellK, CellL Indexes of the cells
  !> @param[out] nb_points Number of points
  !> @param[out] idx_points Indexes of the points
  !> @param[out] x_points Coordinates of the points
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine get_canditates_points(CellK,CellL,nb_points,idx_points,x_points)
      integer, intent(in)               :: CellK, CellL
      integer, intent(out)              :: nb_points
      integer, dimension(:),intent(out) :: idx_points
      real(PR), dimension(:,:),intent(out) :: x_points

      integer :: start, nghbhood, nb_points_old
      integer :: idxJ, CellJ, i_J, CellN
      integer :: i_K, i_L, IdxBndInt

      ! Initialize
      nb_points = 0
      idx_points(:) = -1

      idx_points(1) = CellK
      idx_points(2) = CellL
      nb_points = 2

      ! Then loop over the neighborhood to get other neighbors
      ! by taking care if there are not already in the list of
      ! the candidate points
      start = 1
      do nghbhood = 1, nghb_level
        nb_points_old = nb_points
        do idxJ = start, nb_points_old
          CellJ = idx_points(idxJ)
          do i_J = 1, pmesh%elt(CellJ)%NbInterfaces
            CellN = pmesh%elt(CellJ)%nghb(i_J)
            if (.not. IsPresent(CellN,idx_points) .and. CellN /= 0) then
              nb_points = nb_points + 1
              idx_points(nb_points) = CellN
            end if
          end do
        end do
        start = nb_points_old + 1
      end do

      ! Finally get the coordinates of all the candidate points
      do CellN = 1, nb_points
        x_points(:,CellN) = pmesh%elt(idx_points(CellN))%x(:)
      end do

      ! Also add some boundary points, those boundary points are saved with a negative index
      if (pmesh%elt(CellK)%onb) then
        do i_K = 1, pmesh%elt(CellK)%NbInterfaces
          IdxBndInt = pmesh%elt(CellK)%BndInterfaces(i_K)
          if (IdxBndInt /= 0) then
            nb_points = nb_points + 1
            x_points(:,nb_points) = BndInterfaces(IdxBndInt)%x(:)
            idx_points(nb_points) = - IdxBndInt
          endif
        end do
      end if

      if (pmesh%elt(CellL)%onb) then
        do i_L = 1, pmesh%elt(CellL)%NbInterfaces
          IdxBndInt = pmesh%elt(CellL)%BndInterfaces(i_L)
          if (IdxBndInt /= 0) then
            nb_points = nb_points + 1
            x_points(:,nb_points) = BndInterfaces(IdxBndInt)%x(:)
            idx_points(nb_points) = - IdxBndInt
          endif
        end do
      end if

  end subroutine get_canditates_points

  !--------------------------------------------------------------------!
  !> @brief
  !> Looks in the candidate points for points that could reconstruct the points \f$M\f$
  !>
  !> @param[in] CellK Index of the point from which \f$M\f$ is issued
  !> @param[in] CellL Neighbour of K, which is the first point
  !> @param[in] n_K Outgoing normal of the cell \f$K\f$
  !> @param[in] xK, xL Coordinates
  !> @param[inout] xM Coordinates of the point \f$M\f$
  !> @param[in] nb_points Number of the possible points for the reconstruction
  !> @param[in] idx_points Indexes of the possible points for the reconstruction
  !> @param[in] x_points Coordinates of the possible points for the reconstruction
  !> @param[inout] dKM, dKL Distance
  !> @param[inout] ratio_M Ratio to move closer the point \f$M\f$
  !> @param[in] iter_M Number of iterations done perviously to move the point \f$M\f$
  !> @param[out] idx Indexes of the points used for the reconstruction
  !> @param[out] coord_bar Barycentric coordinates of the point \f$M\f$
  !> @param[in] Card_S Number of points used
  !>
  !> @remarks The first point of \c idx NEED TO BE the neighbor of \f$K\f$.
  !> @remarks \f$K\f$ K is not necessarily in idx.
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mesh_mp_compute_bar_coord_triangle
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mesh_mp_compute_bar_coord_stared_polygon
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_binomialcoeff_rec
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_commons_dlp_mp_get_comb
  subroutine find_points(CellK,xK,CellL,xL,n_K,xM,nb_points,idx_points,x_points,&
          dKM,dKL,ratio_M,iter_M,idx,coord_bar,Card_S)
      integer,  intent(in)  :: CellK, CellL, Card_S
      integer, intent(in)   :: nb_points
      real(PR), intent(in)  :: dKL
      real(PR), dimension(:), intent(in)    :: xK, xL, n_K
      real(PR), dimension(:), intent(inout) :: xM
      real(PR), intent(inout) :: dKM, ratio_M
      integer, intent(in)  :: iter_M
      integer, dimension(:), intent(in)    :: idx_points
      real(PR), dimension(:,:), intent(in) :: x_points
      integer , dimension(:), intent(out)     :: idx
      real(PR), dimension(:), intent(out)     :: coord_bar

      integer  :: iter, idxJ
      logical  :: ok
      real(PR) :: distance, min_distance, ratio_M_save, dKM_save
      real(PR), dimension(dim_,Card_S) :: x_points_candidate
      real(PR), dimension(Card_S) :: distance_points_candidate
      integer, dimension(Card_S) :: idx_candidate
      real(PR), dimension(dim_)           :: xM_save
      real(PR), dimension(card_S)         :: coord_bar_save
      integer :: idx_comb, nb_combs
      integer :: Card_S_minus_L
      logical :: L_inside
#if defined(_DEBUG_)
      !real(PR), dimension(dim_) :: check_M
#endif /* defined(_DEBUG_) */

      ! The firt point is always L
      idx(1) = CellL
      distance_points_candidate(1) = dKL

      Card_S_minus_L = Card_S - 1
      nb_combs = BinomialCoeff_Rec(Card_S_minus_L, nb_points)

      ! Test the points, take the closest choice
      ! by looping over the combinations of points and move M closer to K
      min_distance = InitMin
      ok = .FALSE.
      loop_test: do iter = iter_M, iter_M_max
        loop_comb: do idx_comb = 1, nb_combs

          call get_comb(idx_comb,nb_points, &
              Card_S,CellL,xL,Card_S_minus_L, &
              idx_points,x_points, &
              idx_candidate,x_points_candidate, &
              L_inside)
          if (L_inside) cycle loop_comb

          if (Card_S == 3) then
            call compute_bar_coord_triangle(xM,&
                x_points_candidate(:,1),x_points_candidate(:,2),x_points_candidate(:,3),&
                coord_bar)
          else
            call compute_bar_coord_stared_polygon(xM,Card_S,x_points_candidate,coord_bar)
          endif

          test_inside: if (minval(coord_bar(1:Card_S))>=0._PR) then
            ! The distance is the greatest lenght between the candidate points and the cell center
            distance = distance_points_candidate(1)
            loop_distance: do idxJ = 2, Card_S
              distance = max(distance,NORM2(x_points_candidate(:,idxJ) - xK(:)))
            end do loop_distance
            test_closer: if (distance < min_distance) then
              idx(2:Card_S) = idx_candidate(2:Card_S)
              min_distance = distance
              ok = .TRUE.

              ! Save the values
              coord_bar_save(1:Card_S) = coord_bar(1:Card_S)
              ratio_M_save = ratio_M
              dKM_save = dKM
              xM_save(:) = xM(:)
#if defined(_DEBUG_)
              !check_M(:) = 0._PR
              !do idxJ = 1, Card_S
                !check_M(:) = check_M(:) + coord_bar(idxJ)*x_points_candidate(:,idxJ)
              !end do
              !if (NORM2(check_M - xM) > eps) then
                !write(StdOut,'(2A,I4,A)')__FILE__,':', __LINE__, &
                    !&" -> wrong computations of the barycentric coordinates (xG - xM):"
                !write(StdOut,'(ES23.16,1X,ES23.16)') &
                    !& check_M(:) - xM(:)
                !write(StdOut,'(A,2(I9,1X))') 'On interface between K and L: ', CellK, CellL
              !end if
#endif /* _DEBUG_ */
            end if test_closer
          end if test_inside

        end do loop_comb

        ! Iterate
        ratio_M = ratio_M - step
        dKM = ratio_M*dKL
        xM(:) = xK(:) + dKM*n_K(:)
      end do loop_test

      if (.not. ok) then
        write(StdErr,'(A)') "Trouble to find a third point for M"
        write(StdErr,'(A,I8,1X,I8)')"With K and L: ", CellK, CellL
        write(StdErr,'(A,ES23.16)')"ratio: ", ratio_M
        write(StdErr,'(A,2(ES23.16,1X))')"xK : ", xK
        write(StdErr,'(A,2(ES23.16,1X))')"xL : ", xL
        write(StdErr,'(A,2(ES23.16,1X))')"xM: ", xM
#if defined(_STOP_ON_FAILURE_)
        FLUSH StdErr
        STOP
#endif /* _STOP_ON_FAILURE_ */
      end if

#if defined(_IO_EXTENSIVE_PRINT_)
      do idxJ = 2, Card_S
        write(StdOut,'(A,I8)') 'J:', idx(idxJ)
        if (idx_candidate(idxJ) > 0) then
          write(StdOut,'(A,2ES23.16)')"xJ: ", pmesh%elt(idx(idxJ))%x(:)
        else
          write(StdOut,'(A,2ES23.16)')"xJ: ", BndInterfaces(-idx(idxJ))%x(:)
        endif
      end do
#endif /* _IO_EXTENSIVE_PRINT_ */

      ! Restore the values
      ratio_M = ratio_M_save
      dKM = dKM_save
      xM(:) = xM_save(:)
      coord_bar(1:Card_S) = coord_bar_save(1:Card_S)

  end subroutine find_points

#if defined(_DEBUG_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Check if the normal are well reconstructed
  !>
  !> @param[in] IdxIntInt Current interface
  !> @param[in] n_K, n_L Normal out of \f$K\f$ and \f$L\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine check_normal(IdxIntInt, n_K, n_L)
      integer, intent(in) :: IdxIntInt
      real(PR), dimension(:), intent(in) :: n_K, n_L

      real(PR), dimension(dim_) :: n_h
      integer :: idxJ, idxM

      idxM = 1

#if defined(_SRC_ENABLE_)
#if defined(_DLP_SAVE_xKJ_)
      n_h(:) = 0._PR
      do idxJ = 1, IntInterfaces(IdxIntInt)%Card_S
        n_h(:) = n_h(:) + IntInterfaces(IdxIntInt)%coeff_n_KJ(idxJ,idxM)*IntInterfaces(IdxIntInt)%xKJ(:,idxJ,idxM)
      end do
      if (NORM2(n_K - n_h) > eps) then
        !$OMP CRITICAL (check_normal_CRIT01)
        write(StdOut,'(2A,I4,A,I8)') __FILE__,':', __LINE__,' -> Warning: normal not properly reconstructed on interface: ', IdxIntInt
        write(StdOut,'(A,ES23.16)')            '||n_K - n_h||, with KJ  : ', NORM2(n_K - n_h)
        write(StdOut,'(A,ES23.16,1X,ES23.16)') ' (n_K - n_h) , with KJ  : ', n_K - n_h
        write(StdOut,'(A,I9,A,I9)') 'K: ', IntInterfaces(IdxIntInt)%idx(1,2),&
            ', and L: ', IntInterfaces(IdxIntInt)%idx(1,1)
        write(StdOut,'(A)')'--------------------------'
        !$OMP END CRITICAL (check_normal_CRIT01)
      end if
#endif /* _DLP_SAVE_xKJ_ */
#endif /* _SRC_ENABLE_ */

      n_h(:) = 0._PR
      do idxJ = 1, IntInterfaces(IdxIntInt)%Card_S
        n_h(:) = n_h(:) + IntInterfaces(IdxIntInt)%coeff_n_eta(idxJ,idxM)*IntInterfaces(IdxIntInt)%eta(:,idxJ,idxM)
      end do
      if (NORM2(n_K - n_h) > eps) then
        !$OMP CRITICAL (check_normal_CRIT01)
        write(StdOut,'(2A,I4,A,I8)') __FILE__,':', __LINE__,' -> Warning: normal not properly reconstructed on interface: ', IdxIntInt
        write(StdOut,'(A,ES23.16)')            '||n_K - n_h||, with eta : ', NORM2(n_K - n_h)
        write(StdOut,'(A,ES23.16,1X,ES23.16)') ' (n_K - n_h) , with eta : ', n_K - n_h
        write(StdOut,'(A,I9,A,I9)') 'K: ', IntInterfaces(IdxIntInt)%idx(1,2),&
            ', and L: ', IntInterfaces(IdxIntInt)%idx(1,1)
        write(StdOut,'(A)')'--------------------------'
        !$OMP END CRITICAL (check_normal_CRIT01)
      end if

      idxM = 2

#if defined(_SRC_ENABLE_)
#if defined(_DLP_SAVE_xKJ_)
      n_h(:) = 0._PR
      do idxJ = 1, IntInterfaces(IdxIntInt)%Card_S
        n_h(:) = n_h(:) + IntInterfaces(IdxIntInt)%coeff_n_KJ(idxJ,idxM)*IntInterfaces(IdxIntInt)%xKJ(:,idxJ,idxM)
      end do
      if (NORM2(n_L - n_h) > eps) then
        !$OMP CRITICAL (check_normal_CRIT01)
        write(StdOut,'(2A,I4,A,I8)') __FILE__,':', __LINE__,' -> Warning: normal not properly reconstructed on interface: ', IdxIntInt
        write(StdOut,'(A,ES23.16)')            '||n_L - n_h||, with LJ  : ', NORM2(n_L - n_h)
        write(StdOut,'(A,ES23.16,1X,ES23.16)') ' (n_L - n_h) , with LJ  : ', n_L - n_h
        write(StdOut,'(A,I9,A,I9)') 'K: ', IntInterfaces(IdxIntInt)%idx(1,2),&
            ', and L: ', IntInterfaces(IdxIntInt)%idx(1,1)
        write(StdOut,'(A)')'--------------------------'
        !$OMP END CRITICAL (check_normal_CRIT01)
      end if
#endif /* _DLP_SAVE_xKJ_ */
#endif /* _SRC_ENABLE_ */

      n_h(:) = 0._PR
      do idxJ = 1, IntInterfaces(IdxIntInt)%Card_S
        n_h(:) = n_h(:) + IntInterfaces(IdxIntInt)%coeff_n_eta(idxJ,idxM)*IntInterfaces(IdxIntInt)%eta(:,idxJ,idxM)
      end do
      if (NORM2(n_L - n_h) > eps) then
        !$OMP CRITICAL (check_normal_CRIT01)
        write(StdOut,'(2A,I4,A,I8)') __FILE__,':', __LINE__,' -> Warning: normal not properly reconstructed on interface: ', IdxIntInt
        write(StdOut,'(A,ES23.16)')            '||n_L - n_h||, with eta : ', NORM2(n_L - n_h)
        write(StdOut,'(A,ES23.16,1X,ES23.16)') ' (n_L - n_h) , with eta : ', n_L - n_h
        write(StdOut,'(A,I9,A,I9)') 'K: ', IntInterfaces(IdxIntInt)%idx(1,2),&
            ', and L: ', IntInterfaces(IdxIntInt)%idx(1,1)
        write(StdOut,'(A)')'--------------------------'
        !$OMP END CRITICAL (check_normal_CRIT01)
      end if

  end subroutine check_normal
#endif /* defined(_DEBUG_) */

  !--------------------------------------------------------------------------
  !> @brief
  !> Get a combination of \c Card_S points from the \c nb_points
  !>
  !> Adapted from https://rosettacode.org/wiki/Combinations#Fortran
  !> see also \cite Knu_2005-b for more powerful implementations
  !>
  !> @param[in] idx_comb Index of the current combination
  !> @param[in] nb_points Number of points
  !> @param[in] Card_S Number of points in \f$\mathcal{S}_{K,i}\f$
  !> @param[in] CellL, xL Index and coordinates of cell \f$L\f$
  !> @param[in] Card_S_minus_L Number of points in \f$\mathcal{S}_{K,i}\setminus L\f$
  !> @param[in] idx_points, x_points Index and coordinates of points
  !> @param[out] idx_points_comb, x_points_comb Index and coordinates of the combination
  !> @param[out] L_inside Indicate if L inside the combinations
  !>
  !> @remark The first point of \c idx_points_comb NEED TO BE the neighbor of \f$K\f$.
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_binomialcoeff_rec
  subroutine get_comb(idx_comb,nb_points, &
        Card_S,CellL,xL,Card_S_minus_L, &
        idx_points,x_points, &
        idx_points_comb,x_points_comb, &
        L_inside)
    integer, intent(in) :: idx_comb, nb_points
    integer, intent(in) :: Card_S
    integer, intent(in) :: CellL, Card_S_minus_L
    real(PR), dimension(:), intent(in) :: xL
    integer, dimension(:), intent(in) :: idx_points
    real(PR), dimension(:,:), intent(in) :: x_points
    integer, dimension(:), intent(out) :: idx_points_comb
    real(PR), dimension(:,:), intent(out) :: x_points_comb
    logical, intent(out) :: L_inside

    integer :: s, ix, kx, t

    if (Card_S > nb_points) then
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Too many points asked ."
      STOP
    end if

    L_inside = .FALSE.

    ix = idx_comb - 1
    kx = Card_S_minus_L

    do s = 1, nb_points
      if ( kx == 0 ) exit
      t = BinomialCoeff_Rec(kx - 1, nb_points  - s)
      if ( ix < t ) then
        ! L will be added later, so we do not need any combinations with it
        test_CellL: if (idx_points(s) == CellL) then
          L_inside = .TRUE.
          idx_points_comb(:) = 0
          return
        end if test_CellL
        idx_points_comb(kx+1) = idx_points(s)
        x_points_comb(:,kx+1) = x_points(:,s)
        kx = kx - 1
      else
        ix = ix - t
      end if
    end do

    ! Add cell L in the first place
    idx_points_comb(1) = CellL
    x_points_comb(:,1) = xL(:)

  end subroutine get_comb

  !--------------------------------------------------------------------!
  !> @brief
  !> Save the informations about the nodes \f$M\f$
  !>
  !> @param[in] IdxIntInt Current interface
  !> @param[in] CellK, CellL Indexes of the point from which \f$M\f$ is issued.
  !> @param[in] xK,xM_K,xL,xM_L Coordinates of the points
  !> @param[in] dKM,dLM Distance
  !> @param[in] idx_K, idx_L Indexes of the points used for the reconstruction
  !> @param[in] coord_bar_K, coord_bar_L Barycentric coordinates of the point \f$M\f$
  !> @param[in] Card_S Number of points used
  !>
  !> @remarks The first point of \c idx NEED TO BE the neighbor of \f$K\f$.
  !> @remarks \f$K\f$ K is not necessarily in idx.
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine save_infos_interfaces(IdxIntInt,CellK,xK,xM_K,dKM,idx_K,coord_bar_K,CellL,xL,xM_L,dLM,idx_L,coord_bar_L,Card_S)
      integer, intent(in) :: IdxIntInt, CellK, CellL, Card_S
      real(PR), intent(in) :: dKM, dLM
      real(PR), dimension(:), intent(in) :: xK, xM_K, xL, xM_L
      integer , dimension(:), intent(in) :: idx_K, idx_L
      real(PR), dimension(:), intent(in) :: coord_bar_K, coord_bar_L

      integer :: idxJ, CellJ
      integer :: idxM
      real(PR) :: dKJ, dLJ
      real(PR), dimension(dim_) :: xJ, xKJ, xLJ

      IntInterfaces(IdxIntInt)%xM(:,1) = xM_K(:)
      IntInterfaces(IdxIntInt)%xM(:,2) = xM_L(:)
      IntInterfaces(IdxIntInt)%dist(1) = dKM
      IntInterfaces(IdxIntInt)%dist(2) = dLM

      allocate(IntInterfaces(IdxIntInt)%idx(Card_S,NbM))
#if defined(_DLP_SAVE_COORD_BAR_)
      allocate(IntInterfaces(IdxIntInt)%coord_bar(Card_S,NbM))
#endif /* _DLP_SAVE_COORD_BAR_ */
#if defined(_DLP_SAVE_xKJ_)
      allocate(IntInterfaces(IdxIntInt)%xKJ(dim_,Card_S,NbM))
#endif /* _DLP_SAVE_xKJ_ */
      allocate(IntInterfaces(IdxIntInt)%dKJ(Card_S,NbM))
      allocate(IntInterfaces(IdxIntInt)%eta(dim_,Card_S,NbM))
      allocate(IntInterfaces(IdxIntInt)%tau(dim_,Card_S,NbM))
      allocate(IntInterfaces(IdxIntInt)%coeff_n_eta(Card_S,NbM))
#if defined(_SRC_ENABLE_)
      allocate(IntInterfaces(IdxIntInt)%coeff_n_KJ(Card_S,NbM))
#endif /* _SRC_ENABLE_ */

#if defined(_DLP_SAVE_COORD_BAR_)
      IntInterfaces(IdxIntInt)%coord_bar(1:Card_S,1) = coord_bar_K(1:Card_S)
      IntInterfaces(IdxIntInt)%coord_bar(1:Card_S,2) = coord_bar_L(1:Card_S)
#endif /* _DLP_SAVE_COORD_BAR_ */

      do idxJ = 1, Card_S
        !---------------------------------------------------------------------!
        idxM = 1
        CellJ = idx_K(idxJ)
        if (CellJ > 0) then
          xJ(:) = pmesh%elt(CellJ)%x(:)
        else
          xJ(:) = BndInterfaces(-CellJ)%x(:)
        end if
        xKJ(:) = xJ(:) - xK(:)
        dKJ = NORM2(xKJ)
        IntInterfaces(IdxIntInt)%idx(idxJ,idxM)  = CellJ
#if defined(_DLP_SAVE_xKJ_)
        IntInterfaces(IdxIntInt)%xKJ(:,idxJ,idxM) = xKJ(:)
#endif /* _DLP_SAVE_xKJ_ */
        IntInterfaces(IdxIntInt)%dKJ(idxJ,idxM)   = dKJ
        if (CellJ /= CellK) then
          IntInterfaces(IdxIntInt)%eta(:,idxJ,idxM) = xKJ(:)/dKJ
          IntInterfaces(IdxIntInt)%tau(:,idxJ,idxM) = compute_tangent(xKJ)/dKJ
          IntInterfaces(IdxIntInt)%coeff_n_eta(idxJ,idxM) = coord_bar_K(idxJ)*dKJ/dKM
        else
          IntInterfaces(IdxIntInt)%eta(:,idxJ,idxM) = 0._PR
          IntInterfaces(IdxIntInt)%tau(:,idxJ,idxM) = 0._PR
          IntInterfaces(IdxIntInt)%coeff_n_eta(idxJ,idxM) = 0._PR
        end if
#if defined(_SRC_ENABLE_)
        IntInterfaces(IdxIntInt)%coeff_n_KJ(idxJ,idxM)  = coord_bar_K(idxJ)/dKM
#endif /* _SRC_ENABLE_ */
        !---------------------------------------------------------------------!
        idxM = 2
        CellJ = idx_L(idxJ)
        if (CellJ > 0) then
          xJ(:) = pmesh%elt(CellJ)%x(:)
        else
          xJ(:) = BndInterfaces(-CellJ)%x(:)
        end if
        xLJ(:) = xJ(:) - xL(:)
        dLJ = NORM2(xLJ)
        IntInterfaces(IdxIntInt)%idx(idxJ,idxM)  = CellJ
#if defined(_DLP_SAVE_xKJ_)
        IntInterfaces(IdxIntInt)%xKJ(:,idxJ,idxM) = xLJ(:)
#endif /* _DLP_SAVE_xKJ_ */
        IntInterfaces(IdxIntInt)%dKJ(idxJ,idxM)   = dLJ
        if (CellJ /= CellL) then
          IntInterfaces(IdxIntInt)%eta(:,idxJ,idxM) = xLJ(:)/dLJ
          IntInterfaces(IdxIntInt)%tau(:,idxJ,idxM) = compute_tangent(xLJ)/dLJ
          IntInterfaces(IdxIntInt)%coeff_n_eta(idxJ,idxM) = coord_bar_L(idxJ)*dLJ/dLM
        else
          IntInterfaces(IdxIntInt)%eta(:,idxJ,idxM) = 0._PR
          IntInterfaces(IdxIntInt)%tau(:,idxJ,idxM) = 0._PR
          IntInterfaces(IdxIntInt)%coeff_n_eta(idxJ,idxM) = 0._PR
        end if
#if defined(_SRC_ENABLE_)
        IntInterfaces(IdxIntInt)%coeff_n_KJ(idxJ,idxM)  = coord_bar_L(idxJ)/dLM
#endif /* _SRC_ENABLE_ */
        !---------------------------------------------------------------------!
      end do

      IntInterfaces(IdxIntInt)%beta = min(IntInterfaces(IdxIntInt)%coeff_n_eta(1,1), IntInterfaces(IdxIntInt)%coeff_n_eta(1,2))
#if defined(_SRC_ENABLE_)
      IntInterfaces(IdxIntInt)%beta_bar = min(IntInterfaces(IdxIntInt)%coeff_n_KJ(1,1), IntInterfaces(IdxIntInt)%coeff_n_KJ(1,2))
#endif /* _SRC_ENABLE_ */

  end subroutine save_infos_interfaces

  !--------------------------------------------------------------------!
  !> @brief
  !> Save the informations about the nodes \f$M\f$ (version for admissible interfaces)
  !>
  !> @param[in] IdxIntInt Current interface
  !> @param[in] CellK, CellL Indexes of the point from which \f$M\f$ is issued.
  !> @param[in] xK, xL, xKL Coordinates of the points
  !> @param[in] dKL Distance
  !> @param[in] Card_S Number of points used
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine save_infos_interfaces_adm(IdxIntInt,CellK,xK,CellL,xL,xKL,dKL,Card_S)
      integer, intent(in) :: IdxIntInt, CellK, CellL, Card_S
      real(PR), dimension(:), intent(in) :: xK, xL, xKL
      real(PR), intent(in) :: dKL

      integer :: idxJ
      integer :: idxM
      integer :: i_K, i_L

      if (Card_S /= 1) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> On admissible interface, only point is needed."
        FLUSH StdErr
        STOP
      end if

      idxM = 1
      i_K = IntInterfaces(IdxIntInt)%i_Cell(idxM)
      IntInterfaces(IdxIntInt)%xM(:,idxM) = xL(:)
      IntInterfaces(IdxIntInt)%dist(idxM) = dKL
      idxM = 2
      i_L = IntInterfaces(IdxIntInt)%i_Cell(idxM)
      IntInterfaces(IdxIntInt)%xM(:,idxM) = xK(:)
      IntInterfaces(IdxIntInt)%dist(idxM) = dKL

      allocate(IntInterfaces(IdxIntInt)%idx(Card_S,NbM))
#if defined(_DLP_SAVE_COORD_BAR_)
      allocate(IntInterfaces(IdxIntInt)%coord_bar(Card_S,NbM))
#endif /* _DLP_SAVE_COORD_BAR_ */
#if defined(_DLP_SAVE_xKJ_)
      allocate(IntInterfaces(IdxIntInt)%xKJ(dim_,Card_S,NbM))
#endif /* _DLP_SAVE_xKJ_ */
      allocate(IntInterfaces(IdxIntInt)%dKJ(Card_S,NbM))
      allocate(IntInterfaces(IdxIntInt)%eta(dim_,Card_S,NbM))
      allocate(IntInterfaces(IdxIntInt)%tau(dim_,Card_S,NbM))
      allocate(IntInterfaces(IdxIntInt)%coeff_n_eta(Card_S,NbM))
#if defined(_SRC_ENABLE_)
      allocate(IntInterfaces(IdxIntInt)%coeff_n_KJ(Card_S,NbM))
#endif /* _SRC_ENABLE_ */

      idxJ = 1
      !---------------------------------------------------------------------!
      idxM = 1
      IntInterfaces(IdxIntInt)%idx(idxJ,idxM) = CellL
#if defined(_DLP_SAVE_xKJ_)
      IntInterfaces(IdxIntInt)%xKJ(:,idxJ,idxM) = xKL(:)
#endif /* _DLP_SAVE_xKJ_ */
#if defined(_DLP_SAVE_COORD_BAR_)
      IntInterfaces(IdxIntInt)%coord_bar(idxJ,idxM) = 1._PR
#endif /* _DLP_SAVE_COORD_BAR_ */
      IntInterfaces(IdxIntInt)%dKJ(idxJ,idxM) = dKL
      IntInterfaces(IdxIntInt)%eta(:,idxJ,idxM) = pmesh%elt(CellK)%normal(:,i_K)
      IntInterfaces(IdxIntInt)%tau(:,idxJ,idxM) = pmesh%elt(CellK)%tangent(:,i_K)
      IntInterfaces(IdxIntInt)%coeff_n_eta(idxJ,idxM) = 1._PR
#if defined(_SRC_ENABLE_)
      IntInterfaces(IdxIntInt)%coeff_n_KJ(idxJ,idxM)  = 1._PR/dKL
#endif /* _SRC_ENABLE_ */
      !---------------------------------------------------------------------!
      idxM = 2
      IntInterfaces(IdxIntInt)%idx(idxJ,idxM) = CellK
#if defined(_DLP_SAVE_xKJ_)
      IntInterfaces(IdxIntInt)%xKJ(:,idxJ,idxM) = - xKL(:)
#endif /* _DLP_SAVE_xKJ_ */
#if defined(_DLP_SAVE_COORD_BAR_)
      IntInterfaces(IdxIntInt)%coord_bar(idxJ,idxM) = 1._PR
#endif /* _DLP_SAVE_COORD_BAR_ */
      IntInterfaces(IdxIntInt)%dKJ(idxJ,idxM) = dKL
      IntInterfaces(IdxIntInt)%eta(:,idxJ,idxM) = pmesh%elt(CellL)%normal(:,i_L)
      IntInterfaces(IdxIntInt)%tau(:,idxJ,idxM) = pmesh%elt(CellL)%tangent(:,i_L)
      IntInterfaces(IdxIntInt)%coeff_n_eta(idxJ,idxM) = 1._PR
#if defined(_SRC_ENABLE_)
      IntInterfaces(IdxIntInt)%coeff_n_KJ(idxJ,idxM)  = 1._PR/dKL
#endif /* _SRC_ENABLE_ */
      !---------------------------------------------------------------------!

      IntInterfaces(IdxIntInt)%beta = 1._PR
#if defined(_SRC_ENABLE_)
      IntInterfaces(IdxIntInt)%beta_bar = 1._PR/dKL
#endif /* _SRC_ENABLE_ */

  end subroutine save_infos_interfaces_adm

  !--------------------------------------------------------------------!
  !> @brief
  !> Creation of the connectivity table for the points linked to the
  !> additional points.
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine create_stencil_DLP()
      integer :: CellK, CellL, i_K, IdxIntInt
      integer :: idxM
      integer :: CellJ, idxJ
      integer :: ns
      integer, dimension(:), allocatable :: stencil
      real(PR), dimension(:,:), allocatable :: eta, tau

      !$OMP SINGLE
      write(StdOut,'(A)')'Creating the stencil of DLP...'
      !$OMP END SINGLE NOWAIT

      allocate(stencil(max_Card_S*Max_Nb_Nghb))
      allocate(eta(dim_,max_Card_S*Max_Nb_Nghb))
      allocate(tau(dim_,max_Card_S*Max_Nb_Nghb))

      !$OMP DO &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      loop_elt : do CellK = 1, pmesh%NbCells
        ns = 0
        stencil(:) = -1
        eta(:,:) = 0._PR

        allocate(pmesh%elt(CellK)%idxM(pmesh%elt(CellK)%NbInterfaces))
#if defined(_SRC_ENABLE_)
        allocate(pmesh%elt(CellK)%delta_K(NbEq))
#if defined(_TP_FLUX_)
        pmesh%elt(CellK)%delta_K(:) = pmesh%elt(CellK)%Delta_x
#endif /* _TP_FLUX_ */
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
        allocate(pmesh%elt(CellK)%delta_K_tilde(NbEq,quad_seg(MaxDegree)%nb_points))
#if defined(_TP_FLUX_)
        pmesh%elt(CellK)%delta_K_tilde(:,:) = pmesh%elt(CellK)%Delta_x
#endif /* _TP_FLUX_ */
#endif /*  _HO_MOOD_MAX_DEGREE_ >= 1 */
#endif /* _SRC_ENABLE_ */

        ! We are looking for the points we need to add in the stencil
        loop_interfaces_K: do i_K = 1, pmesh%elt(CellK)%NbInterfaces
          IdxIntInt = pmesh%elt(CellK)%IntInterfaces(i_K)

          test_not_bnd: if (IdxIntInt /= 0) then
            if (IntInterfaces(IdxIntInt)%idx(1,2) == CellK) then
              idxM = 1
            else
              idxM = 2
            end if
            pmesh%elt(CellK)%idxM(i_K) = idxM

            CellL = pmesh%elt(CellK)%nghb(i_K)

            ! Add the points to the stencil of K if there are not already present
            ! remark: remove K from is own stencil
            ! remark: if the interface is admissible, only add the neighbour
            ! remark: always add the nghb
            idxJ = 1
            CellJ = IntInterfaces(IdxIntInt)%idx(idxJ,idxM)
            if (CellJ /= CellL) then
              write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> The points are not neighbours"
              FLUSH StdErr
              STOP
            end if
            ns = ns + 1
            stencil(ns)  = CellJ
            eta(:,ns) = IntInterfaces(IdxIntInt)%eta(:,idxJ,idxM)
            tau(:,ns) = IntInterfaces(IdxIntInt)%tau(:,idxJ,idxM)

            test_not_adm: if (.not. IntInterfaces(IdxIntInt)%admissible) then
              do idxJ = 2, IntInterfaces(IdxIntInt)%Card_S
                CellJ = IntInterfaces(IdxIntInt)%idx(idxJ,idxM)
                if (.not. IsPresent(CellJ,stencil) .and. CellJ /= CellK) then
                  ns = ns + 1
                  stencil(ns)  = CellJ
                  eta(:,ns) = IntInterfaces(IdxIntInt)%eta(:,idxJ,idxM)
                  tau(:,ns) = IntInterfaces(IdxIntInt)%tau(:,idxJ,idxM)
                end if
              end do
            end if test_not_adm

          else
            pmesh%elt(CellK)%idxM(i_K) = 0
          end if test_not_bnd

        end do loop_interfaces_K

        ! Number of points in the stencil of K
        pmesh%elt(CellK)%NbSten = ns
        ! Save the stencil
        allocate(pmesh%elt(CellK)%stencil(ns))
        pmesh%elt(CellK)%stencil(:) = stencil(1:ns)
        ! And the unitary vectors
        allocate(pmesh%elt(CellK)%eta(dim_,ns))
        pmesh%elt(CellK)%eta(:,:) = eta(:,1:ns)
        allocate(pmesh%elt(CellK)%tau(dim_,ns))
        pmesh%elt(CellK)%tau(:,:) = tau(:,1:ns)

      end do loop_elt
      !$OMP END DO

      deallocate(stencil)
      deallocate(eta)
      deallocate(tau)


  end subroutine create_stencil_DLP

  !--------------------------------------------------------------------!
  !> @brief
  !> Free additional nodes arrays
  !>
  !> @authors Florian Blachère
  !--------------------------------------------------------------------!
  subroutine free_IntInterfaces()
      integer :: IdxIntInt

      write(StdOut,'(A)') 'Freeing interior interfaces...'

      do IdxIntInt = 1, pmesh%NbIntInterfaces
        deallocate(IntInterfaces(IdxIntInt)%idx)
#if defined(_DLP_SAVE_COORD_BAR_)
        deallocate(IntInterfaces(IdxIntInt)%coord_bar)
#endif /* _DLP_SAVE_COORD_BAR_ */
#if defined(_DLP_SAVE_xKJ_)
        deallocate(IntInterfaces(IdxIntInt)%xKJ)
#endif /* _DLP_SAVE_xKJ_ */
        deallocate(IntInterfaces(IdxIntInt)%dKJ)
        deallocate(IntInterfaces(IdxIntInt)%eta)
        deallocate(IntInterfaces(IdxIntInt)%tau)
        deallocate(IntInterfaces(IdxIntInt)%coeff_n_eta)
#if defined(_SRC_ENABLE_)
        deallocate(IntInterfaces(IdxIntInt)%coeff_n_KJ)
        deallocate(IntInterfaces(IdxIntInt)%nu_bar)
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
        !deallocate(IntInterfaces(IdxIntInt)%nu_bar_tilde)
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */
#endif /* _SRC_ENABLE_ */
        deallocate(IntInterfaces(IdxIntInt)%nu)
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
        deallocate(IntInterfaces(IdxIntInt)%nu_tilde)
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */
      end do
      deallocate(IntInterfaces)

  end subroutine free_IntInterfaces

end module mod_commons_DLP
