!> MOOD routines
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_HO_MOOD
#if defined(_HO_MOOD_)

  use mod_numerics, only: PR, zero, one, InitMin
  use type_IO, only: StdOut, StdErr
  use type_mesh, only: dim_, Max_Nb_Nghb
  use type_mesh, only: pmesh, nodes
  use type_mesh_DLP, only: IntInterfaces
  use mod_physics_var, only: NbEq
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
  use type_HO_MOOD, only: alphas
  use mod_HO_MOOD_int, only: compute_I_KL
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */
  use type_HO_MOOD, only: MaxDegree, Nd_max, NbS_MOOD
  use mod_tools_array, only: IsPresent, ConditioningNumber
  use mod_HO_MOOD_poly, only: compute_polynomial_coefficients
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells, chunk_IntInterfaces
#endif /* _FC_PAR_OMP_ */

  IMPLICIT NONE
  private
  public InitMOOD
  public InitMOOD_iter

  integer, parameter :: max_candidate_points = 5*NbS_MOOD

contains
  !--------------------------------------------------------------------------
  !> @brief
  !> Initialisation for MOOD
  !>
  !> Get the stencil used for the polynomial reconstruction,
  !> and compute the transpose of the pseudo-inverse of the associated matrix
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_array_mp_ispresent
  subroutine InitMOOD()
      !======================================================================
      ! Method to get the stencil
      !======================================================================
!#define _HO_MOOD_STENCIL_LOOP_COND_
#define _HO_MOOD_STENCIL_DIRECT_
      !======================================================================

      !======================================================================
      ! Method to compute the transpose of the pseudo inverse
      !======================================================================
#define _HO_MOOD_PI_SVD_
!#define _HO_MOOD_PI_QR_
#if defined(_HO_MOOD_PI_QR_)
!#define _HO_MOOD_PI_QR_T_
#endif /* _HO_MOOD_PI_QR_ */
!#define _HO_MOOD_PI_CHOLESKY_
      !======================================================================

#if defined(_HO_MOOD_STENCIL_LOOP_COND_)
      ! Target for the conditioning number of X
      real(PR), parameter :: ConditioningNumber_X = 5._PR
      ! Number of maximal iteration to find a well conditioning matrix
      integer, parameter :: iter_C_max = 100
      real(PR) :: rand
      real(PR) :: kappa
      integer :: iter_C
      integer :: CellL, CellJ
      integer :: idxL, idxJ
      integer :: NbS
#endif /* _HO_MOOD_STENCIL_LOOP_COND_ */

      ! For the SVD computation (using DGESVD)
#if defined(_HO_MOOD_STENCIL_LOOP_COND_) || defined(_HO_MOOD_PI_SVD_)
      real(PR), dimension(:), allocatable :: S
      real(PR), dimension(:,:), allocatable :: U, VT
#endif /* _HO_MOOD_STENCIL_LOOP_COND_ || _HO_MOOD_PI_SVD_ */

#if defined(_HO_MOOD_PI_SVD_)
      integer :: col, row
#endif /* _HO_MOOD_PI_SVD_ */

      ! For the QR decomposition (using DGEQRF)
#if defined(_HO_MOOD_PI_QR_)
      real(PR), dimension(:), allocatable :: TAU
#if defined(_HO_MOOD_PI_QR_T_)
      real(PR), dimension(:,:), allocatable :: XT
#endif /* _HO_MOOD_PI_QR_T_ */
#endif /* _HO_MOOD_PI_QR_ */

#if defined(_HO_MOOD_STENCIL_LOOP_COND_) || defined(_HO_MOOD_PI_SVD_) || defined(_HO_MOOD_PI_QR_)
      integer, parameter :: LWORK = max_candidate_points
      real(PR), dimension(:), allocatable :: WORK
#endif /* _HO_MOOD_STENCIL_LOOP_COND_ || _HO_MOOD_PI_SVD_ || _HO_MOOD_PI_QR_ */

#if defined(_HO_MOOD_PI_CHOLESKY_)
      real(PR), dimension(:,:), allocatable :: A
#endif /* _HO_MOOD_PI_CHOLESKY_ */

      integer :: INFO

      real(PR), dimension(:,:), allocatable :: X, X_cpy

      integer, dimension(:), allocatable :: stencil_MOOD
      integer :: nb_candidate_points
      integer, dimension(:), allocatable :: candidate_points
      integer :: CellK

      !$OMP SINGLE
      write(StdOut,'(A)')'Initializing stencil and matrix of MOOD...'
      !$OMP END SINGLE NOWAIT

      allocate(X(max_candidate_points,Nd_max))
      allocate(X_cpy(max_candidate_points,Nd_max))

      allocate(stencil_MOOD(max_candidate_points))
      allocate(candidate_points(max_candidate_points))

#if defined(_HO_MOOD_STENCIL_LOOP_COND_) || defined(_HO_MOOD_PI_SVD_)
      allocate(S(Nd_max))
#endif /* _HO_MOOD_STENCIL_LOOP_COND_ || _HO_MOOD_PI_SVD_ */
#if (defined(_HO_MOOD_STENCIL_LOOP_COND_) || defined(_HO_MOOD_PI_SVD_)) && defined(_DEBUG_)
      allocate(U(max_candidate_points,max_candidate_points))
#endif /* (_HO_MOOD_STENCIL_LOOP_COND_ || _HO_MOOD_PI_SVD_) && _DEBUG_ */
#if (defined(_DEBUG_) && defined(_HO_MOOD_STENCIL_LOOP_COND_)) || defined(_HO_MOOD_PI_SVD_)
      allocate(VT(Nd_max,Nd_max))
#endif /* (_DEBUG_ && _HO_MOOD_STENCIL_LOOP_COND_) || _HO_MOOD_PI_SVD_ */

#if defined(_HO_MOOD_PI_QR_)
      allocate(TAU(Nd_max))
#if defined(_HO_MOOD_PI_QR_T_)
      allocate(XT(Nd_max,max_candidate_points))
#endif /* _HO_MOOD_PI_QR_T_ */
#endif /* _HO_MOOD_PI_QR_ */

#if defined(_HO_MOOD_STENCIL_LOOP_COND_) || defined(_HO_MOOD_PI_SVD_) || defined(_HO_MOOD_PI_QR_)
      allocate(WORK(LWORK))
#endif /* _HO_MOOD_STENCIL_LOOP_COND_ || _HO_MOOD_PI_SVD_ || _HO_MOOD_PI_QR_ */

#if defined(_HO_MOOD_PI_CHOLESKY_)
      allocate(A(Nd_max,Nd_max))
#endif /* _HO_MOOD_PI_CHOLESKY_ */

      ! Loop over the elements
      !$OMP DO &
      !$OMP SCHEDULE(static, chunk_NbPrimaryCells)
      loop_elt: do CellK = 1, pmesh%NbCells

        pmesh%elt(CellK)%CellPD = MaxDegree

#if (_HO_MOOD_MAX_DEGREE_ >= 1)
        ! Get some candidate points for the stencil
        call GetCandidatePointsMOOD(CellK,nb_candidate_points,candidate_points)

#if defined(_HO_MOOD_STENCIL_LOOP_COND_)
        !======================================================================
        ! Initialise the conditioning number of the matrix X
        kappa = InitMin
        iter_C = 0

        ! Loop until the matrix has a small conditioning number
        loop_conditioning: do while(kappa > ConditioningNumber_X .and. iter_C < iter_C_max)
          NbS = 0
          stencil_MOOD(:) = -1
          ! Get a possible stencil
          loop_stencil: do while (NbS < NbS_MOOD)
            ! First pick a random point in the candidate points
            ! warning: RANDOM_NUMBER give a result in [0;1[
            call RANDOM_NUMBER(rand)
            idxJ = floor(rand*nb_candidate_points) + 1
            CellJ = candidate_points(idxJ)
            ! Then check if this point is not already present,
            ! for sure if it is the first point add it and continue.
            test_not_1st_point: if (NbS > 0) then
              test_not_present: if (.not. IsPresent(CellJ,stencil_MOOD)) then
                ! And check if this point is a neighbor of a previous one
                loop_test_nghb: do idxL = 1, NbS
                  CellL = stencil_MOOD(idxL)
                  test_nghb: if (IsPresent(CellJ,pmesh%elt(CellL)%nghb)) then
                    NbS = NbS + 1
                    stencil_MOOD(NbS) = CellJ
                    exit loop_test_nghb
                  endif test_nghb
                end do loop_test_nghb
              end if test_not_present

            else
              NbS = NbS + 1
              stencil_MOOD(NbS) = CellJ

            endif test_not_1st_point
          end do loop_stencil

          ! First compute the matrix
          call compute_matrix_X(CellK,NbS_MOOD,stencil_MOOD,X)
          ! Then its SVD decomposition (to get the conditioning number)
          ! warning: DGESVD erase the content of X
          ! X = U*S*VT
          call DGESVD('N', 'N', NbS_MOOD, Nd_max, X, max_candidate_points, S, U, max_candidate_points, VT, Nd_max, WORK, LWORK, INFO)
          if (INFO /= 0) then
            write(StdErr,'(2A,I4,A,I4)')__FILE__,':', __LINE__, ' -> DGESVD failed, INFO = ', INFO
            FLUSH StdErr
            STOP
          endif

          ! Conditioning number of X
          kappa = S(1)/S(Nd_max)
          iter_C = iter_C + 1

        end do loop_conditioning

        if (kappa > ConditioningNumber_X) then
          write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, ' -> The search for the matrix with&
              & the smallest conditioning number failed, too much iterations.'
          write(StdErr,'(A,ES23.16)') 'Last conditioning number computed: ', kappa
          FLUSH StdErr
          STOP
        endif
        !======================================================================

#elif defined(_HO_MOOD_STENCIL_DIRECT_)
        !======================================================================
        stencil_MOOD(:) = candidate_points(:)
        !======================================================================

#else /* _HO_MOOD_STENCIL_ */
        !======================================================================
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, ' -> No method defined &
            &to get the stencil of MOOD'
        FLUSH StdErr
        STOP
        !======================================================================
#endif /* _HO_MOOD_STENCIL_ */

        !======================================================================
#if defined(_HO_MOOD_STENCIL_LOOP_COND_)
        pmesh%elt(CellK)%NbS_MOOD = NbS_MOOD
#elif defined(_HO_MOOD_STENCIL_DIRECT_)
        pmesh%elt(CellK)%NbS_MOOD = nb_candidate_points
#endif /* _HO_MOOD_STENCIL_ */

        allocate(pmesh%elt(CellK)%stencil_MOOD(pmesh%elt(CellK)%NbS_MOOD))
        pmesh%elt(CellK)%stencil_MOOD(:) = stencil_MOOD(1:pmesh%elt(CellK)%NbS_MOOD)
        !======================================================================

        allocate(pmesh%elt(CellK)%CoeffPolyMOOD(NbEq,Nd_max))

        call compute_matrix_X(CellK,pmesh%elt(CellK)%NbS_MOOD,pmesh%elt(CellK)%stencil_MOOD,X)
        allocate(pmesh%elt(CellK)%T_PI_X(pmesh%elt(CellK)%NbS_MOOD,Nd_max))

#if defined(_HO_MOOD_PI_SVD_)
        !--------------------------------------------------------------------------
        ! Compute the transposed of the pseudo-inverse using the SVD decomposition
        ! warning: DGESVD erase the content of X
        X_cpy(:,:) = X(:,:)
        ! X = U*S*VT
        call DGESVD('N', 'A', pmesh%elt(CellK)%NbS_MOOD, Nd_max, X, max_candidate_points, S, U, max_candidate_points, VT, Nd_max, WORK, LWORK, INFO)
        X(:,:) = X_cpy(:,:)

        ! (PI_X)^T = (X*VT^T)*((S^2)^-1*VT)

        !X_cpy(:,:) = MATMUL(X,transpose(VT))
        call DGEMM('N', 'T', pmesh%elt(CellK)%NbS_MOOD, Nd_max, Nd_max, one, X, max_candidate_points, VT, Nd_max, zero, X_cpy, max_candidate_points)

        ! VT = (S^2)^-1*VT
        do col = 1, Nd_max
          do row = 1, Nd_max
            VT(row,col) = VT(row,col)/S(row)**2
          end do
        end do

        !pmesh%elt(CellK)%T_PI_X(:,:) = MATMUL(X_cpy(1:pmesh%elt(CellK)%NbS_MOOD,:),VT)
        call DGEMM('N', 'N', pmesh%elt(CellK)%NbS_MOOD, Nd_max, Nd_max, one, X_cpy, max_candidate_points, VT, Nd_max, zero, pmesh%elt(CellK)%T_PI_X, pmesh%elt(CellK)%NbS_MOOD)
        !--------------------------------------------------------------------------

#elif defined(_HO_MOOD_PI_QR_)
        !--------------------------------------------------------------------------
        ! Compute the transposed of the pseudo-inverse using a QR decomposition
        X_cpy(:,:) = X(:,:)

        ! After DGEQRF the upper part of X will contain R
        call DGEQRF(pmesh%elt(CellK)%NbS_MOOD, Nd_max, X, max_candidate_points, TAU, WORK, LWORK, INFO)
        !call DGEQRFP(pmesh%elt(CellK)%NbS_MOOD, Nd_max, X, max_candidate_points, TAU, WORK, LWORK, INFO)
        !call DGEQR2(pmesh%elt(CellK)%NbS_MOOD, Nd_max, X, max_candidate_points, TAU, WORK, INFO)
        !call DGEQR2P(pmesh%elt(CellK)%NbS_MOOD, Nd_max, X, max_candidate_points, TAU, WORK, INFO)
        if (INFO /= 0) then
          write(StdErr,'(2A,I4,A,I4)')__FILE__,':', __LINE__, ' -> DGEQRF failed, INFO = ', INFO
          FLUSH StdErr
          STOP
        endif

#if defined(_HO_MOOD_PI_QR_T_)
        ! (PI_X) = (R^T*R)^-1*X^T
        !-------------------------------
        XT(:,:) = transpose(X_cpy)
        ! (R^T)^-1*X^T, using a forward substitution (R^T: lower)
        !call DTRTRS('U', 'T', 'N', Nd_max, pmesh%elt(CellK)%NbS_MOOD, X, max_candidate_points, XT, Nd_Max, INFO)
        !if (INFO /= 0) then
         !write(StdErr,'(2A,I4,A,I4)')__FILE__,':', __LINE__, ' -> DTRTRS failed, INFO = ', INFO
         !FLUSH StdErr
         !STOP
        !endif
        call DTRSM('L', 'U', 'T', 'N', Nd_max, pmesh%elt(CellK)%NbS_MOOD, one, X, max_candidate_points, XT, Nd_Max)

        ! R^-1*(R^T)^-1*X^T, using a backward substitution (R: upper)
        !call DTRTRS('U', 'N', 'N', Nd_max, pmesh%elt(CellK)%NbS_MOOD, X, max_candidate_points, XT, Nd_Max, INFO)
        !if (INFO /= 0) then
         !write(StdErr,'(2A,I4,A,I4)')__FILE__,':', __LINE__, ' -> DTRTRS failed, INFO = ', INFO
         !FLUSH StdErr
         !STOP
        !endif
        call DTRSM('L', 'U', 'N', 'N', Nd_max, pmesh%elt(CellK)%NbS_MOOD, one, X, max_candidate_points, XT, Nd_Max)

        pmesh%elt(CellK)%T_PI_X(:,:) = transpose(XT(:,1:pmesh%elt(CellK)%NbS_MOOD))
        !-------------------------------

#else /* _HO_MOOD_PI_QR_T_ */
        ! (PI_X)^T = X*(R^T*R)^-1
        !-------------------------------
        ! X*(R^-1), using a backward substitution (R: upper)
        call DTRSM('R', 'U', 'N', 'N', pmesh%elt(CellK)%NbS_MOOD, Nd_max, one, X, max_candidate_points, X_cpy, max_candidate_points)

        ! X*(R^-1)*(R^T^-1), using a forward substitution (R^T: lower)
        call DTRSM('R', 'U', 'T', 'N', pmesh%elt(CellK)%NbS_MOOD, Nd_max, one, X, max_candidate_points, X_cpy, max_candidate_points)

        pmesh%elt(CellK)%T_PI_X(:,:) = X_cpy(1:pmesh%elt(CellK)%NbS_MOOD,:)
        !-------------------------------
#endif /* _HO_MOOD_PI_QR_T_ */
        !--------------------------------------------------------------------------

#elif defined(_HO_MOOD_PI_CHOLESKY_)
        !--------------------------------------------------------------------------
        ! Compute the transposed of the pseudo-inverse using a LL^T decomposition
        X_cpy(:,:) = X(:,:)

        ! A = X^T*X
        call DGEMM('T','N',Nd_max,Nd_max,pmesh%elt(CellK)%NbS_MOOD,one,X,max_candidate_points,X_cpy,max_candidate_points,zero,A,Nd_max)

        ! After DPOTRF the lower part of A will contain L
        ! A = X^T*X = L*L^T
        call DPOTRF('L',Nd_max,A,Nd_max,INFO)
        if (INFO /= 0) then
          write(StdErr,'(2A,I4,A,I4)')__FILE__,':', __LINE__, ' -> DPOTRF failed, INFO = ', INFO
          FLUSH StdErr
          STOP
        endif

        ! (PI_X)^T = X*(L*L^T)^-1
        !-------------------------------
        ! X*(L^T^-1), using a backward substitution (L^T: upper)
        call DTRSM('R', 'L', 'T', 'N', pmesh%elt(CellK)%NbS_MOOD, Nd_max, one, A, Nd_max, X, max_candidate_points)

        ! X*(L^T^-1)*(L^-1), using a forward substitution (L: lower)
        call DTRSM('R', 'L', 'N', 'N', pmesh%elt(CellK)%NbS_MOOD, Nd_max, one, A, Nd_max, X, max_candidate_points)

        pmesh%elt(CellK)%T_PI_X(:,:) = X(1:pmesh%elt(CellK)%NbS_MOOD,:)

#else /* _HO_MOOD_PI_ */
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, ' -> No method defined &
            & to compute the pseudo-inverse of the matrix X'
        FLUSH StdErr
        STOP
#endif /* _HO_MOOD_PI_ */

#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */

      end do loop_elt
      !$OMP END DO

      deallocate(X)
      deallocate(X_cpy)
      deallocate(stencil_MOOD)
      deallocate(candidate_points)

#if defined(_HO_MOOD_STENCIL_LOOP_COND_) || defined(_HO_MOOD_PI_SVD_)
      deallocate(S)
#endif /* _HO_MOOD_STENCIL_LOOP_COND_ || _HO_MOOD_PI_SVD_ */
#if (defined(_HO_MOOD_STENCIL_LOOP_COND_) || defined(_HO_MOOD_PI_SVD_)) && defined(_DEBUG_)
      deallocate(U)
#endif /* (_HO_MOOD_STENCIL_LOOP_COND_ || _HO_MOOD_PI_SVD_) && _DEBUG_ */
#if (defined(_DEBUG_) && defined(_HO_MOOD_STENCIL_LOOP_COND_)) || defined(_HO_MOOD_PI_SVD_)
      deallocate(VT)
#endif /* (_DEBUG_ && _HO_MOOD_STENCIL_LOOP_COND_) || _HO_MOOD_PI_SVD_ */

#if defined(_HO_MOOD_PI_QR_)
      deallocate(TAU)
#if defined(_HO_MOOD_PI_QR_T_)
      deallocate(XT)
#endif /* _HO_MOOD_PI_QR_T_ */
#endif /* _HO_MOOD_PI_QR_ */

#if defined(_HO_MOOD_STENCIL_LOOP_COND_) || defined(_HO_MOOD_PI_SVD_) || defined(_HO_MOOD_PI_QR_)
      deallocate(WORK)
#endif /* _HO_MOOD_STENCIL_LOOP_COND_ || _HO_MOOD_PI_SVD_ || _HO_MOOD_PI_QR_ */

#if defined(_HO_MOOD_PI_CHOLESKY_)
      deallocate(A)
#endif /* _HO_MOOD_PI_CHOLESKY_ */


  end subroutine InitMOOD

#if (_HO_MOOD_MAX_DEGREE_ >= 1)
  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the matrix \f$X_K\f$ of the overdetermined linear system for the
  !> polynomial reconstruction (see (1.11) in \cite Dio_2012 )
  !>
  !> @param[in] CellK Current cell
  !> @param[in] nb_points_stencil Number of points used in the MOOD reconstruction
  !> @param[in] stencil_MOOD Stencil used for the reconstruction
  !> @param[out] X Matrix of the reconstruction
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_int_mp_compute_i_kl
  subroutine compute_matrix_X(CellK,nb_points_stencil,stencil_MOOD,X)
      integer, intent(in) :: CellK, nb_points_stencil
      integer, dimension(:), intent(in) :: stencil_MOOD
      real(PR), dimension(:,:), intent(out) :: X

#if defined(_HO_MOOD_MATRIX_SCALING_)
      real(PR) :: scaling
#endif /* _HO_MOOD_MATRIX_SCALING_ */
      real(PR) :: I_KK
      integer, dimension(dim_) :: alpha
      integer :: idx_alpha, idx_L
      integer :: CellL

#if defined(_HO_MOOD_MATRIX_SCALING_)
      ! Use the scaling from \cite Fri
      scaling = 1._PR/sqrt(pmesh%elt(CellK)%area)
#endif /* _HO_MOOD_MATRIX_SCALING_ */

      loop_alpha: do idx_alpha = 1, Nd_max
        alpha(:) = alphas(:,idx_alpha)
        I_KK = compute_I_KL(CellK,CellK,alpha)

        loop_stencil_MOOD: do idx_L = 1, nb_points_stencil
          CellL = stencil_MOOD(idx_L)

#if defined(_HO_MOOD_MATRIX_SCALING_)
          X(idx_L,idx_alpha) = scaling**(sum(alpha))*(compute_I_KL(CellK,CellL,alpha) - I_KK)
#else /* _HO_MOOD_MATRIX_SCALING_ */
          X(idx_L,idx_alpha) = compute_I_KL(CellK,CellL,alpha) - I_KK
#endif /* _HO_MOOD_MATRIX_SCALING_ */

        end do loop_stencil_MOOD
      end do loop_alpha

  end subroutine compute_matrix_X
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */

  !--------------------------------------------------------------------------
  !> @brief
  !> Get candidate points for the MOOD stencil
  !> (following the algorithm from p. 19 in \cite Dio_2012 )
  !>
  !> @param[in] CellK Working cell
  !> @param[out] nb_candidate_points Number of candidate points found
  !> @param[out] candidate_points Index of the candidate points
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine GetCandidatePointsMOOD(CellK,nb_candidate_points,candidate_points)
      integer, intent(in) :: CellK
      integer, intent(out) :: nb_candidate_points
      integer, dimension(:), intent(out) :: candidate_points

      integer :: start, nb_points_old
      integer :: neighborhood
      integer :: i_K, CellL
      integer :: idxJ, CellJ, i_J, CellN
      integer, parameter :: nghb_level = 5

      ! Initialize
      nb_candidate_points = 0
      candidate_points(:) = -1

      ! First put all the neighbors of the current cell K
      loop_nghb_K: do i_K = 1, pmesh%elt(CellK)%NbInterfaces
        CellL = pmesh%elt(CellK)%nghb(i_K)
        test_not_bnd: if (CellL /= 0) then
          nb_candidate_points = nb_candidate_points + 1
          candidate_points(nb_candidate_points) = CellL
          if (nb_candidate_points >= max_candidate_points) exit loop_nghb_K
        end if test_not_bnd
      end do loop_nghb_K

      ! Then loop over the neighborhood to get other neighbors
      ! by taking care if there are not already in the list of
      ! the candidate points
      start = 1
      loop_nghbhood: do neighborhood = 2, nghb_level
        nb_points_old = nb_candidate_points
        do idxJ = start, nb_points_old
          CellJ = candidate_points(idxJ)
          do i_J = 1, pmesh%elt(CellJ)%NbInterfaces
            CellN = pmesh%elt(CellJ)%nghb(i_J)
            if (.not. IsPresent(CellN,candidate_points) .and. CellN /= 0 .and. CellN /= CellK) then
              nb_candidate_points = nb_candidate_points + 1
              candidate_points(nb_candidate_points) = CellN
              if (nb_candidate_points >= max_candidate_points) exit loop_nghbhood
            end if
          end do
        end do
        start = nb_points_old + 1

#if defined (_HO_MOOD_STENCIL_DIRECT_)
        if (nb_candidate_points >= NbS_MOOD) exit loop_nghbhood
#endif /* _HO_MOOD_STENCIL_DIRECT_ */

      end do loop_nghbhood

      if (nb_candidate_points < NbS_MOOD) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, ' -> Not enought candidate points'
        FLUSH StdErr
        STOP
      end if

  end subroutine GetCandidatePointsMOOD

  !--------------------------------------------------------------------------
  !> @brief
  !> Initialize the degree of the polynomial reconstruction on cells \c CellPD
  !> and compute the coefficients of the polynomial reconstruction
  !>
  !> @param[in] U_n Mean values at current time
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_poly_mp_compute_polynomial_coefficients
  subroutine InitMOOD_iter(U_n)
      real(PR), dimension(:,:), intent(in) :: U_n
      integer  :: CellK

      !$OMP DO &
      !$OMP SCHEDULE(static, chunk_NbPrimaryCells)
      do CellK = 1, pmesh%NbCells
        pmesh%elt(CellK)%CellPD = MaxDegree

#if (_HO_MOOD_MAX_DEGREE_ >= 1)
        call compute_polynomial_coefficients(U_n,CellK,pmesh%elt(CellK)%CoeffPolyMOOD)
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */
      end do
      !$OMP END DO
  end subroutine InitMOOD_iter

#endif /* _HO_MOOD_ */
end module mod_HO_MOOD
