!> MOOD routines to compute integrals over triangles (and polygons)
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_HO_MOOD_int

  use mod_numerics, only: PR
  use type_IO, only: StdErr
  use mod_tools, only: x_to_alpha, plus1
  use mod_tools_mesh, only: TriangleArea
  use type_mesh, only: dim_
  use type_mesh, only: pmesh, nodes
#if defined(_HO_MOOD_)
  use type_HO_MOOD, only: MaxDegree
  use mod_HO_MOOD_quad, only: quad_tri
#endif /* _HO_MOOD_ */

  IMPLICIT NONE
  private
#if defined(_HO_MOOD_)
  public compute_I_KL
#endif /* _HO_MOOD_ */
  public compute_mean_f

contains
#if defined(_HO_MOOD_)
  !--------------------------------------------------------------------------
  !> @brief
  !> Compute one part of of the coefficients of the matrix
  !> \f$X_K\f$: \f$I_{KL}\f$ using quadrature rules on triangles (and
  !> polygons by triangulation) (see Appendix A of \cite Dio_2012 )
  !>
  !> @param[in] CellK Current cell
  !> @param[in] CellL Cell of integration
  !> @param[in] alpha Multi index
  !> @return \f$I_{KL} = \frac{1}{\vert L \vert} \int_L (\boldsymbol{x} - \boldsymbol{x}_K)^{\boldsymbol{\alpha}} d\boldsymbol{x}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_x_to_alpha
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mesh_mp_trianglearea
  function compute_I_KL(CellK,CellL,alpha) result(I_KL)
      integer, intent(in) :: CellK, CellL
      integer, dimension(dim_), intent(in) :: alpha
      real(PR) :: I_KL

      real(PR), dimension(dim_) :: xQ, xK, xL
      real(PR), dimension(dim_) :: x1, x2, x3
      integer :: IdxQuad, i_L

      xK(:) = pmesh%elt(CellK)%x(:)
      I_KL = 0._PR

      if (pmesh%elt(CellL)%NbInterfaces == 3) then
        x1(:) = nodes(pmesh%elt(CellL)%vertices_glob(1))%x(:)
        x2(:) = nodes(pmesh%elt(CellL)%vertices_glob(2))%x(:)
        x3(:) = nodes(pmesh%elt(CellL)%vertices_glob(3))%x(:)

        loop_quad: do IdxQuad = 1, quad_tri(MaxDegree)%nb_points
          xQ(:) = quad_tri(MaxDegree)%bar_coord(1,IdxQuad)*x1(:) &
              + quad_tri(MaxDegree)%bar_coord(2,IdxQuad)*x2(:) &
              + quad_tri(MaxDegree)%bar_coord(3,IdxQuad)*x3(:) &
              - xK(:)
          I_KL = I_KL + quad_tri(MaxDegree)%weights(IdxQuad)*x_to_alpha(xQ,alpha)
        end do loop_quad

      else if (pmesh%elt(CellL)%NbInterfaces > 3) then
        ! The polygon is split into triangles of vertices x1, x2 and xL
        xL(:) = pmesh%elt(CellL)%x(:)

        loop_tri: do i_L = 1, pmesh%elt(CellL)%NbInterfaces - 1
          x1(:) = nodes(pmesh%elt(CellL)%vertices_glob(i_L))%x(:)
          x2(:) = nodes(pmesh%elt(CellL)%vertices_glob(i_L + 1))%x(:)

          loop_quad_tri: do IdxQuad = 1, quad_tri(MaxDegree)%nb_points
            xQ(:) = quad_tri(MaxDegree)%bar_coord(1,IdxQuad)*x1(:) &
                + quad_tri(MaxDegree)%bar_coord(2,IdxQuad)*x2(:) &
                + quad_tri(MaxDegree)%bar_coord(3,IdxQuad)*xL(:) &
                - xK(:)

            I_KL = I_KL + TriangleArea(x1,x2,xL)*quad_tri(MaxDegree)%weights(IdxQuad)*x_to_alpha(xQ,alpha)
          end do loop_quad_tri
        end do loop_tri

        i_L = pmesh%elt(CellL)%NbInterfaces
        x1(:) = nodes(pmesh%elt(CellL)%vertices_glob(i_L))%x(:)
        x2(:) = nodes(pmesh%elt(CellL)%vertices_glob(1))%x(:)

        loop_quad_tri_last: do IdxQuad = 1, quad_tri(MaxDegree)%nb_points
          xQ(:) = quad_tri(MaxDegree)%bar_coord(1,IdxQuad)*x1(:) &
              + quad_tri(MaxDegree)%bar_coord(2,IdxQuad)*x2(:) &
              + quad_tri(MaxDegree)%bar_coord(3,IdxQuad)*xL(:) &
              - xK(:)

          I_KL = I_KL + TriangleArea(x1,x2,xL)*quad_tri(MaxDegree)%weights(IdxQuad)*x_to_alpha(xQ,alpha)
        end do loop_quad_tri_last

        I_KL = I_KL/pmesh%elt(CellL)%area

      else
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, ' -> Quadrature rules &
            &are not implemented for this kind of elements'
        FLUSH StdErr
        STOP
      end if

  end function compute_I_KL
#endif /* _HO_MOOD_ */

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the mean value of the initial condition over a cell using quadrature rules on triangles (and
  !> polygons by triangulation) (see Appendix A of \cite Dio_2012 )
  !>
  !> @param[in] CellK Current cell
  !> @param[in] t_n Current time
  !> @param[in] degree Degree of the quadrature
  !> @param[in] func Function to average over a cell
  !> @param[in] size_func Size of the function
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function compute_mean_f(CellK,t_n,degree,func,size_func) result(mean_f)
      integer, intent(in) :: CellK, degree
      real(PR), intent(in) :: t_n
      interface
        function func(x,t,n)
            use mod_numerics, only: PR
            real(PR), dimension(:), intent(in) :: x
            real(PR), intent(in) :: t
            integer, intent(in) :: n
            real(PR), dimension(n) :: func
        end function
      end interface
      integer, intent(in) :: size_func
      real(PR), dimension(size_func) :: mean_f

      real(PR), dimension(dim_) :: xK
#if (_HO_MOOD_MAX_DEGREE_ >=1)
      real(PR), dimension(dim_) :: xQ, x1, x2, x3
      integer :: i_K, IdxQuad
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */

      xK(:) = pmesh%elt(CellK)%x(:)
      mean_f(:) = 0._PR

      test_degree: if (degree == 0) then
        mean_f(:) = func(xK,t_n,size_func)

#if (_HO_MOOD_MAX_DEGREE_ >=1)
      else if (degree >= 1 .and. degree <= 5) then
        test_nb_faces: if (pmesh%elt(CellK)%NbInterfaces == 3) then
          x1(:) = nodes(pmesh%elt(CellK)%vertices_glob(1))%x(:)
          x2(:) = nodes(pmesh%elt(CellK)%vertices_glob(2))%x(:)
          x3(:) = nodes(pmesh%elt(CellK)%vertices_glob(3))%x(:)

          loop_quad: do IdxQuad = 1, quad_tri(degree)%nb_points
            xQ(:) = quad_tri(degree)%bar_coord(1,IdxQuad)*x1(:) &
                + quad_tri(degree)%bar_coord(2,IdxQuad)*x2(:) &
                + quad_tri(degree)%bar_coord(3,IdxQuad)*x3(:)
            mean_f(:) = mean_f(:) + quad_tri(degree)%weights(IdxQuad)*func(xQ,t_n,size_func)
          end do loop_quad

        else if (pmesh%elt(CellK)%NbInterfaces > 3) then
          ! The polygon is split into triangles of vertices x1, x2 and xK

          loop_tri: do i_K = 1, pmesh%elt(CellK)%NbInterfaces
            x1(:) = nodes(pmesh%elt(CellK)%vertices_glob(i_K))%x(:)
            x2(:) = nodes(pmesh%elt(CellK)%vertices_glob(plus1(i_K,pmesh%elt(CellK)%NbInterfaces)))%x(:)

            loop_quad_tri: do IdxQuad = 1, quad_tri(degree)%nb_points
              xQ(:) = quad_tri(degree)%bar_coord(1,IdxQuad)*x1(:) &
                  + quad_tri(degree)%bar_coord(2,IdxQuad)*x2(:) &
                  + quad_tri(degree)%bar_coord(3,IdxQuad)*xK(:)

              mean_f(:) = mean_f(:) + TriangleArea(x1,x2,xK)*quad_tri(degree)%weights(IdxQuad)*func(xQ,t_n,size_func)
            end do loop_quad_tri
          end do loop_tri

          mean_f(:) = mean_f(:)/pmesh%elt(CellK)%area

        else
          write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, ' -> Quadrature rules &
              &are not implemented for this kind of elements'
          FLUSH StdErr
          STOP
        end if test_nb_faces
#endif /* _HO_MOOD_MAX_DEGREE_ >= 1 */

      else
        write(StdErr,'(2A,I4,A,I1)')__FILE__,':', __LINE__, " -> Wrong degree: ", degree
        FLUSH StdErr
        STOP
      end if test_degree

  end function compute_mean_f

end module mod_HO_MOOD_int
