!> Parameters for quadrature rules \cite Dio_2012
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_HO_MOOD_quad
#if defined(_HO_MOOD_)

  use mod_numerics, only: PR
  use mod_numerics, only: zero, half, third, one, two
  use type_IO, only: StdOut

  implicit none

  private
  public quad_seg
  public quad_tri
  public InitQuadratureRules
  public EndQuadratureRules

  !--------------------------------------------------------------------!
  !> @brief
  !> Type definition for the quadrature rules
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  type quad_rule
    !> Number of quadrature points
    integer :: nb_points
    !> Barycentric coordinates
    real(PR), dimension(:,:), allocatable :: bar_coord
    !> Weights
    real(PR), dimension(:), allocatable :: weights
  end type quad_rule

  !> Quadrature rules for segments, from degree 1 to 5
  type(quad_rule), dimension(:), allocatable :: quad_seg
  !> Quadrature rules for triangles, from degree 1 to 5
  type(quad_rule), dimension(:), allocatable :: quad_tri

  !--------------------------------------------------------------------
  real(PR), parameter :: sqrt_third = 1._PR/sqrt(3._PR)
  real(PR), parameter :: sqrt_3_5 = sqrt(3._PR/5._PR)
  !--------------------------------------------------------------------
  real(PR), parameter :: sqrt_15 = sqrt(15._PR)
  ! Coefficients for the quadrature rules on triangles with degree 3 and 4
  real(PR), parameter :: a1  = 0.091576213509771_PR
  real(PR), parameter :: a2  = 0.445948490915965_PR
  real(PR), parameter :: xi1 = 0.109951743655322_PR
  real(PR), parameter :: xi2 = 0.223381589678011_PR
  ! Coefficients for the quadrature rules on triangles with degree 5
  real(PR), parameter :: a1_deg5  = (5._PR - sqrt_15)/21._PR
  real(PR), parameter :: a2_deg5  = (5._PR + sqrt_15)/21._PR
  real(PR), parameter :: xi1_deg5 = (155._PR - sqrt_15)/1200._PR
  real(PR), parameter :: xi2_deg5 = (155._PR + sqrt_15)/1200._PR
  !--------------------------------------------------------------------

CONTAINS

  !--------------------------------------------------------------------------
  !> @brief
  !> Initialisation of the weights for the quadrature rules for segments and triangles
  !> from \cite Dio_2012
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine InitQuadratureRules()
      integer, parameter :: nb_vert_seg = 2
      integer, parameter :: nb_vert_tri = 3
      integer :: degree

      write(StdOut,'(A)') 'Initializing quadratures...'

      allocate(quad_seg(5))
      allocate(quad_tri(5))

      degree = 1
      quad_seg(degree)%nb_points = 1
      allocate(quad_seg(degree)%bar_coord(nb_vert_seg,quad_seg(degree)%nb_points))
      quad_seg(degree)%bar_coord(:,:) = reshape(SOURCE=(/ &
          half, half &
          /), &
          SHAPE=(/nb_vert_seg,quad_seg(degree)%nb_points/))
      allocate(quad_seg(degree)%weights(quad_seg(degree)%nb_points))
      quad_seg(degree)%weights(:) = [one]

      do degree = 2, 3
        quad_seg(degree)%nb_points = 2
        allocate(quad_seg(degree)%bar_coord(nb_vert_seg,quad_seg(degree)%nb_points))
        quad_seg(degree)%bar_coord(:,:) = reshape(SOURCE=(/ &
            half + half*sqrt_third, half - half*sqrt_third, &
            half - half*sqrt_third, half + half*sqrt_third &
            /), &
            SHAPE=(/nb_vert_seg,quad_seg(degree)%nb_points/))
        allocate(quad_seg(degree)%weights(quad_seg(degree)%nb_points))
        quad_seg(degree)%weights(:) = [half, half]
      end do

      do degree = 4, 5
        quad_seg(degree)%nb_points = 3
        allocate(quad_seg(degree)%bar_coord(nb_vert_seg,quad_seg(degree)%nb_points))
        quad_seg(degree)%bar_coord(:,:) = reshape(SOURCE=(/ &
            half, half, &
            half + half*sqrt_3_5, half - half*sqrt_3_5, &
            half - half*sqrt_3_5, half + half*sqrt_3_5 &
            /), &
            SHAPE=(/nb_vert_seg,quad_seg(degree)%nb_points/))
        allocate(quad_seg(degree)%weights(quad_seg(degree)%nb_points))
        quad_seg(degree)%weights(:) = [8._PR, 5._PR, 5._PR]/18._PR
      end do

      !===================================================================!
      !===================================================================!

      degree = 1
      quad_tri(degree)%nb_points = 1
      allocate(quad_tri(degree)%bar_coord(nb_vert_tri,quad_tri(degree)%nb_points))
      quad_tri(degree)%bar_coord(:,:) = reshape(SOURCE=(/ &
          third, third, third &
          /), &
          SHAPE=(/nb_vert_tri,quad_tri(degree)%nb_points/))
      allocate(quad_tri(degree)%weights(quad_tri(degree)%nb_points))
      quad_tri(degree)%weights(:) = [one]

      degree = 2
      quad_tri(degree)%nb_points = 3
      allocate(quad_tri(degree)%bar_coord(nb_vert_tri,quad_tri(degree)%nb_points))
      quad_tri(degree)%bar_coord(:,:) = reshape(SOURCE=(/ &
          half, half, zero, &
          zero, half, half, &
          half, zero, half &
          /), &
          SHAPE=(/nb_vert_tri,quad_tri(degree)%nb_points/))
      allocate(quad_tri(degree)%weights(quad_tri(degree)%nb_points))
      quad_tri(degree)%weights(:) = [third, third, third]

      do degree = 3, 4
        quad_tri(degree)%nb_points = 6
        allocate(quad_tri(degree)%bar_coord(nb_vert_tri,quad_tri(degree)%nb_points))
        quad_tri(degree)%bar_coord(:,:) = reshape(SOURCE=(/ &
            a1, a1, one - two*a1, &
            one - two*a1, a1, a1, &
            a1, one - two*a1, a1, &
            a2, a2, one - two*a2, &
            one - two*a2, a2, a2, &
            a2, one - two*a2, a2 &
            /), &
            SHAPE=(/nb_vert_tri,quad_tri(degree)%nb_points/))
        allocate(quad_tri(degree)%weights(quad_tri(degree)%nb_points))
        quad_tri(degree)%weights(:) = [xi1, xi1, xi1, xi2, xi2, xi2]
      end do

      degree = 5
      quad_tri(degree)%nb_points = 7
      allocate(quad_tri(degree)%bar_coord(nb_vert_tri,quad_tri(degree)%nb_points))
      quad_tri(degree)%bar_coord(:,:) = reshape(SOURCE=(/ &
          third, third, third, &
          a1_deg5, a1_deg5, one - two*a1_deg5, &
          one - two*a1_deg5, a1_deg5, a1_deg5, &
          a1_deg5, one - two*a1_deg5, a1_deg5, &
          a2_deg5, a2_deg5, one - two*a2_deg5, &
          one - two*a2_deg5, a2_deg5, a2_deg5, &
          a2_deg5, one - two*a2_deg5, a2_deg5 &
          /), &
          SHAPE=(/nb_vert_tri,quad_tri(degree)%nb_points/))
      allocate(quad_tri(degree)%weights(quad_tri(degree)%nb_points))
      quad_tri(degree)%weights(:) = [9._PR/40._PR, xi1, xi1, xi1, xi2, xi2, xi2]

  end subroutine InitQuadratureRules

  !--------------------------------------------------------------------------
  !> @brief
  !> Deallocate the quadrature rules
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine EndQuadratureRules()
      integer :: degree

      write(StdOut,'(A)') 'Freeing quadratures...'

      do degree = 1, 5
        deallocate(quad_seg(degree)%bar_coord)
        deallocate(quad_seg(degree)%weights)
        deallocate(quad_tri(degree)%bar_coord)
        deallocate(quad_tri(degree)%weights)
      end do

      deallocate(quad_seg)
      deallocate(quad_tri)

  end subroutine EndQuadratureRules

#endif /* _HO_MOOD_ */
end module mod_HO_MOOD_quad
