!> MOOD routines to recompute the sum of fluxes
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_HO_MOOD_recompute
#if defined(_HO_MOOD_)

  use mod_numerics, only: PR
  use type_IO, only: StdErr
  use type_mesh, only: pmesh, BndInterfaces, dim_
  use type_mesh_DLP, only: IntInterfaces, max_Card_S, nu_one
  use mod_physics_var, only: NbEq
  use mod_flux_num_HO_MOOD, only: compute_flux_n_HO_MOOD, compute_flux_bnd_n_HO_MOOD
  use type_recompute, only: NbCellsToUpdate
  use type_recompute, only: CellsToUpdate
  use mod_HO_MOOD_degree, only: computeIntFacePD, computeBndFacePD
  use type_HO_MOOD, only: MaxDegree
#if !defined(_TP_FLUX_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
  use mod_coeff_DLP, only: compute_nu_tilde_inter
#if defined(_SRC_ENABLE_)
  use mod_delta, only: compute_delta_tilde
#endif /* _SRC_ENABLE_ */
#endif /* .not. _TP_FLUX_ && _HO_MOOD_MAX_DEGREE_ >= 1 */

  IMPLICIT NONE

  private
  public RecomputePhiMOOD

contains

  !--------------------------------------------------------------------!
  !> @brief
  !> Re-compute \f$\Phi\f$ (sum of numerical fluxes on a cell)
  !>
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[inout] Phi Sum of numerical flux on all cells
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_num_ho_mood_mp_compute_flux_n_ho_mood
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_num_ho_mood_mp_compute_flux_bnd_n_ho_mood
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_degree_mp_computeintfacepd
  subroutine RecomputePhiMOOD(U_n,t_n,Phi)
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n
      real(PR), dimension(:,:), intent(inout) :: Phi

      real(PR), dimension(NbEq) :: flux_n
      integer :: CellK, i_K
      integer :: idxK, idxM
      integer :: IdxIntInt, IdxBndInt

      !====================================================================!
      !$OMP DO
      loop_elt_CellsToUpdate: do idxK = 1, NbCellsToUpdate
        CellK = CellsToUpdate(idxK)
        Phi(:,CellK) = 0._PR
        loop_interfaces_CellsToUpdate: do i_K = 1, pmesh%elt(CellK)%NbInterfaces
          IdxIntInt = pmesh%elt(CellK)%IntInterfaces(i_K)

          test_bnd_CellsInvalid: if (IdxIntInt /= 0) then
            idxM = pmesh%elt(CellK)%idxM(i_K)

            call computeIntFacePD(IdxIntInt)

#if !defined(_TP_FLUX_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
            if (IntInterfaces(IdxIntInt)%FacePD < MaxDegree .and. IntInterfaces(IdxIntInt)%FacePD >= 1) then
              ! If the degree changed, but not to 0 (or -1), the HO nu coefficients need to be recomputed

              ! But for now do not enable this feature, as theoretically there is no intermediary degree.
              ! For now the decrease process is MaxDegree (HO) => 0 (order one) => -1 (TP flux)
              write(StdErr,'(2A,I4,A,I1)')__FILE__,':', __LINE__, &
                  &" -> Intermediary degree are not implemented: ", IntInterfaces(IdxIntInt)%FacePD
              FLUSH StdErr
              STOP

              !call compute_nu_tilde_inter(IdxIntInt,U_n,t_n,&
                  !IntInterfaces(IdxIntInt)%FacePD,IntInterfaces(IdxIntInt)%nu_tilde)
#if defined(_SRC_ENABLE_)
              !call compute_delta_HO_MOOD(CellK,IntInterfaces(IdxIntInt)%FacePD,pmesh%elt(CellK)%delta_tilde)
#endif /* _SRC_ENABLE_ */
            end if
#endif /* .not. TP_FLUX_ && _HO_MOOD_MAX_DEGREE_ >= 1 */

            call compute_flux_n_HO_MOOD(CellK,IdxIntInt,idxM,U_n,t_n,flux_n)

          else
            IdxBndInt = pmesh%elt(CellK)%BndInterfaces(i_K)
            call computeBndFacePD(IdxBndInt)
            call compute_flux_bnd_n_HO_MOOD(CellK,IdxBndInt,i_K,U_n,t_n,flux_n)

          end if test_bnd_CellsInvalid

          Phi(:,CellK) = Phi(:,CellK) + pmesh%elt(CellK)%le(i_K)*flux_n(:)

        end do loop_interfaces_CellsToUpdate
        Phi(:,CellK) = Phi(:,CellK)/pmesh%elt(CellK)%area
      end do loop_elt_CellsToUpdate
      !$OMP END DO
      !====================================================================!

  end subroutine RecomputePhiMOOD

#endif /* _HO_MOOD_ */
end module mod_HO_MOOD_recompute
