!> Dump solution to a file (for instance to use in a restart)
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_dump
#if defined (_IO_ENABLE_DUMP_)

  use type_mesh, only: pmesh, dim_
  use type_IO, only: StdOut
  use mod_numerics, only: PR, MaxLen
  use mod_physics_var, only: NbEq, primitive_name, conservative_name
  use mod_physics_var, only: test_case
  use mod_physics_var, only: Output_Directory

#if defined(_LIB_HDF5_)
  use hdf5
#elif defined(_LIB_NC_)
  use netcdf
#endif /* _BINARY_FORMAT_ */

  IMPLICIT NONE

  private
  public dump_solution

CONTAINS
  !--------------------------------------------------------------------!
  !> @brief
  !> Wrapper to dump solution to a plain file (for instance to use in a restart)
  !>
  !> @param[in] U_n_p1 Solution at time \f$t^n\f$
  !> @param[in] info Informations about the run
  !> @param[in] iter Number of iterations
  !> @param[in] t_n_p1 Time \f$t^{n+1}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine dump_solution(U_n_p1,info,iter,t_n_p1)
      real(PR), dimension(:,:), intent(in) :: U_n_p1
      character(len=*), intent(in) :: info
      integer, intent(in) :: iter
      real(PR), intent(in) :: t_n_p1
#if defined(_LIB_HDF5_)
      call dump_solution_h5(U_n_p1,info,iter,t_n_p1)
#elif defined(_LIB_NC_)
      call dump_solution_nc(U_n_p1,info,iter,t_n_p1)
#else
      call dump_solution_plain(U_n_p1,info,iter,t_n_p1)
#endif /* _BINARY_FORMAT_ */
  end subroutine dump_solution

  !--------------------------------------------------------------------!
  !> @brief
  !> Dump solution to a plain file (for instance to use in a restart)
  !>
  !> @param[in] U_n_p1 Solution at time \f$t^n\f$
  !> @param[in] info Informations about the run
  !> @param[in] iter Number of iterations
  !> @param[in] t_n_p1 Time \f$t^{n+1}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine dump_solution_plain(U_n_p1,info,iter,t_n_p1)
      real(PR), dimension(:,:), intent(in) :: U_n_p1
      character(len=*), intent(in) :: info
      integer, intent(in) :: iter
      real(PR), intent(in) :: t_n_p1

      integer, parameter :: dump_id = 13
      integer :: CellK, eq

      OPEN(dump_id,file=trim(Output_Directory)//'/dump_'//trim(info)//'.dat',action='write')
      !=================================================================!
      WRITE(dump_id,'(A)') '# Infos'
      WRITE(dump_id,'(A)') trim(info)
      WRITE(dump_id,'(A)') '# Nb. iterations, time'
      WRITE(dump_id,'(I9,1X,ES23.16)') iter, t_n_p1
      !=================================================================!
      write(dump_id,*) ((U_n_p1(eq,CellK),eq=1,NbEq),CellK=1,pmesh%NbCells)
      CLOSE(dump_id)

  end subroutine dump_solution_plain

#if defined(_LIB_NC_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Dump solution to a NetCDF \cite NetCDF, \cite NetCDF90
  !> file (for instance to use in a restart)
  !>
  !> @param[in] U_n_p1 Solution at time \f$t^n\f$
  !> @param[in] info Informations about the run
  !> @param[in] iter Number of iterations
  !> @param[in] t_n_p1 Time \f$t^{n+1}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine dump_solution_nc(U_n_p1,info,iter,t_n_p1)
      real(PR), dimension(:,:), intent(in) :: U_n_p1
      character(len=*), intent(in) :: info
      integer, intent(in) :: iter
      real(PR), intent(in) :: t_n_p1

      integer :: eq

      character(len=MaxLen) :: ncFile,ncLibVers
      integer :: ncErr, ncId, nc_dim_NbCells_Id, nc_dim_scalar_Id
      integer :: nc_var_iterations_Id, nc_var_time_Id
      integer, dimension(NbEq) :: nc_var_U_Id

      ncFile = trim(Output_Directory)//'/dump_'//trim(info)//'.nc'
      !==================================================================!
      ! Create the netcdf file (it will be in define mode)
      ncErr = nf90_create(trim(ncFile),nf90_clobber,ncId)

      ! Set dimensions
      ncErr = nf90_def_dim(ncId, 'NbCells', pmesh%NbCells, nc_dim_NbCells_Id)
      ncErr = nf90_def_dim(ncId, 'scalar', 1, nc_dim_scalar_Id)

      ! Set names
      ncErr = nf90_def_var(ncId, "Iterations", nf90_int, nc_dim_scalar_Id, nc_var_iterations_Id)
      ncErr = nf90_def_var(ncId, 'Time', nf90_double, nc_dim_scalar_Id, nc_var_time_Id)
      do eq=1, NbEq
        ncErr = nf90_def_var(ncId, trim(conservative_name(eq)), nf90_double, &
            nc_dim_NbCells_Id, nc_var_U_Id(eq))
      end do

      ! Stop the definition and go in data mode
      ncErr = nf90_enddef(ncId)

      ! Put value of the variable
      ncErr = nf90_put_var(ncId, nc_var_iterations_Id, iter)
      ncErr = nf90_put_var(ncId, nc_var_time_Id, t_n_p1)

      do eq=1, NbEq
        ncErr = nf90_put_var(ncId, nc_var_U_Id(eq), U_n_p1(eq,:))
      end do

      ! Close the netcdf file
      ncErr = nf90_close(ncId)
      !==================================================================!
  end subroutine dump_solution_nc

  !--------------------------------------------------------------------!
  !> @brief
  !> Check if the return code of the NetCDF \cite NetCDF, \cite NetCDF90
  !> routines are good or not
  !>
  !> @param[in] ncErr Error code
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine check_nc(ncErr)
      integer, intent(in) :: ncErr
      if (ncErr /= nf90_noerr) then
        write(StdOut,'(A)') trim (adjustl(nf90_strerror(ncErr)))
      end if
  end subroutine check_nc
#endif /* _LIB_NC_ */


#if defined(_LIB_HDF5_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Dump solution to an HDF5 \cite HDF5 file (for instance to use in a restart)
  !>
  !> @param[in] U_n_p1 Solution at time \f$t^n\f$
  !> @param[in] info Informations about the run
  !> @param[in] iter Number of iterations
  !> @param[in] t_n_p1 Time \f$t^{n+1}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine dump_solution_h5(U_n_p1,info,iter,t_n_p1)
      real(PR), dimension(:,:), intent(in) :: U_n_p1
      character(len=*), intent(in) :: info
      integer, intent(in) :: iter
      real(PR), intent(in) :: t_n_p1

      integer :: eq

      character(len=MaxLen) :: h5File
      integer :: h5Err
      integer(HSIZE_T), dimension(1), parameter :: one = 1
      integer(HSIZE_T), dimension(1) :: NbCells

      integer(HID_T) :: dspace_id  ! Dataspace identifier
      integer(HID_T) :: dset_id    ! Dataset identifier
      integer(HID_T) :: h5Id       ! File identifier

      NbCells = pmesh%NbCells

      h5File = trim(Output_Directory)//'/dump_'//trim(info)//'.h5'
      !==================================================================!
      ! Initialize FORTRAN interface.
      call h5open_f(h5Err)

      ! Create a new file using default properties.
      call h5fcreate_f(h5File, H5F_ACC_TRUNC_F, h5Id, h5Err)

      call h5screate_simple_f(1, one, dspace_id, h5Err)
      call h5dcreate_f(h5Id, 'Iterations', H5T_NATIVE_INTEGER, dspace_id, dset_id, h5Err)
      call h5dwrite_f(dset_id, H5T_NATIVE_INTEGER, iter, one, h5Err)
      call h5dclose_f(dset_id, h5Err)

      call h5screate_simple_f(1, one, dspace_id, h5Err)
      call h5dcreate_f(h5Id, 'Time', H5T_NATIVE_DOUBLE, dspace_id, dset_id, h5Err)
      call h5dwrite_f(dset_id, H5T_NATIVE_DOUBLE, t_n_p1, one, h5Err)
      call h5dclose_f(dset_id, h5Err)

      do eq = 1, NbEq
        ! Create the dataspace.
        call h5screate_simple_f(1, NbCells, dspace_id, h5Err)

        ! Create the dataset with default properties.
        call h5dcreate_f(h5Id, trim(conservative_name(eq)), H5T_NATIVE_DOUBLE, dspace_id, &
            dset_id, h5Err)

        ! Write the dataset.
        call h5dwrite_f(dset_id, H5T_NATIVE_DOUBLE, U_n_p1(eq,:), NbCells, h5Err)

        ! End access to the dataset and release resources used by it.
        call h5dclose_f(dset_id, h5Err)
      end do

      ! Terminate access to the data space.
      call h5sclose_f(dspace_id, h5Err)

      ! Close the file.
      call h5fclose_f(h5Id, h5Err)

      ! Close FORTRAN interface.
      call h5close_f(h5Err)
      !==================================================================!
  end subroutine dump_solution_h5
#endif /* _LIB_HDF5_ */

#endif /* _IO_ENABLE_DUMP_ */
end module mod_dump
