!> Re-computations done with a two point flux, when the cells are not admissible
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_TPCorrec_recompute
#if defined(_TP_CORREC_) && !defined(_HO_MOOD_)

  use mod_numerics, only: PR, eps
  use mod_numerics, only: InitMin, InitMax
  use type_IO, only: StdErr
  use type_mesh, only: pmesh, BndInterfaces, dim_
  use type_mesh_DLP, only: IntInterfaces, NbM, max_Card_S, nu_one, NbM
  use mod_physics_var, only: NbEq
  use mod_physics, only: compute_greatest_wave_speed
  use mod_flux_num, only: compute_flux_n, compute_flux_bnd_n
  use mod_flux_1D, only: flux_1D
  use mod_bnd_values, only: compute_value_bnd
#if defined(_SRC_ENABLE_)
  use mod_asymptotic, only  : compute_alpha_KK, compute_alpha_KK_NoAP
#if !defined(_TP_FLUX_)
  use mod_coeff_DLP, only: compute_nu_bar_inter
#endif /* .not. _TP_FLUX_ */
  use mod_flux_1D_source, only: flux_1D_source_NoAP
  use mod_flux_num, only: compute_flux_bnd_n_NoAP
#endif /* _SRC_ENABLE_ */
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells, chunk_IntInterfaces
#endif /* _FC_PAR_OMP_ */
  use type_recompute, only: NbCellsInvalid, NbCellsToUpdate
  use type_recompute, only: CellsToUpdate
  use type_recompute, only: FlagCells, FlagIntInterfaces

  IMPLICIT NONE

  private
  public Recompute_dt_TPCorrec
  public RecomputePhiTPCorrec

CONTAINS
  !--------------------------------------------------------------------------
  !> @brief
  !> Re-compute the time step \f$\Delta t\f$, which fulfils a classical CFL
  !>
  !> \f[\max_{\stackrel{K \in \mathcal{M}}{J \in \overline{{\mathcal E}_K}}}
  !>      \left(b_{KJ} \frac{\Delta t}{\delta_{KJ}}\right)\leq 1\f]
  !>
  !> @param[in] U_n Solution at time \f$t^n\f$
  !> @param[in] t_n Current time
  !> @param[inout] t_n_p1 Next time
  !> @param[inout] dt Computed time step
  !> @param[out] dt_changed Indicate if the routine changed \f$\Delta t\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_advection_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_burgers_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_compute_greatest_wave_speed
  subroutine Recompute_dt_TPCorrec(U_n,t_n,t_n_p1,dt,dt_changed)
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in)  :: t_n
      real(PR), intent(inout) :: t_n_p1
      real(PR), intent(inout) :: dt
      logical, intent(out) :: dt_changed

      real(PR), dimension(dim_) :: eta, tau
      real(PR), dimension(NbEq) :: U_K, U_J, sum_d
      real(PR), dimension(NbEq) :: delta_K
      real(PR) :: b_KJ
      real(PR), save :: dt_tmp ! "save" attribute for OMP
      integer :: CellK, CellJ
      integer :: idxM, idxK
      integer :: IdxIntInt, idxJ, i_K
      integer :: IdxBndInt

      !$OMP SINGLE
      dt_tmp = InitMax
      !$OMP END SINGLE

      !$OMP DO &
      !$OMP REDUCTION(MAX:dt_tmp)
      loop_elt: do idxK=1, NbCellsToUpdate
        CellK = CellsToUpdate(idxK)

        !-------------------------------------------------------------------------
        sum_d(:) = 0._PR
        loop_interfaces_K_1: do i_K = 1, pmesh%elt(CellK)%NbInterfaces
          IdxIntInt = pmesh%elt(CellK)%IntInterfaces(i_K)

          test_bnd_1: if (IdxIntInt /= 0) then
            test_flag_1: if (FlagIntInterfaces(IdxIntInt) == 0) then
              idxM = pmesh%elt(CellK)%idxM(i_K)
              loop_S_Ki: do idxJ = 1, IntInterfaces(IdxIntInt)%Card_S
                sum_d(:) = sum_d(:) + pmesh%elt(CellK)%le(i_K)*IntInterfaces(IdxIntInt)%nu(:,idxJ,idxM)
              end do loop_S_Ki

            else !if (FlagIntInterfaces(IdxIntInt) == 1 .or. FlagIntInterfaces(IdxIntInt) == 2) then

              !sum_d(:) = sum_d(:) + pmesh%elt(CellK)%le(i_K)*nu_one(:)
              sum_d(:) = sum_d(:) + pmesh%elt(CellK)%le(i_K)

            end if test_flag_1

          else
            !sum_d(:) = sum_d(:) + pmesh%elt(CellK)%le(i_K)*nu_one(:)
            sum_d(:) = sum_d(:) + pmesh%elt(CellK)%le(i_K)
          end if test_bnd_1
        end do loop_interfaces_K_1
        !-------------------------------------------------------------------------

        delta_K(:) = pmesh%elt(CellK)%area/sum_d(:)
#if defined(_SRC_ENABLE_)
        pmesh%elt(CellK)%delta_K(:) = delta_K(:)
#endif /* _SRC_ENABLE_ */

        !-------------------------------------------------------------------------
        U_K(:) = U_n(:,CellK)
        loop_interfaces_K_2: do i_K = 1, pmesh%elt(CellK)%NbInterfaces
          IdxIntInt = pmesh%elt(CellK)%IntInterfaces(i_K)

          test_bnd_2: if (IdxIntInt /= 0) then

            test_flag_2: if (FlagIntInterfaces(IdxIntInt) == 0) then

              idxM = pmesh%elt(CellK)%idxM(i_K)

              do idxJ = 1, IntInterfaces(IdxIntInt)%Card_S
                CellJ = IntInterfaces(IdxIntInt)%idx(idxJ,idxM)
                eta(:) = IntInterfaces(IdxIntInt)%eta(:,idxJ,idxM)
                tau(:) = IntInterfaces(IdxIntInt)%tau(:,idxJ,idxM)

                if (CellJ > 0) then
                  U_J(:) = U_n(:,CellJ)
                else
                  IdxBndInt = - CellJ
                  U_J(:) = compute_value_bnd(IdxBndInt,U_n,BndInterfaces(IdxBndInt)%x,t_n)
                endif
                b_KJ = compute_greatest_wave_speed(U_K,U_J,eta,tau)

                dt_tmp = max(dt_tmp,maxval(b_KJ/delta_K(:)))
              end do

            else !if (FlagIntInterfaces(IdxIntInt) == 2 .or. FlagIntInterfaces(IdxIntInt) == 1) then

              CellJ = pmesh%elt(CellK)%nghb(i_K)
              eta(:) = pmesh%elt(CellK)%normal(:,i_K)
              tau(:) = pmesh%elt(CellK)%tangent(:,i_K)
              U_J(:) = U_n(:,CellJ)
              b_KJ = compute_greatest_wave_speed(U_K,U_J,eta,tau)
              dt_tmp = max(dt_tmp,maxval(b_KJ/delta_K(:)))

            end if test_flag_2
          end if test_bnd_2
        end do loop_interfaces_K_2
        !-------------------------------------------------------------------------

      end do loop_elt
      !$OMP END DO

      !$OMP SINGLE
      dt_tmp = 1._PR/dt_tmp
      ! Take the minimum between the new and the old time step
      test_dt_tmp_greater: if (dt_tmp > dt) then
        dt_changed = .FALSE.
      else
        dt_changed = .TRUE.
        dt = dt_tmp
        t_n_p1 = t_n + dt
      end if test_dt_tmp_greater

      !if (IsZero(dt)) then
      if (abs(dt) < eps) then
        write(StdErr,'(2A,I4,A,ES23.16)')__FILE__,':', __LINE__, " -> Time step too small, dt=", dt
        FLUSH StdErr
        STOP
      end if
      !$OMP END SINGLE

  end subroutine Recompute_dt_TPCorrec

  !--------------------------------------------------------------------!
  !> @brief
  !> Re-compute \f$\Phi\f$ (sum of numerical fluxes on a cell)
  !>
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[inout] Phi Sum of numerical flux on all cells
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_mp_flux_1d
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_mp_flux_1d_source
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_mp_flux_1d_source_noap

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_coeff_dlp_mp_compute_nu_bar_inter
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_mp_compute_alpha_kk_noap
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_mp_compute_alpha_kk
  subroutine RecomputePhiTPCorrec(U_n,t_n,Phi)
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n
      real(PR), dimension(:,:), intent(inout) :: Phi

      real(PR), dimension(NbEq) :: flux_n, U_K, U_L
      real(PR), dimension(dim_) :: normal, tangent
      integer :: CellK, CellL, i_K, idxK, idxM, IdxIntInt, IdxBndInt

#if defined(_SRC_ENABLE_)
      real(PR) :: alpha_KK
#endif /* _SRC_ENABLE_ */

      !====================================================================!
      ! First recompute the cells that were not admissible,
      ! in those cells the AP correction is never use.
      !====================================================================!
      !$OMP DO
      loop_elt_CellsInvalid: do idxK = 1, NbCellsInvalid
        CellK = CellsToUpdate(idxK)
        Phi(:,CellK) = 0._PR
        U_K(:) = U_n(:,CellK)
        loop_interfaces_CellsInvalid: do i_K = 1, pmesh%elt(CellK)%NbInterfaces
          IdxIntInt = pmesh%elt(CellK)%IntInterfaces(i_K)

          test_bnd_CellsInvalid: if (IdxIntInt /= 0) then

              test_flag_CellsInvalid: if (FlagIntInterfaces(IdxIntInt) == 2 .or. FlagIntInterfaces(IdxIntInt) == 1) then
              ! Interface between two not admissible cells,
              ! we use the flux with K,L on \vec{n},
              ! in order to preserve the set of admissible states.
              ! OR
              ! Interface between an admissible cell and a not admissible cell,
              ! we use the flux with K,L on \vec{n},
              ! in order to enforce the conservation property.

              CellL = pmesh%elt(CellK)%nghb(i_K)
              U_L(:) = U_n(:,CellL)

              normal(:)  = pmesh%elt(CellK)%normal(:,i_K)
              tangent(:) = pmesh%elt(CellK)%tangent(:,i_K)

#if defined(_SRC_ENABLE_)
              ! The alpha_KK is computed without AP correction,
              ! in order to be constant over the cell, as the other interfaces
              ! use an alpha_KK without AP correction.
              alpha_KK = compute_alpha_KK_NoAP(CellK,U_K,normal,tangent)
              flux_n(:) = flux_1D_source_NoAP(CellK,pmesh%elt(CellK)%delta_K,U_K,U_L,normal,tangent,alpha_KK)
#else
              flux_n(:) = flux_1D(U_K,U_L,normal,tangent)
#endif /* _SRC_ENABLE_ */

            else
              write(StdErr,'(2A,I4,A,I1)')__FILE__,':', __LINE__, " -> Wrong flag: ", FlagIntInterfaces(IdxIntInt)
              FLUSH StdErr
              STOP

            end if test_flag_CellsInvalid

          else
            IdxBndInt = pmesh%elt(CellK)%BndInterfaces(i_K)
#if defined(_SRC_ENABLE_)
            normal(:)  = pmesh%elt(CellK)%normal(:,i_K)
            tangent(:) = pmesh%elt(CellK)%tangent(:,i_K)
            alpha_KK = compute_alpha_KK_NoAP(CellK,U_K,normal,tangent)
            call compute_flux_bnd_n_NoAP(CellK,IdxBndInt,i_K,U_n,t_n,alpha_KK,flux_n)
#else
            call compute_flux_bnd_n(CellK,IdxBndInt,i_K,U_n,t_n,flux_n)
#endif /* _SRC_ENABLE_ */
          end if test_bnd_CellsInvalid

          Phi(:,CellK) = Phi(:,CellK) + pmesh%elt(CellK)%le(i_K)*flux_n(:)

        end do loop_interfaces_CellsInvalid
        Phi(:,CellK) = Phi(:,CellK)/pmesh%elt(CellK)%area
      end do loop_elt_CellsInvalid
      !$OMP END DO
      !====================================================================!

      !====================================================================!
      ! Then recompute the nghb of the not admissible cells,
      ! for those the AP correction is used and especially on alpha_KK,
      ! in order that is the same on the cell.
      !====================================================================!
      !$OMP DO
      loop_elt_CellsToRecompute: do idxK = NbCellsInvalid + 1, NbCellsToUpdate
        CellK = CellsToUpdate(idxK)
        Phi(:,CellK) = 0._PR
        U_K(:) = U_n(:,CellK)
        loop_interfaces_CellsToRecompute: do i_K = 1, pmesh%elt(CellK)%NbInterfaces
          IdxIntInt = pmesh%elt(CellK)%IntInterfaces(i_K)

          test_bnd_CellsToRecompute: if (IdxIntInt /= 0) then

            test_flag_CellsToRecompute: if (FlagIntInterfaces(IdxIntInt) == 0) then
              ! Interface between an admissible cell and a nghb of a not admissible cell,
              ! the computation can be done classically

              idxM = pmesh%elt(CellK)%idxM(i_K)
#if defined(_SRC_ENABLE_) && !defined(_TP_FLUX_)
              call compute_nu_bar_inter(IdxIntInt,U_n,t_n,IntInterfaces(IdxIntInt)%nu_bar)
#endif /* _SRC_ENABLE_ && .not. _TP_FLUX_ */
              call compute_flux_n(CellK,IdxIntInt,idxM,U_n,t_n,flux_n)

            else if (FlagIntInterfaces(IdxIntInt) == 1) then
              ! Interface between an admissible cell and a not admissible cell,
              ! we use the flux with K,L on \vec{n},
              ! in order to enforce the conservation property.

              CellL = pmesh%elt(CellK)%nghb(i_K)
              U_L(:) = U_n(:,CellL)

              normal(:)  = pmesh%elt(CellK)%normal(:,i_K)
              tangent(:) = pmesh%elt(CellK)%tangent(:,i_K)
#if defined(_SRC_ENABLE_)
              ! The alpha_KK is computed is computed with AP correction,
              ! in order to be constant over the cell, as the other interfaces
              ! use an alpha_KK with AP correction.
              alpha_KK = compute_alpha_KK(CellK,U_K,normal,tangent)
              flux_n(:) = flux_1D_source_NoAP(CellK,pmesh%elt(CellK)%delta_K,U_K,U_L,normal,tangent,alpha_KK)
#else
              flux_n(:) = flux_1D(U_K,U_L,normal,tangent)
#endif /* _SRC_ENABLE_ */
            else
              write(StdErr,'(2A,I4,A,I1)')__FILE__,':', __LINE__, " -> Wrong flag: ", FlagIntInterfaces(IdxIntInt)
              FLUSH StdErr
              STOP

            end if test_flag_CellsToRecompute

          else
            IdxBndInt = pmesh%elt(CellK)%BndInterfaces(i_K)
            call compute_flux_bnd_n(CellK,IdxBndInt,i_K,U_n,t_n,flux_n)
          end if test_bnd_CellsToRecompute

          Phi(:,CellK) = Phi(:,CellK) + pmesh%elt(CellK)%le(i_K)*flux_n(:)

        end do loop_interfaces_CellsToRecompute
        Phi(:,CellK) = Phi(:,CellK)/pmesh%elt(CellK)%area
      end do loop_elt_CellsToRecompute
      !$OMP END DO
      !====================================================================!

  end subroutine RecomputePhiTPCorrec

#endif /* _TP_CORREC_ && .not. _HO_MOOD_ */
end module mod_TPCorrec_recompute
