!> Compute things related to source term and AP for the hyperbolic heat model.
!>
!> \f[\left\{\begin{array}{rcl}
!>   \partial_t p   + \partial_x u_x  + \partial_y u_y  & = & 0 \\
!>   \partial_t u_x + \partial_x p                      & = & -\sigma u_x \\
!>   \partial_t u_y                   + \partial_y p    & = & -\sigma u_y
!> \end{array}\right.\f]
!>
!> \f[\textbf{U} = \left(p, u_x, u_y\right)^T \f]
!> \f[\textbf{R}(\textbf{U}) = \left(p, 0, 0,\right)^T \f]
!> \f[\gamma(\textbf{U}) = \sigma(\textbf{U}) \f]
!>
!> Limit equation:
!> \f[\partial_t p - \text{div}\left(\frac{1}{\sigma} \nabla p \right) = 0 \f]
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_asymptotic_HypHeat
#if defined(_SRC_ENABLE_) && defined(_PHY_HYPHEAT_)

  use mod_numerics, only: PR, eps, zero

  use type_mesh,  only: pmesh, dim_
  use mod_physics_HypHeat, only: compute_greatest_wave_speed, compute_wave_speed
  use mod_physics_HypHeat_var, only: NbEq, test_case
  use mod_physics_HypHeat_var, only: sigma

  IMPLICIT NONE

  private
  public gamma_, gamma_bar_KJ, gamma_bar_KK
  public g_u
  public R

CONTAINS
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$\gamma(U)=\sigma\f$.
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] U_K Solution on cell \f$K\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hypheat_mp_compute_sigma_m
  function gamma_(CellK,U_K)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(in) :: U_K
      real(PR) :: gamma_

      gamma_ = compute_sigma(U_K)

  end function gamma_

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$\bar{\gamma}(U) = \bar{\sigma}\f$.
  !>
  !> * For Rusanov:
  !> \f[\bar{\sigma}_m = 3\sigma \frac{\nu^J_{K,i}b_{KJ}^2}
  !>   {2c^2\bar{\nu}^J_{K,i} \delta_{KJ}} - \sigma \f]
  !>
  !> * For HLL:
  !> \f[\bar{\sigma}_m = 3\sigma \frac{\nu^J_{K,i}(-b^+_{KJ}b^-_{KJ})}
  !>   {2c^2\bar{\nu}^J_{K,i} \delta_{KJ}} - \sigma \f]
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] nu_J, nu_bar_J, delta_K Coefficients
  !> @param[in] U_K, U_J Solution on cells \f$K\f$ and \f$J\f$
  !> @param[in] eta Unitary vector in the direction \f$KJ\f$
  !> @param[in] tau Unitary vector in the direction orthogonal to \f$KJ\f$
  !> @param[in] gamma_K \f$\gamma(U_K)\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_greatest_wave_speed
  function gamma_bar_KJ(CellK,nu_J,nu_bar_J,delta_K,U_K,U_J,eta,tau,gamma_K)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(in) :: nu_J
      real(PR), intent(in)               :: nu_bar_J
      real(PR), dimension(:), intent(in) :: delta_K
      real(PR), dimension(:), intent(in) :: U_K, U_J
      real(PR), dimension(:), intent(in) :: eta, tau
      real(PR), intent(in) :: gamma_K
      real(PR) :: gamma_bar_KJ

#if defined(_SRC_DISABLE_AP_CORREC_)
      gamma_bar_KJ = 0._PR
#else
      real(PR) :: speed_square
      integer, parameter  :: eq_lim = 1

#if defined(_FLUX_RUS_)
      real(PR) :: b_KJ

      b_KJ = compute_greatest_wave_speed(U_K, U_J, eta, tau)
      speed_square = b_KJ**2

#elif defined(_FLUX_HLL_)
      real(PR) :: wave_speed_J, wave_speed_K

      call compute_wave_speed(U_K, U_J, eta, tau, wave_speed_K, wave_speed_J)
      speed_square = -wave_speed_J*wave_speed_K
#endif /* _FLUX_ */

      if (abs(nu_J(eq_lim)) < eps .or. abs(nu_bar_J) < eps) then
        gamma_bar_KJ = 0._PR
      else
        gamma_bar_KJ = gamma_K*(nu_J(eq_lim)*speed_square&
            /(2._PR*nu_bar_J*delta_K(eq_lim)) - 1._PR)
      end if
#endif /* _SRC_DISABLE_AP_CORREC_ */

  end function gamma_bar_KJ

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$\bar{\gamma}(U) = \bar{\sigma}\f$.
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] U_K Solution on cell \f$K\f$
  !> @param[in] eta Unitary vector in the direction \f$KJ\f$
  !> @param[in] tau Unitary vector in the direction orthogonal to \f$KJ\f$
  !> @param[in] gamma_K \f$\gamma(U_K)\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_hypheat_mp_gamma_
  function gamma_bar_KK(CellK,U_K,eta,tau,gamma_K)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), dimension(:), intent(in) :: eta, tau
      real(PR), intent(in) :: gamma_K
      real(PR) :: gamma_bar_KK

#if defined(_SRC_DISABLE_AP_CORREC_)
      gamma_bar_KK = 0._PR
#else

      real(PR) :: speed_square
#if defined(_FLUX_RUS_)
      real(PR) :: b_KK

      b_KK = compute_greatest_wave_speed(U_K, U_K, eta, tau)
      speed_square = b_KK**2

#elif defined(_FLUX_HLL_)
      real(PR) :: wave_speed_Kp, wave_speed_Km

      call compute_wave_speed(U_K, U_K, eta, tau, wave_speed_Km, wave_speed_Kp)
      speed_square = -wave_speed_Km*wave_speed_Kp
#endif /* _FLUX_ */

      gamma_bar_KK = gamma_K*(2._PR*speed_square - 1._PR)
#endif /* _SRC_DISABLE_AP_CORREC_ */

  end function gamma_bar_KK

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$R(U)\f$.
  !>
  !> \f[\textbf{R}(\textbf{U}) = \left(E, 0, 0\right)^T \f]
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] U_K Solution on cell \f$K\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function R(CellK,U_K)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), dimension(NbEq) :: R


      R(1)   = U_K(1)
      R(2:3) = zero

  end function R

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute \f$g(u)\f$ of the limit equation.
  !>
  !> Generic limit equation:
  !> \f[\partial_t(f(u)) - \text{div}(\alpha(u) \nabla g(u)) = 0 \f]
  !>
  !> Limit equation for \f$M_1\f$:
  !> \f[\partial_t p - \text{div}\left(\frac{1}{\sigma} \nabla p \right) = 0 \f]
  !>
  !> @param[in] U Conservative variable of the hyperbolic system
  !> @return \f$g(u)\f$, with \f$u\f$ the variable of the limit equation
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function g_u(U)
      real(PR), dimension(:), intent(in) :: U
      real(PR) :: g_u

      g_u = U(1)
  end function g_u

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the value of \f$\sigma\f$
  !>
  !> @param[in] U Conservative variable
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function compute_sigma(U)
      real(PR), dimension(:), intent(in) :: U
      real(PR) :: compute_sigma

      select case(test_case)
      case  default
        compute_sigma = sigma
      end select

  end function compute_sigma

#endif /* _PHY_HYPHEAT_ */
end module mod_asymptotic_HypHeat
