!> Compute things related to source term and AP for the multigroup M1 model.
!>
!> \f[\left\{\begin{array}{rcl}
!>   \partial_t E_q       +     \text{div} ({\bf F}_q) & = & c (\sigma^e_q a \theta_q^4(T) - \sigma^a_q E_q) \\
!>   \partial_t {\bf F}_q + c^2 \text{div} ({\bf P_q}) & = & -c \sigma^f_q {\bf F}_q \\
!>   \rho C_v \partial_t T                             & = & -c \sum_{q=1}^Q (\sigma^e_q a \theta_q^4(T) - \sigma^a_q E_q)
!> \end{array}\right.\f]
!>
!> where the index \f$q\f$ defines each group.
!>
!> with :
!> * \f$ E_q \f$ the radiative energy,
!> * \f$ {\bf F}_q \f$ the radiative flux,
!> * \f$ T \f$ the temperature,
!> * \f$ {\bf P}_q = \frac{E_q}{2}\left( (1-\chi_q) {\bf I}  + (3\chi_q-1)\frac{{\bf F}_q\otimes {\bf F}_q}{\Vert {\bf F}_q \Vert_2^2}\right) \f$ the radiative pressure,
!> * \f$\chi_q\f$ is the Eddington factor and it is no more explicit as in the grey M1 model, but as a fist approxamtion we use it:
!>     \f$\chi_q = \frac{3 + 4f_q^2}{5+ \sqrt{4-3f_q^2}}\f$
!> * \f$f_q\f$ is the anistropy factor: \f$f_q=\frac{\Vert {\bf F}_q \Vert}{c E_q} \leq 1 \f$,
!> * \f$ c \f$ the speed of light.
!>
!> @remark See \cite Tur_2002 \cite Tur_2003 \cite Sar_2013 for the definition.
!>
!> The limit equation is:
!> \f[ \partial_t (\rho C_V T + a T^4) - \text{div} \Bigg(\sum_q \Big(\frac{c}{3 \sigma^R_q} \nabla a \theta_q^4(T)\Big) \Bigg) \f]
!>
!> @author Florian Blachère>
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.

module mod_asymptotic_M1MG
#if defined(_SRC_ENABLE_) && defined(_PHY_M1MG_)

  use mod_numerics, only: PR, eps, InitMax
  use type_IO, only: StdErr

  use type_mesh,  only: pmesh, dim_
  use mod_tools, only: max_NaN
  use mod_tools_array, only: FindIndex
  use mod_tools_interp, only: LinearInterp, TriLinearInterp
  use mod_physics_M1MG, only: compute_greatest_wave_speed, compute_wave_speed
  use mod_physics_M1MG, only: compute_f
  use mod_physics_M1MG, only: compute_a_theta_q_4, compute_dT_a_theta_q_4
  use mod_physics_M1MG_var, only: NbEq, test_case
  use mod_physics_M1MG_var, only: NbGroup
  use mod_physics_M1MG_var, only: c, a, rho, C_v
  use mod_physics_M1MG_var, only: tab_T
  use mod_physics_M1MG_var, only: tab_f
  use mod_physics_M1MG_var, only: tab_T_R
  use mod_physics_M1MG_var, only: tab_sigma_a, tab_sigma_e, tab_sigma_f, tab_sigma_R
#if defined(_FC_INTEL_) || (defined(_FC_GNU_) && __GNUC__>=5) || defined(_FC_PGI_)
  use, INTRINSIC :: IEEE_ARITHMETIC, only: IEEE_IS_NAN
#endif /* _FC_ */


  IMPLICIT NONE

  private
  public gamma_, gamma_bar_KJ, gamma_bar_KK
  public g_u
  public R

CONTAINS
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$\gamma(U)=c \sigma^m\f$.
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] U_K Solution on cell \f$K\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_compute_sigma_m
  function gamma_(CellK,U_K)
    integer, intent(in) :: CellK
    real(PR), dimension(:), intent(in) :: U_K
    real(PR) :: gamma_

    gamma_ = c*compute_sigma_m(CellK,U_K)

  end function gamma_

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$\bar{\gamma}(U) = c\bar{\sigma}^m\f$.
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] nu_J, nu_bar_J, delta_K Coefficients
  !> @param[in] U_K, U_J Solution on cells \f$K\f$ and \f$J\f$
  !> @param[in] eta Unitary vector in the direction \f$KJ\f$
  !> @param[in] tau Unitary vector in the direction orthogonal to \f$KJ\f$
  !> @param[in] gamma_K \f$\gamma(U_K)\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_compute_sigma_R
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_a_theta_q_4
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_dt_a_theta_q_4
  function gamma_bar_KJ(CellK,nu_J,nu_bar_J,delta_K,U_K,U_J,eta,tau,gamma_K)
    integer, intent(in) :: CellK
    real(PR), dimension(:), intent(in) :: nu_J
    real(PR), intent(in)               :: nu_bar_J
    real(PR), dimension(:), intent(in) :: delta_K
    real(PR), dimension(:), intent(in) :: U_K, U_J
    real(PR), dimension(:), intent(in) :: eta, tau
    real(PR), intent(in) :: gamma_K
    real(PR) :: gamma_bar_KJ

    real(PR) :: sigma_R
    real(PR) :: T_K, T_J
    real(PR) :: speed_square
    integer :: group, eq0

#if defined(_FLUX_RUS_)
    real(PR) :: b_KJ
#elif defined(_FLUX_HLL_)
    real(PR) :: wave_speed_J, wave_speed_K
#endif /* _FLUX_ */

    gamma_bar_KJ = 0._PR

#if !defined(_SRC_DISABLE_AP_CORREC_)
    sigma_R = compute_sigma_R(U_K)
    test_nu_bar: if (.not. (abs(nu_bar_J) < eps)) then
#if defined(_FLUX_RUS_)
      b_KJ = compute_greatest_wave_speed(U_K, U_J, eta, tau)
      speed_square = b_KJ**2
#elif defined(_FLUX_HLL_)
      call compute_wave_speed(U_K, U_J, eta, tau, wave_speed_K, wave_speed_J)
      speed_square = -wave_speed_J*wave_speed_K
#endif /* _FLUX_ */

      T_K = U_K(NbEq)
      T_J = U_J(NbEq)

      test_temp: if (abs(T_J - T_K) > sqrt(eps)) then
        loop_group: do group = 1, NbGroup
          eq0 = (1 + dim_)*(group-1)
          gamma_bar_KJ = gamma_bar_KJ + nu_J(eq0+1)/delta_K(eq0+1)&
              *(compute_a_theta_q_4(group,T_J) - compute_a_theta_q_4(group,T_K))/(a*T_J**4 - a*T_K**4)
        end do loop_group

      else

        loop_group_same_temp: do group = 1, NbGroup
          eq0 = (1 + dim_)*(group-1)
          gamma_bar_KJ = gamma_bar_KJ + nu_J(eq0+1)/delta_K(eq0+1)&
              *compute_dT_a_theta_q_4(group,T_K)
        end do loop_group_same_temp

      end if test_temp

      gamma_bar_KJ = gamma_bar_KJ*speed_square*3._PR*sigma_R/(2._PR*c*nu_bar_J) - gamma_K
    end if test_nu_bar

#endif /* _SRC_DISABLE_AP_CORREC_ */

  end function gamma_bar_KJ

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$\bar{\gamma}(U) = c\bar{\sigma}^m\f$.
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] U_K Solution on cell \f$K\f$
  !> @param[in] eta Unitary vector in the direction \f$KJ\f$
  !> @param[in] tau Unitary vector in the direction orthogonal to \f$KJ\f$
  !> @param[in] gamma_K \f$\gamma(U_K)\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_compute_sigma_R
  function gamma_bar_KK(CellK,U_K,eta,tau,gamma_K)
    integer, intent(in) :: CellK
    real(PR), dimension(:), intent(in) :: U_K
    real(PR), dimension(:), intent(in) :: eta, tau
    real(PR), intent(in) :: gamma_K
    real(PR) :: gamma_bar_KK

    real(PR) :: speed_square
    real(PR) :: sigma_R
#if defined(_FLUX_RUS_)
    real(PR) :: b_KK
#elif defined(_FLUX_HLL_)
    real(PR) :: wave_speed_Kp, wave_speed_Km
#endif /* _FLUX_ */

    gamma_bar_KK = 0._PR

#if !defined(_SRC_DISABLE_AP_CORREC_)
    sigma_R = compute_sigma_R(U_K)
#if defined(_FLUX_RUS_)
    b_KK = compute_greatest_wave_speed(U_K, U_K, eta, tau)
    speed_square = b_KK**2
#elif defined(_FLUX_HLL_)
    call compute_wave_speed(U_K, U_K, eta, tau, wave_speed_Km, wave_speed_Kp)
    speed_square = -wave_speed_Km*wave_speed_Kp
#endif /* _FLUX_ */

    gamma_bar_KK = speed_square*sigma_R/c*6._PR*(4._PR*a*U_K(Nbeq)**3) - gamma_K

#endif /* _SRC_DISABLE_AP_CORREC_ */

  end function gamma_bar_KK

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the coefficient \f$\sigma^m(U)\f$.
  !>
  !> \f[\sigma^m = \max\left(\sigma^a_q,\sigma^f_q,\sum_q\frac{\sigma^e_q a \theta_q^4(T)}{\rho C_v T}\right)\f]
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] U_K Solution on cell \f$K\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_compute_sigma_a
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_compute_sigma_e
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_compute_sigma_f
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_a_theta_q_4
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_max_nan
  function compute_sigma_m(CellK,U_K)
    integer, intent(in) :: CellK
    real(PR), dimension(:), intent(in) :: U_K
    real(PR) :: compute_sigma_m

    real(PR) :: T, sum_sigma_e
    integer :: group

    T = U_K(NbEq)
    sum_sigma_e = 0._PR

    compute_sigma_m = InitMax
    do group = 1, NbGroup
      compute_sigma_m = max_NaN(compute_sigma_m,max_NaN(compute_sigma_a(group,U_K),compute_sigma_f(group,U_K)))
      sum_sigma_e = sum_sigma_e + compute_sigma_e(group,U_K)*compute_a_theta_q_4(group,T)
    end do
    compute_sigma_m = max_NaN(compute_sigma_m,sum_sigma_e/(rho(CellK)*C_v(CellK)*T))

  end function compute_sigma_m

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the coefficient \f$\sigma_{1,q}(U) = \sigma^m - \sigma^a_q\f$.
  !>
  !> @param[in] group Index of the group
  !> @param[in] sigma_m \f$\sigma^m\f$
  !> @param[in] U_K Solution on cell \f$K\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_compute_sigma_a
  function compute_sigma_1(group,sigma_m,U_K)
    integer, intent(in) :: group
    real(PR), intent(in)               :: sigma_m
    real(PR), dimension(:), intent(in) :: U_K
    real(PR) :: compute_sigma_1

    compute_sigma_1 = sigma_m - compute_sigma_a(group,U_K)
  end function compute_sigma_1

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the coefficient \f$\sigma_{2,q}(U) = \sigma^m - \sigma^f_q\f$.
  !>
  !> @param[in] group Index of the group
  !> @param[in] sigma_m \f$\sigma^m\f$
  !> @param[in] U_K Solution on cell \f$K\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_compute_sigma_f
  function compute_sigma_2(group,sigma_m,U_K)
    integer, intent(in) :: group
    real(PR), intent(in)               :: sigma_m
    real(PR), dimension(:), intent(in) :: U_K
    real(PR) :: compute_sigma_2

    compute_sigma_2 = sigma_m - compute_sigma_f(group,U_K)
  end function compute_sigma_2

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the coefficient \f$\sigma_3(U) = \sigma^m - \sum_q\frac{\sigma^e_q a \theta_q^4(T)}{\rho C_v T}\f$.
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] sigma_m \f$\sigma^m\f$
  !> @param[in] U_K Solution on cell \f$K\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_compute_sigma_e
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_a_theta_q_4
  function compute_sigma_3(CellK,sigma_m,U_K)
    integer, intent(in) :: CellK
    real(PR), intent(in)               :: sigma_m
    real(PR), dimension(:), intent(in) :: U_K
    real(PR) :: compute_sigma_3

    real(PR) :: T, sum_sigma_e
    integer :: group

    T = U_K(NbEq)
    sum_sigma_e = 0._PR

    do group = 1, NbGroup
      sum_sigma_e = sum_sigma_e + compute_sigma_e(group,U_K)*compute_a_theta_q_4(group,T)
    end do
    compute_sigma_3 = sigma_m - sum_sigma_e/(rho(CellK)*C_v(CellK)*T)

  end function compute_sigma_3

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the function \f$R(U)\f$.
  !>
  !> \f[\textbf{R}(\textbf{U}) = \frac{1}{\sigma^m}\left(\sigma^e a T^4 + \sigma_1 E,
  !>                  \sigma_2 F_x, \sigma_2 F_y,
  !>                  \frac{\sigma^a E}{\rho C_v} + \sigma_3 T\right)^T \f]
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] U_K Solution on cell \f$K\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_compute_sigma_m
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_compute_sigma_1
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_compute_sigma_2
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_compute_sigma_3
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_compute_sigma_a
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_m1mg_mp_compute_sigma_e
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_a_theta_q_4
  function R(CellK,U_K)
    integer, intent(in) :: CellK
    real(PR), dimension(:), intent(in) :: U_K
    real(PR), dimension(NbEq) :: R

    real(PR), dimension(dim_) :: F
    real(PR) :: sigma_m, sigma_1, sigma_2, sigma_3
    real(PR) :: E, T
    real(PR) :: sum_sigma_a_E
    integer :: group, eq0

    sigma_m = compute_sigma_m(CellK,U_K)

    if (.not. (abs(sigma_m) < eps)) then
      T = U_K(NbEq)
      sum_sigma_a_E = 0._PR

      do group = 1, NbGroup
        eq0 = (1 + dim_)*(group-1)

        E = U_K(eq0+1)
        F(1) = U_K(eq0+2)
        F(2) = U_K(eq0+3)

        sigma_1 = compute_sigma_1(group,sigma_m,U_K)
        sigma_2 = compute_sigma_2(group,sigma_m,U_K)

        R(eq0+1) = (compute_sigma_e(group,U_K)*compute_a_theta_q_4(group,T) + sigma_1*E)
        R(eq0+2) = sigma_2*F(1)
        R(eq0+3) = sigma_2*F(2)

        sum_sigma_a_E = sum_sigma_a_E + compute_sigma_a(group,U_K)*E
      end do

      sigma_3 = compute_sigma_3(CellK,sigma_m,U_K)
      R(NbEq) = sum_sigma_a_E/(rho(CellK)*C_v(CellK)) + sigma_3*T

      R(:) = R(:)/sigma_m
    else
      R(:) = 0._PR
    end if

  end function R

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute \f$g(u)\f$ of the limit equation.
  !>
  !> Generic limit equation:
  !> \f[\partial_t(f(u)) - \text{div}(\alpha(u) \nabla g(u)) = 0 \f]
  !>
  !> Limit equation for \f$M_1\f$:
  !> \f[\partial_t(\rho C_v + aT^4) - \text{div}(\frac{c}{3\sigma^f}\nabla (a T^4)) = 0 \f]
  !>
  !> @param[in] U Conservative variable of the hyperbolic system
  !> @return \f$g(u)\f$, with \f$u\f$ the variable of the limit equation
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function g_u(U)
    real(PR), dimension(:), intent(in) :: U
    real(PR) :: g_u

    g_u = a*U(NbEq)**4
  end function g_u

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the value of \f$\sigma^a_q\f$
  !>
  !> @param[in] group Index of the group
  !> @param[in] U_K Conservative variable
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_array_mp_findindex
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_interp_mp_trilinearinterp
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_f
  function compute_sigma_a(group,U_K)
    integer, intent(in) :: group
    real(PR), dimension(:), intent(in) :: U_K
    real(PR) :: compute_sigma_a

    real(PR) :: T, ff, T_R, E
    real(PR), dimension(dim_) :: F
    integer :: eq0
    integer :: idx_T, idx_f, idx_T_R

    T = U_K(NbEq)
    eq0 = (1 + dim_)*(group-1)
    E = U_K(eq0+1)
    F(:) = U_K(eq0+2:eq0+3)
    ff = compute_f(E,F)
    T_R = sqrt(sqrt(E/a))

    idx_f = FindIndex(ff,tab_f)
    idx_T = FindIndex(T,tab_T)
    idx_T_R = FindIndex(T_R,tab_T_R)

#if defined(_TP_CORREC_) || ((_HO_MOOD_DEGREE_) >=1)
    ! The values are not valid, generate a NaN to limit
    if (idx_f < 0) then
      compute_sigma_a = sqrt(1._PR - abs(ff))
      return
    else if (idx_T < 0) then
      compute_sigma_a = sqrt(T)
      return
    else if (idx_T_R < 0) then
      compute_sigma_a = sqrt(U_K(eq0+1))
      return
    end if
#endif /* defined(_TP_CORREC_) || ((_HO_MOOD_DEGREE_) >=1) */

    ! Linear interpolation
    compute_sigma_a = TriLinearInterp(tab_f(idx_f),tab_f(idx_f+1),&
        tab_T_R(idx_T_R), tab_T_R(idx_T_R+1),&
        tab_T(idx_T),tab_T(idx_T+1),&
        tab_sigma_a(idx_f,idx_T_R,idx_T,group),&
        tab_sigma_a(idx_f+1,idx_T_R,idx_T,group),&
        tab_sigma_a(idx_f+1,idx_T_R+1,idx_T,group), &
        tab_sigma_a(idx_f,idx_T_R+1,idx_T,group),&
        tab_sigma_a(idx_f,idx_T_R,idx_T+1,group),&
        tab_sigma_a(idx_f+1,idx_T_R,idx_T+1,group),&
        tab_sigma_a(idx_f+1,idx_T_R+1,idx_T+1,group),&
        tab_sigma_a(idx_f,idx_T_R+1,idx_T+1,group),&
        ff,T_R,T)

    ! In case of NaN in the tabulated values
#if defined(_FC_INTEL_) || (defined(_FC_GNU_) && __GNUC__>=5) || defined(_FC_PGI_)
    if (IEEE_IS_NAN(compute_sigma_a)) then
      compute_sigma_a = 0._PR
    end if
#else
    if (ISNAN(compute_sigma_a)) then
      compute_sigma_a = 0._PR
    end if
#endif /* _FC_ */

    ! Cut-off in case of bad interpolation
    if (compute_sigma_a < 0._PR) compute_sigma_a = 0._PR

  end function compute_sigma_a

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the value of \f$\sigma^e_q\f$
  !>
  !> @param[in] group Index of the group
  !> @param[in] U_K Conservative variable
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_array_mp_findindex
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_interp_mp_linearinterp
  function compute_sigma_e(group,U_K)
    integer, intent(in) :: group
    real(PR), dimension(:), intent(in) :: U_K
    real(PR) :: compute_sigma_e

    real(PR) :: T
    integer :: idx_T

    T = U_K(NbEq)

    ! Find nearest index of T in the precomputed values tab_T
    idx_T = FindIndex(T,tab_T)

#if defined(_TP_CORREC_) || ((_HO_MOOD_DEGREE_) >=1)
    ! The values are not valid, generate a NaN to limit
    if (idx_T < 0) then
      compute_sigma_e = sqrt(T)
      return
    end if
#endif /* defined(_TP_CORREC_) || ((_HO_MOOD_DEGREE_) >=1) */

    ! Linear interpolation between the two temperatures
    compute_sigma_e = LinearInterp(tab_T(idx_T),tab_sigma_e(idx_T,group),&
        tab_T(idx_T+1),tab_sigma_e(idx_T+1,group),T)

    ! Cut-off in case of bad interpolation
    if (compute_sigma_e < 0._PR) compute_sigma_e = 0._PR

  end function compute_sigma_e

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the value of \f$\sigma^f_q\f$
  !>
  !> @param[in] group Index of the group
  !> @param[in] U_K Conservative variable
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_array_mp_findindex
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_interp_mp_trilinearinterp
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_f
  function compute_sigma_f(group,U_K)
    integer, intent(in) :: group
    real(PR), dimension(:), intent(in) :: U_K
    real(PR) :: compute_sigma_f

    real(PR) :: T, ff, T_R, E
    real(PR), dimension(dim_) :: F
    integer :: eq0
    integer :: idx_T, idx_f, idx_T_R

    T = U_K(NbEq)
    eq0 = (1 + dim_)*(group-1)
    E = U_K(eq0+1)
    F(:) = U_K(eq0+2:eq0+3)
    ff = compute_f(E,F)
    T_R = sqrt(sqrt(E/a))

    idx_f = FindIndex(ff,tab_f)
    idx_T = FindIndex(T,tab_T)
    idx_T_R = FindIndex(T_R,tab_T_R)

#if defined(_TP_CORREC_) || ((_HO_MOOD_DEGREE_) >=1)
    ! The values are not valid, generate a NaN to limit
    if (idx_f < 0) then
      compute_sigma_f = sqrt(1._PR - abs(ff))
      return
    else if (idx_T < 0) then
      compute_sigma_f = sqrt(T)
      return
    else if (idx_T_R < 0) then
      compute_sigma_f = sqrt(U_K(eq0+1))
      return
    end if
#endif /* defined(_TP_CORREC_) || ((_HO_MOOD_DEGREE_) >=1) */

    ! Linear interpolation
    compute_sigma_f = TriLinearInterp(tab_f(idx_f),tab_f(idx_f+1),&
        tab_T_R(idx_T_R), tab_T_R(idx_T_R+1),&
        tab_T(idx_T),tab_T(idx_T+1),&
        tab_sigma_f(idx_f,idx_T_R,idx_T,group),&
        tab_sigma_f(idx_f+1,idx_T_R,idx_T,group),&
        tab_sigma_f(idx_f+1,idx_T_R+1,idx_T,group), &
        tab_sigma_f(idx_f,idx_T_R+1,idx_T,group),&
        tab_sigma_f(idx_f,idx_T_R,idx_T+1,group),&
        tab_sigma_f(idx_f+1,idx_T_R,idx_T+1,group),&
        tab_sigma_f(idx_f+1,idx_T_R+1,idx_T+1,group),&
        tab_sigma_f(idx_f,idx_T_R+1,idx_T+1,group),&
        ff,T_R,T)

    ! In case of NaN in the tabulated values
#if defined(_FC_INTEL_) || (defined(_FC_GNU_) && __GNUC__>=5) || defined(_FC_PGI_)
    if (IEEE_IS_NAN(compute_sigma_f)) then
      compute_sigma_f = 0._PR
    end if
#else
    if (ISNAN(compute_sigma_f)) then
      compute_sigma_f = 0._PR
    end if
#endif /* _FC_ */

    ! Cut-off in case of bad interpolation
    if (compute_sigma_f < 0._PR) compute_sigma_f = 0._PR

  end function compute_sigma_f

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the value of \f$\sigma^R\f$
  !>
  !> @param[in] U_K Conservative variable
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_array_mp_findindex
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_interp_mp_linearinterp
  function compute_sigma_R(U_K)
    real(PR), dimension(:), intent(in) :: U_K
    real(PR) :: compute_sigma_R

    real(PR) :: T
    integer :: idx_T

    T = U_K(NbEq)

    idx_T = FindIndex(T,tab_T)

#if defined(_TP_CORREC_) || ((_HO_MOOD_DEGREE_) >=1)
    ! The values are not valid, generate a NaN to limit
    if (idx_T < 0) then
      compute_sigma_R = sqrt(T)
      return
    end if
#endif /* defined(_TP_CORREC_) || ((_HO_MOOD_DEGREE_) >=1) */

    compute_sigma_R = LinearInterp(tab_T(idx_T),tab_sigma_R(idx_T),&
        tab_T(idx_T+1),tab_sigma_R(idx_T+1),T)

    ! Cut-off in case of bad interpolation
    if (compute_sigma_R < 0._PR) compute_sigma_R = 0._PR

  end function compute_sigma_R

#endif /* _PHY_M1MG_ */
end module mod_asymptotic_M1MG
