!> Computations of the \f$\delta_{KJ}\f$ coefficients
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_delta
#if !defined(_TP_FLUX_) && defined(_SRC_ENABLE_)

  use mod_numerics, only: PR

  use type_mesh, only: pmesh, BndInterfaces, dim_
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells
#endif /* _FC_PAR_OMP_ */
  use type_mesh_DLP, only: IntInterfaces, nu_one
  use mod_physics_var, only: NbEq
#if (_HO_MOOD_MAX_DEGREE_ >=1)
  use type_IO, only: StdErr
  use type_HO_MOOD, only: MaxDegree
  use mod_HO_MOOD_quad, only: quad_seg
#endif /* _HO_MOOD_MAX_DEGREE_ >=1 */

  IMPLICIT NONE

  private
  public compute_delta_all
  public compute_delta
#if (_HO_MOOD_MAX_DEGREE_ >=1)
  public compute_delta_tilde
#endif /* _HO_MOOD_MAX_DEGREE_ >=1 */

CONTAINS
  !--------------------------------------------------------------------------
  !> @brief
  !> Compute all the coefficients \f$\delta_{KJ}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_delta_mp_compute_delta
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_delta_mp_compute_delta_tilde
  subroutine compute_delta_all()
      integer :: CellK

      !$OMP DO &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      loop_elt: do CellK = 1, pmesh%NbCells
        call compute_delta(CellK,pmesh%elt(CellK)%delta_K(:))
#if (_HO_MOOD_MAX_DEGREE_ >=1)
        call compute_delta_tilde(CellK,MaxDegree,pmesh%elt(CellK)%delta_K_tilde(:,:))
#endif /* _HO_MOOD_MAX_DEGREE_ >=1 */
      end do loop_elt

  end subroutine compute_delta_all

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the coefficient \f$\delta_{KJ}\f$ for one cell
  !>
  !> \f{align*}{\delta_{KJ} & = \frac{\vert K \vert}{\sum_{i \in \mathcal{E}_K} \vert e_i \vert \sum_{J \in \mathcal{S}_{K,i}}\nu^J_{K,i}} \\
  !>                        & = \frac{\vert K \vert}{\sum_{J \in \overline{\mathcal{E}}_K} d_{KJ}}\f}
  !>
  !> @param[in] CellK Current Cell
  !> @param[out] delta_K Coefficients
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine compute_delta(CellK,delta_K)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(out) :: delta_K

      real(PR), dimension(NbEq) :: sum_d
      integer :: IdxIntInt, idxJ, i_K
      integer :: idxM

      sum_d(:) = 0._PR

      loop_nghb: do i_K=1, pmesh%elt(CellK)%NbInterfaces
        IdxIntInt = pmesh%elt(CellK)%IntInterfaces(i_K)

        test_bnd: if (IdxIntInt /= 0) then
          idxM = pmesh%elt(CellK)%idxM(i_K)

          loop_S_Ki: do idxJ = 1, IntInterfaces(IdxIntInt)%Card_S
            sum_d(:) = sum_d(:) + pmesh%elt(CellK)%le(i_K)*IntInterfaces(IdxIntInt)%nu(:,idxJ,idxM)
          end do loop_S_Ki

        else
          sum_d(:) = sum_d(:) + pmesh%elt(CellK)%le(i_K)*nu_one(:)
        end if test_bnd

      end do loop_nghb

      delta_K(:) = pmesh%elt(CellK)%area/sum_d(:)

  end subroutine compute_delta

#if (_HO_MOOD_MAX_DEGREE_ >=1)
  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the coefficient \f$\delta_{KJ}\f$ for one cell with the HO coefficients
  !>
  !> @param[in] CellK Current Cell
  !> @param[in] degree Degree of the polynomial reconstruction
  !> @param[out] delta_K_tilde Coefficients
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine compute_delta_tilde(CellK,degree,delta_K_tilde)
      integer, intent(in) :: CellK
      integer, intent(in) :: degree
      real(PR), dimension(:,:), intent(out) :: delta_K_tilde

      real(PR), dimension(NbEq,quad_seg(degree)%nb_points) :: sum_d_tilde
      integer :: IdxIntInt, idxJ, i_K, IdxQuad
      integer :: idxM

      test_degree: if (degree >= 1 .and. degree <= 5) then

        sum_d_tilde(:,:) = 0._PR

          loop_nghb: do i_K=1, pmesh%elt(CellK)%NbInterfaces
            IdxIntInt = pmesh%elt(CellK)%IntInterfaces(i_K)

            test_bnd: if (IdxIntInt /= 0) then
              idxM = pmesh%elt(CellK)%idxM(i_K)

              loop_S_Ki: do idxJ = 1, IntInterfaces(IdxIntInt)%Card_S
                sum_d_tilde(:,:) = sum_d_tilde(:,:) + pmesh%elt(CellK)%le(i_K)*IntInterfaces(IdxIntInt)%nu_tilde(:,idxJ,:,idxM)
              end do loop_S_Ki

            else
              loop_quad: do IdxQuad = 1, quad_seg(degree)%nb_points
                sum_d_tilde(:,IdxQuad) = sum_d_tilde(:,IdxQuad) + pmesh%elt(CellK)%le(i_K)*nu_one(:)
              end do loop_quad
            end if test_bnd

          end do loop_nghb

      else
        write(StdErr,'(2A,I4,A,I1)')__FILE__,':', __LINE__, " -> Wrong degree: ", degree
        FLUSH StdErr
        STOP
      end if test_degree


      delta_K_tilde(:,:) = pmesh%elt(CellK)%area/sum_d_tilde(:,:)

  end subroutine compute_delta_tilde
#endif /* _HO_MOOD_MAX_DEGREE_ >=1 */

#endif /* .not. _TP_FLUX_ && _SRC_ENABLE_ */
end module mod_delta
