!> Save the arrays necessary for the recomputations
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module type_recompute
#if defined(_TP_CORREC_) || defined(_HO_MOOD_)

  use mod_numerics, only: PR
  use type_mesh, only: pmesh
  use type_IO, only: StdOut

  IMPLICIT NONE

  !> Number of invalid cells
  integer, save :: NbCellsInvalid

  !> Number of the cells that need to be recomputed
  !> (invalid cells or neighbors of an invalid cell)
  integer, save :: NbCellsToUpdate

  !> Indexes of the cells that need to be recomputed
  !> (invalid cells or neighbors of an invalid cell)
  integer, dimension(:), allocatable, save :: CellsToUpdate

  !> Flag for the cells:
  !> * 0: valid
  !> * 1: invalid
  !> * 2: neighbour of an invalid cell (only use when defined(_TP_CORREC_) && !defined(_HO_MOOD_))
  integer, dimension(:), allocatable, save :: FlagCells

#if defined(_TP_CORREC_) && !defined(_HO_MOOD_)
  !> Flag for the interior interface:
  !> * 0: interface between two admissible cells
  !> * 1: interface between a valid cell and an invalid admissible one
  !> * 2: interface between two invalid cells
  integer, dimension(:), allocatable, save :: FlagIntInterfaces
#endif /* _TP_CORREC_ !_HO_MOOD_ */

  !> Maximum number of iterations possible in the loop of the re-computations
  integer, parameter :: MaxIterPossibleReComp = 1000

  !> Total number of activation of the re-computation
  integer, save :: TotReComp
  !> Total number of iterations of the re-computation
  integer, save :: TotIterReComp

  !> Maximum number of invalid cells
  integer, save :: MaxInvalidCells
  !> Maximum number of iterations done in the loop of the re-computation
  integer, save :: MaxIterRecomp

#if defined(_TP_CORREC_) && defined(_HO_MOOD_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
  !> Number of invalid cells that need to be recomputed with the TP flux
  integer, save :: NbCellsInvalidTP
  !> Total number of activation of the re-computation involving the TP flux
  integer, save :: TotReCompTP
  !> Maximum number of invalid cells that need to be recomputed with the TP flux
  integer, save :: MaxInvalidCellsTP
  !> Flag for the re-computation involving the TP flux
  integer, save :: FlagReCompTP
#endif /* _TP_CORREC_ && _HO_MOOD_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */

  !> Flag for the re-computation
  integer :: FlagReComp
  !> Number of re-computation
  integer :: NbIterReComp

CONTAINS

  !--------------------------------------------------------------------!
  !> @brief
  !> Allocate different arrays needed to save the cells to update
  !> and initialise counters for the re-computations
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine InitCellsToUpdate()

      write(StdOut,'(A)')'Initialising array counters for recomputations...'

      allocate(CellsToUpdate(pmesh%NbCells))
      allocate(FlagCells(pmesh%NbCells))
#if defined(_TP_CORREC_) && !defined(_HO_MOOD_)
      allocate(FlagIntInterfaces(pmesh%NbIntInterfaces))
#endif /* _TP_CORREC_ !_HO_MOOD_ */

      TotReComp = 0
      TotIterReComp = 0
      MaxInvalidCells = 0
      MaxIterReComp = 0

#if defined(_TP_CORREC_) && defined(_HO_MOOD_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
      TotReCompTP = 0
      MaxInvalidCellsTP = 0
#endif /* _TP_CORREC_ && _HO_MOOD_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */

      write(StdOut,'(A)')'-------------------------------------------------------'

  end subroutine InitCellsToUpdate

  !--------------------------------------------------------------------!
  !> @brief
  !> Deallocate the arrays allocated by type_recompute::initcellstoupdate()
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine EndCellsToUpdate

      write(StdOut,'(A)') 'Freeing arrays for recomputations...'

      deallocate(CellsToUpdate)
      deallocate(FlagCells)
#if defined(_TP_CORREC_) && !defined(_HO_MOOD_)
      deallocate(FlagIntInterfaces)
#endif /* _TP_CORREC_ !_HO_MOOD_ */
  end subroutine EndCellsToUpdate

  !--------------------------------------------------------------------!
  !> @brief
  !> Print recomputations stats
  !>
  !> integer[in] iter Number of time iterations
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine PrintRecompStats(iter)
    integer, intent(in) :: iter

    write(StdOut,'(A,I9,A,F7.2,A)') 'Total number of recomp   : ', TotReComp,&
        & ', ', real(TotReComp,PR)/real(iter,PR)*100._PR, '%'
#if defined(_TP_CORREC_) && defined(_HO_MOOD_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
    write(StdOut,'(A,I9,A,F7.2,A)') 'Tot. number of TP recomp : ', TotReCompTP,&
        & ', ', real(TotReCompTP,PR)/real(iter,PR)*100._PR, '%'
#endif /* _TP_CORREC_ && _HO_MOOD_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
    write(StdOut,'(A,I9,A,F7.2,A)') 'Total number of iter     : ', TotIterReComp
    write(StdOut,'(A,I9)')          'Maximum nb of iterations : ', MaxIterReComp
    write(StdOut,'(A,I9,A,F7.2,A)') 'Maximum nb invalid cells : ', MaxInvalidCells,&
        & ', ', real(MaxInvalidCells,PR)/real(pmesh%NbCells,PR)*100._PR, '%'
#if defined(_TP_CORREC_) && defined(_HO_MOOD_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
    write(StdOut,'(A,I9,A,F7.2,A)') 'Max. nb invalid TP cells : ', MaxInvalidCellsTP,&
        & ', ', real(MaxInvalidCellsTP,PR)/real(pmesh%NbCells,PR)*100._PR, '%'
#endif /* _TP_CORREC_ && _HO_MOOD_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
    write(StdOut,'(A)')'--------------------------'

  end subroutine PrintRecompStats

#endif /* _TP_CORREC_ || _HO_MOOD_ */
end module type_recompute
