!> Computations of the boundary values
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_bnd_values

  use mod_numerics, only: PR
  use type_IO, only: StdErr
  use type_mesh, only: pmesh, BndInterfaces, dim_
  use type_mesh, only: Neumann_BND, Dirichlet_BND, Wall_BND
#if defined(_MESH_PERIODIC_)
  use type_mesh, only: Periodic_BND
#endif /* _MESH_PERIODIC */
  use mod_physics, only: U_Dirichlet
  use mod_physics_var, only: NbEq
#if defined(_HO_MOOD_)
  use type_mesh_DLP, only: IntInterfaces
  use mod_HO_MOOD_poly, only: compute_polynomial_reconstruction
  use mod_HO_MOOD_poly, only: check_polynomial_reconstruction
#endif /* _HO_MOOD_ */
#if (_HO_MOOD_MAX_DEGREE_ >= 1) && defined(_SRC_ENABLE_)
  use mod_asymptotic, only: compute_beta
#endif /* (_HO_MOOD_MAX_DEGREE_ >= 1) && _SRC_ENABLE_ */
#if defined(_PHY_M1MG_)
  use mod_physics_M1MG_var, only: NbGroup
#endif /* defined(_PHY_M1MG_) */

  IMPLICIT NONE

  private
  public compute_value_bnd

CONTAINS
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the value on the boundary interface.
  !>
  !> @param[in] IdxBndInt Index of the boundary interface
  !> @param[in] U_n Solution at current time
  !> @param[in] x Coordinates of the evaluation point
  !> @param[in] t_n Current time
  !> @return U_J Value on the boundary interface
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_bnd_values_mp_compute_value_bnd_n
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_bnd_values_mp_compute_value_bnd_d
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_bnd_values_mp_compute_value_bnd_w
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_bnd_values_mp_compute_value_bnd_p
  function compute_value_bnd(IdxBndInt,U_n,x,t_n) result(U_J)
    integer, intent(in)                    :: IdxBndInt
    real(PR), dimension(:,:), intent(in)   :: U_n
    real(PR), dimension(:), intent(in)     :: x
    real(PR), intent(in)                   :: t_n
    real(PR), dimension(NbEq) :: U_J

    ! Define the value at the neighbor points
    ! depending on the boundary condition
    select case (BndInterfaces(IdxBndInt)%BndTag(1))
    case (Neumann_BND)
      U_J(:) = compute_value_bnd_N(IdxBndInt,U_n,x,t_n)

    case (Dirichlet_BND)
      U_J(:) = compute_value_bnd_D(IdxBndInt,U_n,x,t_n)

#if defined(_PHY_M1_) || defined(_PHY_EULER_) || defined(_PHY_SW_) || defined(_PHY_HYDRORAD_) || defined(_PHY_M1MG_)
    case (Wall_BND)
      U_J(:) = compute_value_bnd_W(IdxBndInt,U_n,x,t_n)
#endif /* _PHY_ */

#if defined(_MESH_PERIODIC_)
    case (Periodic_BND)
      U_J(:) = compute_value_bnd_P(IdxBndInt,U_n,x,t_n)
#endif /* _MESH_PERIODIC */

    case default
      write(StdErr,'(2A,I4,A,I9)')__FILE__,':', __LINE__, " -> Unknown boundary conditions: ", BndInterfaces(IdxBndInt)%BndTag(1)
      FLUSH StdErr
      STOP
    end select

  end function compute_value_bnd

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the boundary value for Neumann
  !>
  !> @param[in] IdxBndInt Index of the boundary interface
  !> @param[in] U_n Solution at current time
  !> @param[in] x Coordinates of the evaluation point
  !> @param[in] t_n Current time
  !> @return U_J Value on the boundary interface
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_poly_mp_compute_polynomial_reconstruction
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_poly_mp_check_polynomial_reconstruction
  function compute_value_bnd_N(IdxBndInt,U_n,x,t_n) result(U_J)
    integer, intent(in)                    :: IdxBndInt
    real(PR), dimension(:,:), intent(in)   :: U_n
    real(PR), dimension(:), intent(in)     :: x
    real(PR), intent(in)                   :: t_n
    real(PR), dimension(NbEq) :: U_J

    integer :: CellK, i_K
#define _BC_NEUMANN_RECONSTRUCT_NGHB_
#if defined(_BC_NEUMANN_RECONSTRUCT_NGHB_)
    real(PR) :: NbNghb
    integer :: CellL
#endif /* _BC_NEUMANN_RECONSTRUCT_NGHB_ */

#if defined(_HO_MOOD_)
    !integer :: IdxIntInt
    !real(PR), dimension(NbEq) :: U_K_tilde, U_L_tilde
#endif /* _HO_MOOD_ */

    ! Index of the cell owning the boundary interface
    CellK = BndInterfaces(IdxBndInt)%IdxCell
    ! Index of the boundary interface in the cell K
    i_K = BndInterfaces(IdxBndInt)%IdxInterface

#if defined(_BC_NEUMANN_RECONSTRUCT_NGHB_)
    ! Here \f$U_J = \frac{\text{NbNghb} U_K + \sum_{L \in \mathcal{E}_K} U_L}{2 \text{NbNghb}}\f$,
    ! in order to have a small numerical diffusion on the boundary interface
    ! and avoid some numerical instabilities.

    NbNghb = 0._PR
    U_J(:) = 0._PR

    do i_K = 1, pmesh%elt(CellK)%NbInterfaces
      CellL = pmesh%elt(CellK)%nghb(i_K)
      if (CellL /= 0) then
        NbNghb = NbNghb + 1._PR
!#if defined(_HO_MOOD_)
        !IdxIntInt = pmesh%elt(CellK)%IntInterfaces(i_K)
        !call compute_polynomial_reconstruction(U_n,CellL,IntInterfaces(IdxIntInt)%FacePD,pmesh%elt(CellL)%CoeffPolyMOOD,x,t_n,U_L_tilde)
!#if !defined(_HO_MOOD_CHECK_POLY_)
        !call check_polynomial_reconstruction(U_n,CellL,U_L_tilde)
!#endif /* _HO_MOOD_CHECK_POLY_ */
        !U_J(:) = U_J(:) + U_L_tilde(:)

!#else
        U_J(:) = U_J(:) + U_n(:,CellL)
!#endif /* _HO_MOOD_ */
      endif
    end do

!#if defined(_HO_MOOD_)
    !call compute_polynomial_reconstruction(U_n,CellK,BndInterfaces(IdxBndInt)%FacePD,pmesh%elt(CellK)%CoeffPolyMOOD,x,t_n,U_K_tilde)
!#if !defined(_HO_MOOD_CHECK_POLY_)
    !call check_polynomial_reconstruction(U_n,CellK,U_K_tilde)
!#endif /* _HO_MOOD_CHECK_POLY_ */
    !U_J(:) = 0.5_PR*(U_J(:)/NbNghb + U_K_tilde(:))

!#else
    U_J(:) = 0.5_PR*(U_J(:)/NbNghb + U_n(:,CellK))
!#endif /* _HO_MOOD_ */


#else /* _BC_NEUMANN_RECONSTRUCT_NGHB_ */
!#if defined(_HO_MOOD_)
    !call compute_polynomial_reconstruction(U_n,CellK,BndInterfaces(IdxBndInt)%FacePD,pmesh%elt(CellK)%CoeffPolyMOOD,x,t_n,U_J)
!#if !defined(_HO_MOOD_CHECK_POLY_)
    !call check_polynomial_reconstruction(U_n,CellK,U_J)
!#endif /* _HO_MOOD_CHECK_POLY_ */

!#else
    U_J(:) = U_n(:,CellK)
!#endif /* _HO_MOOD_ */
#endif /* _BC_NEUMANN_RECONSTRUCT_NGHB_ */

  end function compute_value_bnd_N

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the boundary value for Dirichlet
  !>
  !> @param[in] IdxBndInt Index of the boundary interface
  !> @param[in] U_n Solution at current time
  !> @param[in] x Coordinates of the evaluation point
  !> @param[in] t_n Current time
  !> @return U_J Value on the boundary interface
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_advection_mp_u_dirichlet
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_burgers_mp_u_dirichlet
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_u_dirichlet
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_u_dirichlet
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_u_dirichlet
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_u_dirichlet
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_u_dirichlet
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_u_dirichlet
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_u_dirichlet
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_u_dirichlet
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_mp_compute_beta
  function compute_value_bnd_D(IdxBndInt,U_n,x,t_n) result(U_J)
    integer, intent(in)                    :: IdxBndInt
    real(PR), dimension(:,:), intent(in)   :: U_n
    real(PR), dimension(:), intent(in)     :: x
    real(PR), intent(in)                   :: t_n
    real(PR), dimension(NbEq) :: U_J

    integer :: CellK, i_K
#if (_HO_MOOD_MAX_DEGREE_ >= 1) && defined(_SRC_ENABLE_)
    real(PR) :: beta_K
#endif /* (_HO_MOOD_MAX_DEGREE_ >= 1) && _SRC_ENABLE_ */

    ! Index of the cell owning the boundary interface
    CellK = BndInterfaces(IdxBndInt)%IdxCell
    ! Index of the boundary interface in the cell K
    i_K = BndInterfaces(IdxBndInt)%IdxInterface

#if (_HO_MOOD_MAX_DEGREE_ >= 1) && defined(_SRC_ENABLE_)
    beta_K = compute_beta(CellK,U_n(:,CellK),t_n)
    ! beta*U_HO + (1-beta)*U
    U_J(:) = beta_K*U_Dirichlet(x,t_n) + (1._PR - beta_K)*U_Dirichlet(BndInterfaces(IdxBndInt)%x,t_n)
#else
    U_J(:) = U_Dirichlet(x,t_n)
#endif /* (_HO_MOOD_MAX_DEGREE_ >= 1) && _SRC_ENABLE_ */

  end function compute_value_bnd_D

#if defined(_PHY_M1_) || defined(_PHY_EULER_) || defined(_PHY_SW_) || defined(_PHY_HYDRORAD_) || defined(_PHY_M1MG_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the boundary value for a wall
  !>
  !> @param[in] IdxBndInt Index of the boundary interface
  !> @param[in] U_n Solution at current time
  !> @param[in] x Coordinates of the evaluation point
  !> @param[in] t_n Current time
  !> @return U_J Value on the boundary interface
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_poly_mp_compute_polynomial_reconstruction
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_poly_mp_check_polynomial_reconstruction
  function compute_value_bnd_W(IdxBndInt,U_n,x,t_n) result(U_J)
    integer, intent(in)                    :: IdxBndInt
    real(PR), dimension(:,:), intent(in)   :: U_n
    real(PR), dimension(:), intent(in)     :: x
    real(PR), intent(in)                   :: t_n
    real(PR), dimension(NbEq) :: U_J

    integer :: CellK, i_K
#define _BC_WALL_MATRIX_
#if defined(_BC_WALL_MATRIX_)
    real(PR), dimension(dim_) :: normal
    real(PR), dimension(dim_,dim_) :: RotationMatrix
#endif /* defined(_BC_WALL_MATRIX_) */
#if defined(_PHY_M1MG_)
    integer :: group, eq0
#endif /* defined(_PHY_M1MG_) */

    ! Index of the cell owning the boundary interface
    CellK = BndInterfaces(IdxBndInt)%IdxCell
    ! Index of the boundary interface in the cell K
    i_K = BndInterfaces(IdxBndInt)%IdxInterface

!#if defined(_HO_MOOD_)
    !call compute_polynomial_reconstruction(U_n,CellK,BndInterfaces(IdxBndInt)%FacePD,pmesh%elt(CellK)%CoeffPolyMOOD,x,t_n,U_J)
!#if !defined(_HO_MOOD_CHECK_POLY_)
    !call check_polynomial_reconstruction(U_n,CellK,U_J)
!#endif /* _HO_MOOD_CHECK_POLY_ */
!#else /* _HO_MOOD_ */
    U_J(:) = U_n(:,CellK)
!#endif /* _HO_MOOD_ */

#if defined(_BC_WALL_MATRIX_)
    normal(:)  = pmesh%elt(CellK)%normal(:,i_K)

    ! Compute the matrix of the rotation to get the local basis
    RotationMatrix(1,1) = -normal(1)**2 + normal(2)**2
    RotationMatrix(2,1) = -2._PR*normal(1)*normal(2)
    RotationMatrix(1,2) = -2._PR*normal(1)*normal(2)
    RotationMatrix(2,2) = -(-normal(1)**2 + normal(2)**2)

    U_J(2:3) = MATMUL(RotationMatrix,U_J(2:3))
#if defined(_PHY_HYDRORAD_)
    U_J(6:7) = MATMUL(RotationMatrix,U_J(6:7))
#elif defined(_PHY_M1MG_)
    do group = 2, NbGroup
      eq0 = (1 + dim_)*(group-1)
      U_J(eq0+2:eq0+3) = MATMUL(RotationMatrix,U_J(eq0+2:eq0+3))
    end do
#endif /* defined(_PHY_) */

#else
    ! See \cite ClaFig_2014
    U_J(2:3) = U_J(2:3) - 2._PR*dot_product(U_J(2:3),pmesh%elt(CellK)%normal(:,i_K))*pmesh%elt(CellK)%normal(:,i_K)
#if defined(_PHY_HYDRORAD_)
    U_J(6:7) = U_J(6:7) - 2._PR*dot_product(U_J(6:7),pmesh%elt(CellK)%normal(:,i_K))*pmesh%elt(CellK)%normal(:,i_K)
#elif defined(_PHY_M1MG_)
    do group = 2, NbGroup
      eq0 = (1 + dim_)*(group-1)
      U_J(eq0+2:eq0+3) = U_J(eq0+2:eq0+3) - 2._PR*dot_product(U_J(eq0+2:eq0+3),pmesh%elt(CellK)%normal(:,i_K))*pmesh%elt(CellK)%normal(:,i_K)
    end do
#endif /* defined(_PHY_) */
#endif /* _BC_WALL_MATRIX_ */

  end function compute_value_bnd_W
#endif /* _PHY_ */

#if defined(_MESH_PERIODIC_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the boundary value for a periodic boundary
  !>
  !> @param[in] IdxBndInt Index of the boundary interface
  !> @param[in] U_n Solution at current time
  !> @param[in] x Coordinates of the evaluation point
  !> @param[in] t_n Current time
  !> @return U_J Value on the boundary interface
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_poly_mp_compute_polynomial_reconstruction
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_poly_mp_check_polynomial_reconstruction
  function compute_value_bnd_P(IdxBndInt,U_n,x,t_n) result(U_J)
    integer, intent(in)                    :: IdxBndInt
    real(PR), dimension(:,:), intent(in)   :: U_n
    real(PR), dimension(:), intent(in)     :: x
    real(PR), intent(in)                   :: t_n
    real(PR), dimension(NbEq) :: U_J

    integer :: CellK, i_K
#if defined(_HO_MOOD_)
    real(PR), dimension(dim_) :: x_periodic
#endif /* defined(_HO_MOOD_) */

    integer :: CellKPeriodic
#if (_HO_MOOD_MAX_DEGREE_ >= 1) && defined(_SRC_ENABLE_)
    real(PR) :: beta_K
#endif /* (_HO_MOOD_MAX_DEGREE_ >= 1) && _SRC_ENABLE_ */

    ! Index of the cell owning the boundary interface
    CellK = BndInterfaces(IdxBndInt)%IdxCell
    ! Index of the boundary interface in the cell K
    i_K = BndInterfaces(IdxBndInt)%IdxInterface

    CellKPeriodic = BndInterfaces(IdxBndInt)%IdxCellPeriodic
#if defined(_HO_MOOD_)
    if (BndInterfaces(IdxBndInt)%BndTag(2) == 5) then
      ! South
      x_periodic(1) = x(1)
      x_periodic(2) = x(2) + pmesh%length(2)
    else if (BndInterfaces(IdxBndInt)%BndTag(2) == 6) then
      ! East
      x_periodic(1) = x(1) - pmesh%length(1)
      x_periodic(2) = x(2)
    else if (BndInterfaces(IdxBndInt)%BndTag(2) == 7) then
      ! North
      x_periodic(1) = x(1)
      x_periodic(2) = x(2) - pmesh%length(2)
    else if (BndInterfaces(IdxBndInt)%BndTag(2) == 8) then
      ! West
      x_periodic(1) = x(1) + pmesh%length(1)
      x_periodic(2) = x(2)
    else
      write(StdErr,'(2A,I4,A,I9)')__FILE__,':', __LINE__, " -> side unkown: ", BndInterfaces(IdxBndInt)%BndTag(2)
      FLUSH StdErr
      STOP
    end if
    call compute_polynomial_reconstruction(U_n,CellKPeriodic,BndInterfaces(IdxBndInt)%FacePD,&
        pmesh%elt(CellKPeriodic)%CoeffPolyMOOD,x_periodic,t_n,U_J)
#if (_HO_MOOD_MAX_DEGREE_ >= 1) && defined(_SRC_ENABLE_)
    beta_K = compute_beta(CellK,U_n(:,CellK),t_n)
    ! beta*U_HO + (1-beta)*U
    U_J(:) = beta_K*U_J(:) + (1._PR - beta_K)*U_n(:,CellKPeriodic)
#endif /* (_HO_MOOD_MAX_DEGREE_ >= 1) && _SRC_ENABLE_ */
#else
    U_J(:) = U_n(:,CellKPeriodic)
#endif /* _HO_MOOD_ */

  end function compute_value_bnd_P
#endif /* _MESH_PERIODIC */

end module mod_bnd_values
