!> Computations of the numerical fluxes in one direction.
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_flux_1D

  use mod_numerics, only: PR, eps
  use mod_numerics, only: NegPart, PosPart
  use type_IO, only: StdErr
  use type_mesh, only: dim_
  use mod_physics, only: physic_flux, compute_wave_speed, compute_greatest_wave_speed
#if defined(_PHY_HYDRORAD_)
  use mod_physics_HydroRad, only: compute_wave_speed_hydro, compute_greatest_wave_speed_hydro
  use mod_physics_HydroRad, only: compute_wave_speed_rad, compute_greatest_wave_speed_rad
#endif /* defined(_PHY_HYDRORAD_) */
#if defined(_FLUX_HLLC_)
#if defined(_PHY_EULER_)
  use mod_physics_Euler, only: compute_HLLC_values
#elif defined(_PHY_M1_)
  use mod_physics_M1, only: compute_HLLC_values
#endif /* _PHY_ */
#endif /* _FLUX_HLLC_*/
#if defined(_FLUX_UPWIND_) && defined(_PHY_ADV_)
  use mod_physics_Advection_var, only: speed_adv
#endif /* _FLUX_UPWIND_*/

  use mod_physics_var, only: NbEq

  IMPLICIT NONE

  private
  public flux_1D

CONTAINS
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the 1D numerical flux in the direction of \f$ \vec{\eta}_{KJ}\f$
  !> (Rusanov, HLL , HLLC  ...).
  !>
  !> * Rusanov \cite Rus_1961 :
  !>  \f[ \boldsymbol{\mathcal{F}}_{KJ} \cdot \boldsymbol{\eta}_{KJ} =
  !>    \frac{{\bf F}({\bf U}_J) + {\bf F}({\bf U}_K)}{2} \cdot  \boldsymbol{\eta}_{KJ}
  !>    - \frac{b_{KJ}}{2}({\bf U}_J - {\bf U}_K) \f]
  !> * HLL \cite HarLaxLee_1983 :
  !>   \f[ \boldsymbol{\mathcal{F}}_{KJ} \cdot \boldsymbol{\eta}_{KJ} =
  !>     \frac{1}{b^-_{KJ} - b^+_{KJ}}\left(
  !>       b^+_{KJ}{\bf F}({\bf U}_K) \cdot \boldsymbol{\eta}_{KJ}
  !>       - b^-_{KJ} {\bf F}({\bf U}_J) \cdot \boldsymbol{\eta}_{KJ}
  !>       + b^-_{KJ} b^+_{KJ} ({\bf U}_J - {\bf U}_K) \right) \f]
  !>
  !> * HLC \cite TorSprSpe_1994 \cite BerChaDub_2007 \cite BerDubDubNguTur_2010 :
  !>
  !>
  !> @param[in] U_K, U_J States at each sides of the interface
  !> @param[in] eta Direction to compute the flux
  !> @param[in] tau Tangent direction
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_advection_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_burgers_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_compute_greatest_wave_speed

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_advection_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_burgers_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_compute_wave_speed

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_hllc_values
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_hllc_values

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_greatest_wave_speed_hydro
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_greatest_wave_speed_rad
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_wave_speed_hydro
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_wave_speed_rad

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_advection_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_burgers_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_physic_flux
  function flux_1D(U_K,U_J,eta,tau)
      real(PR), dimension(:), intent(in) :: U_K, U_J
      real(PR), dimension(:), intent(in) :: eta, tau
      real(PR), dimension(NbEq) :: flux_1D

      real(PR), dimension(NbEq,dim_) :: F_K, F_J
#if defined(_FLUX_RUS_)
      real(PR) :: b_KJ
#elif defined(_FLUX_HLL_)
      real(PR) :: wave_speed_K, wave_speed_J
      !real(PR) :: wave_speed_K_m, wave_speed_K_p
      !real(PR) :: wave_speed_J_m, wave_speed_J_p
#elif defined(_FLUX_HLLC_) && defined(_PHY_EULER_)
      real(PR), dimension(NbEq) :: U_K_star, U_J_star
      real(PR) :: wave_speed_K, wave_speed_J, wave_speed_star
#endif /* _FLUX_ */

#if defined(_FLUX_UPWIND_) && defined(_PHY_ADV_)

      flux_1D(:) = max(0._PR,dot_product(speed_adv,eta))*U_K(:) + &
          min(0._PR,dot_product(speed_adv,eta))*U_J(:)

#elif defined(_FLUX_RUS_)
      F_K(:,:) = physic_flux(U_K)
      F_J(:,:) = physic_flux(U_J)

#if defined(_PHY_HYDRORAD_)
      b_KJ = compute_greatest_wave_speed_hydro(U_K, U_J, eta, tau)
      flux_1D(1:4) = 0.5_PR*(MATMUL(F_K(1:4,:) + F_J(1:4,:),eta) - b_KJ*(U_J(1:4) - U_K(1:4)))
      b_KJ = compute_greatest_wave_speed_rad(U_K, U_J, eta, tau)
      flux_1D(5:7) = 0.5_PR*(MATMUL(F_K(5:7,:) + F_J(5:7,:),eta) - b_KJ*(U_J(5:7) - U_K(5:7)))
#else
      b_KJ = compute_greatest_wave_speed(U_K, U_J, eta, tau)
      flux_1D(:) = 0.5_PR*(MATMUL(F_K + F_J,eta) - b_KJ*(U_J(:) - U_K(:)))
#endif /* defined(_PHY_HYDRORAD_) */

#elif defined(_FLUX_HLL_)
      F_K(:,:) = physic_flux(U_K)
      F_J(:,:) = physic_flux(U_J)

#if defined(_PHY_HYDRORAD_)
      call compute_wave_speed_hydro(U_K, U_J, eta, tau, wave_speed_K, wave_speed_J)
      if (0._PR <= wave_speed_K) then
        flux_1D(1:4) = MATMUL(F_K(1:4,:),eta)
      else if (wave_speed_K <= 0._PR .and. 0._PR <= wave_speed_J) then
        flux_1D(1:4) = (wave_speed_J*MATMUL(F_K(1:4,:),eta) - wave_speed_K*MATMUL(F_J(1:4,:),eta) &
            + (wave_speed_K*wave_speed_J)*(U_J(1:4) - U_K(1:4)))/(wave_speed_J - wave_speed_K)
      else if (wave_speed_J <= 0._PR) then
        flux_1D(1:4) = MATMUL(F_J(1:4,:),eta)
      end if

      call compute_wave_speed_rad(U_K, U_J, eta, tau, wave_speed_K, wave_speed_J)
      if (0._PR <= wave_speed_K) then
        flux_1D(5:7) = MATMUL(F_K(5:7,:),eta)
      else if (wave_speed_K <= 0._PR .and. 0._PR <= wave_speed_J) then
        flux_1D(5:7) = (wave_speed_J*MATMUL(F_K(5:7,:),eta) - wave_speed_K*MATMUL(F_J(5:7,:),eta) &
            + (wave_speed_K*wave_speed_J)*(U_J(5:7) - U_K(5:7)))/(wave_speed_J - wave_speed_K)
      else if (wave_speed_J <= 0._PR) then
        flux_1D(5:7) = MATMUL(F_J(5:7,:),eta)
      end if

#else
      call compute_wave_speed(U_K, U_J, eta, tau, wave_speed_K, wave_speed_J)
      if (0._PR <= wave_speed_K) then
        flux_1D(:) = MATMUL(F_K,eta)
      else if (wave_speed_K <= 0._PR .and. 0._PR <= wave_speed_J) then
        flux_1D(:) = (wave_speed_J*MATMUL(F_K,eta) - wave_speed_K*MATMUL(F_J,eta) &
            + (wave_speed_K*wave_speed_J)*(U_J(:) - U_K(:)))/(wave_speed_J - wave_speed_K)
      else if (wave_speed_J <= 0._PR) then
        flux_1D(:) = MATMUL(F_J,eta)
      end if

      !wave_speed_K_m = NegPart(wave_speed_K)
      !wave_speed_K_p = PosPart(wave_speed_K)
      !wave_speed_J_m = NegPart(wave_speed_J)
      !wave_speed_J_p = PosPart(wave_speed_J)

      !flux_1D(:) = ((wave_speed_J_m - wave_speed_K_m)*MATMUL(F_J,eta) &
          !+  (wave_speed_J_p - wave_speed_K_p)*MATMUL(F_K,eta) &
          !- 0.5_PR*(wave_speed_J*abs(wave_speed_K) - wave_speed_K*abs(wave_speed_J))*(U_J(:) - U_K(:)))&
          !/(wave_speed_J - wave_speed_K)
#endif /* defined(_PHY_HYDRORAD_) */

#elif defined(_FLUX_HLLC_) && defined(_PHY_EULER_)
      call compute_HLLC_values(U_K, U_J, eta, tau, wave_speed_K, wave_speed_star, wave_speed_J, U_K_star, U_J_star)

      F_K(:,:) = physic_flux(U_K)
      F_J(:,:) = physic_flux(U_J)

      if (0._PR <= wave_speed_K) then
        flux_1D(:) = MATMUL(F_K,eta)
      else if (wave_speed_K <= 0._PR .and. 0._PR <= wave_speed_star) then
        flux_1D(:) = MATMUL(F_K,eta) + wave_speed_K*(U_K_star(:) - U_K(:))
      else if (wave_speed_star <= 0._PR .and. 0._PR <= wave_speed_J) then
        flux_1D(:) = MATMUL(F_J,eta) + wave_speed_J*(U_J_star(:) - U_J(:))
        !flux_1D(:) = MATMUL(F_K,eta) + wave_speed_K*(U_K_star(:) - U_K(:)) + wave_speed_star*(U_J_star(:) - U_K_star(:))
      else if (0._PR >= wave_speed_J) then
        flux_1D(:) = MATMUL(F_J,eta)
      end if

#elif defined(_FLUX_HLLC_) && defined(_PHY_M1_)
      call compute_HLLC_values(U_K, U_J, eta, tau, flux_1D)

#else
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> flux not defined."
      FLUSH StdErr
      STOP
#endif /* _FLUX_ */

#if (defined(_PHY_M1_) || defined(_PHY_M1MG_)) && !defined(_FLUX_HLLC_)
      ! No numerical flux on the last equation of the M1 model (temperature),
      ! for HLLC it is already done in mod_physics_M1::compute_HLLC_values.
      flux_1D(NbEq) = 0._PR
#endif /* (defined(_PHY_M1_) || defined(_PHY_M1MG_)) && .not. defined(_FLUX_HLLC_) */

  end function flux_1D

end module mod_flux_1D
