!> Computations of the numerical fluxes in one direction (HO version).
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_flux_1D_HO_MOOD
#if defined(_HO_MOOD_)

  use mod_numerics, only: PR
  use type_IO, only: StdErr
  use type_mesh, only: pmesh, BndInterfaces, dim_
  use type_mesh_DLP, only: IntInterfaces
  use mod_physics_var, only: NbEq
  use mod_flux_1D, only: flux_1D

  use mod_bnd_values, only: compute_value_bnd
  use mod_tools, only: plus1
  use type_mesh, only: nodes
  use type_HO_MOOD, only: Nd_max
  use mod_HO_MOOD_quad, only: quad_seg
  use mod_HO_MOOD_poly, only: compute_polynomial_reconstruction

  IMPLICIT NONE

  private
  public flux_1D_HO_MOOD
  public flux_1D_bnd_HO_MOOD

CONTAINS
  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the flux using the polynomial reconstruction (see Appendix A of \cite Dio_2012
  !> for the quadrature rules)
  !>
  !> @param[in] U_n Mean values at current time
  !> @param[in] t_n Current time
  !> @param[in] CellK, CellJ Index of the cell at each side of the interface
  !> @param[in] IdxIntInt Index of the interior interface
  !> @param[in] i_K Index of the interface in CellK
  !> @param[in] eta Direction to compute the flux
  !> @param[in] tau Tangent direction
  !> @return \f$\sum\limits_{r=1}^R \xi_{K,i,r} \ \boldsymbol{\mathcal{F}}\left({\bf U}_{K,r}^n, {\bf U}_{J,r}^n\right)\cdot\boldsymbol{\eta}_{KJ}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_plus1
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_poly_mp_compute_polynomial_reconstruction
  function flux_1D_HO_MOOD(U_n,t_n,CellK,CellJ,IdxIntInt,i_K,eta,tau)
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n
      integer, intent(in) :: CellK, CellJ, IdxIntInt, i_K
      real(PR), dimension(:), intent(in) :: eta, tau
      real(PR), dimension(NbEq) :: flux_1D_HO_MOOD

      real(PR), dimension(NbEq) :: U_K_tilde, U_J_tilde
      real(PR), dimension(NbEq,Nd_max) :: R_K, R_J
      real(PR), dimension(dim_) :: xQ, xA, xB
      integer ::  i_K_p1, degree, IdxQuad

      degree = IntInterfaces(IdxIntInt)%FacePD

      test_degree: if (degree >= 1 .and. degree <= 5) then
        R_K(:,:) = pmesh%elt(CellK)%CoeffPolyMOOD(:,:)
        R_J(:,:) = pmesh%elt(CellJ)%CoeffPolyMOOD(:,:)

        i_K_p1 = plus1(i_K,pmesh%elt(CellK)%NbInterfaces)

        xA(:) = nodes(pmesh%elt(CellK)%vertices_glob(i_K))%x(:)
        xB(:) = nodes(pmesh%elt(CellK)%vertices_glob(i_K_p1))%x(:)

        flux_1D_HO_MOOD(:) = 0._PR

        loop_quad: do IdxQuad = 1, quad_seg(degree)%nb_points
          xQ(:) = quad_seg(degree)%bar_coord(1,IdxQuad)*xA(:) &
              + quad_seg(degree)%bar_coord(2,IdxQuad)*xB(:)

          call compute_polynomial_reconstruction(U_n,CellK,degree,R_K,xQ,t_n,U_K_tilde)
          call compute_polynomial_reconstruction(U_n,CellJ,degree,R_J,xQ,t_n,U_J_tilde)

          flux_1D_HO_MOOD(:) = flux_1D_HO_MOOD(:) &
              + quad_seg(degree)%weights(IdxQuad)*flux_1D(U_K_tilde,U_J_tilde,eta,tau)
        end do loop_quad

#if defined(_TP_CORREC_)
      else if (degree == 0 .or. degree == -1) then
#else
      else if (degree == 0) then
#endif /* _TP_CORREC_ */
        flux_1D_HO_MOOD(:) = flux_1D(U_n(:,CellK),U_n(:,CellJ),eta,tau)

      else
        write(StdErr,'(2A,I4,A,I1)')__FILE__,':', __LINE__, " -> Wrong degree: ", degree
        FLUSH StdErr
        STOP
      end if test_degree

  end function flux_1D_HO_MOOD

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the flux using the polynomial reconstruction (see Appendix A of \cite Dio_2012
  !> for the quadrature rules) (routine for flux which involves boundary values)
  !>
  !> @param[in] U_n Mean values at current time
  !> @param[in] t_n Current time
  !> @param[in] CellK Current cell
  !> @param[in] IdxBndInt Index of the boundary interface
  !> @param[in] i_K Index of the interface in CellK
  !> @param[in] eta Direction to compute the flux
  !> @param[in] tau Tangent direction
  !> @return \f$\sum\limits_{r=1}^R \xi_{K,i,r} \ \boldsymbol{\mathcal{F}}\left({\bf U}_{K,r}^n, {\bf U}_{J,r}^n\right)\cdot\boldsymbol{\eta}_{KJ}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_plus1
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_poly_mp_compute_polynomial_reconstruction
  function flux_1D_bnd_HO_MOOD(U_n,t_n,CellK,IdxBndInt,i_K,eta,tau)
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n
      integer, intent(in) :: CellK, IdxBndInt, i_K
      real(PR), dimension(:), intent(in) :: eta, tau
      real(PR), dimension(NbEq) :: flux_1D_bnd_HO_MOOD

      real(PR), dimension(NbEq) :: U_K_tilde, U_J_tilde
      real(PR), dimension(NbEq,Nd_max) :: R_K
      real(PR), dimension(dim_) :: xQ, xA, xB
      integer :: IdxQuad, i_K_p1, degree

      degree = BndInterfaces(IdxBndInt)%FacePD

      test_degree: if (degree >= 1 .and. degree <= 5) then
        R_K(:,:) = pmesh%elt(CellK)%CoeffPolyMOOD(:,:)

        i_K_p1 = plus1(i_K,pmesh%elt(CellK)%NbInterfaces)

        xA(:) = nodes(pmesh%elt(CellK)%vertices_glob(i_K))%x(:)
        xB(:) = nodes(pmesh%elt(CellK)%vertices_glob(i_K_p1))%x(:)

        flux_1D_bnd_HO_MOOD(:) = 0._PR

        loop_quad: do IdxQuad = 1, quad_seg(degree)%nb_points
          xQ(:) = quad_seg(degree)%bar_coord(1,IdxQuad)*xA(:) &
              + quad_seg(degree)%bar_coord(2,IdxQuad)*xB(:)

          call compute_polynomial_reconstruction(U_n,CellK,degree,R_K,xQ,t_n,U_K_tilde)
          U_J_tilde(:) = compute_value_bnd(IdxBndInt,U_n,xQ,t_n)

          flux_1D_bnd_HO_MOOD(:) = flux_1D_bnd_HO_MOOD(:) + quad_seg(degree)%weights(IdxQuad)*flux_1D(U_K_tilde,U_J_tilde,eta,tau)
        end do loop_quad

#if defined(_TP_CORREC_)
      else if (degree == 0 .or. degree == -1) then
#else
      else if (degree == 0) then
#endif /* _TP_CORREC_ */
        U_J_tilde(:) = compute_value_bnd(IdxBndInt,U_n,BndInterfaces(IdxBndInt)%x,t_n)
        flux_1D_bnd_HO_MOOD(:) = flux_1D(U_n(:,CellK),U_J_tilde,eta,tau)

      else
        write(StdErr,'(2A,I4,A,I1)')__FILE__,':', __LINE__, " -> Wrong degree: ", degree
        FLUSH StdErr
        STOP
      end if test_degree
  end function flux_1D_bnd_HO_MOOD

#endif /* _HO_MOOD_ */
end module mod_flux_1D_HO_MOOD
