!> Computations of the numerical fluxes in one direction (with source term).
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_flux_1D_source
#if defined(_SRC_ENABLE_)

  use mod_numerics, only: PR
  use type_IO, only: StdErr
  use type_mesh, only: dim_
  use mod_physics, only: physic_flux, compute_wave_speed, compute_greatest_wave_speed
  use mod_physics_var, only: NbEq
  use mod_flux_1D, only: flux_1D

  use mod_asymptotic, only: compute_alpha_KJ, compute_alpha_KK, R_bar
  use mod_asymptotic, only: compute_alpha_KJ_NoAP, compute_alpha_KK_NoAP, R
#if defined(_FLUX_SRC_CENTER_)
  use type_mesh, only: pmesh
  use mod_asymptotic, only: gamma_
#endif /* defined(_FLUX_SRC_CENTER_) */

  IMPLICIT NONE

  private
  public flux_1D_source
  public flux_1D_source_NoAP

CONTAINS
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the 1D numerical flux in the direction of \f$\vec{\eta}_{KJ}\f$,
  !> with the source term by using the method described in \cite BerTur_2011 .
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] nu_J Coefficient of the HLL-DLP scheme
  !> @param[in] nu_bar_J Coefficient of the DLP scheme
  !> @param[in] delta_K Coefficient
  !> @param[in] U_K, U_J States at each sides of the interface
  !> @param[in] eta Direction to compute the flux
  !> @param[in] tau Vector orthogonal to eta
  !> @param[in] U_K_mean, U_J_mean Mean values on cells \f$K\f$ and \f$J\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_advection_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_burgers_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_compute_greatest_wave_speed

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_advection_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_burgers_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_compute_wave_speed

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_advection_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_burgers_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_physic_flux

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_mp_flux_1d

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_mp_compute_alpha_kj
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_mp_compute_alpha_kk
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_mp_r_bar
  function flux_1D_source(CellK,nu_J,nu_bar_J,delta_K,U_K,U_J,eta,tau,U_K_mean,U_J_mean)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(in) :: nu_J
      real(PR), intent(in)               :: nu_bar_J
      real(PR), dimension(:), intent(in) :: delta_K
      real(PR), dimension(:), intent(in) :: U_K, U_J
      real(PR), dimension(:), intent(in) :: eta, tau
      real(PR), dimension(NbEq) :: flux_1D_source
      real(PR), dimension(:), optional, intent(in) :: U_K_mean, U_J_mean

      real(PR), dimension(NbEq,dim_) :: F_K
      real(PR), dimension(NbEq) :: alpha_KJ, R_bar_KJ
      real(PR) :: alpha_KK
#if defined(_FLUX_RUS_)
      real(PR) :: b_KJ
#elif defined(_FLUX_HLL_)
      real(PR) :: wave_speed_K, wave_speed_J
#endif /* _FLUX */

#if defined(_FLUX_SRC_CENTER_)
      flux_1D_source(:) = flux_1D(U_K,U_J,eta,tau) - pmesh%elt(CellK)%area*gamma_(CellK,U_K)*(R(CellK,U_K) - U_K(:))

#else /* defined(_FLUX_SRC_CENTER_) */
      F_K(:,:) = physic_flux(U_K)

#if defined(_HO_MOOD_)
      alpha_KJ(:) = compute_alpha_KJ(CellK,nu_J,nu_bar_J,delta_K,U_K,U_J,eta,tau,U_K_mean,U_J_mean)
      R_bar_KJ(:) = R_bar(CellK,nu_J,nu_bar_J,delta_K,U_K,U_J,eta,tau,U_K_mean,U_J_mean)
      alpha_KK = compute_alpha_KK(CellK,U_K,eta,tau,U_K_mean)
#else
      alpha_KJ(:) = compute_alpha_KJ(CellK,nu_J,nu_bar_J,delta_K,U_K,U_J,eta,tau)
      R_bar_KJ(:) = R_bar(CellK,nu_J,nu_bar_J,delta_K,U_K,U_J,eta,tau)
      alpha_KK = compute_alpha_KK(CellK,U_K,eta,tau)
#endif /* _HO_MOOD_ */


#if defined(_FLUX_RUS_)
      b_KJ = compute_greatest_wave_speed(U_K, U_J, eta, tau)

      flux_1D_source(:) = alpha_KJ(:)*flux_1D(U_K,U_J,eta,tau) &
          - (alpha_KJ(:) - alpha_KK)*MATMUL(F_K,eta) &
          - (1._PR - alpha_KJ(:))*b_KJ*(R_bar_KJ(:) - U_K(:))

#elif defined(_FLUX_HLL_)
      call compute_wave_speed(U_K, U_J, eta, tau, wave_speed_K, wave_speed_J)

      if ((wave_speed_K < 0._PR .and. wave_speed_J < 0._PR) .or. (wave_speed_K > 0._PR .and. wave_speed_J > 0._PR)) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> routine not defined"
        FLUSH StdErr
        STOP
      end if

      flux_1D_source(:) = alpha_KJ(:)*flux_1D(U_K,U_J,eta,tau) &
          - (alpha_KJ(:) - alpha_KK)*MATMUL(F_K,eta) &
          - (1._PR - alpha_KJ(:))*0.5_PR*(wave_speed_J - wave_speed_K)*(R_bar_KJ(:) - U_K(:))

#elif defined(_FLUX_HLLC_)
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> HLLC w/ source term not yet implemented"
      FLUSH StdErr
      STOP
#endif /* _FLUX_ */

#endif /* defined(_FLUX_SRC_CENTER_) */

  end function flux_1D_source

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the 1D numerical flux in the direction of \f$\vec{\eta}_{KJ}\f$,
  !> with the source term by using the method described in \cite BerTur_2011
  !> without the AP correction.
  !>
  !> @param[in] CellK Index of the cell
  !> @param[in] delta_K Coefficient
  !> @param[in] U_K, U_J States at each sides of the interface
  !> @param[in] eta Direction to compute the flux
  !> @param[in] tau Vector orthogonal to eta
  !> @param[in] alpha_KK Coefficient \f$\alpha_{KK}\f$ with or without AP corection
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_advection_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_burgers_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_compute_greatest_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_compute_greatest_wave_speed

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_advection_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_burgers_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_compute_wave_speed

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_advection_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_burgers_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hydrorad_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_hypheat_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_p1_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_sw_mp_physic_flux
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_telegraph_mp_physic_flux

  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_1d_mp_flux_1d
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_mp_compute_alpha_kj_noap
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_asymptotic_mp_r
  function flux_1D_source_NoAP(CellK,delta_K,U_K,U_J,eta,tau,alpha_KK)
      integer, intent(in) :: CellK
      real(PR), dimension(:), intent(in) :: delta_K
      real(PR), dimension(:), intent(in) :: U_K, U_J
      real(PR), dimension(:), intent(in) :: eta, tau
      real(PR), intent(in) :: alpha_KK
      real(PR), dimension(NbEq) :: flux_1D_source_NoAP

      real(PR), dimension(NbEq,dim_) :: F_K
      real(PR), dimension(NbEq) :: alpha_KJ_NoAP
#if defined(_FLUX_RUS_)
      real(PR) :: b_KJ
#elif defined(_FLUX_HLL_)
      real(PR) :: wave_speed_K, wave_speed_J
#endif /* _FLUX_ */

#if defined(_FLUX_SRC_CENTER_)
      flux_1D_source_NoAP(:) = flux_1D(U_K,U_J,eta,tau) - pmesh%elt(CellK)%area*gamma_(CellK,U_K)*(R(CellK,U_K) - U_K(:))

#else /* defined(_FLUX_SRC_CENTER_) */
      F_K(:,:) = physic_flux(U_K)

      alpha_KJ_NoAP(:) = compute_alpha_KJ_NoAP(CellK,delta_K,U_K,U_J,eta,tau)

#if defined(_FLUX_RUS_)
      b_KJ = compute_greatest_wave_speed(U_K, U_J, eta, tau)

      flux_1D_source_NoAP(:) = alpha_KJ_NoAP(:)*flux_1D(U_K,U_J,eta,tau) &
          - (alpha_KJ_NoAP(:) - alpha_KK)*MATMUL(F_K,eta) &
          - (1._PR - alpha_KJ_NoAP(:))*b_KJ*(R(CellK,U_K) - U_K(:))

#elif defined(_FLUX_HLL_)
      call compute_wave_speed(U_K, U_J, eta, tau, wave_speed_K, wave_speed_J)

      flux_1D_source_NoAP(:) = alpha_KJ_NoAP(:)*flux_1D(U_K,U_J,eta,tau) &
          - (alpha_KJ_NoAP(:) - alpha_KK)*MATMUL(F_K,eta) &
          - (1._PR - alpha_KJ_NoAP(:))*0.5_PR*(wave_speed_J - wave_speed_K)*(R(CellK,U_K) - U_K(:))

#elif defined(_FLUX_HLLC_)
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> HLLC w/ source term not yet implemented"
      FLUSH StdErr
      STOP
#endif /* _FLUX_ */

#endif /* defined(_FLUX_SRC_CENTER_) */

  end function flux_1D_source_NoAP

#endif /* _SRC_ENABLE_ */
end module mod_flux_1D_source
