!> Computations of the numerical fluxes.
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_operators

  use mod_numerics, only: PR
  use type_mesh, only: pmesh, BndInterfaces, dim_
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells, chunk_IntInterfaces, chunk_BndInterfaces
#endif /* _FC_PAR_OMP_ */
  use type_mesh_DLP, only: IntInterfaces, max_Card_S, NbM
  use mod_physics_var, only: NbEq
  use mod_flux_num, only: compute_flux_n
#if defined(_SRC_ENABLE_) && !defined(_TP_FLUX_)
  use mod_coeff_DLP, only: compute_nu_bar_inter
#endif /* _SRC_ENABLE_ && .not. _TP_FLUX_ */
  use mod_flux_num, only: compute_flux_bnd_n
#if defined(_HO_MOOD_)
  use mod_HO_MOOD_degree, only: computeIntFacePD, computeBndFacePD
  use mod_flux_num_HO_MOOD, only: compute_flux_n_HO_MOOD
  use mod_flux_num_HO_MOOD, only: compute_flux_bnd_n_HO_MOOD
#endif /* _HO_MOOD_ */

  IMPLICIT NONE

  private
  public compute_Phi

CONTAINS
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute \f$\Phi\f$ (sum of numerical fluxes on a cell)
  !>
  !> \f$\Phi\f$ such as: \f$U^{(n+1)} =  U^{(n)} - \Delta t \Phi(U^{(n)},t^n)\f$, with:
  !>
  !> \f[\Phi_K = \sum\limits_{i \in \mathcal{E}_K}\left(\frac{|e_i|}{|K|} \mathcal{F}_{K,i}\cdot \vec{n}_{K,i}\right)\f]
  !>
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[out] Phi Sum of numerical flux on all cells
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_num_mp_compute_flux_n
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_num_mp_compute_flux_bnd_n
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_num_mp_compute_flux_bnd_n_noap
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_num_ho_mood_mp_compute_flux_n_ho_mood
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_flux_num_ho_mood_mp_compute_flux_bnd_n_ho_mood
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_degree_mp_computeintfacepd
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_ho_mood_degree_mp_computebndfacepd
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_coeff_dlp_mp_compute_nu_bar_inter
  subroutine compute_Phi(U_n,t_n,Phi)
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n
      real(PR), dimension(:,:), intent(out) :: Phi

      real(PR), dimension(NbEq) :: flux_n
      real(PR) :: le
      integer :: CellK, CellL
      integer :: i_K
      integer :: IdxIntInt, IdxBndInt
      integer :: idxM

      !==================================================================!
      !                          Initialisation
      !==================================================================!
      !$OMP DO &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      do CellK = 1, pmesh%NbCells
        Phi(:,CellK) = 0._PR
      end do
      !$OMP END DO

      !==================================================================!
      !         Loop over the interior interfaces to compute the flux
      !==================================================================!
      !$OMP DO &
      !$OMP REDUCTION(+:Phi) &
      !$OMP SCHEDULE(STATIC,chunk_IntInterfaces)
      do IdxIntInt = 1, pmesh%NbIntInterfaces
        CellK = IntInterfaces(IdxIntInt)%nghb(1)
        CellL = IntInterfaces(IdxIntInt)%nghb(2)
        le = IntInterfaces(IdxIntInt)%le
#if defined(_HO_MOOD_)
        call computeIntFacePD(IdxIntInt)
#endif /* _HO_MOOD_ */

#if defined(_SRC_ENABLE_) && !defined(_TP_FLUX_)
        call compute_nu_bar_inter(IdxIntInt,U_n,t_n,IntInterfaces(IdxIntInt)%nu_bar)
#endif /* _SRC_ENABLE_ && .not. _TP_FLUX_ */

        idxM = 1
#if defined(_HO_MOOD_)
        call compute_flux_n_HO_MOOD(CellK,IdxIntInt,idxM,U_n,t_n,flux_n)
#else
        call compute_flux_n(CellK,IdxIntInt,idxM,U_n,t_n,flux_n)
#endif /* _HO_MOOD_ */

        Phi(:,CellK) = Phi(:,CellK) + le/pmesh%elt(CellK)%area*flux_n(:)

        ! If there is no source term the flux is conservative,
        ! so we could avoid some computations
#if defined(_SRC_ENABLE_)
        idxM = 2
#if defined(_HO_MOOD_)
        call compute_flux_n_HO_MOOD(CellL,IdxIntInt,idxM,U_n,t_n,flux_n)
#else /* _HO_MOOD_ */
        call compute_flux_n(CellL,IdxIntInt,idxM,U_n,t_n,flux_n)
#endif /* _HO_MOOD_ */

#else
        flux_n(:) = - flux_n(:)
#endif /* _SRC_ENABLE_ */
        Phi(:,CellL) = Phi(:,CellL) + le/pmesh%elt(CellL)%area*flux_n(:)

      end do
      !$OMP END DO
      !==================================================================!


      !==================================================================!
      !                Loop over the boundary interfaces
      !==================================================================!
      !$OMP DO &
      !$OMP REDUCTION(+:Phi) &
      !$OMP SCHEDULE(STATIC,chunk_BndInterfaces)
      do IdxBndInt = 1, pmesh%NbBndInterfaces
        CellK = BndInterfaces(IdxBndInt)%IdxCell
        i_K = BndInterfaces(IdxBndInt)%IdxInterface

#if defined(_HO_MOOD_)
        call computeBndFacePD(IdxBndInt)
        call compute_flux_bnd_n_HO_MOOD(CellK,IdxBndInt,i_K,U_n,t_n,flux_n)
#else
        call compute_flux_bnd_n(CellK,IdxBndInt,i_K,U_n,t_n,flux_n)
#endif /* _HO_MOOD_ */

        Phi(:,CellK) = Phi(:,CellK) + pmesh%elt(CellK)%le(i_K)/pmesh%elt(CellK)%area*flux_n(:)
      end do
      !$OMP END DO
      !==================================================================!

  end subroutine compute_Phi

end module mod_operators
