!> Reverse Cuthill-McKee algorithm \cite CutMcK_1969
!>
!> @remarks http://ciprian-zavoianu.blogspot.fr/2009/01/project-bandwidth-reduction.html
!> @remarks http://www.fortran.com/qsort_c.f95
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_RCM
#if defined(_MESH_P2D_)

  use type_mesh, only: type_centroid, pmesh, Max_Nb_nghb
  use mod_numerics, only: PR, one

  implicit none

  private
  public RCM, apply_RCM

  integer, dimension(:), allocatable, save :: R

contains

  !--------------------------------------------------------------------!
  !> @brief
  !> Reverse Cuthill-McKee algorithm \cite CutMcK_1969
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine RCM()

    ! Q : queue
    integer, dimension(:), allocatable :: Q
    ! UsedForR : array to know if K is in R
    logical, dimension(:), allocatable :: UsedForR
    ! order of a node (nb of neighbour)
    integer, dimension(Max_Nb_nghb) :: order

    integer, dimension(:,:), allocatable :: nghb
    integer, dimension(:), allocatable :: NbNghb

    ! idxR: is the first non zero in R
    integer :: idxR
    ! idxQ: is the first non zero in Q
    integer :: idxQ

    integer :: CellK, CellL, idx, i_K, CellP, CellC
    integer :: NbCells
    integer :: NbNghb_P, NbNghb_C, min_NbNghb
    integer, dimension(Max_Nb_nghb) :: nghb_P, nghb_C

    NbCells = pmesh%NbCells

    allocate(R(NbCells))

    ! Get the real number of nghb (removing the boundary interfaces)
    allocate(nghb(Max_Nb_nghb,pmesh%NbCells))
    allocate(NbNghb(pmesh%NbCells))
    do CellK = 1, NbCells
      NbNghb(CellK) = 0
      do i_K = 1, pmesh%elt(CellK)%NbInterfaces
        CellL = pmesh%elt(CellK)%nghb(i_K)
        if (CellL /= 0) then
          NbNghb(CellK) = NbNghb(CellK) + 1
          nghb(NbNghb(CellK),CellK) = CellL
        endif
      end do
    end do

    allocate(Q(NbCells))
    allocate(UsedForR(NbCells))

    idxR = 1
    idxQ = 1

    R(:) = 0
    Q(:) = 0
    UsedForR(:) = .FALSE.

    ! While all the cells are not in R, we loop
    loop_R : do while (R(NbCells) == 0)
      ! We are looking for the cell which have the less nghb
      ! and not already in R
      min_NbNghb = huge(0)
      do CellK = 1, NbCells
        if (NbNghb(CellK) < min_NbNghb .and. UsedForR(CellK) .eqv. .FALSE.) then
          min_NbNghb = NbNghb(CellK)
          CellP = CellK
        end if
      end do

      ! Getting the cell with less nghb and not in R := CellP
      UsedForR(CellP) = .TRUE.
      R(idxR) = CellP
      idxR = idxR + 1

      ! Get the order of each nghb of P
      NbNghb_P = NbNghb(CellP)
      nghb_P(1:NbNghb_P) = nghb(1:NbNghb_P,CellP)
      do idx = 1, NbNghb_P
        order(idx) = NbNghb(nghb(idx,CellP))
      end do

      ! Sort those nghb of P by their order
      call qsort(order(1:NbNghb_P),nghb_P(1:NbNghb_P))

      ! Stock those nghb in the queue
      Q(idxQ:idxQ+NbNghb_P-1) = nghb_P(1:NbNghb_P)
      idxQ = idxQ + NbNghb_P

      ! While the queue is not empty
      loop_Q : do while (Q(1) /= 0)
        ! Get the first node in the queue
        CellC = Q(1)

        ! Move all the nodes
        do CellK = 1, idxQ - 1
          Q(CellK) =  Q(CellK + 1)
        end do
        Q(idxQ) = 0
        idxQ = idxQ - 1

        ! If the nodes is not in R
        if (UsedForR(CellC) .eqv. .FALSE.) then
          ! Stock C in R
          UsedForR(CellC) = .TRUE.
          R(idxR) = CellC
          idxR = idxR + 1
          ! Compute how many nghb have C that are not in R
          ! then stock those nghb and their order
          NbNghb_C = 0
          do idx = 1, NbNghb(CellC)
            if (UsedForR(nghb(idx,CellC)) .eqv. .FALSE.) then
              NbNghb_C = NbNghb_C + 1
              nghb_C(NbNghb_C) = nghb(idx,CellC)
              order(NbNghb_C) = NbNghb(nghb(idx,CellC))
            end if
          end do

          ! Sort the nghb of C that are not in R
          call qsort(order(1:NbNghb_C),nghb_C(1:NbNghb_C))

          ! Finally stock the nghb of C that are not in R in the queue
          Q(idxQ:idxQ+NbNghb_C-1) = nghb_C(1:NbNghb_C)
          idxQ = idxQ + NbNghb_C
        end if

      end do loop_Q
    end do loop_R

    ! Inverse the permutation to get the Reverse algorithm
    !R(:) = R(NbCells:1:-1)
    do CellK = 1, ISHFT(NbCells,-1)
      idx = R(CellK)
      R(CellK) = R(NbCells - CellK + 1)
      R(NbCells - CellK + 1) = idx
    end do

    deallocate(Q)
    deallocate(UsedForR)

    deallocate(nghb)
    deallocate(NbNghb)

  end subroutine RCM

  !--------------------------------------------------------------------!
  !> @brief
  !> Apply the permutations of th Reverse Cuthill-McKee algorithm \cite CutMcK_1969
  !> ( see \cite Knu_1998-a )
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine apply_RCM()

      type(type_centroid) :: elt_tmp
      integer :: CellK, perm

      do CellK = 1, pmesh%NbCells
        perm = R(CellK)
        do while (perm > CellK)
          perm = R(perm)
        end do
        if (perm ==  CellK) then
          elt_tmp = pmesh%elt(CellK)
          do while (R(perm) /= CellK)
            pmesh%elt(perm) = pmesh%elt(R(perm))
            perm = R(perm)
          end do
          pmesh%elt(perm) = elt_tmp
        end if
      end do

      deallocate(R)

  end subroutine apply_RCM

  !--------------------------------------------------------------------!
  !> @brief
  !> Qsort algorithm to sort an integer array A, with b as tag. b is sorted
  !> in the same way as A.
  !>
  !> @param[inout] A integer array to sort
  !> @param[inout] b integer tag array sorted in the same way as A
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  recursive subroutine qsort(A,b)
    integer, intent(in out), dimension(:) :: A, b
    integer :: iq

    if(size(A) > 1) then
      call qsort_partition(A, b, iq)
      call qsort(A(:iq-1),b(:iq-1))
      call qsort(A(iq:),b(iq:))
    endif
  end subroutine qsort

  !--------------------------------------------------------------------!
  !> @brief
  !> Auxiliary routine for qsort
  !>
  !> @param[inout] A integer array to sort
  !> @param[inout] b integer tag array sorted in the same way as A
  !> @param[out] marker where to start
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine qsort_partition(A, b, marker)
    integer, intent(in out), dimension(:) :: A, b
    integer, intent(out) :: marker
    integer :: i, j
    integer :: temp
    integer :: x      ! pivot point
    x = A(1)
    i= 0
    j= size(A) + 1

    do while (.TRUE.)
      j = j-1
      do while (.TRUE.)
        if (A(j) <= x) exit
        j = j-1
      end do
      i = i+1
      do while (.TRUE.)
        if (A(i) >= x) exit
        i = i+1
      end do
      if (i < j) then
        ! exchange A(i) and A(j)
        temp = A(i)
        A(i) = A(j)
        A(j) = temp
        ! exchange b(i) and b(j)
        temp = b(i)
        b(i) = b(j)
        b(j) = temp
      elseif (i == j) then
        marker = i+1
        return
      else
        marker = i
        return
      endif
    end do

  end subroutine qsort_partition

#endif /* _MESH_ */
end module mod_RCM
