!> Definition of the type used to save the informations
!> about the interfaces and the nodes used in the DLP scheme \cite DroLeP_2011 .
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module type_mesh_DLP
  use mod_numerics, only: PR
  use type_mesh,  only: dim_
  use mod_physics_var, only: NbEq

  implicit none

  !> Maximum number of points used to compute the value in M
  integer, save :: max_Card_S

  !> Number of points M, there is one on both sides of the interfaces
  integer,  parameter :: NbM = 2
  !> Initial position of the point M from its origin
  !> @warning for now keep \c ratio_init <= 1
  real(PR), parameter :: ratio_init = 1._PR
  !> Define how much we move closer M to its origin at each step
  real(PR), parameter :: step = 0.1_PR
  !> Number of maximum iterations to find the supplementary points
  integer,  parameter :: iter_M_max = floor(ratio_init/step) - 1
  !> Number of levels to get the candidates points (1=direct neighbors,
  !> 2=neighbors of the neighbors, ...)
  integer,  parameter :: nghb_level = 2

  !> \f$\nu\f$ for boundary interfaces, and the two points flux
  real(PR), dimension(NbEq), parameter :: nu_one(1:NbEq) = 1._PR

  !--------------------------------------------------------------------!
  !> @brief
  !> Type definition for the interior interfaces and the additional nodes
  !> of Droniou and Le Potier \cite DroLeP_2011 .
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  type type_int_interface
    !> Cardinal of \f$S_{K,i}\f$ and \f$S_{L,i}\f$, number of points used for the reconstruction
    !> usually \f$\#S_{K,i}=3\f$ but it could higher and in the case of admissible interface
    !> \f$\#S_{K,i}=2\f$
    integer :: Card_S
    !> Coordinates of the two points \f$M\f$
    real(PR), dimension(dim_,NbM) :: xM
    !> Distance \f$KM_K\f$ and \f$LM_L\f$
    real(PR), dimension(NbM) :: dist
    !> Local indexes of the points used to reconstruct \f$KM_K\f$ and \f$LM_L\f$
    !> <tt> idx(Card_S,NbM) </tt>
    integer,  dimension(:,:), allocatable :: idx
    !> Barycentric coordinates in the polygon
    !> <tt> coord_bar(Card_S,NbM) </tt>
    real(PR), dimension(:,:), allocatable :: coord_bar
#if defined(_SRC_ENABLE_)
    !> Coefficient of the normal in the polygon regarding to \f$\overrightarrow{KJ}\f$
    !> which are \f$\frac{\textrm{CoordBar}}{\textrm{dist}}\f$
    !> <tt> coeff_n_KJ(Card_S,NbM) </tt>
    real(PR), dimension(:,:), allocatable :: coeff_n_KJ
#endif /* _SRC_ENABLE_ */
    !> Coefficient of the normal in the polygon regarding to \f$\overrightarrow{\eta}\f$
    !> which are \f$\frac{\textrm{CoordBar}}{\textrm{dist}}\Vert\overrightarrow{KJ}\Vert\f$
    !> <tt> coeff_n_eta(Card_S,NbM) </tt>
    real(PR), dimension(:,:), allocatable :: coeff_n_eta
    !> Vector in the direction \f$KJ\f$ and \f$LJ\f$
    !> <tt> xKJ(dim_,Card_S,NbM) </tt>
    real(PR), dimension(:,:,:), allocatable :: xKJ
    !> Norm of the vector \f$\vec{KJ}\f$ and \f$\vec{LJ}\f$
    !> <tt> dKJ(Card_S,NbM) </tt>
    real(PR), dimension(:,:), allocatable :: dKJ
    !> Unitary vector in the direction \f$KJ\f$ andr \f$LJ\f$
    !> <tt> eta(dim_,Card_S,NbM) </tt>
    real(PR), dimension(:,:,:), allocatable :: eta
    !> Unitary vector in the direction orthogonal to \f$KJ\f$ and \f$LJ\f$
    !> <tt> tau(dim_,Card_S,NbM) </tt>
    real(PR), dimension(:,:,:), allocatable :: tau
    !> \f$\beta\f$ coefficient:
    !> \f$\beta = \Vert\overrightarrow{KL}\Vert \min(\frac{c_{L,M_K}}{|KM_K|}\Vert,\frac{c_{K,M_L}}{|LM_L|})\f$
    real(PR) :: beta
#if defined(_SRC_ENABLE_)
    !> \f$\bar{\beta}\f$ coefficient (DLP):
    !> \f$\bar{\beta} = \min(\frac{c_{L,M_K}}{|KM_K|}\Vert,\frac{c_{K,M_L}}{|LM_L|})\f$
    real(PR) :: beta_bar
#endif /* _SRC_ENABLE_ */
    !> Length of the interface
    real(PR) :: le
    !> Index of the two neighbors cells \f$K\f$ and \f$L\f$
    integer, dimension(NbM) :: nghb
    !> Index in \f$K\f$ or \f$L\f$ of the interface used
    integer, dimension(NbM) :: i_Cell
    !> Define if the interface is admissible (i.e. if \f$ \vec{KL}\wedge\vec{n} = \vec{0} \f$)
    logical  :: admissible
    !> \f$\nu_{K,i}^{J,j}\f$ coefficients
    !> <tt> nu(NbEq,Card_S,NbM) </tt>
    real(PR), dimension(:,:,:), allocatable :: nu
    !> \f$\tilde{\nu}_{K,i,r}^{J,j}\f$ coefficients
    !> <tt> nu_tilde(NbEq,Card_S,R_i,NbM) </tt>
    real(PR), dimension(:,:,:,:), allocatable :: nu_tilde
#if defined(_SRC_ENABLE_)
    !> \f$\bar{\nu}_{K,i}^{J}\f$ coefficients
    !> <tt> nu_bar(Card_S,NbM) </tt>
    real(PR), dimension(:,:), allocatable :: nu_bar
    !> \f$\tilde{\bar{\nu}}_{K,i,r}^{J}\f$ coefficients
    !> <tt> nu_bar_tilde(Card_S,R_i,NbM) </tt>
    real(PR), dimension(:,:,:), allocatable :: nu_bar_tilde
#endif /* _SRC_ENABLE_ */
#if defined(_HO_MOOD_)
    !> <tt>FacePD</tt>: degree of the polynomial reconstruction on the interfaces
    integer :: FacePD
#endif /* _HO_MOOD_ */
  end type type_int_interface

  !> Contains the additional nodes for the DLP scheme
  type(type_int_interface), dimension(:), allocatable, save :: IntInterfaces

end module type_mesh_DLP
