!> Physics of the Euler model with ideal gas and friction \cite Tor_2009 .
!>
!> \f[\left\{\begin{array}{rcl}
!>   \partial_t \rho   + \partial_x \rho u               + \partial_y \rho v                    & = & 0 \\
!>   \partial_t \rho u + \partial_x (\rho u^2 + p(\rho)) + \partial_y \rho uv                   & = & - \kappa \rho u \\
!>   \partial_t \rho v + \partial_x \rho uv              + \partial_y (\rho v^2 + p(\rho))      & = & - \kappa \rho v  \\
!>   \partial_t E      + \partial_x u (E+p)              + \partial_y v (E +p)                  & = & 0
!> \end{array}\right.\f]
!> with:
!> * \f$ \rho \f$: density of the fluid,
!> * \f$ \vec{V} =(u,v) \f$: speed of the fluid,
!> * \f$ E = \rho(\frac{u^2 + v^2}{2} + e) \f$: total energy per unit volume, \f$e\f$ specific internal energy,
!> * \f$ \kappa > 0 \f$,
!> * the equation of state (EOS): \f$e=\frac{p}{\rho(\gamma-1)}\f$, \f$p'(\rho) > 0\f$,
!> * and the speed of the sound in the fluid is defined as: \f$c = \sqrt{\frac{\gamma p}{\rho}}\f$.
!>
!> In test cases with source terms we are using the isentropic Euler model:
!> \f[\left\{\begin{array}{rcl}
!>   \partial_t \rho   + \partial_x \rho u               + \partial_y \rho v               & = & 0 \\
!>   \partial_t \rho u + \partial_x (\rho u^2 + p(\rho)) + \partial_y \rho uv              & = & - \kappa \rho u \\
!>   \partial_t \rho v + \partial_x \rho uv              + \partial_y (\rho v^2 + p(\rho)) & = & - \kappa \rho v  \\
!> \end{array}\right.\f]
!> with:
!> * the equation of state (EOS): \f$p=\kappa_p\rho^{\gamma}\f$, \f$p'(\rho) = \kappa_p \gamma p^{\gamma - 1} > 0\f$,
!>
!> @remarks See \cite MarMil_1990 and \cite Vaz_1987 for the asymptotic behaviour.
!> @remarks See \cite KurTad_2002 and \cite Des_2013 for the 2D test cases.
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_physics_Euler
#if defined(_PHY_EULER_)

  ! Choice for the computations of the wave speeds
#define _PHY_EULER_DAVIS_SPEED_
!#define _PHY_EULER_ROE_SPEED_
!#define _PHY_EULER_EINFELDT_SPEED_
!#define _PHY_EULER_PRESSURE_BASED_SPEED_

  use mod_numerics, only: PR, pi, eps
  use type_IO, only: StdErr
  use mod_tools, only: RotationMatrix
  use mod_tools, only: min_NaN, max_NaN
  use type_mesh, only: pmesh, dim_
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells
#endif /* _FC_PAR_OMP_ */
  use mod_read, only: read_solution
  use mod_read, only: Restarting

  use mod_physics_Euler_var, only: NbEq, test_case
  use mod_physics_Euler_var, only: T_end
  use mod_physics_Euler_var, only: adiabatic_index, adiabatic_index_m1
#if defined(_PHY_EULER_PRESSURE_BASED_SPEED_)
  use mod_physics_Euler_var, only: adiabatic_index_p1
#endif /* _PHY_EULER_WAVE_SPEED_ */
#if defined(_PHY_EULER_ISENTROPIC_)
  use mod_physics_Euler_var, only: kappa_p
#endif /* _PHY_EULER_ENERGY_ */

#if defined(_HO_MOOD_)
  use type_HO_MOOD, only: MaxDegree
  use mod_HO_MOOD_int, only: compute_mean_f
#endif /* _HO_MOOD_ */

  IMPLICIT NONE

  private
  public physic_flux, U_Dirichlet
  public compute_wave_speed,  compute_greatest_wave_speed
#if defined(_FLUX_HLLC_)
  public compute_HLLC_values
#endif /* _FLUX_HLLC_ */
  public init_U
  public primitive_to_conservative, conservative_to_primitive
  public IsPhysicallyAdmissible

  public compute_pressure
#if defined(_SRC_ENABLE_)
  public p_prime_rho
#endif /* _SRC_ENABLE_ */


CONTAINS
  !--------------------------------------------------------------------!
  !> @brief
  !> Give the initial condition according to the test case
  !>
  !> @param[out] U0 array containing the initial condition
  !> @param[out] iter  Number of iteration previously done
  !> @param[out] T_init  Initial time
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_primitive_to_conservative
  subroutine init_U(U0,iter,T_init)
      real(PR), dimension(:,:), intent(out) :: U0
      integer , intent(out) :: iter
      real(PR), intent(out) :: T_init

      real(PR), dimension(NbEq) :: U_K, V_K
      real(PR), dimension(dim_) :: xK
      integer :: CellK

#if defined(_PHY_EULER_IDEAL_GAS_)
      real(PR), dimension(dim_,dim_) :: R
      real(PR), dimension(dim_) :: x_c
      real(PR) :: radius, theta
#endif /* defined(_PHY_EULER_IDEAL_GAS_) */

      !$OMP SINGLE
      select case(test_case)
      case ('cst', 'cst_w_speed')
        T_end = 1._PR
#if defined(_PHY_EULER_ISENTROPIC_)
      case ('Gauss_x', 'Gauss', 'circle')
        T_end = 10.0_PR
      case ('Gauss_large')
        T_end = 100.0_PR
      case ('Square_x')
        T_end = 5.0_PR
      case ('sod_x')
        T_end = 0.3_PR
      case ('StepM3')
#if defined(_SRC_ENABLE_)
        T_end = 2.0_PR
#else
        T_end = 4.0_PR
#endif /* _SRC_ENABLE_ */
#elif defined(_PHY_EULER_IDEAL_GAS_)
      case ('KT3', 'KT5', 'KT6', &
            'KT6_x1', 'KT6_x2')
        T_end = 0.3_PR
      case ('KT6_xy')
        T_end = sqrt(2._PR)*0.3_PR
      case ('KT')
        T_end = 0.15_PR
      case ('implosion')
        T_end = 5.0_PR
      case ('explosion') ! 2D sod
        T_end = 0.125_PR
      case ('KH')
        T_end = 5.0_PR
      case ('Gresho')
        T_end = 3.0_PR
      case ('sod_x', 'sod_y')
        T_end = 0.15_PR
      case ('rarefaction_x')
        T_end = 0.08_PR
      case ('2shock_x')
        T_end = 0.1_PR
      case ('sod_xy')
        T_end = sqrt(2._PR)*0.15_PR
      case ('lax_x')
        T_end = 0.15_PR
      case ('CisM1_x', 'CisM3_x')
        T_end = 0.3_PR
      case ('StepM3')
        T_end = 4.0_PR
      case ('Ramp')
        T_end = 0.2_PR
      case ('sonde_80km')
        T_end = 0.1_PR
      case ('ball')
        T_end = 1.E-5_PR
#endif /* PHY */
      case default
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
        FLUSH StdErr
        STOP
      end select
      !$OMP END SINGLE NOWAIT

      test_restart: if (Restarting) then
        !$OMP SINGLE
        call read_solution(U0,iter,T_init)
        !$OMP END SINGLE

      else
        !$OMP SINGLE
        iter = 0
        T_init = 0._PR
        !$OMP END SINGLE

        select case(test_case)
        case ('cst')
          !$OMP SINGLE
          V_K(1) = 1._PR
          V_K(2) = 0._PR
          V_K(3) = 0._PR
#if defined(_PHY_EULER_IDEAL_GAS_)
          V_K(4) = 1._PR
#endif /* _PHY_EULER_ENERGY_ */
          call primitive_to_conservative(V_K,U_K)
          !$OMP END SINGLE COPYPRIVATE(U_K)

          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            U0(:,CellK) = U_K(:)
          end do
          !$OMP END DO

        case ('cst_w_speed')
          !$OMP SINGLE
          V_K(1) = 1._PR
          V_K(2) = 1._PR
          V_K(3) = 1._PR
#if defined(_PHY_EULER_IDEAL_GAS_)
          V_K(4) = 1._PR
#endif /* _PHY_EULER_ENERGY_ */
          call primitive_to_conservative(V_K,U_K)
          !$OMP END SINGLE COPYPRIVATE(U_K)

          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            U0(:,CellK) = U_K(:)
          end do
          !$OMP END DO

#if defined(_PHY_EULER_ISENTROPIC_)
        case ('Gauss_x', 'Gauss', 'Gauss_large')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
            V_K(:) = compute_mean_f(CellK,T_init,MaxDegree,InitialCondition,NbEq)
#else
            xK(:) = pmesh%elt(CellK)%x(:)
            V_K(:) = InitialCondition(xK,T_init,NbEq)
#endif
            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)
          end do
          !$OMP END DO

        case ('Square_x')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
            if (xK(1) > 0.45_PR .and. xK(1) < 0.55_PR) then
              U0(1,CellK) = 10._PR
            else
              U0(1,CellK) = 1.E-14_PR
            endif
            U0(2:3,CellK) = 0._PR
          end do
          !$OMP END DO

        case ('circle')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
            if ((xK(1) - 0.5_PR)**2 + (xK(2) - 0.5_PR)**2 < (0.1_PR)**2) then
              U0(1,CellK) = 1.0_PR
            else
              U0(1,CellK) = 0.1_PR
            end if
            U0(2,CellK) = 0._PR
            U0(3,CellK) = 0._PR
          end do
          !$OMP END DO

        case ('sod_x')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if (xK(1) < 0.5_PR) then
              V_K(1) = 1._PR
              V_K(2) = 0._PR
              V_K(3) = 0._PR
            else
              V_K(1) = 0.125_PR
              V_K(2) = 0._PR
              V_K(3) = 0._PR
            end if

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)

          end do
          !$OMP END DO

        case ('StepM3')
          !$OMP SINGLE
          V_K(1) = 1.4_PR
          V_K(2) = 3._PR
          V_K(3) = 0._PR
          call primitive_to_conservative(V_K,U_K)
          !$OMP END SINGLE COPYPRIVATE(U_K)
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            U0(:,CellK) = U_K(:)
          end do
          !$OMP END DO
#endif /* _PHY_EULER_ENERGY_ */

#if defined(_PHY_EULER_IDEAL_GAS_)
        case ('KT3')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
            if (xK(2) <= 0.5_PR .and. xK(1) <= 0.5_PR) then
              V_K(1) = .138_PR
              V_K(2) = 1.206_PR
              V_K(3) = 1.206_PR
              V_K(4) = 0.029_PR
            else if( xK(2) <= 0.5_PR .and. xK(1) >= 0.5_PR) then
              V_K(1) = .5323_PR
              V_K(2) = 0._PR
              V_K(3) = 1.206_PR
              V_K(4) = .3_PR
            else if( xK(2) >= 0.5_PR .and. xK(1) <= 0.5_PR) then
              V_K(1) = 0.5323_PR
              V_K(2) = 1.206_PR
              V_K(3) = 0.0_PR
              V_K(4) = 0.3_PR
            else !if( xK(2) >= 0.5_PR .and. xK(1) >= 0.5_PR) then
              V_K(1) = 1.5_PR
              V_K(2) = 0._PR
              V_K(3) = 0._PR
              V_K(4) = 1.5_PR
            end if

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)
          end do
          !$OMP END DO

        case ('KT5')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if (xK(2) <= 0.5_PR .and. xK(1) <= 0.5_PR) then
              V_K(1) = 1._PR
              V_K(2) = 0.75_PR
              V_K(3) = .5_PR
              V_K(4) = 1._PR
            else if( xK(2) <= 0.5_PR .and. xK(1) >= 0.5_PR) then
              V_K(1) = 3._PR
              V_K(2) = 0.75_PR
              V_K(3) = -0.5_PR
              V_K(4) = 1._PR
            else if( xK(2) >= 0.5_PR .and. xK(1) <= 0.5_PR) then
              V_K(1) = 2._PR
              V_K(2) = -0.75_PR
              V_K(3) = 0.5_PR
              V_K(4) = 1._PR
            else !if( xK(2) >= 0.5_PR .and. xK(1) >= 0.5_PR) then
              V_K(1) = 1._PR
              V_K(2) = -.75_PR
              V_K(3) = -0.5_PR
              V_K(4) = 1._PR
            end if

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)
          end do
          !$OMP END DO

        case ('KT6')
          x_c(:) = 0.5_PR !+ pmesh%Delta_x ! Trick for fully admissible meshes
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if (xK(2) <= x_c(2) .and. xK(1) <= x_c(1)) then
              V_K(1) = 1._PR
              V_K(2) = -0.75_PR
              V_K(3) = 0.5_PR
              V_K(4) = 1._PR
            else if ( xK(2) <= x_c(2) .and. xK(1) >= x_c(1)) then
              V_K(1) = 3._PR
              V_K(2) = -0.75_PR
              V_K(3) = -0.5_PR
              V_K(4) = 1._PR
            else if ( xK(2) >= x_c(2) .and. xK(1) <= x_c(1)) then
              V_K(1) = 2._PR
              V_K(2) = 0.75_PR
              V_K(3) = 0.5_PR
              V_K(4) = 1._PR
            else !if ( xK(2) >= x_c(2) .and. xK(1) >= x_c(1)) then
              V_K(1) = 1._PR
              V_K(2) = 0.75_PR
              V_K(3) = -0.5_PR
              V_K(4) = 1._PR
            end if

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)
          end do
          !$OMP END DO

        case ('KT6_xy')
          call RotationMatrix(0.25_PR*pi,R)
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if (xK(2) <= xK(1) .and. xK(2) <= 1._PR - xK(1)) then
              V_K(1) = 1._PR
              V_K(2) = -0.75_PR
              V_K(3) = 0.5_PR
              V_K(4) = 1._PR
            else if ( xK(2) <= xK(1) .and. xK(2) >= 1._PR - xK(1)) then
              V_K(1) = 3._PR
              V_K(2) = -0.75_PR
              V_K(3) = -0.5_PR
              V_K(4) = 1._PR
            else if ( xK(2) >= xK(1) .and. xK(2) <= 1._PR - xK(1)) then
              V_K(1) = 2._PR
              V_K(2) = 0.75_PR
              V_K(3) = 0.5_PR
              V_K(4) = 1._PR
            else !if ( xK(2) >= xK(1) .and. xK(2) >= 1._PR - xK(1)) then
              V_K(1) = 1._PR
              V_K(2) = 0.75_PR
              V_K(3) = -0.5_PR
              V_K(4) = 1._PR
            end if

            V_K(2:3) = MATMUL(R,V_K(2:3))

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)
          end do
          !$OMP END DO

      case ('KT6_x1')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if (xK(1) <= 0.5_PR) then
              V_K(1) = 1._PR
              V_K(2) = -0.75_PR
              V_K(3) = 0.5_PR
              V_K(4) = 1._PR
            else
              V_K(1) = 3._PR
              V_K(2) = -0.75_PR
              V_K(3) = 0.5_PR
              V_K(4) = 1._PR
            end if

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)
          end do
          !$OMP END DO

      case ('KT6_x2')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if (xK(1) <= 0.5_PR) then
              V_K(1) = 1._PR
              V_K(2) = -0.75_PR
              V_K(3) = 0.5_PR
              V_K(4) = 1._PR
            else
              V_K(1) = 3._PR
              V_K(2) = -0.75_PR
              V_K(3) = -0.5_PR
              V_K(4) = 1._PR
            end if

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)
          end do
          !$OMP END DO

        case ('KT')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if (xK(2) <= 0.5_PR .and. xK(1) <= 0.5_PR) then
              V_K(1) = 1._PR
              V_K(2) = 0._PR
              V_K(3) = 0._PR
              V_K(4) = 1._PR
            else if ( xK(2) <= 0.5_PR .and. xK(1) >= 0.5_PR) then
              V_K(1) = 3._PR
              V_K(2) = 0._PR
              V_K(3) = 0._PR
              V_K(4) = 1._PR
            else if ( xK(2) >= 0.5_PR .and. xK(1) <= 0.5_PR) then
              V_K(1) = 2._PR
              V_K(2) = 0._PR
              V_K(3) = 0._PR
              V_K(4) = 1._PR
            else !if ( xK(2) >= 0._PR .and. xK(1) >= 0._PR) then
              V_K(1) = 1._PR
              V_K(2) = 0._PR
              V_K(3) = 0._PR
              V_K(4) = 1._PR
            end if

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)
          end do
          !$OMP END DO

        case ('implosion')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if (xK(1) + xk(2) > 0.5_PR) then
              V_K(1) = 1._PR
              V_K(2) = 0._PR
              V_K(3) = 0._PR
              V_K(4) = 1._PR
            else
              V_K(1) = 0.125_PR
              V_K(2) = 0._PR
              V_K(3) = 0._PR
              V_K(4) = 0.4_PR
            end if

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)

          end do
          !$OMP END DO

        case ('explosion') ! 2D sod
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if ((xK(1) - 0.5_PR)**2 + (xK(2) - 0.5_PR)**2 < (0.2_PR)**2) then
              V_K(1) = 1._PR
              V_K(2) = 0._PR
              V_K(3) = 0._PR
              V_K(4) = 1._PR
            else
              V_K(1) = 0.125_PR
              V_K(2) = 0._PR
              V_K(3) = 0._PR
              V_K(4) = 0.1_PR
            end if

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)

          end do
          !$OMP END DO

        case ('KH')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if ( 0.75_PR < xK(2) .and.  xk(2) < 0.5_PR) then
              V_K(1) = 1._PR
              V_K(2) = - 0.5_PR
              V_K(3) = 0._PR
              V_K(4) = 2.5_PR
            else if ( 0.25_PR < xK(2) .and.  xk(2) < 0.75_PR) then
              V_K(1) = 2._PR
              V_K(2) = 0.5_PR
              V_K(3) = 0._PR
              V_K(4) = 2.5_PR
            else !if ( 0._PR < xK(2) .and.  xk(2) < 0.25_PR) then
              V_K(1) = 1._PR
              V_K(2) = - 0.5_PR
              V_K(3) = 0._PR
              V_K(4) = 2.5_PR
            end if

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)

          end do
          !$OMP END DO

        case ('Gresho')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            V_K(1) = 1._PR

            radius = sqrt((xk(1)-0.5_PR)**2 + (xk(2)-0.5)**2)
            theta = atan2(xK(2),xK(1))
            call RotationMatrix(theta,R)

            if (0.0_PR <= radius .and. radius < 0.2_PR) then
              V_K(2) = 0.0_PR
              V_K(3) = 5.0_PR*radius
              V_K(4) = 5.0_PR + 12.5_PR*radius**2
            else if (0.2 <= radius  .and. radius < 0.4_PR) then
              V_K(2) = 0.0_PR
              V_K(3) = 2.0_PR - 5.0_PR*radius
              V_K(4) = 9._PR - 4._PR*log(0.2_PR) + 12.5_PR*radius**2 &
                  - 20._PR*radius + 4._PR*log(radius)
            else !if (0.4_PR <= radius) then
              V_K(2) = 0.0_PR
              V_K(3) = 0.0_PR
              V_K(4) = 3._PR + 4.0*log(2._PR)
            end if

            V_K(2:3) = MATMUL(R,V_K(2:3))

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)

          end do
          !$OMP END DO

        case ('sod_x')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if (xK(1) < 0.5_PR) then
              V_K(1) = 1._PR
              V_K(2) = 0._PR
              V_K(3) = 0._PR
              V_K(4) = 1._PR
            else
              V_K(1) = 0.125_PR
              V_K(2) = 0._PR
              V_K(3) = 0._PR
              V_K(4) = 0.1_PR
            end if

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)

          end do
          !$OMP END DO

        case ('rarefaction_x')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if (xK(1) < 0.5_PR) then
              V_K(1) = 1._PR
              V_K(2) = - 1._PR
              V_K(3) = 0._PR
              V_K(4) = 1.5_PR
            else
              V_K(1) = 0.1989_PR
              V_K(2) = 1._PR
              V_K(3) = 0._PR
              V_K(4) = 0.1564_PR
            end if

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)

          end do
          !$OMP END DO

        case ('2shock_x')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if (xK(1) < 0.5_PR) then
              V_K(1) = 1._PR
              V_K(2) = 10._PR
              V_K(3) = 0._PR
              V_K(4) = 1._PR
            else
              V_K(1) = 1._PR
              V_K(2) = - 10._PR
              V_K(3) = 0._PR
              V_K(4) = 1._PR
            end if

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)

          end do
          !$OMP END DO

        case ('sod_y')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if (xK(2) < 0.5_PR) then
              V_K(1) = 1._PR
              V_K(2) = 0._PR
              V_K(3) = 0._PR
              V_K(4) = 1._PR
            else
              V_K(1) = 0.125_PR
              V_K(2) = 0._PR
              V_K(3) = 0._PR
              V_K(4) = 0.1_PR
            end if

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)

          end do
          !$OMP END DO

        case ('sod_xy')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if (xK(2) < xK(1)) then
              V_K(1) = 1._PR
              V_K(2) = 0._PR
              V_K(3) = 0._PR
              V_K(4) = 1._PR
            else
              V_K(1) = 0.125_PR
              V_K(2) = 0._PR
              V_K(3) = 0._PR
              V_K(4) = 0.1_PR
            end if

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)

          end do
          !$OMP END DO

        case ('lax_x')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if (xK(1) < 0.5_PR) then
              V_K(1) = 0.445_PR
              V_K(2) = 0.698_PR
              V_K(3) = 0._PR
              V_K(4) = 3.528_PR
            else
              V_K(1) = 0.5_PR
              V_K(2) = 0._PR
              V_K(3) = 0._PR
              V_K(4) = 0.571_PR
            end if

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)

          end do
          !$OMP END DO

        case ('CisM1_x')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if (xK(1) < 0.5_PR) then
              V_K(1) = 1.4_PR
              V_K(2) = 0._PR
              V_K(3) = 1._PR
              V_K(4) = 1.0_PR
            else
              V_K(1) = 1.4_PR
              V_K(2) = 0._PR
              V_K(3) = - 1._PR
              V_K(4) = 1.0_PR
            end if

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)

          end do
          !$OMP END DO

        case ('CisM3_x')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if (xK(1) < 0.5_PR) then
              V_K(1) = 1.4_PR
              V_K(2) = 0._PR
              V_K(3) = 3._PR
              V_K(4) = 1.0_PR
            else
              V_K(1) = 1.4_PR
              V_K(2) = 0._PR
              V_K(3) = - 3._PR
              V_K(4) = 1.0_PR
            end if

            call primitive_to_conservative(V_K,U_K)
            U0(:,CellK) = U_K(:)

          end do
          !$OMP END DO

        case ('StepM3')
          !$OMP SINGLE
          V_K(1) = 1.4_PR
          V_K(2) = 3._PR
          V_K(3) = 0._PR
          V_K(4) = 1._PR
          call primitive_to_conservative(V_K,U_K)
          !$OMP END SINGLE COPYPRIVATE(U_K)
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            U0(:,CellK) = U_K(:)
          end do
          !$OMP END DO

        case ('Ramp')
          !$OMP SINGLE
          V_K(1) = 1.4_PR
          V_K(1) = 1.4_PR
          V_K(2) = 0._PR
          V_K(3) = 0._PR
          V_K(4) = 1._PR
          call primitive_to_conservative(V_K,U_K)
          !$OMP END SINGLE COPYPRIVATE(U_K)
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            U0(:,CellK) = U_K(:)
          end do
          !$OMP END DO

        case ('sonde_80km')
          !$OMP SINGLE
          V_K(1) = 1.57005E-5_PR
          V_K(2) = 0._PR
          V_K(3) = 0._PR
          V_K(4) = 8.86280E-1_PR
          call primitive_to_conservative(V_K,U_K)
          !$OMP END SINGLE COPYPRIVATE(U_K)
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            U0(:,CellK) = U_K(:)
          end do
          !$OMP END DO

        case ('ball')
          !$OMP SINGLE
          V_K(1) = 0.00042_PR
          V_K(2) = 16.E3_PR
          V_K(3) = 0._PR
          V_K(4) = 30.6305_PR
          call primitive_to_conservative(V_K,U_K)
          !$OMP END SINGLE COPYPRIVATE(U_K)
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            U0(:,CellK) = U_K(:)
          end do
          !$OMP END DO

#endif /* _PHY_EULER_ENERGY_ */
        end select

      end if test_restart

  end subroutine init_U

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the initial condition
  !>
  !> @param[in] xK Coordinates
  !> @param[in] t_n Current time
  !> @param[in] NbEq Number of equations
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function InitialCondition(xK,t_n,NbEq) result(U0)
      real(PR), dimension(:), intent(in) :: xK
      real(PR), intent(in) :: t_n
      integer, intent(in) :: NbEq
      real(PR), dimension(NbEq) :: U0

      select case(test_case)
#if defined(_PHY_EULER_ISENTROPIC_)
      case ('Gauss_x')
        U0(1) = 1.E-1_PR*exp(-0.5_PR*(xK(1) - 0.5_PR)**2/(0.05_PR)**2) + 1.E-1_PR
        U0(2:3) = 0._PR

      case ('Gauss')
        U0(1) = 0.1_PR*exp(- 0.5_PR*(xK(1) - 0.5_PR)**2/(0.05_PR)**2 &
            - 0.5_PR*(xK(2) - 0.5_PR)**2/(0.05_PR)**2) &
            + 1.E-1_PR
        U0(2:3) = 0._PR

      case ('Gauss_large')
        U0(1) = exp(- (xK(1) - 50._PR)**2 - (xK(2) - 50._PR)**2) + 1._PR
        U0(2:3) = 0._PR
#endif /* _PHY_EULER_ENERGY_ */
      case default
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
        FLUSH StdErr
        STOP
      end select

  end function InitialCondition

  !--------------------------------------------------------------------!
  !> @brief
  !> Give the limit boundary value
  !>
  !> @param[in] x Coordinates of the point on the boundary
  !> @param[in] t_n Current time
  !>
  !> @return Value of the solution on the boundary at point
  !> \f$x\f$ and time \f$t^n\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_primitive_to_conservative
  function U_Dirichlet(x,t_n)
      real(PR), dimension(:), intent(in) :: x
      real(PR), intent(in) :: t_n
      real(PR), dimension(NbEq) :: U_Dirichlet

      real(PR), dimension(NbEq) :: U_K, V_K

      select case(test_case)
      case ('cst')
        V_K(1) = 1._PR
        V_K(2) = 0._PR
        V_K(3) = 0._PR
#if defined(_PHY_EULER_IDEAL_GAS_)
        V_K(4) = 1._PR
#endif /* _PHY_EULER_ENERGY_ */
      case ('cst_w_speed')
        V_K(1) = 1._PR
        V_K(2) = 1._PR
        V_K(3) = 1._PR
#if defined(_PHY_EULER_IDEAL_GAS_)
        V_K(4) = 1._PR
#endif /* _PHY_EULER_ENERGY_ */

#if defined(_PHY_EULER_ISENTROPIC_)
      case ('StepM3')
        V_K(1) = 1.4_PR
        V_K(2) = 3._PR
        V_K(3) = 0._PR
#endif /* _PHY_EULER_ENERGY_ */

#if defined(_PHY_EULER_IDEAL_GAS_)
      case ('sod_x')
        if (x(1) < 0.0_PR) then
          V_K(1) = 1._PR
          V_K(2) = 0._PR
          V_K(3) = 0._PR
          V_K(4) = 1._PR
        else
          V_K(1) = 0.125_PR
          V_K(2) = 0._PR
          V_K(3) = 0._PR
          V_K(4) = 0.1_PR
        end if

      case ('StepM3')
        V_K(1) = 1.4_PR
        V_K(2) = 3._PR
        V_K(3) = 0._PR
        V_K(4) = 1._PR

      case ('Ramp')
        V_K(1) = 8._PR
        V_K(2) = 8.25_PR
        V_K(3) = 0._PR
        V_K(4) = 116.5_PR

      case ('sonde_80km')
        V_K(1) = 1.57005E-5_PR
        ! Linear ramp for the speed
        if (t_n < 0.05_PR) then
          V_K(2) = (1.E4_PR - 1.E3_PR)/0.05_PR*t_n + 1.E3_PR
        else
          V_K(2) = 1.E4_PR
        end if
        V_K(3) = 0._PR
        V_K(4) = 8.86280E-1_PR

      case ('ball')
        V_K(1) = 0.00042_PR
        V_K(2) = 16.E3_PR
        V_K(3) = 0._PR
        V_K(4) = 30.6305_PR
#endif /* _PHY_EULER_ENERGY_ */
      case default
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
        FLUSH StdErr
        STOP
      end select

      call primitive_to_conservative(V_K,U_K)
      U_Dirichlet(:) = U_K(:)

  end function U_Dirichlet

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the wave speeds on a interface
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent  Normal and tangent to the interface
  !> @param[out] a_l, a_r  Wave speeds on the left and right
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_sound_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_pressure
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_min_nan
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_max_nan
  subroutine compute_wave_speed(Ul,Ur,normal,tangent,a_l,a_r)
      real(PR), dimension(:), intent(in) :: Ul, Ur
      real(PR), dimension(:), intent(in) :: normal, tangent
      real(PR), intent(out) :: a_l, a_r

      ! Speed of the fluid
      real(PR), dimension(dim_) :: u_l, u_r
      ! Normal speed of the fluid
      real(PR) :: u_l_n, u_r_n
      ! Density of the fluid
      real(PR) :: rho_l, rho_r
#if defined(_PHY_EULER_DAVIS_SPEED_)
      ! Speed of sound in the fluid
      real(PR) :: c_l, c_r
#elif defined(_PHY_EULER_ROE_SPEED_)
      real(PR) :: u_tilde, a_tilde
      real(PR) :: sqrt_rho_l, sqrt_rho_r
      ! Pressure in the fluid
      real(PR) :: p_l, p_r
      ! Enthalpy
      real(PR) :: H_l, H_r, H_tilde
#elif defined(_PHY_EULER_EINFELDT_SPEED_)
      real(PR) :: sqrt_rho_l, sqrt_rho_r
      real(PR) :: u_bar, d_bar, eta_2
      ! Speed of sound in the fluid
      real(PR) :: c_l, c_r
#elif defined(_PHY_EULER_PRESSURE_BASED_SPEED_)
      real(PR) :: rho_bar, c_bar, c_l, c_r, p_pvrs, p_star, p_l, p_r, q_l, q_r
#endif /* _PHY_EULER_WAVE_SPEED_ */

      rho_l = Ul(1)
      u_l(:) = Ul(2:3)/rho_l
      u_l_n = dot_product(u_l,normal)

      rho_r = Ur(1)
      u_r(:) = Ur(2:3)/rho_r
      u_r_n = dot_product(u_r,normal)

#if defined(_PHY_EULER_DAVIS_SPEED_)
      c_l = sound_speed(Ul)
      c_r = sound_speed(Ur)

      !a_l = min_NaN(u_l_n - c_l, u_r_n - c_r)
      !a_r = max_NaN(u_l_n + c_l, u_r_n + c_r)

      !a_l = min_NaN(u_l_n - c_l, min_NaN(u_r_n - c_r, 0._PR))
      !a_r = max_NaN(u_l_n + c_l, max_NaN(u_r_n + c_r, 0._PR))

      a_l = min_NaN(- abs(u_l_n) - c_l, - abs(u_r_n) - c_r)
      a_r = max_NaN(  abs(u_l_n) + c_l,   abs(u_r_n) + c_r)

#elif defined(_PHY_EULER_PRESSURE_BASED_SPEED_)
      c_l = sound_speed(Ul)
      c_r = sound_speed(Ur)

      p_l = compute_pressure(Ul)
      p_r = compute_pressure(Ur)

      rho_bar = 0.5_PR*(rho_l + rho_r)
      c_bar = 0.5_PR*(c_l + c_r)

      p_pvrs = 0.5_PR*((p_l + p_r) - (u_r_n - u_l_n)*rho_bar*c_bar)
      p_star = max(0._PR,p_pvrs)

      if (p_star <= p_l) then
        q_l = 1._PR
      else
        q_l = sqrt(1._PR + (adiabatic_index_p1)/(2._PR*adiabatic_index)*(p_star/p_l - 1._PR))
      end if

      if (p_star <= p_r) then
        q_r = 1._PR
      else
        q_r = sqrt(1._PR + (adiabatic_index_p1)/(2._PR*adiabatic_index)*(p_star/p_r - 1._PR))
      end if

      a_l = u_l_n - c_l*q_l
      a_r = u_r_n + c_r*q_r

      !a_l = - abs(u_l_n) - c_l*q_l
      !a_r =   abs(u_r_n) + c_r*q_r

#elif defined(_PHY_EULER_ROE_SPEED_)
#if defined(_PHY_EULER_IDEAL_GAS_)
      p_l = compute_pressure(Ul)
      p_r = compute_pressure(Ur)
      ! Enthalpy
      H_l = (Ul(4) + p_l)/rho_l
      H_r = (Ur(4) + p_r)/rho_r

      sqrt_rho_l = sqrt(rho_l)
      sqrt_rho_r = sqrt(rho_r)

      H_tilde = (sqrt_rho_l*H_l + sqrt_rho_l*H_r)/(sqrt_rho_l + sqrt_rho_l)
      u_tilde = (sqrt_rho_l*u_l_n + sqrt_rho_l*u_r_n)/(sqrt_rho_l + sqrt_rho_l)
      a_tilde = sqrt((adiabatic_index_m1)*(H_tilde - 0.5_PR*u_tilde**2))

      a_l = u_tilde - a_tilde
      a_r = u_tilde + a_tilde

      !a_l = - abs(u_tilde) - a_tilde
      !a_r =   abs(u_tilde) + a_tilde

#elif defined(_PHY_EULER_ISENTROPIC_)
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Incompatible method defined to compute wave speed."
      FLUSH StdErr
      STOP
#endif /* _PHY_EULER_ENERGY_ */

#elif defined(_PHY_EULER_EINFELDT_SPEED_)
      c_l = sound_speed(Ul)
      c_r = sound_speed(Ur)

      sqrt_rho_l = sqrt(rho_l)
      sqrt_rho_r = sqrt(rho_r)

      eta_2 = 0.5_PR*sqrt_rho_l*sqrt_rho_r/(sqrt_rho_l+sqrt_rho_r)**2

      d_bar = sqrt((sqrt_rho_l*c_l**2 + sqrt_rho_r*c_r**2)/(sqrt_rho_l + sqrt_rho_r) &
          + eta_2*(u_r_n - u_l_n)**2)

      u_bar = (sqrt_rho_l*u_l_n + sqrt_rho_l*u_r_n)/(sqrt_rho_l + sqrt_rho_l)

      a_l = u_bar - d_bar
      a_r = u_bar + d_bar

      !a_l = - abs(u_bar) - d_bar
      !a_r =   abs(u_bar) + d_bar
#else
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> No method defined to compute wave speed."
      FLUSH StdErr
      STOP
#endif /* _PHY_EULER_WAVE_SPEED_ */

  end subroutine compute_wave_speed

#if defined(_FLUX_HLLC_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the wave speeds and intermediary state for the HLLC flux
  !>
  !> @remark see \cite Tor_2009 and \cite TorSprSpe_1994
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent Normal and tangent to the interface
  !> @param[out] a_l, a_star, a_r  HLLC wave speed
  !> @param[out] U_l_star, U_r_star intermediary state
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_pressure
  subroutine compute_HLLC_values(Ul,Ur,normal,tangent,a_l,a_star,a_r,U_l_star,U_r_star)
      real(PR), dimension(:), intent(in) :: Ul, Ur
      real(PR), dimension(:), intent(in) :: normal, tangent
      real(PR), intent(out) :: a_l, a_star, a_r
      real(PR), dimension(:), intent(out) :: U_l_star, U_r_star
#if defined(_PHY_EULER_IDEAL_GAS_)
      real(PR) :: rho_l, rho_r
      real(PR) :: p_l, p_r
      real(PR), dimension(dim_) :: u_l, u_r
      real(PR) :: u_l_n, u_r_n, u_l_t, u_r_t

      real(PR), dimension(dim_,dim_) :: RotationMatrix

      call compute_wave_speed(Ul,Ur,normal,tangent,a_l,a_r)

      rho_l = Ul(1)
      rho_r = Ur(1)

      u_l(1) = Ul(2)
      u_l(2) = Ul(3)
      u_l(:) = u_l(:)/rho_l
      u_r(1) = Ur(2)
      u_r(2) = Ur(3)
      u_r(:) = u_r(:)/rho_r

      u_l_n = dot_product(u_l,normal)
      u_l_t = dot_product(u_l,tangent)

      u_r_n = dot_product(u_r,normal)
      u_r_t = dot_product(u_r,tangent)

      p_l = compute_pressure(Ul)
      p_r = compute_pressure(Ur)

      a_star = ((p_r - p_l) + rho_l*u_l_n*(a_l - u_l_n) - rho_r*u_r_n*(a_r - u_r_n))&
          /(rho_l*(a_l - u_l_n) - rho_r*(a_r - u_r_n))

      RotationMatrix(:,1) = normal(:)
      RotationMatrix(:,2) = tangent(:)

      U_l_star(1) = 1._PR
      U_l_star(2:3) = MATMUL(RotationMatrix,[a_star, u_l_t])
      U_l_star(4) = Ul(4)/rho_l + (a_star - u_l_n)*(a_star + p_l/(rho_l*(a_l - u_l_n)))
      U_l_star(:) = rho_l*(a_l - u_l_n)/(a_l - a_star)*U_l_star(:)

      U_r_star(1) = 1._PR
      U_r_star(2:3) = MATMUL(RotationMatrix,[a_star, u_r_t])
      U_r_star(4) = Ur(4)/rho_r + (a_star - u_r_n)*(a_star + p_r/(rho_r*(a_r - u_r_n)))
      U_r_star(:) = rho_r*(a_r - u_r_n)/(a_r - a_star)*U_r_star(:)

#else
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> HLLC wave speeds not defined for Euler isentropic"
      FLUSH StdErr
      STOP
#endif /* _PHY_EULER_ENERGY_ */

  end subroutine compute_HLLC_values
#endif /* _FLUX_HLLC_ */

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the largest wave speed
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent Normal and tangent to the interface
  !> @return Largest wave speed
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_max_nan
  function compute_greatest_wave_speed(Ul,Ur,normal,tangent)
      real(PR), dimension(:), intent(in) :: Ul, Ur
      real(PR), dimension(:), intent(in) :: normal, tangent

      real(PR) :: compute_greatest_wave_speed

      real(PR) :: a_l, a_r

      call compute_wave_speed(Ul,Ur,normal,tangent,a_l,a_r)

      compute_greatest_wave_speed = max_NaN(abs(a_l),abs(a_r))

  end function compute_greatest_wave_speed

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the sound speed in the fluid
  !>
  !> @param[in] U Conservative variables
  !> @return the sound speed:
  !> \f$c  = \sqrt{\frac{\gamma p}{\rho}}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_pressure
  function sound_speed(U)
      real(PR),  dimension(:), intent(in) :: U
      real(PR) :: sound_speed

      sound_speed = sqrt(adiabatic_index*compute_pressure(U))/sqrt(U(1))

  end function sound_speed

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the physic flux
  !>
  !> @remark
  !> For the classical Euler model:
  !> \f${\bf F}({\bf U}) = \left(\begin{array}{c}(\rho u)^t \\ \rho (p\ Id + u \otimes u) \\ (E+p) u^t \end{array}\right)\f$
  !> and for the isentropic Euler model:
  !> \f${\bf F}({\bf U}) = \left(\begin{array}{c}(\rho u)^t \\ \rho (p\ Id + u \otimes u) \end{array}\right)\f$
  !>
  !> @param[in] U_K solution on cell CellK
  !> @return the physic flux on the cell CellK
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_pressure
  function physic_flux(U_K)
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), dimension(NbEq,dim_) :: physic_flux

      real(PR) :: rho, rho_u, rho_v, p
#if defined(_PHY_EULER_IDEAL_GAS_)
      real(PR) :: E
#endif /* defined(_PHY_EULER_IDEAL_GAS_) */

      rho   = U_K(1)
      rho_u = U_K(2)
      rho_v = U_K(3)
#if defined(_PHY_EULER_IDEAL_GAS_)
      E     = U_K(4)
#endif /* _PHY_EULER_ENERGY_ */
      p = compute_pressure(U_K)

      physic_flux(1,1) = rho_u
      physic_flux(2,1) = rho_u**2/rho + p
      physic_flux(3,1) = rho_u*rho_v/rho
#if defined(_PHY_EULER_IDEAL_GAS_)
      physic_flux(4,1) = rho_u/rho*(E+p)
#endif /* _PHY_EULER_ENERGY_ */

      physic_flux(1,2) = rho_v
      physic_flux(2,2) = rho_u*rho_v/rho
      physic_flux(3,2) = rho_v**2/rho + p
#if defined(_PHY_EULER_IDEAL_GAS_)
      physic_flux(4,2) = rho_v/rho*(E+p)
#endif /* _PHY_EULER_ENERGY_ */

  end function physic_flux

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the pressure
  !>
  !> @param[in] U conservative variables
  !> @return the pressure:
  !>   * for a ideal gas model \f$p  = (\gamma - 1) (E - \frac{\rho (u^2 + v^2)}{2}) \f$
  !>   * for an isentropic model \f$p  = \kappa_p \rho^\gamma \f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function compute_pressure(U)
      real(PR), dimension(:), intent(in) :: U
      real(PR) :: compute_pressure

#if defined(_PHY_EULER_IDEAL_GAS_)
      real(PR) :: rho_e

      rho_e = U(4) - 0.5_PR*dot_product(U(2:3),U(2:3))/U(1)
      compute_pressure = adiabatic_index_m1*rho_e
#elif defined(_PHY_EULER_ISENTROPIC_)
      compute_pressure = kappa_p*U(1)**adiabatic_index
#endif /* _PHY_EULER_ENERGY_ */

  end function compute_pressure

#if defined(_PHY_EULER_ISENTROPIC_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the derivative of the pressure regarding to the density: \f$p'(\rho)\f$
  !>
  !> @param[in] U conservative variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function p_prime_rho(U)
      real(PR), dimension(:), intent(in) :: U
      real(PR) :: p_prime_rho

#if defined(_PHY_EULER_IDEAL_GAS_)
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> function not defined"
      FLUSH StdErr
      STOP
#elif defined(_PHY_EULER_ISENTROPIC_)
      p_prime_rho = adiabatic_index*kappa_p*U(1)**adiabatic_index_m1
#endif /* _PHY_EULER_ENERGY_ */

  end function p_prime_rho
#endif /* _PHY_EULER_ENERGY_ */

  !--------------------------------------------------------------------!
  !> @brief
  !> Convert the primitive to conservative variables
  !>
  !> @remark
  !> For the classical Euler model the primitive variables are:
  !> * \f$ \rho \f$  the density
  !> * \f$ u \f$ the speed
  !> * \f$ p \f$ the pressure
  !>
  !> @remark
  !> For the classical Euler model the conservative variables are:
  !> * \f$ \rho \f$  the density
  !> * \f$ \rho u \f$ the momentum
  !> * \f$ E \f$ the energy
  !>
  !> @param[in] V_K primitive variables
  !> @param[out] U_K conservative variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine primitive_to_conservative(V_K,U_K)
      ! [rho, u, v(, p)] => [rho, rho*u, rho*v(, E)]
      real(PR), dimension(:), intent(in) :: V_K
      real(PR), dimension(:), intent(out) :: U_K

      ! rho
      U_K(1) = V_K(1)
      ! rho*u
      U_K(2) = V_K(2)*V_K(1)
      ! rho*v
      U_K(3) = V_K(3)*V_K(1)
#if defined(_PHY_EULER_IDEAL_GAS_)
      ! E = rho*0.5*(u**2 + v**2) + p/(gamma-1)
      U_K(4) = V_K(1)*0.5_PR*(V_K(2)**2 + V_K(3)**2) + V_K(4)/adiabatic_index_m1
#endif /* _PHY_EULER_ENERGY_ */

  end subroutine primitive_to_conservative

  !--------------------------------------------------------------------!
  !> @brief
  !> Convert the conservative to primitive variables
  !>
  !> @remark
  !> For the classical Euler model the primitive variables are:
  !> * \f$ \rho \f$  the density
  !> * \f$ u \f$ the speed
  !> * \f$ p \f$ the pressure
  !>
  !> @remark
  !> For the classical Euler model the conservative variables are:
  !> * \f$ \rho \f$  the density
  !> * \f$ \rho u \f$ the momentum
  !> * \f$ E \f$ the energy
  !>
  !> @param[in] U_K conservative variables
  !> @param[out] V_K primitive variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine conservative_to_primitive(U_K,V_K)
      ! [rho, rho*u, rho*v(, E)] =>  [rho, u, v(, p)]
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), dimension(:), intent(out) :: V_K

      ! rho
      V_K(1) = U_K(1)
      ! u
      V_K(2) = U_K(2)/U_K(1)
      ! v
      V_K(3) = U_K(3)/U_K(1)
#if defined(_PHY_EULER_IDEAL_GAS_)
      ! p
      V_K(4) = compute_pressure(U_K)
#endif /* _PHY_EULER_ENERGY_ */

  end subroutine conservative_to_primitive

  !--------------------------------------------------------------------!
  !> @brief
  !> Check if the value of the solution on a cell is physically admissible.
  !>
  !> @param[in] U_K conservative variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function IsPhysicallyAdmissible(U_K)
      real(PR), dimension(:), intent(in) :: U_K

      logical :: IsPhysicallyAdmissible

      if (U_K(1) < 0._PR) then
        IsPhysicallyAdmissible = .FALSE.
      else if (compute_pressure(U_K) < 0._PR) then
        IsPhysicallyAdmissible = .FALSE.
      else
        IsPhysicallyAdmissible = .TRUE.
      end if

  end function IsPhysicallyAdmissible

#endif /* _PHY_EULER_ */
end module mod_physics_Euler
