!> Routines for the entropy for Euler \cite BerDes_2014
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.

module mod_physics_Euler_entropy
#if defined(_PHY_EULER_) && defined(_HO_MOOD_ENTROPIC_)

  use mod_numerics, only: PR, eps
  use type_mesh, only: dim_
  use mod_physics_Euler, only: compute_pressure
  use mod_physics_Euler_var, only: adiabatic_index, NbEq

  IMPLICIT NONE

  private
  public compute_entropy_phy
  public compute_entropy_math
  public compute_entropy_r
  public entropy_flux

CONTAINS
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the physical entropy: \f$s = \frac{p}{\rho^{\gamma}}\f$
  !>
  !> @param[in] U_K Mean values
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_pressure
  function compute_entropy_phy(U_K)
      real(PR), dimension(:), intent(in) :: U_K
      real(PR) :: compute_entropy_phy

      compute_entropy_phy = compute_pressure(U_K)/(U_K(1)**adiabatic_index)

  end function compute_entropy_phy

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute \f$\mathcal{F}(ln(s))\f$
  !>
  !> With \f$\mathcal{F}\f$ such that: \f$\mathcal{F}'(y) < 0\f$
  !> and \f$\mathcal{F}'(y) < \gamma \mathcal{F}''(y)\f$
  !>
  !> @param[in] entropy_phy Physical entropy
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function F_ln(entropy_phy)
      real(PR), intent(in) :: entropy_phy
      real(PR) :: F_ln

      !F_ln = exp(- log(entropy_phy))
      F_ln = 1._PR/entropy_phy
  end function F_ln

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the mathematical entropy: \f$\eta = \rho \mathcal{F}(\ln(s))\f$
  !>
  !> Where \f$s\f$ is the physical entropy and \f$\mathcal{F}\f$ such that
  !> \f$\mathcal{F}'(y) < 0\f$ and \f$\mathcal{F}'(y) < \gamma \mathcal{F}''(y)\f$
  !> @param[in] U_K Mean values
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_entropy_mp_f_ln
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_entropy_mp_compute_entropy_phy
  function compute_entropy_math(U_K)
      real(PR), dimension(:), intent(in) :: U_K
      real(PR) :: compute_entropy_math

      compute_entropy_math = U_K(1)*F_ln(compute_entropy_phy(U_K))

  end function compute_entropy_math

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the "r" entropy: \f$r = -s^{1/\gamma} = - \frac{p^{1/\gamma}}{\rho}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_mp_compute_pressure
  function compute_entropy_r(U_K)
      real(PR), dimension(:), intent(in) :: U_K
      real(PR) :: compute_entropy_r

      compute_entropy_r = - compute_pressure(U_K)**(1._PR/adiabatic_index)/U_K(1)

  end function compute_entropy_r

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the entropy flux: \f$G(U) = \rho u \mathcal{F}(\ln(s))\f$
  !>
  !> @param[in] U_K Mean values
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_entropy_mp_f_ln
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_euler_entropy_mp_compute_entropy_phy
  function entropy_flux(U_K)
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), dimension(dim_) :: entropy_flux

      entropy_flux(:) = U_K(2:3)*F_ln(compute_entropy_phy(U_K))

  end function entropy_flux

#endif /* _PHY_EULER_ && _HO_MOOD_ENTROPIC_ */
end module mod_physics_Euler_entropy
