!> Physics of the Euler model with ideal gas and friction \cite Tor_2009 .
!>
!> \f[\left\{\begin{array}{rcl}
!>   \partial_t \rho   + \partial_x \rho u               + \partial_y \rho v               & = & 0 \\
!>   \partial_t \rho u + \partial_x (\rho u^2 + p(\rho)) + \partial_y \rho uv              & = & - \kappa \rho u \\
!>   \partial_t \rho v + \partial_x \rho uv              + \partial_y (\rho v^2 + p(\rho)) & = & - \kappa \rho v  \\
!>   \partial_t E      + \partial_x u (E+p)              + \partial_y v (E +p)             & = & 0
!> \end{array}\right.\f]
!> with:
!> * \f$ \rho \f$: density of the fluid,
!> * \f$ \vec{V} =(u,v) \f$: speed of the fluid,
!> * \f$ E = \rho(\frac{u^2 + v^2}{2} + e) \f$: total energy per unit volume, \f$e\f$ specific internal energy,
!> * \f$ \kappa > 0 \f$,
!> * the equation of state (EOS): \f$e=\frac{p}{\rho(\gamma-1)}\f$, \f$p'(\rho) > 0\f$,
!> * and the speed of the sound in the fluid is defined as: \f$c = \sqrt{\frac{\gamma p}{\rho}}\f$.
!>
!> In test cases with source terms we are using the isentropic Euler model:
!> \f[\left\{\begin{array}{rcl}
!>   \partial_t \rho   + \partial_x \rho u               + \partial_y \rho v               & = & 0 \\
!>   \partial_t \rho u + \partial_x (\rho u^2 + p(\rho)) + \partial_y \rho uv              & = & - \kappa \rho u \\
!>   \partial_t \rho v + \partial_x \rho uv              + \partial_y (\rho v^2 + p(\rho)) & = & - \kappa \rho v  \\
!> \end{array}\right.\f]
!> with:
!> * the equation of state (EOS): \f$p=\kappa_p\rho^{\gamma}\f$, \f$p'(\rho) = \kappa_p > 0\f$,
!>
!> @remarks See \cite MarMil_1990 and \cite Vaz_1987 for the asymptotic behaviour.
!> @remarks See \cite KurTad_2002 and \cite Des_2013 for the 2D test cases.
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_physics_Euler_var
#if defined(_PHY_EULER_)

  use mod_numerics, only: PR, MaxLen

  IMPLICIT NONE

  private
  public NbEq, NbEq_str,  T_end
  public len_model, len_primitive, len_conservative
  public model_name, primitive_name, conservative_name, test_case

  public adiabatic_index, adiabatic_index_m1, adiabatic_index_p1
  public kappa, kappa_p

  integer, parameter                  :: len_model = 5

  !> Name of the model
  character(len=len_model), parameter :: model_name = 'Euler'

  !> Number of equations (unknowns) in the model
#if defined(_PHY_EULER_IDEAL_GAS_)
  integer, parameter                  :: NbEq = 4
#elif defined(_PHY_EULER_ISENTROPIC_)
  integer, parameter                  :: NbEq = 3
#endif /* _PHY_EULER_ENERGY_ */

  !> Number of equations (unknowns) in the model, in a string
#if defined(_PHY_EULER_IDEAL_GAS_)
  character(len=1), parameter :: NbEq_str = '4'
#elif defined(_PHY_EULER_ISENTROPIC_)
  character(len=1), parameter :: NbEq_str = '3'
#endif /* _PHY_EULER_ENERGY_ */


  !> \f$\gamma\f$: adiabatic index, for diatomic gazes \f$\gamma = 1.4\f$
  !> and for a more general gazes \f$\gamma \in ]1;3]\f$
  real(PR), parameter :: adiabatic_index = 1.4_PR
  !> \f$\gamma - 1\f$
  real(PR), parameter :: adiabatic_index_m1 = adiabatic_index - 1._PR
  !> \f$\gamma + 1\f$
  real(PR), parameter :: adiabatic_index_p1 = adiabatic_index + 1._PR

  !> \f$\kappa > 0\f$ (\f$s^{-1}\f$): coefficient of friction
  real(PR), parameter :: kappa = 2000._PR

  !> \f$\kappa_p > 0\f$: coefficient
  !> in the computation of the pressure for an isentropic model
  !> \f$p  = \kappa_p \rho^\gamma \f$
  real(PR), parameter :: kappa_p = 1._PR

  !> Final time of the simulation
  real(PR), save :: T_end

  integer, parameter :: len_primitive    = 3
  integer, parameter :: len_conservative = 5

  !> Primitive variables used in the model
#if defined(_PHY_EULER_IDEAL_GAS_)
  character(len=len_primitive), dimension(NbEq), parameter :: primitive_name = ['rho', 'u  ', 'v  ', 'p  ']
#elif defined(_PHY_EULER_ISENTROPIC_)
  character(len=len_primitive), dimension(NbEq), parameter :: primitive_name = ['rho', 'u  ', 'v  ']
#endif /* _PHY_EULER_ENERGY_ */

  !> Conservative variables used in the model
#if defined(_PHY_EULER_IDEAL_GAS_)
  character(len=len_conservative), dimension(NbEq), parameter :: conservative_name = ['rho  ', 'rho_u', 'rho_v', 'E    ']
#elif defined(_PHY_EULER_ISENTROPIC_)
  character(len=len_conservative), dimension(NbEq), parameter :: conservative_name = ['rho  ', 'rho_u', 'rho_v']
#endif /* _PHY_EULER_ENERGY_ */

  !> Name of the test case
  character(len=MaxLen), save :: test_case

#endif /* _PHY_EULER_ */
end module mod_physics_Euler_var
