!> Physics of the radiative hydrodynamic model \cite Pom_1973 \cite Tur_2003 .
!>
!> \f[
!>   \left\{
!>     \begin{array}{rcl}
!>       \partial_t \rho         + \text{div}(\rho {\bf u})                                      & = &  0 \\
!>       \partial_t \rho {\bf u} + \text{div}(\rho {\bf u}\otimes {\bf u}+ {\bf P}_R) + \nabla p & = &  \frac{1}{c} \sigma^f {\bf F}_R \\
!>       \partial_t E            + \text{div}((E+p){\bf u})                                      & = &  c \left(\sigma^a E_R - \sigma^e a T^4 \right)\\
!>       \partial_t E_R          + \text{div}({\bf F}_R)                                         & = &  c \left(\sigma^e a T^4 - \sigma^a E_R \right)\\
!>       \partial_t {\bf F}_R    + c^2 \text{div}(({\bf P}_R)                                    & = & -c \sigma^f {\bf F}_R
!>     \end{array}
!>   \right..
!> \f]
!>
!> with:
!> * \f$ \rho \f$: density of the fluid,
!> * \f$ {\bf u} \f$: speed of the fluid,
!> * \f$ E = \rho(\frac{u^2}{2} + e) \f$: total energy per unit volume, \f$e\f$ specific internal energy,
!> * \f$ \kappa > 0 \f$,
!> * the equation of state (EOS): \f$e=\frac{p}{\rho(\gamma-1)}\f$,
!> * the speed of the sound in the fluid is defined as: \f$c = \sqrt{\frac{\gamma p}{\rho}}\f$,
!> * \f$ E_R \f$ the radiative energy,
!> * \f$ {\bf F}_R \f$ the radiative flux,
!> * \f$ T \f$ the temperature,
!> * \f$ {\bf P}_R = \frac{E_R}{2}\left( (1-\chi)  + (3\chi-1)\frac{{\bf F}_R \otimes {\bf F}}{\Vert {\bf F} \Vert_2^2}\right) \f$ the radiative pressure,
!> * \f$\chi = \frac{3 + 4f^2}{5+ \sqrt{4-3f^2}} \f$ the Eddington factor, in order that \f$\chi\f$ is well defined:
!>     \f$ f=\frac{\Vert {\bf F} \Vert}{c E} \leq 1 \f$,
!> * \f$ c \f$ the speed of light.
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.

module mod_physics_HydroRad_var
#if defined(_PHY_HYDRORAD_)

  use mod_numerics, only: PR, pi, MaxLen

  IMPLICIT NONE

  private
  public NbEq, NbEq_str, T_end
  public len_model, len_primitive, len_conservative
  public model_name, primitive_name, conservative_name, test_case

  public adiabatic_index, adiabatic_index_m1, adiabatic_index_p1
  public molar_mass
  public c, sigma, sigma_a, sigma_f, sigma_e, a, ideal_gas_constant
#if defined(_SRC_ENABLE_)
  public C_v
#endif /* _SRC_ENABLE_ */

  integer, parameter :: len_model = 9

  !> Name of the model
  character(len=len_model), parameter :: model_name = 'HydroRad'

  !> Number of equations (unknowns) in the model
  integer, parameter                  :: NbEq = 7

  !> Number of equations (unknowns) in the model, in a string
  character(len=1), parameter :: NbEq_str = '7'

  !> \f$\gamma\f$: adiabatic index, for diatomic gazes \f$\gamma = 1.4\f$
  !> and for a more general gazes \f$\gamma \in ]1;3]\f$
  real(PR), parameter :: adiabatic_index = 1.4_PR ! hydrogen
  !real(PR), parameter :: adiabatic_index = 5._PR/3._PR ! helium
  !> \f$\gamma - 1\f$
  real(PR), parameter :: adiabatic_index_m1 = adiabatic_index - 1._PR
  !> \f$\gamma + 1\f$
  real(PR), parameter :: adiabatic_index_p1 = adiabatic_index + 1._PR

  !> Speed of light (\f$c \approx 299\ 792\ 458\ m.s^{-1}\f$)
  real(PR), parameter :: c = 299792458._PR

  !> Gas constant (\f$R \approx 3.314\ 462\ 1 \ J.mol^{-1}.K^{-1}\f$)
  real(PR), parameter :: ideal_gas_constant = 8.3144621_PR

  !> Avogadro constant (\f$N_A \approx 6.022\ 141\ 29\ \times 10^{23}\ mol^{-1}\f$)
  real(PR), parameter :: N_A = 6.02214129E+23_PR

  !> Boltzmann constant (\f$k_B = \frac{R}{N_A} \approx 1.380\ 648\ 8 \times 10^{-23}\ J.K^{-1}\f$)
  real(PR), parameter :: k_B = ideal_gas_constant/N_A

  !> Planck constant (\f$h \approx 6.626\ 069\ 57 \times10^{-34}\ J.s^{-1}\f$)
  real(PR), parameter :: h = 6.62606957E-34_PR

  !> Reduced Planck constant (\f$\bar{h}=\frac{h}{2\pi}\ J.s^{-1}\f$)
  real(PR), parameter :: h_bar = h/(2._PR*pi)

  !> Stefan-Boltzmann constant (\f$\sigma =\frac{2\pi^5R^4}{15h^3c^2N_A^4} \approx 5.670\ 373\ 21\times10^{-8}\ W.m^{-2}.K^{-4}\f$)
  real(PR), parameter :: sigma = (2._PR*pi**5*ideal_gas_constant**4)/(15._PR*h**3*c**2*N_A**4)

  !> Radiation constant (or radiation density constant) \f$a = \frac{4\sigma}{c} \approx 7.5657 \times10^{-16}\ J.m^{-3}.K^{-4}\f$)
  real(PR), parameter :: a = 4._PR*sigma/c

  !> Molar mass of the fluid (\f$ kg.m^{-3} \f$)
  real(PR), parameter :: molar_mass = 1.E-3_PR ! hydrogen
  !real(PR), parameter :: molar_mass = 4.E-3_PR ! helium
  !real(PR), parameter :: molar_mass = 30.E-3_PR ! air

#if defined(_SRC_ENABLE_)
  !> Volumic heat capacity (\f$J.K^{-1}\f$)
  real(PR) , parameter:: C_v = ideal_gas_constant/adiabatic_index_m1
#endif /* _SRC_ENABLE_ */

  !> Mean opacity of absorption
  real(PR), parameter :: sigma_a = 3.1E-8_PR
  !> Mean opacity of absorption of flux
  real(PR), parameter :: sigma_f = 3.1E-8_PR
  !> Mean opacity of emission
  real(PR), parameter :: sigma_e = 3.1E-8_PR

  !> Final time of the simulation
  real(PR), save :: T_end

  integer, parameter :: len_primitive    = 4
  integer, parameter :: len_conservative = 5

  !> Primitive variables used in the model
  character(len=len_primitive), dimension(NbEq), parameter :: primitive_name = ['rho ', 'u   ', 'v   ', 'p   ','E_R ', 'F_Rx', 'F_Ry']

  !> Conservative variables used in the model
  character(len=len_conservative), dimension(NbEq), parameter :: conservative_name = ['rho  ', 'rho_u', 'rho_v', 'E    ','E_R  ', 'F_Rx ', 'F_Ry ']

  !> Name of the test case
  character(len=MaxLen), save :: test_case

#endif /* _PHY_HYDRORAD_ */
end module mod_physics_HydroRad_var
