!> Physics of the multigroup M1 model for radiative transfer.
!>
!> \f[\left\{\begin{array}{rcl}
!>   \partial_t E_q       +     \text{div} ({\bf F}_q) & = & c (\sigma^e_q a \theta_q^4(T) - \sigma^a_q E_q) \\
!>   \partial_t {\bf F}_q + c^2 \text{div} ({\bf P_q}) & = & -c \sigma^f_q {\bf F}_q \\
!>   \rho C_v \partial_t T                             & = & -c \sum_{q=1}^Q (\sigma^e_q a \theta_q^4(T) - \sigma^a_q E_q)
!> \end{array}\right.\f]
!>
!> where the index \f$q\f$ defines each group.
!>
!> with :
!> * \f$ E_q \f$ the radiative energy,
!> * \f$ {\bf F}_q \f$ the radiative flux,
!> * \f$ T \f$ the temperature,
!> * \f$ {\bf P}_q = \frac{E_q}{2}\left( (1-\chi_q) {\bf I}  + (3\chi_q-1)\frac{{\bf F}_q\otimes {\bf F}_q}{\Vert {\bf F}_q \Vert_2^2}\right) \f$ the radiative pressure,
!> * \f$\chi_q\f$ is the Eddington factor and it is no more explicit as in the grey M1 model, but as a fist approxamtion we use it:
!>     \f$\chi_q = \frac{3 + 4f_q^2}{5+ \sqrt{4-3f_q^2}}\f$
!> * \f$f_q\f$ is the anistropy factor: \f$f_q=\frac{\Vert {\bf F}_q \Vert}{c E_q} \leq 1 \f$,
!> * \f$ c \f$ the speed of light.
!>
!> @remark See \cite Tur_2002 \cite Tur_2003 \cite Sar_2013 for the definition.
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_physics_M1MG
#if defined(_PHY_M1MG_)

  use mod_numerics, only: PR, pi, eps
  use mod_numerics, only: zero, third, one
  use mod_numerics, only: sqrt_2, sqrt_3
  use type_IO, only: StdOut, StdErr
  use mod_tools, only: SolvePolynomial2
  use mod_tools, only: min_NaN, max_NaN
  use mod_tools_array, only: FindIndex
  use mod_tools_array, only: tensor_product
  use mod_tools_interp, only: TriLinearInterp
  use type_mesh, only: pmesh, dim_
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells
#endif /* _FC_PAR_OMP_ */
  use mod_read, only: read_solution
  use mod_read, only: Restarting

  use mod_physics_M1MG_var, only: NbEq, test_case
  use mod_physics_M1MG_var, only: T_end
  use mod_physics_M1MG_var, only: c, a, h, k_B
  use mod_physics_M1MG_var, only: NbGroup
  use mod_physics_M1MG_var, only: frequencies
#if defined(_SRC_ENABLE_)
  use mod_physics_M1MG_var, only: rho, C_v
#endif /* _SRC_ENABLE_ */
  use mod_physics_M1MG_var, only: tab_f
  use mod_physics_M1MG_var, only: tab_T
  use mod_physics_M1MG_var, only: tab_T_R
  use mod_physics_M1MG_var, only: tab_chi

#if defined(_FC_PGI_)
  use mod_tools_array, only: NORM2
#endif /* defined(_FC_PGI_) */

  IMPLICIT NONE

  private
  public physic_flux, U_Dirichlet
  public compute_wave_speed, compute_greatest_wave_speed
  public init_U
  public primitive_to_conservative, conservative_to_primitive
  public IsPhysicallyAdmissible

  public compute_f
  public compute_a_theta_q_4
  public compute_dT_a_theta_q_4

CONTAINS
  !--------------------------------------------------------------------!
  !> @brief
  !> Give the initial condition according to the test case
  !>
  !> @param[out] U0 Initial condition
  !> @param[out] iter Number of iteration previously done
  !> @param[out] T_init Initial time
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine init_U(U0,iter,T_init)
    real(PR), dimension(:,:), intent(out) :: U0
    integer , intent(out) :: iter
    real(PR), intent(out) :: T_init

    real(PR), dimension(dim_) :: xK
    integer :: CellK
    integer :: group, eq0

    !$OMP SINGLE
    select case(test_case)
    case ('cst')
      T_end = 1.0_PR/c
    case ('tophat')
      T_end = 8.14E-8_PR
      !T_end = 5.70E-6_PR
    case ('square_hole')
      T_end = 3.17E-8_PR
      !T_end = 1.58E-6_PR
    case ('x_1', 'x_1_group')
      T_end = 0.5_PR/c
    case default
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
      FLUSH StdErr
      STOP
    end select
    !$OMP END SINGLE NOWAIT

    test_restart: if (Restarting) then
      !$OMP SINGLE
      call read_solution(U0,iter,T_init)
      !$OMP END SINGLE

    else
      !$OMP SINGLE
      iter = 0
      T_init = 0._PR
      !$OMP END SINGLE NOWAIT

      select case(test_case)
      case ('cst')
        !$OMP DO &
        !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
        do CellK = 1, pmesh%NbCells
          U0(NbEq,CellK) = 6.E4_PR
          do group = 1, NbGroup
            eq0 = (1 + dim_)*(group-1)
            U0(eq0+1,CellK) = compute_a_theta_q_4(group,U0(NbEq,CellK))
            U0(eq0+2,CellK) = 0.0_PR
            U0(eq0+3,CellK) = 0.0_PR
          end do
        end do
        !$OMP END DO

      case ('tophat')
        ! See \cite Sar_2013 for the definition of this test case
        !$OMP DO &
        !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
        do CellK = 1, pmesh%NbCells
          U0(NbEq,CellK) = 6.E4_PR
          do group = 1, NbGroup
            eq0 = (1 + dim_)*(group-1)
            U0(eq0+1,CellK) = compute_a_theta_q_4(group,U0(NbEq,CellK))
            U0(eq0+2,CellK) = 0.0_PR
            U0(eq0+3,CellK) = 0.0_PR
          end do
        end do
        !$OMP END DO

      case ('square_hole')
        ! See \cite Sar_2013 for the definition of this test case
        !$OMP DO &
        !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
        do CellK = 1, pmesh%NbCells
          U0(NbEq,CellK) = 3.E4_PR
          do group = 1, NbGroup
            eq0 = (1 + dim_)*(group-1)
            U0(eq0+1,CellK) = compute_a_theta_q_4(group,U0(NbEq,CellK))
            U0(eq0+2,CellK) = 0.0_PR
            U0(eq0+3,CellK) = 0.0_PR
          end do
        end do
        !$OMP END DO

      case ('x_1')
        !$OMP DO &
        !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
        do CellK = 1, pmesh%NbCells
          xK(:) = pmesh%elt(CellK)%x(:)
          U0(NbEq,CellK) = 300._PR
          do group = 1, NbGroup
            eq0 = (1 + dim_)*(group-1)
            if (xK(1) < 0.5_PR) then
              U0(eq0+1,CellK) = 1._PR
              U0(eq0+2,CellK) = c*0.1_PR
              U0(eq0+3,CellK) = 0._PR
            else
              U0(eq0+1,CellK) = 10._PR
              U0(eq0+2,CellK) = c*(-1._PR)
              U0(eq0+3,CellK) = 0._PR
            end if
          end do
        end do
        !$OMP END DO

      case ('x_1_group')
        !$OMP DO &
        !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
        do CellK = 1, pmesh%NbCells
          xK(:) = pmesh%elt(CellK)%x(:)
          U0(NbEq,CellK) = 6.E4_PR
          do group = 1, NbGroup
            eq0 = (1 + dim_)*(group-1)
            if (xK(1) < 0.5_PR) then
              U0(eq0+1,CellK) = 1._PR*compute_a_theta_q_4(group,U0(NbEq,CellK))
              U0(eq0+2,CellK) = c*0.1_PR*U0(eq0+1,CellK)
              U0(eq0+3,CellK) = 0._PR
            else
              U0(eq0+1,CellK) = 10._PR*compute_a_theta_q_4(group,U0(NbEq,CellK))
              U0(eq0+2,CellK) = c*(-0.1_PR)*U0(eq0+1,CellK)
              U0(eq0+3,CellK) = 0._PR
            end if
          end do
        end do
        !$OMP END DO
      end select

    end if test_restart

#if defined(_SRC_ENABLE_)
    !$OMP SINGLE
    allocate(rho(pmesh%NbCells))
    allocate(C_v(pmesh%NbCells))
    !$OMP END SINGLE

    !$OMP DO &
    !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
    do CellK = 1, pmesh%NbCells
      rho(CellK) = 1._PR
      C_v(CellK) = 1.E-3_PR
    end do
    !$OMP END DO
#endif /* _SRC_ENABLE_ */


  end subroutine init_U

  !--------------------------------------------------------------------!
  !> @brief
  !> Give the limit boundary value
  !>
  !> @param[in] x Coordinates of the point on the boundary
  !> @param[in] t_n Current time
  !>
  !> @return Value of the solution on the boundary at point
  !> \f$x\f$ and time \f$t^n\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function U_Dirichlet(x,t_n)
    real(PR), dimension(:), intent(in) :: x
    real(PR), intent(in) :: t_n
    real(PR), dimension(NbEq) :: U_Dirichlet

    integer :: group, eq0

    select case(test_case)
    case ('cst')
      U_Dirichlet(NbEq) = 6.E4_PR
      do group = 1, NbGroup
        eq0 = (1 + dim_)*(group-1)
        U_Dirichlet(eq0+1) = compute_a_theta_q_4(group,U_Dirichlet(NbEq))
        U_Dirichlet(eq0+2) = 0._PR
        U_Dirichlet(eq0+3) = 0._PR
      end do

    case ('tophat')
      U_Dirichlet(NbEq) = 1.E5_PR
      do group = 1, NbGroup
        eq0 = (1 + dim_)*(group-1)
        U_Dirichlet(eq0+1) = compute_a_theta_q_4(group,U_Dirichlet(NbEq))
        U_Dirichlet(eq0+2) = 0.8_PR*c*U_Dirichlet(eq0+1)
        U_Dirichlet(eq0+3) = 0._PR
      end do

    case ('square_hole')
      U_Dirichlet(NbEq) = 9.E4_PR
      do group = 1, NbGroup
        eq0 = (1 + dim_)*(group-1)
        U_Dirichlet(eq0+1) = compute_a_theta_q_4(group,U_Dirichlet(NbEq))
        U_Dirichlet(eq0+2) = 0.8_PR*c*U_Dirichlet(eq0+1)
        U_Dirichlet(eq0+3) = 0._PR
      end do

    case default
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
      FLUSH StdErr
      STOP
    end select

  end function U_Dirichlet

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the wave speeds waves on a interface
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent Normal and tangent to the interface
  !> @param[out] a_l, a_r  Wave speeds on the left and right
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine compute_wave_speed(Ul,Ur,normal,tangent,a_l,a_r)
    real(PR), dimension(:), intent(in) :: Ul, Ur
    real(PR), dimension(:), intent(in) :: normal, tangent
    real(PR), intent(out) :: a_l, a_r

    a_l = -c
    a_r = c

  end subroutine compute_wave_speed

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the largest wave speed
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent Normal and tangent to the interface
  !> @return Largest wave speed
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function compute_greatest_wave_speed(Ul,Ur,normal,tangent)
    real(PR), dimension(:), intent(in) :: Ul, Ur
    real(PR), dimension(:), intent(in) :: normal, tangent

    real(PR) :: compute_greatest_wave_speed

    compute_greatest_wave_speed = c

  end function compute_greatest_wave_speed

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the physic flux
  !>
  !> @remark
  !> For the M1 model:
  !> \f${\bf F}({\bf U}) = \left(\begin{array}{c}F^t \\ c^2 P \\ 0\end{array}\right)\f$
  !>
  !> @param[in] U_K solution on cell CellK
  !> @return the physic flux on the cell CellK
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_radiative_pressure
  function physic_flux(U_K)
    real(PR), dimension(:), intent(in) :: U_K
    real(PR), dimension(NbEq,dim_)     :: physic_flux

    real(PR) :: E, T
    real(PR), dimension(dim_,dim_) :: c2_P
    real(PR), dimension(dim_) :: F

    integer :: group, eq0

    T = U_K(NbEq)

    loop_groups: do group = 1, NbGroup
      eq0 = (1 + dim_)*(group-1)

      E = U_K(eq0+1)
      F(1) = U_K(eq0+2)
      F(2) = U_K(eq0+3)

      c2_P(:,:) = c**2*radiative_pressure(group,E,F,T)

      physic_flux(eq0+1,1) = F(1)
      physic_flux(eq0+2,1) = c2_P(1,1)
      physic_flux(eq0+3,1) = c2_P(2,1)

      physic_flux(eq0+1,2) = F(2)
      physic_flux(eq0+2,2) = c2_P(1,2)
      physic_flux(eq0+3,2) = c2_P(2,2)
    end do loop_groups

    ! Physic flux for the last equation (temperature)
    physic_flux(NbEq,:) = zero

  end function physic_flux

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the radiative pressure
  !>
  !> @param[in] group Index of the group
  !> @param[in] E radiative energy
  !> @param[in] F radiative flux
  !> @param[in] T Temperature
  !> @return the radiative pressure:
  !> \f$P = \frac{E}{2}\left( (1-\chi)Id  + (3\chi-1)\frac{F \otimes F}{\Vert F \Vert_2^2}\right) \f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_chi
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_array_mp_tensor_product
  function radiative_pressure(group,E,F,T) result(P)
    integer, intent(in) :: group
    real(PR), intent(in)                  :: E, T
    real(PR), dimension(:), intent(in) :: F

    real(PR), dimension(dim_,dim_) :: P

    real(PR) :: chi, NORM2_F
    integer  :: i

    NORM2_F = NORM2(F)

    if (NORM2_F < eps) then
      P(:,:) = zero
      do i = 1, dim_
        P(i,i) = third
      end do

    else
      chi = compute_chi(group,E,F,T)

      P(:,:) = 0.5_PR*(3._PR*chi - 1._PR)*tensor_product(F,F)/NORM2_F**2

      do i = 1, dim_
        P(i,i) = P(i,i) + 0.5_PR*(1._PR - chi)
      end do

    end if

    P(:,:) = E*P(:,:)

  end function radiative_pressure

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the Eddington factor
  !>
  !> @param[in] group Index of the group
  !> @param[in] E radiative energy
  !> @param[in] F radiative flux
  !> @param[in] T Temperature
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_array_mp_findindex
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_interp_mp_trilinearinterp
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_compute_f
  function compute_chi(group,E,F,T)
    integer, intent(in) :: group
    real(PR), intent(in)                  :: E, T
    real(PR), dimension(:), intent(in) :: F
    real(PR) :: compute_chi

    real(PR) :: ff, T_R
    integer :: idx_f, idx_T, idx_T_R

    ff = compute_f(E,F)

    ! Value from the grey model
    !compute_chi = (3._PR + 4._PR*ff**2)/(5._PR + 2._PR*sqrt(4._PR - 3._PR*ff**2))
    !return

    T_R = sqrt(sqrt(E/a))

    idx_f = FindIndex(ff,tab_f)
    idx_T = FindIndex(T,tab_T)
    idx_T_R = FindIndex(T_R,tab_T_R)

#if defined(_TP_CORREC_) || ((_HO_MOOD_DEGREE_) >=1)
    ! The values are not valid, generate a NaN to limit
    if (idx_f < 0) then
      compute_chi = sqrt(1._PR - abs(ff))
      return
    else if (idx_T < 0) then
      compute_chi = sqrt(T)
      return
    else if (idx_T_R < 0) then
      compute_chi = sqrt(E)
      return
    end if
#endif /* defined(_TP_CORREC_) || ((_HO_MOOD_DEGREE_) >=1) */

    ! Linear interpolation
    compute_chi = TriLinearInterp(tab_f(idx_f),tab_f(idx_f+1),&
        tab_T_R(idx_T_R), tab_T_R(idx_T_R+1),&
        tab_T(idx_T),tab_T(idx_T+1),&
        tab_chi(idx_f,idx_T_R,idx_T,group),&
        tab_chi(idx_f+1,idx_T_R,idx_T,group),&
        tab_chi(idx_f+1,idx_T_R+1,idx_T,group), &
        tab_chi(idx_f,idx_T_R+1,idx_T,group),&
        tab_chi(idx_f,idx_T_R,idx_T+1,group),&
        tab_chi(idx_f+1,idx_T_R,idx_T+1,group),&
        tab_chi(idx_f+1,idx_T_R+1,idx_T+1,group),&
        tab_chi(idx_f,idx_T_R+1,idx_T+1,group),&
        ff,T_R,T)

    ! In case of bad interpolation, limit chi
    if (compute_chi < 1._PR/3._PR) then
      compute_chi = 1._PR/3._PR
    else if (compute_chi > 1._PR) then
      compute_chi = 1._PR
    endif

  end function compute_chi

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the anistropy factor
  !>
  !> \f[ f = \frac{|| {\bf F} ||}{c E} \f]
  !>
  !> @param[in] E Radiative energy
  !> @param[in] F Radiative flux
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function compute_f(E,F)
    real(PR), intent(in) :: E
    real(PR), dimension(:), intent(in) :: F
    real(PR) :: compute_f

    if (.not. (abs(E) < eps)) then
      compute_f = NORM2(F)/(c*E)
    else
      compute_f = 0._PR
    endif

  end function compute_f

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute \f$ a \theta_q^4(T) = {<B_{\nu}(T)>}_q\f$
  !>
  !> with \f$ <.> = \frac{1}{c} \int_{S^2} \int_0^\infty . d\nu d\Omega\f$
  !> and \f$B_{\nu}(T) = \frac{2 h \nu^3}{c^2} {\left[\exp\left(\frac{h \nu}{k_B T}\right) - 1 \right]}^{-1}\f$
  !>
  !> The computations of \cite Tur_2003 and \cite Sar_2013 leads to \f$ a \theta_q^4(T) = \frac{15 a T^4}{\pi^4} \left(\Xi(z_{q+1}) - \Xi(z_q)\right) \f$.
  !>
  !> See mod_physics_m1mg::xi for the approximations of \f$\Xi(z) = \int_1^z \frac{x^3}{\exp(x)-1}dx \f$ by \f$\xi(z)\f$
  !>
  !> @param[in] q Index of the group
  !> @param[in] T Temperature
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_xi
  function compute_a_theta_q_4(q,T) result(a_theta_q_4)
    integer, intent(in) :: q
    real(PR), intent(in) :: T
    real(PR) :: a_theta_q_4

    real(PR) :: z_q, z_qp1

    z_q   = (h*frequencies(q))/(k_B*T)
    z_qp1 = (h*frequencies(q+1))/(k_B*T)

    a_theta_q_4 = 15._PR*a*T**4/pi**4*(xi(z_qp1) - xi(z_q))

  end function compute_a_theta_q_4

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute \f$ \partial_T a \theta_q^4(T) \f$
  !>
  !> \f[\partial_T a \theta_q^4(T) = \frac{60aT^3}{\pi^4}\left(\Xi(z_{q+1}) - \Xi(z_q)\right)
  !>     + \frac{15aT^3}{\pi^4}\left(\frac{-z_{q+1}^4}{e^{z_{q+1}}-1} + \frac{z_{q}^4}{e^{z_{q}}-1}\right) \f]
  !>
  !> As \f$ \partial_T \Xi(z(T)) = \partial_T z(T) \partial_z \Xi(z) = -\frac{h\nu}{k_B T^2} \frac{z^3}{e^z-1}
  !>                             = - \frac{1}{T}\frac{z^4}{e^z-1}\f$
  !>
  !> See mod_physics_m1mg::xi for the approximations of \f$\Xi(z) = \int_1^z \frac{x^3}{\exp(x)-1}dx \f$ by \f$\xi(z)\f$
  !>
  !> @param[in] q Index of the group
  !> @param[in] T Temperature
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1mg_mp_xi
  function compute_dT_a_theta_q_4(q,T) result(dT_a_theta_q_4)
    integer, intent(in) :: q
    real(PR), intent(in) :: T
    real(PR) :: dT_a_theta_q_4

    real(PR) :: z_q, z_qp1

    z_q   = (h*frequencies(q))/(k_B*T)
    z_qp1 = (h*frequencies(q+1))/(k_B*T)

    dT_a_theta_q_4 = 4._PR*(xi(z_qp1) - xi(z_q))

    if (z_q < 100._PR) then
      dT_a_theta_q_4 = dT_a_theta_q_4 + z_q**4/(exp(z_q) - 1._PR)
    end if
    if (z_qp1 < 100._PR) then
      dT_a_theta_q_4 = dT_a_theta_q_4 - z_qp1**4/(exp(z_qp1) - 1._PR)
    end if

    dT_a_theta_q_4 = 15._PR*a*T**3/pi**4*dT_a_theta_q_4

  end function compute_dT_a_theta_q_4

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute \f$ \xi(z) \approx \Xi(z) = \int_1^z \frac{x^3}{\exp(x)-1}dx \f$
  !>
  !> @param[in] z \f$ \frac{h \nu}{k_B T} \f$
  !>
  !> @remark See \cite Tur_2003 and \cite Sar_2013 for the computations of the coefficients \f$d_i\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function xi(z)
    real(PR), intent(in) :: z
    real(PR) :: xi

    real(PR), parameter :: pi4_15 = pi**4/15._PR
    real(PR), parameter :: d_inf = pi4_15
    real(PR), parameter :: d_0 = pi4_15
    real(PR), parameter :: d_1 = pi4_15
    real(PR), parameter :: d_2 = 0.5_PR*pi4_15
    real(PR), parameter :: d_3 = -(30._PR - pi**4)/90._PR
    real(PR), parameter :: d_4 =  0.06354289909759_PR
    real(PR), parameter :: d_5 = -0.00654100302672_PR
    real(PR), parameter :: d_6 =  2.375978179550176E-04_PR


    ! Cut-off to avoid numerical instabilities with the computations
    ! of the exponential function
    if (z < 100._PR) then
      ! naïve
      !xi = d_inf - exp(-z)*(d_0 + d_1*z + d_2*z**2 + d_3*z**3 + d_4*z**4 + d_5*z**5 + d_6*z**6)
      ! Horner (https://en.wikipedia.org/wiki/Horner%27s_method)
      xi = d_inf - exp(-z)*(d_0 + z*(d_1 + z*(d_2 + z*(d_3 + z*(d_4 + z*(d_5 + z*d_6))))))
    else
      xi = d_inf
    end if

  end function xi

  !--------------------------------------------------------------------!
  !> @brief
  !> Convert the primitive to conservative variables
  !>
  !> @remark
  !> For the M1 model the conservative and primitive variables are the same:
  !> * \f$ E \f$ the radiative energy
  !> * \f$ F \f$ the radiative flux
  !> * \f$ T \f$ the temperature
  !>
  !> @param[in] V_K primitive variables
  !> @param[out] U_K conservative variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine primitive_to_conservative(V_K,U_K)
    real(PR), dimension(:), intent(in)     :: V_K
    real(PR), dimension(:), intent(out) :: U_K

    U_K(:) = V_K(:)

  end subroutine primitive_to_conservative

  !--------------------------------------------------------------------!
  !> @brief
  !> Convert the conservative to primitive variables
  !>
  !> @remark
  !> For the M1 model the conservative and primitive variables are the same:
  !> * \f$ E \f$ the radiative energy
  !> * \f$ F \f$ the radiative flux
  !> * \f$ T \f$ the temperature
  !>
  !> @param[in] U_K conservative variables
  !> @param[out] V_K primitive variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine conservative_to_primitive(U_K,V_K)
    real(PR), dimension(:), intent(in)     :: U_K
    real(PR), dimension(:), intent(out) :: V_K

    V_K(:) = U_K(:)

  end subroutine conservative_to_primitive

  !--------------------------------------------------------------------!
  !> @brief
  !> Check if the value of the solution on a cell is physically admissible.
  !>
  !> @param[in] U_K conservative variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function IsPhysicallyAdmissible(U_K)
    real(PR), dimension(:), intent(in) :: U_K
    logical :: IsPhysicallyAdmissible

    integer :: group, eq0

    IsPhysicallyAdmissible = .TRUE.

    loop_groups: do group = 1, NbGroup
      eq0 = (1 + dim_)*(group-1)
      if (U_K(eq0+1) < zero) then
        IsPhysicallyAdmissible = .FALSE.
        exit loop_groups
      else if (NORM2(U_K(eq0+2:eq0+3)) > c*U_K(eq0+1)) then
        IsPhysicallyAdmissible = .FALSE.
        exit loop_groups
      end if
    end do loop_groups

    if (U_K(NbEq) < zero) then
      IsPhysicallyAdmissible = .FALSE.
    end if

  end function IsPhysicallyAdmissible

#endif /* _PHY_M1MG_ */
end module mod_physics_M1MG
