!> Physics of the M1 model for radiative transfer \cite DubFeu_1999 .
!>
!> \f[\left\{\begin{array}{rcl}
!>   \partial_t E   +     \partial_x F_x         +     \partial_y F_y         & = & c \sigma^e a T^4 - c \sigma^a E \\
!>   \partial_t F_x + c^2 \partial_x P_{xx}(E,F) + c^2 \partial_y P_{xy}(E,F) & = & -c \sigma^f F_x \\
!>   \partial_t F_y + c^2 \partial_x P_{yx}(E,F) + c^2 \partial_y P_{yy}(E,F) & = & -c \sigma^f F_y \\
!>   \partial_t T                                                             & = & \frac{c \sigma^a E - c \sigma^e a T^4}{\rho C_v}
!> \end{array}\right.\f]
!>
!> with:
!> * \f$ E \f$ the radiative energy,
!> * \f$ F \f$ the radiative flux,
!> * \f$ T \f$ the temperature,
!> * \f$P = \frac{E}{2}\left( (1-\chi)Id  + (3\chi-1)\frac{F \otimes F}{\Vert F \Vert_2^2}\right) \f$ the radiative pressure,
!> * \f$\chi = \frac{3 + 4f^2}{5+ \sqrt{4-3f^2}} \f$ the Eddington factor, in order that \f$\chi\f$ is well defined:
!>     \f$ f=\frac{\Vert F \Vert}{c E} \leq 1 \f$,
!> * \f$ c \f$ the speed of light.
!>
!> @remark See \cite Sar_2013 for the test cases.
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_physics_M1
#if defined(_PHY_M1_)

  use mod_numerics, only: PR, pi, eps
  use mod_numerics, only: zero, third, one
  use mod_numerics, only: sqrt_2, sqrt_3
  use type_IO, only: StdOut, StdErr
  use mod_tools, only: SolvePolynomial2
  use mod_tools, only: min_NaN, max_NaN
  use mod_tools_array, only: tensor_product
  use type_mesh, only: pmesh, dim_
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells
#endif /* _FC_PAR_OMP_ */
  use mod_read, only: read_solution
  use mod_read, only: Restarting

  use mod_physics_M1_var, only: NbEq, test_case
  use mod_physics_M1_var, only: T_end
  use mod_physics_M1_var, only: c, a
#if defined(_SRC_ENABLE_)
  use mod_physics_M1_var, only: rho, C_v, ideal_gas_constant
#endif /* _SRC_ENABLE_ */

#if defined(_HO_MOOD_)
  use type_HO_MOOD, only: MaxDegree
  use mod_HO_MOOD_int, only: compute_mean_f
#endif /* _HO_MOOD_ */

#if defined(_FC_PGI_)
  use mod_tools_array, only: NORM2
#endif /* defined(_FC_PGI_) */

  IMPLICIT NONE

  private
  public physic_flux, U_Dirichlet
  public compute_wave_speed, compute_greatest_wave_speed
#if defined(_FLUX_HLLC_)
  public compute_HLLC_values
#endif /* _FLUX_HLLC_ */
  public init_U
  public primitive_to_conservative, conservative_to_primitive
  public IsPhysicallyAdmissible


#if !defined(_FLUX_HLLC_)
!#define _PHY_M1_COMPUTE_WAVE_SPEED_
#endif /* _FLUX_HLLC_ */

CONTAINS
  !--------------------------------------------------------------------!
  !> @brief
  !> Give the initial condition according to the test case
  !>
  !> @param[out] U0 Initial condition
  !> @param[out] iter Number of iteration previously done
  !> @param[out] T_init Initial time
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine init_U(U0,iter,T_init)
      real(PR), dimension(:,:), intent(out) :: U0
      integer , intent(out) :: iter
      real(PR), intent(out) :: T_init

      real(PR), dimension(dim_) :: xK
      integer :: CellK

      !$OMP SINGLE
      select case(test_case)
      case ('cst', 'cont',  &
             'xy_1', &
            'shock_x', 'Big_x')
        T_end = 1.0_PR/c
      case ('x_1', 'x_2', &
            'y_1', 'x_1_inv')
        T_end = 0.5_PR/c
      case ('Gauss_x', 'Gauss')
        T_end = 1.E-5_PR
      case ('Square_x')
        T_end = 1._PR/c
      case ('circle')
        T_end = 1.0_PR/c
      case ('Marshak_x')
        T_end = 1.E-4_PR
      case ('dis_x')
        T_end = 5.E-6_PR
      case ('sonde_80km')
        T_end = 1._PR/c !1.E-8_PR
      case ('sonde')
        T_end = 1.E-4_PR
      case ('shadow_cone_wall', 'shadow_cone')
        T_end = 5.E-8_PR
      case ('KT')
        T_end = 2.E-9_PR
      case ('tophat')
        T_end = 5.7E-6_PR
      case default
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
        FLUSH StdErr
        STOP
      end select
      !$OMP END SINGLE NOWAIT

#if defined(_SRC_ENABLE_)
      !$OMP SINGLE
      allocate(rho(pmesh%NbCells))
      allocate(C_v(pmesh%NbCells))
      !$OMP END SINGLE
#endif /* _SRC_ENABLE_ */

      test_restart: if (Restarting) then
        !$OMP SINGLE
        call read_solution(U0,iter,T_init)
        !$OMP END SINGLE

#if defined(_SRC_ENABLE_)
        select case(test_case)
        case ('Gauss_x', 'Gauss', 'Marshak_x')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            rho(CellK) =  1.E-2_PR
            C_v(CellK) =  1.E-2_PR
          end do
          !$OMP END DO
        case ('sonde_80km')
          ! The source file from the Euler simulation was read in read_solution()
          ! and now we make some adjustments to use it for the M1 model

          ! First, the initial time is set to zero (and the number of iterations)
          !$OMP SINGLE
          T_init = 0._PR
          iter = 0
          !$OMP END SINGLE NOWAIT

          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            ! Next, we save rho from the Euler simulation
            rho(CellK) = U0(1,CellK)
            C_v(CellK) = 1.E-2_PR/rho(CellK)
            ! And finally, compute the temperature: T = p/rho*M/R
            ! with p = \rho (\gamma - 1) (E - \frac{\rho (u^2 + v^2)}{2})
            U0(4,CellK) = ((1.4_PR - 1._PR)*(U0(4,CellK) - 0.5_PR*dot_product(U0(2:3,CellK),U0(2:3,CellK))/U0(1,CellK))*30.E-3_PR)/(rho(CellK)*ideal_gas_constant)
            U0(1,CellK) = a*U0(4,CellK)**4
            U0(2,CellK) = 0._PR
            U0(3,CellK) = 0._PR
          end do
          !$OMP END DO
        end select
#endif /* _SRC_ENABLE_ */

      else
        !$OMP SINGLE
        iter = 0
        T_init = 0._PR
        !$OMP END SINGLE

        select case(test_case)
        case ('cst')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            U0(4,CellK) = 300._PR
            U0(1,CellK) = a*U0(4,CellK)**4
            U0(2,CellK) = 0._PR
            U0(3,CellK) = 0._PR
          end do
          !$OMP END DO

        case ('cont')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
            U0(1,CellK) = xK(1)*xK(2)+2._PR
            U0(2,CellK) = 0.1_PR*xK(1)
            U0(3,CellK) = 0._PR
            U0(4,CellK) = 300._PR
          end do
          !$OMP END DO

        case ('x_1')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
            if (xK(1) < 0.5_PR) then
              U0(1,CellK) = 1._PR
              U0(2,CellK) = 0.1_PR
              U0(3,CellK) = 0._PR
              U0(4,CellK) = 300._PR
            else
              U0(1,CellK) = 10._PR
              U0(2,CellK) = -1._PR
              U0(3,CellK) = 0._PR
              U0(4,CellK) = 300._PR
            end if
          end do
          !$OMP END DO

        case ('y_1')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
            if (xK(2) < 0.5_PR) then
              U0(1,CellK) = 1._PR
              U0(2,CellK) = 0.0_PR
              U0(3,CellK) = 0.1_PR
              U0(4,CellK) = 300._PR
            else
              U0(1,CellK) = 10._PR
              U0(2,CellK) = 0._PR
              U0(3,CellK) = -1._PR
              U0(4,CellK) = 300._PR
            end if
          end do
          !$OMP END DO

        case ('x_1_inv')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
            if (xK(1) < 0.5_PR) then
              U0(1,CellK) = 10._PR
              U0(2,CellK) = 1._PR
              U0(3,CellK) = 0._PR
              U0(4,CellK) = 300._PR
            else
              U0(1,CellK) = 1._PR
              U0(2,CellK) = -0.1_PR
              U0(3,CellK) = .0_PR
              U0(4,CellK) = 300._PR
            end if
          end do
          !$OMP END DO

        case ('xy_1')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
            if (xK(2) < xK(1)) then
              U0(1,CellK) = 10._PR
              U0(2,CellK) = 0.5_PR*sqrt(2._PR)*1._PR
              U0(3,CellK) = -0.5_PR*sqrt(2._PR)*1._PR
              U0(4,CellK) = 300._PR
            else
              U0(1,CellK) = 1._PR
              U0(2,CellK) = 0.5_PR*sqrt(2._PR)*(- 0.1_PR)
              U0(3,CellK) = -0.5_PR*sqrt(2._PR)*(- 0.1_PR)
              U0(4,CellK) = 300._PR
            end if
          end do
          !$OMP END DO

        case ('x_2')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
            if (xK(1) < 0.5_PR) then
              U0(1,CellK) = 2._PR
              U0(2,CellK) = -0.5_PR
              U0(3,CellK) = 0.0_PR
              U0(4,CellK) = 300._PR
            else
              U0(1,CellK) = 2._PR
              U0(2,CellK) = 1._PR
              U0(3,CellK) = 0._PR
              U0(4,CellK) = 300._PR
            end if
          end do
          !$OMP END DO

        case ('Gauss_x', 'Gauss')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
#if (_HO_MOOD_MAX_DEGREE_ >= 1)
            U0(:,CellK) = compute_mean_f(CellK,T_init,MaxDegree,InitialCondition,NbEq)
#else
            xK(:) = pmesh%elt(CellK)%x(:)
            U0(:,CellK) = InitialCondition(xK,T_init,NbEq)
#endif

          end do
          !$OMP END DO

        case ('Square_x')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)

            if (xK(1) > 0.45_PR .and. xK(1) < 0.55_PR) then
              U0(4,CellK) = 300._PR
              U0(1,CellK) = a*U0(4,CellK)**4
              U0(2:3,CellK) = 0._PR
            else
              U0(4,CellK) = 0.1_PR
              U0(1,CellK) = a*U0(4,CellK)**4
              U0(2:3,CellK) = 0._PR
            endif
          end do
          !$OMP END DO

        case ('circle')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
            if ((xK(1) - 0.5_PR)**2 + (xK(2) - 0.5_PR)**2 < (0.1_PR)**2) then
              U0(4,CellK) = 1000._PR
            else
              U0(4,CellK) = 1._PR
            end if
            U0(1,CellK) = a*U0(4,CellK)**4
            U0(2,CellK) = 0._PR
            U0(3,CellK) = 0._PR
          end do
          !$OMP END DO

        case ('Marshak_x')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            U0(4,CellK) = 300._PR
            U0(1,CellK) = a*U0(4,CellK)**4
            U0(2,CellK) = 0._PR
            U0(3,CellK) = 0._PR
          end do
          !$OMP END DO

        case ('dis_x')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
            if (xK(1) < 0.5_PR) then
              U0(4,CellK) = 1.E3_PR
            else
              U0(4,CellK) = 1._PR
            end if
            U0(1,CellK) = a*U0(4,CellK)**4
            U0(2,CellK) = 0._PR
            U0(3,CellK) = 0._PR
          end do
          !$OMP END DO

        case ('shock_x')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
            if (xK(1) < 0.5_PR) then
              U0(4,CellK) = 1000._PR
              U0(1,CellK) = a*U0(4,CellK)**4
              U0(2,CellK) = 0._PR
              U0(3,CellK) = 0._PR
            else
              U0(4,CellK) = 300._PR
              U0(1,CellK) = a*U0(4,CellK)**4
              U0(2,CellK) = 0._PR
              U0(3,CellK) = 0._PR
            end if
          end do
          !$OMP END DO

        case ('Big_x')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
            if (xK(1) < 0.5_PR) then
              U0(4,CellK) = 1.E9_PR
              U0(1,CellK) = a*U0(4,CellK)**4
              U0(2,CellK) = 1.E5_PR
              U0(3,CellK) = 0._PR
            else
              U0(4,CellK) = 1._PR
              U0(1,CellK) = a*U0(4,CellK)**4
              U0(2,CellK) = 0._PR
              U0(3,CellK) = 0._PR
            end if
          end do
          !$OMP END DO

        case ('sonde')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            U0(4,CellK) = 3.E2_PR
            U0(1,CellK) = a*U0(4,CellK)**4
            U0(2,CellK) = 0._PR
            U0(3,CellK) = 0._PR
          end do
          !$OMP END DO

        case ('shadow_cone_wall')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            U0(4,CellK) = 300._PR
            U0(1,CellK) = a*U0(4,CellK)**4
            U0(2,CellK) = 0._PR
            U0(3,CellK) = 0._PR
          end do
          !$OMP END DO

        case ('shadow_cone')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
            if (xK(1) > 0.5_PR .and. xK(2) > 0.5_PR) then
              U0(4,CellK) = 300._PR
              U0(1,CellK) = a*U0(4,CellK)**4
              U0(2,CellK) = 0._PR
              U0(3,CellK) = 0._PR
            else
              U0(4,CellK) = 5.E4_PR
              U0(1,CellK) = a*U0(4,CellK)**4
              U0(2,CellK) = 0._PR
              U0(3,CellK) = 0._PR
            end if
          end do
          !$OMP END DO

        case ('KT')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            xK(:) = pmesh%elt(CellK)%x(:)
            U0(4,CellK) = 1.E3_PR
            U0(1,CellK) = a*U0(4,CellK)**4
            if (xK(2) <= 0.5_PR .and. xK(1) <= 0.5_PR) then
              ! SO
              U0(2,CellK) = 0.99999999_PR*c*U0(1,CellK)
              U0(3,CellK) = 0._PR
            else if( xK(2) <= 0.5_PR .and. xK(1) >= 0.5_PR) then
              ! SE
              U0(2,CellK) = 0._PR
              U0(3,CellK) = -0.99999999_PR*c*U0(1,CellK)
            else if( xK(2) >= 0.5_PR .and. xK(1) <= 0.5_PR) then
              ! NO
              U0(2,CellK) = 0._PR
              U0(3,CellK) = 0.99999999_PR*c*U0(1,CellK)
            else !if( xK(2) >= 0.5_PR .and. xK(1) >= 0.5_PR) then
              ! NE
              U0(2,CellK) = -0.99999999_PR*c*U0(1,CellK)
              U0(3,CellK) = 0._PR
            end if

          end do
          !$OMP END DO

        case ('tophat')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            U0(4,CellK) = 6.E4_PR
            U0(1,CellK) = a*U0(4,CellK)**4
            U0(2,CellK) = 0._PR
            U0(3,CellK) = 0._PR
          end do
          !$OMP END DO
        end select

#if defined(_SRC_ENABLE_)
        select case(test_case)
        case ('tophat')
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            rho(CellK) = 1._PR
            C_v(CellK) = 1.E-3_PR
          end do
          !$OMP END DO
        case default
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          do CellK = 1, pmesh%NbCells
            rho(CellK) = 1.E-2_PR
            C_v(CellK) = 1.E-2_PR
          end do
          !$OMP END DO
        end select
#endif /* _SRC_ENABLE_ */

      end if test_restart

  end subroutine init_U

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the initial condition
  !>
  !> @param[in] xK Coordinates
  !> @param[in] t_n Current time
  !> @param[in] NbEq Number of equations
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function InitialCondition(xK,t_n,NbEq) result(U0)
      real(PR), dimension(:), intent(in) :: xK
      real(PR), intent(in) :: t_n
      integer, intent(in) :: NbEq
      real(PR), dimension(NbEq) :: U0

      select case(test_case)
      case ('Gauss_x')
        U0(4) = 300._PR*exp(- 0.5_PR*(xK(1) - 0.5_PR)**2/(0.05_PR)**2) + 300._PR
        U0(1) = a*U0(4)**4
        U0(2:3) = 0._PR
      case ('Gauss')
        U0(4) = 300._PR*exp(- 0.5_PR*sum((xK(:) - 0.5_PR*pmesh%length(:))**2)/(0.05_PR)**2) + 300._PR
        U0(1) = a*U0(4)**4
        U0(2:3) = 0._PR
      case default
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
        FLUSH StdErr
        STOP
      end select

  end function InitialCondition

  !--------------------------------------------------------------------!
  !> @brief
  !> Give the limit boundary value
  !>
  !> @param[in] x Coordinates of the point on the boundary
  !> @param[in] t_n Current time
  !>
  !> @return Value of the solution on the boundary at point
  !> \f$x\f$ and time \f$t^n\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function U_Dirichlet(x,t_n)
      real(PR), dimension(:), intent(in) :: x
      real(PR), intent(in) :: t_n
      real(PR), dimension(NbEq) :: U_Dirichlet

      select case(test_case)
      case ('Marshak_x')
        U_Dirichlet(4)   = 1000._PR
        U_Dirichlet(1)   = a*U_Dirichlet(NbEq)**4
        U_Dirichlet(2:3) = 0._PR
      case ('sonde')
        U_Dirichlet(4)   = 1.E4_PR
        U_Dirichlet(1)   = a*U_Dirichlet(NbEq)**4
        U_Dirichlet(2:3) = 0._PR
      case ('shadow_cone_wall', 'shadow_cone')
        U_Dirichlet(4) = 5.802E6_PR
        U_Dirichlet(1) = a*U_Dirichlet(NbEq)**4
        U_Dirichlet(2) = c*U_Dirichlet(1)
        U_Dirichlet(3) = 0._PR
      case ('cst')
        U_Dirichlet(4)   = 300._PR
        U_Dirichlet(1)   = a*U_Dirichlet(NbEq)**4
        U_Dirichlet(2:3) = 0._PR
      case ('tophat')
        U_Dirichlet(4) = 1.E6_PR
        U_Dirichlet(1) = a*U_Dirichlet(4)**4
        U_Dirichlet(2) = 0.8_PR*c*U_Dirichlet(1)
        U_Dirichlet(3) = 0._PR
      case default
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Wrong test case"
        FLUSH StdErr
        STOP
      end select


  end function U_Dirichlet

#if defined(_FLUX_HLLC_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the HLLC flux for the \f$M_1\f$ model
  !>
  !> @remark see \cite BerChaDub_2007, \cite BerDubDubNguTur_2010, and \cite Dub_2009
  !>
  !> @warning Computations done with wave speeds equal to \f$\pm c\f$
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent Normal and tangent to the interface
  !> @param[out] flux_1D HLLC flux
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_radiative_pressure
  subroutine compute_HLLC_values(Ul,Ur,normal,tangent,flux_1D)
      real(PR), dimension(:), intent(in) :: Ul, Ur
      real(PR), dimension(:), intent(in) :: normal, tangent
      real(PR), dimension(:), intent(out) :: flux_1D

#if !defined(_PHY_M1_COMPUTE_WAVE_SPEED_)
      real(PR), dimension(dim_,dim_) :: RotationMatrix

      real(PR) :: E_l, F_l_n, F_l_t
      real(PR), dimension(dim_) :: F_l, F_l_nt
      real(PR), dimension(dim_,dim_) :: P_l_nt

      real(PR) :: E_r, F_r_n, F_r_t
      real(PR), dimension(dim_) :: F_r, F_r_nt
      real(PR), dimension(dim_,dim_) :: P_r_nt

      real(PR) :: E_star_HLL, F_star_HLL_n, F_tilde_HLL_n, P_tilde_HLL_n
      real(PR) :: beta_m, beta_p, beta_star
      integer  :: IERR
      real(PR) :: pi_star, E_star_l, E_star_r

      E_l    = Ul(1)
      F_l(:) = Ul(2:3)
      F_l_n  = dot_product(F_l,normal)
      F_l_t  = dot_product(F_l,tangent)
      F_l_nt(:) = [F_l_n, F_l_t]
      P_l_nt(:,:) = radiative_pressure(E_l, F_l_nt)

      E_r    = Ur(1)
      F_r(:) = Ur(2:3)
      F_r_n  = dot_product(F_r,normal)
      F_r_t  = dot_product(F_r,tangent)
      F_r_nt(:) = [F_r_n, F_r_t]
      P_r_nt(:,:) = radiative_pressure(E_r, F_r_nt)

      E_star_HLL    = 0.5_PR*(E_r + E_l - (F_r_n - F_l_n)/c)
      F_star_HLL_n  = 0.5_PR*(F_r_n + F_l_n  - c*(P_r_nt(1,1) - P_l_nt(1,1)))
      F_tilde_HLL_n = 0.5_PR*(F_r_n + F_l_n  - c*(E_r - E_l))
      P_tilde_HLL_n = 0.5_PR*(P_r_nt(1,1) + P_l_nt(1,1)  - (F_r_n - F_l_n)/c)

      call SolvePolynomial2(F_tilde_HLL_n,-c*(P_tilde_HLL_n + E_star_HLL), F_star_HLL_n, beta_m, beta_p, IERR)

      if (abs(beta_m) < one) then
        beta_star = beta_m
      else !if (abs(beta_p) < one) then
        beta_star = beta_p
      endif

      pi_star = P_tilde_HLL_n - beta_star/c*F_tilde_HLL_n
      E_star_l = 1._PR/(1._PR + beta_star)*(E_l - beta_star*pi_star + F_l_n/c)
      E_star_r = 1._PR/(1._PR - beta_star)*(E_r + beta_star*pi_star + F_r_n/c)

      ! P is symmetric, so P(2,1) = P(1,2)
      if (beta_star > zero) then
        flux_1D(1) = c*beta_star*(E_star_l + pi_star)
        flux_1D(2) = c**2*pi_star + c*beta_star*flux_1D(1)
        flux_1D(3) =  beta_star/(1._PR + beta_star)*c*(c*P_l_nt(2,1) + F_l_t)
      else
        flux_1D(1) = c*beta_star*(E_star_r + pi_star)
        flux_1D(2) = c**2*pi_star + c*beta_star*flux_1D(1)
        flux_1D(3) = -beta_star/(1._PR - beta_star)*c*(c*P_r_nt(2,1) + F_r_t)
      endif
      RotationMatrix(:,1) = normal(:)
      RotationMatrix(:,2) = tangent(:)
      flux_1D(2:3) = MATMUL(RotationMatrix,flux_1D(2:3))

      ! No flux on temperature
      flux_1D(NbEq) = zero

#else
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> HLLC not defined w/ wave speeds /= +/- c"
      FLUSH StdErr
      STOP
#endif /* _PHY_M1_COMPUTE_WAVE_SPEED_ */


  end subroutine compute_HLLC_values
#endif /* _FLUX_HLLC_ */

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the wave speeds waves on a interface
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent Normal and tangent to the interface
  !> @param[out] a_l, a_r  Wave speeds on the left and right
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_min_nan
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_max_nan
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_chi
  subroutine compute_wave_speed(Ul,Ur,normal,tangent,a_l,a_r)
      real(PR), dimension(:), intent(in) :: Ul, Ur
      real(PR), dimension(:), intent(in) :: normal, tangent
      real(PR), intent(out) :: a_l, a_r

#if defined(_PHY_M1_COMPUTE_WAVE_SPEED_)
      real(PR), dimension(dim_) :: F_l, ff_l
      real(PR), dimension(dim_) :: F_r, ff_r
      real(PR) :: E_l, xi_l, chi_l, ff_l_n, ff_l_t, norm_ff_l
      real(PR) :: E_r, xi_r, chi_r, ff_r_n, ff_r_t, norm_ff_r
      real(PR) :: lambda_m_l, lambda_p_l
      real(PR) :: lambda_m_r, lambda_p_r
#endif /* _PHY_M1_COMPUTE_WAVE_SPEED_ */

#if defined(_PHY_M1_COMPUTE_WAVE_SPEED_)
      E_l    = Ul(1)
      F_l(:) = Ul(2:3)
      chi_l  = compute_chi(E_l,F_l)
      ff_l(:)= F_l(:)/(c*E_l)
      ff_l_n = dot_product(ff_l,normal)
      ff_l_t = dot_product(ff_l,tangent)
      norm_ff_l = NORM2(ff_l)
      xi_l   = sqrt(4._PR - 3._PR*norm_ff_l**2)

      lambda_p_l = c*(ff_l_n/xi_l + sqrt_2*sqrt((xi_l - 1._PR)*(xi_l + 2._PR)*&
         (2._PR*(xi_l - 1._PR)*(xi_l + 2._PR) + 3._PR*ff_l_t**2))/(sqrt_3*xi_l*(xi_l + 2._PR)))

      lambda_m_l = c*(ff_l_n/xi_l - sqrt_2*sqrt((xi_l - 1._PR)*(xi_l + 2._PR)*&
         (2._PR*(xi_l - 1._PR)*(xi_l + 2._PR) + 3._PR*ff_l_t**2))/(sqrt_3*xi_l*(xi_l + 2._PR)))

      E_r    = Ur(1)
      F_r(:) = Ur(2:3)
      chi_r  = compute_chi(E_r,F_r)
      ff_r(:)= F_r(:)/(c*E_r)
      ff_r_n = dot_product(ff_r,normal)
      ff_r_t = dot_product(ff_r,tangent)
      norm_ff_r = NORM2(ff_r)
      xi_r   = sqrt(4._PR - 3._PR*norm_ff_r**2)

      lambda_p_r = c*(ff_r_n/xi_r + sqrt_2*sqrt((xi_r - 1._PR)*(xi_r + 2._PR)*&
         (2._PR*(xi_r - 1._PR)*(xi_r + 2._PR) + 3._PR*ff_r_t**2))/(sqrt_3*xi_r*(xi_r + 2._PR)))

      lambda_m_r = c*(ff_r_n/xi_r - sqrt_2*sqrt((xi_r - 1._PR)*(xi_r + 2._PR)*&
         (2._PR*(xi_r - 1._PR)*(xi_r + 2._PR) + 3._PR*ff_r_t**2))/(sqrt_3*xi_r*(xi_r + 2._PR)))

      a_l = min_NaN(lambda_m_l, lambda_m_r)
      a_r = max_NaN(lambda_p_l, lambda_p_r)

      !a_l = min_NaN(lambda_m_l, lambda_m_r, 0._PR)
      !a_r = max_NaN(lambda_p_l, lambda_p_r, 0._PR)

      !a_l = min_NaN(c*norm_ff_l, c*norm_ff_r, &
          !(norm_ff_l - chi_l)/(1._PR - norm_ff_l), (norm_ff_r - chi_r)/(1._PR - norm_ff_r), &
          !(norm_ff_l + chi_l)/(1._PR + norm_ff_l), (norm_ff_r + chi_r)/(1._PR + norm_ff_r), &
          !lambda_m_l, lambda_m_r, 0._PR)
      !a_r = max_NaN(c*norm_ff_l, c*norm_ff_r, &
          !(norm_ff_l - chi_l)/(1._PR - norm_ff_l), (norm_ff_r - chi_r)/(1._PR - norm_ff_r), &
          !(norm_ff_l + chi_l)/(1._PR + norm_ff_l), (norm_ff_r + chi_r)/(1._PR + norm_ff_r), &
          !lambda_p_l, lambda_p_r, 0._PR)

#else
      a_l = -c
      a_r = c
#endif /* _PHY_M1_COMPUTE_WAVE_SPEED_ */
  end subroutine compute_wave_speed

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the largest wave speed
  !>
  !> @param[in] Ul, Ur Solution on the left and right of the interface
  !> @param[in] normal, tangent Normal and tangent to the interface
  !> @return Largest wave speed
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_wave_speed
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_mp_max_nan
  function compute_greatest_wave_speed(Ul,Ur,normal,tangent)
      real(PR), dimension(:), intent(in) :: Ul, Ur
      real(PR), dimension(:), intent(in) :: normal, tangent

      real(PR) :: compute_greatest_wave_speed

#if defined(_PHY_M1_COMPUTE_WAVE_SPEED_)
      real(PR) :: a_l, a_r

      call compute_wave_speed(Ul,Ur,normal,tangent,a_l,a_r)
      compute_greatest_wave_speed = max_NaN(abs(a_l),abs(a_r))
#else
      compute_greatest_wave_speed = c
#endif /* _PHY_M1_COMPUTE_WAVE_SPEED_ */

  end function compute_greatest_wave_speed

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the physic flux
  !>
  !> @remark
  !> For the M1 model:
  !> \f${\bf F}({\bf U}) = \left(\begin{array}{c}F^t \\ c^2 P \\ 0\end{array}\right)\f$
  !>
  !> @param[in] U_K solution on cell CellK
  !> @return the physic flux on the cell CellK
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_radiative_pressure
  function physic_flux(U_K)
      real(PR), dimension(:), intent(in) :: U_K
      real(PR), dimension(NbEq,dim_)     :: physic_flux

      real(PR) :: E
      real(PR), dimension(dim_,dim_) :: c2_P
      real(PR), dimension(dim_) :: F

      E = U_K(1)
      F(:) = U_K(2:3)

      c2_P(:,:) = c**2*radiative_pressure(E,F)

      physic_flux(1,1) = F(1)
      physic_flux(2:3,1) = c2_P(1:2,1)
      physic_flux(4,1) = zero

      physic_flux(1,2) = F(2)
      physic_flux(2:3,2) = c2_P(1:2,2)
      physic_flux(4,2) = zero

  end function physic_flux

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the radiative pressure
  !>
  !> @param[in] E radiative energy
  !> @param[in] F radiative flux
  !> @return the radiative pressure:
  !> \f$P = \frac{E}{2}\left( (1-\chi)Id  + (3\chi-1)\frac{F \otimes F}{\Vert F \Vert_2^2}\right) \f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_physics_m1_mp_compute_chi
  !DIR$ ATTRIBUTES FORCEINLINE :: mod_tools_array_mp_tensor_product
  function radiative_pressure(E,F) result(P)
      real(PR), intent(in)                  :: E
      real(PR), dimension(:), intent(in) :: F

      real(PR), dimension(dim_,dim_) :: P

      real(PR) :: chi, NORM2_F
      integer  :: i

      NORM2_F = NORM2(F)

      !if (IsZero(NORM2_F)) then
      if (NORM2_F < eps) then
        P(:,:) = zero
        do i = 1, dim_
          P(i,i) = third
        end do

      else
        chi = compute_chi(E,F)

        P(:,:) = 0.5_PR*(3._PR*chi - 1._PR)*tensor_product(F,F)/NORM2_F**2

        do i = 1, dim_
          P(i,i) = P(i,i) + 0.5_PR*(1._PR - chi)
        end do

      end if

      P(:,:) = E*P(:,:)

  end function radiative_pressure

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the Eddington factor
  !>
  !> @remark
  !> In order that \f$\chi\f$ is well defined the anistropy factor
  !> need to satisfy \f$f=\frac{\Vert F \Vert}{c E} \leq 1\f$
  !>
  !> @param[in] E radiative energye
  !> @param[in] F radiative flux
  !> @return the Eddington factor: \f$\chi = \frac{3 + 4f^2}{5+ \sqrt{4-3f^2}} \f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function compute_chi(E,F)
      real(PR), intent(in)                  :: E
      real(PR), dimension(:), intent(in) :: F

      real(PR) :: compute_chi

      real(PR) :: ff

      ff = NORM2(F)/(c*E)

#if defined(_DEBUG_)
      if (ff > one) then
        write(StdOut,'(A,ES23.16,A)') '||F||/(c*E) =', ff, ' > 1'
        !read(*,*)
        !FLUSH StdOut
        !STOP
      end if
#endif /* _DEBUG_ */

      compute_chi = (3._PR + 4._PR*ff**2)/(5._PR + 2._PR*sqrt(4._PR - 3._PR*ff**2))
  end function compute_chi


  !--------------------------------------------------------------------!
  !> @brief
  !> Convert the primitive to conservative variables
  !>
  !> @remark
  !> For the M1 model the conservative and primitive variables are the same:
  !> * \f$ E \f$ the radiative energy
  !> * \f$ F \f$ the radiative flux
  !> * \f$ T \f$ the temperature
  !>
  !> @param[in] V_K primitive variables
  !> @param[out] U_K conservative variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine primitive_to_conservative(V_K,U_K)
      real(PR), dimension(:), intent(in)     :: V_K
      real(PR), dimension(:), intent(out) :: U_K

      U_K(:) = V_K(:)

  end subroutine primitive_to_conservative

  !--------------------------------------------------------------------!
  !> @brief
  !> Convert the conservative to primitive variables
  !>
  !> @remark
  !> For the M1 model the conservative and primitive variables are the same:
  !> * \f$ E \f$ the radiative energy
  !> * \f$ F \f$ the radiative flux
  !> * \f$ T \f$ the temperature
  !>
  !> @param[in] U_K conservative variables
  !> @param[out] V_K primitive variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine conservative_to_primitive(U_K,V_K)
      real(PR), dimension(:), intent(in)     :: U_K
      real(PR), dimension(:), intent(out) :: V_K

      V_K(:) = U_K(:)

  end subroutine conservative_to_primitive

  !--------------------------------------------------------------------!
  !> @brief
  !> Check if the value of the solution on a cell is physically admissible.
  !>
  !> @param[in] U_K conservative variables
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function IsPhysicallyAdmissible(U_K)
      real(PR), dimension(:), intent(in) :: U_K

      logical :: IsPhysicallyAdmissible

      if (U_K(1) < zero) then
        IsPhysicallyAdmissible = .FALSE.
      else if (NORM2(U_K(2:3)) > c*U_K(1)) then
        IsPhysicallyAdmissible = .FALSE.
      else if (U_K(4) < zero) then
        IsPhysicallyAdmissible = .FALSE.
      else
        IsPhysicallyAdmissible = .TRUE.
      end if

  end function IsPhysicallyAdmissible

#endif /* _PHY_M1_ */
end module mod_physics_M1
