!> Physics of the M1 model for radiative transfer \cite DubFeu_1999 .
!>
!> \f[\left\{\begin{array}{rcl}
!>   \partial_t E   +     \partial_x F_x         +     \partial_y F_y         & = & c \sigma^e a T^4 - c \sigma^a E \\
!>   \partial_t F_x + c^2 \partial_x P_{xx}(E,F) + c^2 \partial_y P_{xy}(E,F) & = & -c \sigma^f F_x \\
!>   \partial_t F_y + c^2 \partial_x P_{yx}(E,F) + c^2 \partial_y P_{yy}(E,F) & = & -c \sigma^f F_y \\
!>   \partial_t T                                                             & = & \frac{c \sigma^a E - c \sigma^e a T^4}{\rho C_v}
!> \end{array}\right.\f]
!>
!> with:
!> * \f$ E \f$ the radiative energy,
!> * \f$ F \f$ the radiative flux,
!> * \f$ T \f$ the temperature,
!> * \f$P = \frac{E}{2}\left( (1-\chi)Id  + (3\chi-1)\frac{F \otimes F}{\Vert F \Vert_2^2}\right) \f$ the radiative pressure,
!> * \f$\chi = \frac{3 + 4f^2}{5+ \sqrt{4-3f^2}} \f$ the Eddington factor, in order that \f$\chi\f$ is well defined:
!>     \f$ f=\frac{\Vert F \Vert}{c E} \leq 1 \f$,
!> * \f$ c \f$ the speed of light.
!>
!> @remark See \cite Sar_2013 for the test cases.
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_physics_M1_var
#if defined(_PHY_M1_)

  use mod_numerics, only: PR, pi, MaxLen

  IMPLICIT NONE

  private
  public NbEq, NbEq_str,  T_end
  public len_model, len_primitive, len_conservative
  public model_name, primitive_name, conservative_name, test_case

  public c, sigma, sigma_a, sigma_f, sigma_e, a, ideal_gas_constant
#if defined(_SRC_ENABLE_)
  public rho, C_v
#endif /* _SRC_ENABLE_ */

  integer, parameter                  :: len_model = 2
  !> Name of the model
  character(len=len_model), parameter :: model_name = 'M1'
  !> Number of equations (unknowns) in the model
  integer, parameter                  :: NbEq = 4
  !> Number of equations (unknowns) in the model, in a string
  character(len=1), parameter :: NbEq_str = '4'

  !> Speed of light (\f$c \approx 299\ 792\ 458\ m.s^{-1}\f$)
  real(PR), parameter :: c = 299792458._PR
  !real(PR), parameter :: c = 1._PR

  !> Gas constant (\f$R \approx 3.314\ 462\ 1 \ J.mol^{-1}.K^{-1}\f$)
  real(PR), parameter :: ideal_gas_constant = 8.3144621_PR

  !> Avogadro constant (\f$N_A \approx 6.022\ 141\ 29\ \times 10^{23}\ mol^{-1}\f$)
  real(PR), parameter :: N_A = 6.02214129E+23_PR

  !> Boltzmann constant (\f$k_B = \frac{R}{N_A} \approx 1.380\ 648\ 8 \times 10^{-23}\ J.K^{-1}\f$)
  real(PR), parameter :: k_B = ideal_gas_constant/N_A

  !> Planck constant (\f$h \approx 6.626\ 069\ 57 \times10^{-34}\ J.s^{-1}\f$)
  real(PR), parameter :: h = 6.62606957E-34_PR

  !> Reduced Planck constant (\f$\bar{h}=\frac{h}{2\pi}\ J.s^{-1}\f$)
  real(PR), parameter :: h_bar = h/(2._PR*pi)

  !> Stefan-Boltzmann constant (\f$\sigma =\frac{2\pi^5R^4}{15h^3c^2N_A^4} \approx 5.670\ 373\ 21\times10^{-8}\ W.m^{-2}.K^{-4}\f$)
  real(PR), parameter :: sigma = (2._PR*pi**5*ideal_gas_constant**4)/(15._PR*h**3*c**2*N_A**4)

  !> Mean opacity of absorption
  real(PR), parameter :: sigma_a = 1500._PR
  !> Mean opacity of absorption of flux
  real(PR), parameter :: sigma_f = 1500._PR
  !> Mean opacity of emission
  real(PR), parameter :: sigma_e = 1500._PR

  !> Radiation constant (or radiation density constant) \f$a = \frac{4\sigma}{c} \approx 7.5657 \times10^{-16}\ J.m^{-3}.K^{-4}\f$)
  real(PR), parameter :: a = 4._PR*sigma/c

#if defined(_SRC_ENABLE_)
  !> Density of the fluid (\f$kg.m^{-3}\f$)
  real(PR), dimension(:), allocatable, save :: rho
  !> Volumic heat capacity (\f$J.K^{-1}\f$)
  real(PR), dimension(:), allocatable, save :: C_v
#endif /* _SRC_ENABLE_ */

  !> Final time of the simulation
  real(PR), save :: T_end

  integer, parameter :: len_primitive    = 2
  integer, parameter :: len_conservative = 2

  !> Primitive variables used in the model
  character(len=len_primitive),    dimension(NbEq), parameter :: primitive_name    = ['E ', 'Fx', 'Fy', 'T ']
  !> Conservative variables used in the model
  character(len=len_conservative), dimension(NbEq), parameter :: conservative_name = ['E ', 'Fx', 'Fy', 'T ']

  !> Name of the test case
  character(len=MaxLen), save :: test_case

#endif /* _PHY_M1_ */
end module mod_physics_M1_var
