!> Perform time iterations
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_iterate

  ! Test if the Butcher array can be used
#if !defined(_TIME_RK3TVD_)
#define _TIME_BUTCHER_
#endif /* !defined(_TIME_RK3TVD_) */

  ! If the following line is uncommented the code will
  ! write a file for each time out_disk
!#define _TIME_NO_TIME_STEP_OUTPUTS_
#if !defined(_TIME_NO_TIME_STEP_OUTPUTS_)
#define _TIME_STEP_OUTPUTS_
#endif /* _TIME_NO_TIME_STEP_OUTPUTS_ */

  use mod_numerics, only: PR, MaxLen
  use type_IO, only: StdOut

  use mod_tools_time, only: GetTime, Print_Time_Stats
  use mod_tools_time, only: iter, TimeCounters

  use mod_tools_error, only: PrintError
#if !defined(_FC_PGI_)
  use mod_tools_sys, only: DisplayMemoryUsage
#endif /* !defined(_FC_PGI_) */

  use type_mesh, only: pmesh, dim_
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells
#endif /* _FC_PAR_OMP_ */

  use mod_physics_var, only: NbEq, T_end
  use mod_physics, only: init_U

#if defined(_TP_FLUX_)
  use mod_CFL, only: compute_dt_adm
#else
  use mod_coeff_DLP, only: compute_nu_all
  use mod_CFL, only: compute_dt
#endif /* _TP_FLUX_ */
#if defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_)
  use mod_time_explicit, only: not_last_time_step
#endif /* defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_) */
  use mod_time_explicit, only: compute_U_n_p1
#if defined (_TIME_SSPRK22_)
  use mod_time_explicit, only: U1
#elif defined (_TIME_SSPRK33_)
  use mod_time_explicit, only: U1, U2
#elif defined (_TIME_RK3TVD_)
  use mod_time_explicit, only: U13, U2
#endif /* _TIME_SCHEME_ */

  use mod_Butcher_Tableau, only: ButcherTableau
  use mod_Butcher_Tableau, only: InitButcherTableau, EndButcherTableau
  use mod_time_explicit_Butcher, only: compute_U_n_p1_Butcher
  use mod_time_explicit_Butcher, only: U_But, Phi_But

#if defined(_IO_ENABLE_VTK_)
  use mod_write_VTK, only: write_VTK_primary
  use type_IO, only: out_disk, n_out_disk
#endif /* _IO_ENABLE_VTK_ */
  use type_IO, only: out_screen, n_out_screen
#if defined(_IO_ENABLE_DUMP_)
  use mod_dump, only: dump_solution
  use type_IO, only: dump_disk, n_dump_disk
#endif /* _IO_ENABLE_DUMP_ */
#if defined(_LIB_SILO_)
  use mod_write_silo, only: write_silo_primary
#endif /* _LIB_SILO_ */
#if defined(_TP_CORREC_) || defined(_HO_MOOD_)
  use type_recompute, only: PrintRecompStats
#endif /* _TP_CORREC_ || _HO_MOOD_ */

#if defined(_HO_MOOD_)
  use mod_HO_MOOD, only: InitMOOD_iter
#endif /* _HO_MOOD_ */
#if !defined(_TP_FLUX_) && defined(_SRC_ENABLE_)
  use mod_delta, only: compute_delta_all
#endif /* .not. _TP_FLUX_ && _SRC_ENABLE_ */

  IMPLICIT NONE

  private
  public iterate

CONTAINS
  !--------------------------------------------------------------------------
  !> @brief
  !> Perform the different time steps to get the final time
  !>
  !> @param[in] info Informations about the run
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine iterate(info)
      character(len=*), intent(in) :: info

      integer :: CellK

      real(PR), dimension(:,:), allocatable, save :: U_n, U_n_p1
#if !defined(_TIME_BUTCHER_)
      real(PR), dimension(:,:), allocatable, save :: Phi
#endif /* !defined(_TIME_BUTCHER_) */
      real(PR), save :: T_init, t_n, t_n_p1, dt
      real(PR), save :: completion
      character(len=MaxLen), save :: info_time
      character(len=9), save :: iter_str

      !================================================================!
      !                   Allocations
      !================================================================!
      !$OMP SINGLE
      allocate(U_n(NbEq,pmesh%NbCells))
      allocate(U_n_p1(NbEq,pmesh%NbCells))
#if defined(_TIME_BUTCHER_)
      call InitButcherTableau()
      allocate(U_But(NbEq,pmesh%NbCells,0:ButcherTableau%NbStages))
      allocate(Phi_But(NbEq,pmesh%NbCells,0:ButcherTableau%NbStages-1))
#else
      allocate(Phi(NbEq,pmesh%NbCells))
#if defined (_TIME_SSPRK22_)
      allocate(U1(NbEq,pmesh%NbCells))
#elif defined (_TIME_SSPRK33_)
      allocate(U1(NbEq,pmesh%NbCells))
      allocate(U2(NbEq,pmesh%NbCells))
#elif defined (_TIME_RK3TVD_)
      allocate(U13(NbEq,pmesh%NbCells))
      allocate(U2(NbEq,pmesh%NbCells))
#endif /* _TIME_SCHEME_ */
#endif /* _TIME_BUTCHER_ */
      !================================================================!

#if !defined(_FC_PGI_)
      call DisplayMemoryUsage()
#endif /* !defined(_FC_PGI_) */
      write(StdOut,'(A)') "-------------------------------------------------------"
      !$OMP END SINGLE

      !================================================================!
      !                        Initialisation
      !================================================================!
      ! Initialize U_n, iter and T_init
      call Init_U(U_n,iter,T_init)

      !$OMP SINGLE
      write(StdOut,'(A,ES23.16)') 'Initial time: ', T_init
      write(StdOut,'(A,ES23.16)') 'Final time  : ', T_end
      write(StdOut,'(A)') "-------------------------------------------------------"

#if defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_)
      not_last_time_step = .TRUE.
#endif /* defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_) */
      t_n = T_init
      completion = 0._PR ! = 100._PR*(t - T_init)/(T_end - T_init)
#if defined(_IO_ENABLE_DUMP_)
      n_dump_disk = 0
#endif /* _IO_ENABLE_DUMP_ */
#if defined(_IO_ENABLE_VTK_)
      n_out_disk = 0
#endif /* _IO_ENABLE_VTK_ */
      n_out_screen = 0
      !$OMP END SINGLE

#if defined(_IO_ENABLE_VTK_)
      !--------------------------------------------------------------------!
      !$OMP SINGLE
      TimeCounters%IO(1) = GetTime()
      write(iter_str,'(I9.9)') iter
      info_time = trim(info)//'_'//iter_str
      call write_VTK_primary(U_n,info_time,completion,iter,t_n)
      TimeCounters%IO(2) = GetTime() - TimeCounters%IO(1)
      !$OMP END SINGLE NOWAIT
      !--------------------------------------------------------------------!
#endif /* _IO_ENABLE_VTK_ */

#if defined(_LIB_SILO_)
      !--------------------------------------------------------------------!
      !$OMP SINGLE
      TimeCounters%IO(1) = GetTime()
      write(iter_str,'(I9.9)') iter
      info_time = trim(info)//'_'//iter_str
      call write_silo_primary(U_n,info_time,completion,iter,t_n)
      TimeCounters%IO(2) = GetTime() - TimeCounters%IO(1)
      !$OMP END SINGLE NOWAIT
      !--------------------------------------------------------------------!
#endif /* _LIB_SILO_ */

      !================================================================!

      !================================================================!
      !                         First iteration
      !================================================================!
#if defined(_HO_MOOD_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      !$OMP END SINGLE NOWAIT
      call InitMOOD_iter(U_n)
      !$OMP SINGLE
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE NOWAIT
#endif /* _HO_MOOD_ */
#if defined(_TP_FLUX_)
      call compute_dt_adm(U_n,t_n,t_n_p1,dt)
#else
      call compute_nu_all(U_n,t_n)
#if defined(_SRC_ENABLE_)
      call compute_delta_all()
#endif /* _SRC_ENABLE_ */
      call compute_dt(U_n,t_n,t_n_p1,dt)
#endif /* _TP_FLUX_ */

      !$OMP SINGLE
      TimeCounters%Printing(1) = GetTime()
      call PrintError(iter,dt,t_n,U_n)
      TimeCounters%Printing(2) = TimeCounters%Printing(2) + GetTime() - TimeCounters%Printing(1)
      !$OMP END SINGLE NOWAIT
      !================================================================!

      !================================================================!
      !                 Loop over the time steps
      !================================================================!
      loop_time: do while (t_n_p1 < T_end)
#if defined(_TIME_BUTCHER_)
        call compute_U_n_p1_Butcher(U_n,t_n,t_n_p1,dt,U_n_p1)
#else
        call compute_U_n_p1(U_n,t_n,t_n_p1,dt,Phi,U_n_p1)
#endif /* _TIME_BUTCHER_ */

        !$OMP SINGLE
        iter = iter + 1
        write(iter_str,'(I9.9)') iter
        info_time = trim(info)//'_'//iter_str
        completion = 100._PR*(t_n_p1 - T_init)/(T_end - T_init)
        !$OMP END SINGLE

        !$OMP SECTIONS
        !--------------------------------------------------------------------!
        !                        Output on disk
        !--------------------------------------------------------------------!
#if defined(_IO_ENABLE_VTK_)
        !$OMP SECTION
#if defined(_TIME_STEP_OUTPUTS_)
        if (floor(completion) >= n_out_disk) then
#endif /* _TIME_STEP_OUTPUTS_ */
          TimeCounters%IO(1) = GetTime()
          n_out_disk = n_out_disk + out_disk
          call write_VTK_primary(U_n_p1,info_time,completion,iter,t_n_p1)
          TimeCounters%IO(2) = TimeCounters%IO(2) + GetTime() - TimeCounters%IO(1)
#if defined(_TIME_STEP_OUTPUTS_)
        end if
#endif /* _TIME_STEP_OUTPUTS_ */
#endif /* _IO_ENABLE_VTK_ */
        !--------------------------------------------------------------------!

        !--------------------------------------------------------------------!
        !              Dump solution for a (possible) restart
        !--------------------------------------------------------------------!
#if defined(_IO_ENABLE_DUMP_)
        !$OMP SECTION
#if defined(_TIME_STEP_OUTPUTS_)
        if (floor(completion) >= n_dump_disk) then
#endif /* _TIME_STEP_OUTPUTS_ */
          TimeCounters%Dump(1) = GetTime()
          n_dump_disk = n_dump_disk + dump_disk
          call dump_solution(U_n_p1,info_time,iter,t_n_p1)
          TimeCounters%Dump(2) = TimeCounters%Dump(2) + GetTime() - TimeCounters%Dump(1)
#if defined(_TIME_STEP_OUTPUTS_)
        end if
#endif /* _TIME_STEP_OUTPUTS_ */
#endif /* _IO_ENABLE_DUMP_ */
        !--------------------------------------------------------------------!

        !--------------------------------------------------------------------!
        !                        Output on screen
        !--------------------------------------------------------------------!
        !$OMP SECTION
#if defined(_TIME_STEP_OUTPUTS_)
        if (floor(completion) >= n_out_screen) then
#endif /* _TIME_STEP_OUTPUTS_ */
          TimeCounters%Printing(1) = GetTime()
          n_out_screen = n_out_screen + out_screen
          call PrintError(iter,dt,t_n_p1,U_n_p1)
#if defined(_TP_CORREC_) || defined(_HO_MOOD_)
          call PrintRecompStats(iter)
#endif /* _TP_CORREC_ || _HO_MOOD_ */
          call Print_Time_Stats(t_n_p1 - T_init, completion)
          FLUSH StdOut
          TimeCounters%Printing(2) = TimeCounters%Printing(2) + GetTime() - TimeCounters%Printing(1)
#if defined(_TIME_STEP_OUTPUTS_)
        end if
#endif /* _TIME_STEP_OUTPUTS_ */
        !--------------------------------------------------------------------!

        !--------------------------------------------------------------------!
        !$OMP SECTION
        t_n = t_n_p1
        !--------------------------------------------------------------------!
        !$OMP END SECTIONS NOWAIT

        !$OMP DO SCHEDULE(STATIC,chunk_NbPrimaryCells)
        do CellK = 1, pmesh%NbCells
          U_n(:,CellK) = U_n_p1(:,CellK)
        end do
        !$OMP END DO

#if defined(_HO_MOOD_)
        !$OMP SINGLE
        TimeCounters%ReComp(1) = GetTime()
        !$OMP END SINGLE NOWAIT
        call InitMOOD_iter(U_n)
        !$OMP SINGLE
        TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
        !$OMP END SINGLE NOWAIT
#endif /* _HO_MOOD_ */
#if defined(_TP_FLUX_)
        call compute_dt_adm(U_n,t_n,t_n_p1,dt)
#else
        call compute_nu_all(U_n,t_n)
#if defined(_SRC_ENABLE_)
        call compute_delta_all()
#endif /* _SRC_ENABLE_ */
        call compute_dt(U_n,t_n,t_n_p1,dt)
#endif /* _TP_FLUX_ */
      end do loop_time
      !================================================================!

      !================================================================!
      !          Last time step to have exactly t_n_p1=T_end
      !================================================================!
      !$OMP SINGLE
      dt = T_end - t_n
      t_n_p1 = t_n + dt ! should be equal to T_end
#if defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_)
      not_last_time_step = .FALSE.
#endif /* defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_) */

      iter = iter + 1
      write(iter_str,'(I9.9)') iter
      info_time = trim(info)//'_'//iter_str
      completion = 100._PR ! = 100._PR*(t - T_init)/(T_end - T_init)
      !$OMP END SINGLE

#if defined(_TIME_BUTCHER_)
      call compute_U_n_p1_Butcher(U_n,t_n,t_n_p1,dt,U_n_p1)
#else
      call compute_U_n_p1(U_n,t_n,t_n_p1,dt,Phi,U_n_p1)
#endif /* _TIME_BUTCHER_ */

      !$OMP SECTIONS
      !--------------------------------------------------------------------!
#if defined(_IO_ENABLE_VTK_)
      !$OMP SECTION
      TimeCounters%IO(1) = GetTime()
      call write_VTK_primary(U_n_p1,info_time,completion,iter,t_n_p1)
      TimeCounters%IO(2) = TimeCounters%IO(2) + GetTime() - TimeCounters%IO(1)
      n_out_disk = n_out_disk + out_disk
#endif /* _IO_ENABLE_VTK_ */
      !--------------------------------------------------------------------!

      !--------------------------------------------------------------------!
#if defined(_IO_ENABLE_DUMP_)
      !$OMP SECTION
      TimeCounters%Dump(1) = GetTime()
      call dump_solution(U_n_p1,info_time,iter,t_n_p1)
      TimeCounters%Dump(2) = TimeCounters%Dump(2) + GetTime() - TimeCounters%Dump(1)
      n_dump_disk = n_dump_disk + dump_disk
#endif /* _IO_ENABLE_DUMP_ */
      !--------------------------------------------------------------------!

      !--------------------------------------------------------------------!
      !$OMP SECTION
      TimeCounters%Printing(1) = GetTime()
      call PrintError(iter,dt,t_n_p1,U_n_p1)
      TimeCounters%Printing(2) = TimeCounters%Printing(2) + GetTime() - TimeCounters%Printing(1)
      !--------------------------------------------------------------------!
      !$OMP END SECTIONS

      !================================================================!

      !$OMP SINGLE
      deallocate(U_n_p1)
      deallocate(U_n)
#if defined(_TIME_BUTCHER_)
      call EndButcherTableau()
      deallocate(U_But)
      deallocate(Phi_But)
#else
      deallocate(Phi)
#if defined (_TIME_SSPRK22_)
      deallocate(U1)
#elif defined (_TIME_SSPRK33_)
      deallocate(U1)
      deallocate(U2)
#elif defined (_TIME_RK3TVD_)
      deallocate(U13)
      deallocate(U2)
#endif /* _TIME_SCHEME_ */
#endif /* _TIME_BUTCHER_ */
      !$OMP END SINGLE NOWAIT

  end subroutine iterate
end module mod_iterate
