!> Performs a time iteration with the right time scheme
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_time_explicit

  use mod_numerics, only: PR
  use type_IO, only: StdErr

  use type_mesh, only: pmesh

#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells
#endif /* _FC_PAR_OMP_ */

  use mod_operators, only: compute_Phi

#if defined(_TP_CORREC_) && !defined(_HO_MOOD_)
  use mod_TPCorrec, only: GetCellsToCorrec
  use mod_TPCorrec_recompute, only: Recompute_dt_TPCorrec, RecomputePhiTPCorrec
#endif /* _TP_CORREC_ && !_HO_MOOD_ */

#if defined(_HO_MOOD_)
  use mod_HO_MOOD, only: InitMOOD_iter
  use mod_HO_MOOD_degree, only: GetCellsToUpdate
  use mod_HO_MOOD_recompute, only: RecomputePhiMOOD
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
  use type_recompute, only: NbCellsInvalidTP
  use type_recompute, only: TotReCompTP
  use type_recompute, only: MaxInvalidCellsTP
  use type_recompute, only: FlagReCompTP
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
#endif /* _HO_MOOD_ */

#if defined(_TP_CORREC_) || defined(_HO_MOOD_)
  use type_recompute, only: NbCellsInvalid, NbCellsToUpdate
  use type_recompute, only: CellsToUpdate
  use mod_tools_time, only: GetTime
  use mod_tools_time, only: TimeCounters
  use type_recompute, only: TotReComp, TotIterReComp
  use type_recompute, only: MaxInvalidCells, MaxIterReComp
  use type_recompute, only: MaxIterPossibleReComp
  use type_recompute, only: FlagReComp, NbIterReComp
#endif /* _TP_CORREC_ || _HO_MOOD_ */

#if !defined(_TP_FLUX_)
  use mod_coeff_DLP, only: compute_nu_all
#if defined(_SRC_ENABLE_)
  use mod_delta, only: compute_delta_all
#endif /* _SRC_ENABLE_ */
#endif /* .not. _TP_FLUX_ */

  IMPLICIT NONE

  private
#if defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_)
  public not_last_time_step
#endif /* defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_) */
  public compute_U_n_p1
#if defined (_TIME_SSPRK22_)
  public U1
#elif defined (_TIME_SSPRK33_)
  public U1
  public U2
#elif defined (_TIME_RK3TVD_)
  public U13
  public U2
#endif /* _TIME_SCHEME_ */

#if defined (_TIME_SSPRK22_)
  real(PR), dimension(:,:), allocatable :: U1
#elif defined (_TIME_SSPRK33_)
  real(PR), dimension(:,:), allocatable :: U1
  real(PR), dimension(:,:), allocatable :: U2
#elif defined (_TIME_RK3TVD_)
  real(PR), dimension(:,:), allocatable :: U13
  real(PR), dimension(:,:), allocatable :: U2
#endif /* _TIME_SCHEME_ */

#if defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_)
  logical, save ::  not_last_time_step
#if defined (_TIME_FE_) || defined (_TIME_SSPRK22_) || defined (_TIME_SSPRK33_) || defined (_TIME_RK3TVD_)
  logical :: dt_changed
#endif /* defined (_TIME_FE_) || defined (_TIME_SSPRK22_) || defined (_TIME_SSPRK33_) || defined (_TIME_RK3TVD_) */
#endif /* defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_) */

CONTAINS
  !--------------------------------------------------------------------------
  !> @brief
  !> Performs a time iteration with the right time scheme
  !>
  !> \f[U^{n} \Rightarrow U^{n+1}\f]
  !>
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[inout]  t_n_p1, dt Next time and time step
  !> @param[inout] Phi Working array
  !> @param[out] U_n_p1 solution at next time step
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine compute_U_n_p1(U_n,t_n,t_n_p1,dt,Phi,U_n_p1)
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n
      real(PR), intent(inout) :: t_n_p1, dt
      real(PR), dimension(:,:), intent(inout) :: Phi
      real(PR), dimension(:,:), intent(out) :: U_n_p1

#if defined (_TIME_FE_)
      call Forward_Euler(U_n,t_n,t_n_p1,dt,Phi,U_n_p1)
#elif defined (_TIME_SSPRK22_)
      call SSPRK22(U_n,t_n,t_n_p1,dt,Phi,U_n_p1)
#elif defined (_TIME_SSPRK33_)
      call SSPRK33(U_n,t_n,t_n_p1,dt,Phi,U_n_p1)
#elif defined (_TIME_RK3TVD_)
      call RK3TVD(U_n,t_n,t_n_p1,dt,Phi,U_n_p1)
#else
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> No time scheme defined..."
      FLUSH StdErr
      STOP
#endif /* _TIME_SCHEME_ */

  end subroutine compute_U_n_p1

#if defined (_TIME_FE_)
  !--------------------------------------------------------------------------
  !> @brief
  !> Forward Euler scheme
  !>
  !> \f[U^{n+1} = U^n - \Delta t \ \Phi(U^n)\f]
  !>
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[inout]  t_n_p1, dt Next time and time step
  !> @param[inout] Phi Working array
  !> @param[out] U_n_p1 solution at next time step
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine Forward_Euler(U_n,t_n,t_n_p1,dt,Phi,U_n_p1)
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n
      real(PR), intent(inout) :: t_n_p1, dt
      real(PR), dimension(:,:), intent(out) :: Phi
      real(PR), dimension(:,:), intent(out) :: U_n_p1

      integer :: CellK
#if defined(_TP_CORREC_) || defined(_HO_MOOD_)
      integer :: idxK
#endif /*_TP_CORREC_ || _HO_MOOD_ */

      !======================================================================
      call compute_Phi(U_n,t_n,Phi)

      !$OMP DO &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      loop_elt: do CellK = 1, pmesh%NbCells
        U_n_p1(:,CellK) = U_n(:,CellK) - dt*Phi(:,CellK)
      end do loop_elt
      !$OMP END DO

#if defined(_HO_MOOD_) && !defined(_HO_MOOD_UNLIMITED_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      FlagReComp = 0
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
      FlagReCompTP = 0
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
      NbIterReComp = 0
      !$OMP END SINGLE
      call GetCellsToUpdate(U_n,U_n_p1,t_n,dt,NbIterReComp)
      loop_MOOD: do while (NbCellsInvalid > 0 .and. NbIterReComp < MaxIterPossibleReComp)
        !$OMP SINGLE
        FlagReComp = 1
        NbIterReComp = NbIterReComp + 1
        MaxInvalidCells = max(MaxInvalidCells,NbCellsInvalid)
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
        test_has_invalidTP_cells: if (NbCellsInvalidTP > 0) then
          FlagReCompTP = 1
        end if test_has_invalidTP_cells
        MaxInvalidCellsTP = max(MaxInvalidCellsTP,NbCellsInvalidTP)
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
        !$OMP END SINGLE NOWAIT
        call RecomputePhiMOOD(U_n,t_n,Phi)
        ! Recompute only on decremented cell and their neighbors
        ! as dt does not change
        !$OMP DO
        loop_elt_MOOD: do idxK = 1, NbCellsToUpdate
          CellK = CellsToUpdate(idxK)
          U_n_p1(:,CellK) = U_n(:,CellK) - dt*Phi(:,CellK)
        end do loop_elt_MOOD
        !$OMP END DO
        call GetCellsToUpdate(U_n,U_n_p1,t_n,dt,NbIterReComp)
      end do loop_MOOD
      !$OMP SINGLE
      if (NbCellsInvalid > 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Bailing out ..."
        write(StdErr,'(I9,A)') NbCellsInvalid, ' invalid cells,'
        write(StdErr,'(I9,A)') NbCellsToUpdate, ' cells to update,'
        write(StdErr,'(I9,A)') NbIterReComp, ' loop of re-computations done.'
        FLUSH StdErr
        STOP
      endif
      MaxIterReComp = max(MaxIterReComp,NbIterReComp)
      TotIterReComp = TotIterReComp + NbIterReComp
      TotReComp = TotReComp + FlagReComp
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
      TotReCompTP = TotReCompTP + FlagReCompTP
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* defined(_HO_MOOD_) && defined(_HO_MOOD_UNLIMITED_) */

#if defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      FlagReComp = 0
      NbIterReComp = 0
      !$OMP END SINGLE
      call GetCellsToCorrec(U_n,U_n_p1,NbIterReComp)
      loop_correc: do while (NbCellsInvalid > 0 .and. NbIterReComp < MaxIterPossibleReComp)
        !$OMP SINGLE
        FlagReComp = 1
        NbIterReComp = NbIterReComp + 1
        MaxInvalidCells = max(MaxInvalidCells,NbCellsInvalid)
        !$OMP END SINGLE NOWAIT

        if (not_last_time_step) then
          ! Recompute the time step by recomputing only on corrected cells
          call Recompute_dt_TPCorrec(U_n,t_n,t_n_p1,dt,dt_changed)
        else
          !$OMP SINGLE
          dt_changed = .FALSE.
          !$OMP END SINGLE
        end if

        ! Recompute the sum of flux only on corrected cells
        call RecomputePhiTPCorrec(U_n,t_n,Phi)

        if (dt_changed) then
          ! Recompute the solution on all the cells, as dt changed
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          loop_elt_correc: do CellK = 1, pmesh%NbCells
            U_n_p1(:,CellK) = U_n(:,CellK) - dt*Phi(:,CellK)
          end do loop_elt_correc
          !$OMP END DO
        else
          ! Recompute the solution only on updated cells as dt did not change
          !$OMP DO
          loop_CellsToUpdate_correc: do idxK = 1, NbCellsToUpdate
            CellK = CellsToUpdate(idxK)
            U_n_p1(:,CellK) = U_n(:,CellK) - dt*Phi(:,CellK)
          end do loop_CellsToUpdate_correc
          !$OMP END DO
        endif

        call GetCellsToCorrec(U_n,U_n_p1,NbIterReComp)
      end do loop_correc
      !$OMP SINGLE
      if (NbCellsInvalid > 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Bailing out ..."
        write(StdErr,'(I9,A)') NbCellsInvalid, ' invalid cells,'
        write(StdErr,'(I9,A)') NbCellsToUpdate, ' cells to update,'
        write(StdErr,'(I9,A)') NbIterReComp, ' loop of re-computations done.'
        FLUSH StdErr
        STOP
      endif
      MaxIterReComp = max(MaxIterReComp,NbIterReComp)
      TotIterReComp = TotIterReComp + NbIterReComp
      TotReComp = TotReComp + FlagReComp
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* defined(_TP_CORREC_) && defined(_TP_FLUX_) && defined(_HO_MOOD_) */
      !======================================================================

  end subroutine Forward_Euler

#elif defined (_TIME_SSPRK22_)
  !--------------------------------------------------------------------------
  !> @brief
  !> Optimal SSPRK2 for nonlinear operators (see (4.1) in \cite GotShuTad_2001 )
  !>
  !> \f[\begin{array}{rcl}
  !>    U^{(1)} &=& U^n - \Delta t \ \Phi(U^n)\\
  !>    U^{n+1} &=& \frac{1}{2} \left( U^{n} + U^{(1)} - \Delta t \ \Phi(U^{(1)}) \right)
  !> \end{array}\f]
  !>
  !> see also (2.4) is \cite GotShu_1998 , and (2.15,2.16) in \cite ShuOsh_1988
  !>
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[inout]  t_n_p1, dt Next time and time step
  !> @param[inout] Phi Working array
  !> @param[out] U_n_p1 solution at next time step
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine SSPRK22(U_n,t_n,t_n_p1,dt,Phi,U_n_p1)
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n
      real(PR), intent(inout) :: t_n_p1, dt
      real(PR), dimension(:,:), intent(inout) :: Phi
      real(PR), dimension(:,:), intent(out) :: U_n_p1

      real(PR) :: t2
      integer :: CellK
#if defined(_TP_CORREC_) || defined(_HO_MOOD_)
      integer :: idxK
#endif /*_TP_CORREC_ || _HO_MOOD_ */

      !======================================================================
      call compute_Phi(U_n,t_n,Phi)

      !$OMP DO &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      loop_elt_1: do CellK = 1, pmesh%NbCells
        U1(:,CellK) = U_n(:,CellK) - dt*Phi(:,CellK)
      end do loop_elt_1
      !$OMP END DO

#if defined(_HO_MOOD_) && !defined(_HO_MOOD_UNLIMITED_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      FlagReComp = 0
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
      FlagReCompTP = 0
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
      NbIterReComp = 0
      !$OMP END SINGLE
      call GetCellsToUpdate(U_n,U1,t_n,dt,NbIterReComp)
      loop_MOOD_1: do while (NbCellsInvalid > 0 .and. NbIterReComp < MaxIterPossibleReComp)
        !$OMP SINGLE
        FlagReComp = 1
        NbIterReComp = NbIterReComp + 1
        MaxInvalidCells = max(MaxInvalidCells,NbCellsInvalid)
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
        test_has_invalidTP_cells_1: if (NbCellsInvalidTP > 0) then
          FlagReCompTP = 1
        end if test_has_invalidTP_cells_1
        MaxInvalidCellsTP = max(MaxInvalidCellsTP,NbCellsInvalidTP)
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
        !$OMP END SINGLE NOWAIT
        call RecomputePhiMOOD(U_n,t_n,Phi)
        ! Recompute only on decremented cell and their neighbors
        ! as dt does not change
        !$OMP DO
        loop_elt_1_MOOD: do idxK = 1, NbCellsToUpdate
          CellK = CellsToUpdate(idxK)
          U1(:,CellK) = U_n(:,CellK) - dt*Phi(:,CellK)
        end do loop_elt_1_MOOD
        !$OMP END DO
        call GetCellsToUpdate(U_n,U1,t_n,dt,NbIterReComp)
      end do loop_MOOD_1
      !$OMP SINGLE
      if (NbCellsInvalid > 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Bailing out ..."
        write(StdErr,'(I9,A)') NbCellsInvalid, ' invalid cells,'
        write(StdErr,'(I9,A)') NbCellsToUpdate, ' cells to update,'
        write(StdErr,'(I9,A)') NbIterReComp, ' loop of re-computations done.'
        FLUSH StdErr
        STOP
      endif
      MaxIterReComp = max(MaxIterReComp,NbIterReComp)
      TotIterReComp = TotIterReComp + NbIterReComp
      TotReComp = TotReComp + FlagReComp
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
      TotReCompTP = TotReCompTP + FlagReCompTP
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _HO_MOOD_ && _HO_MOOD_UNLIMITED_ */

#if defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      FlagReComp = 0
      NbIterReComp = 0
      !$OMP END SINGLE
      call GetCellsToCorrec(U_n,U1,NbIterReComp)
      loop_correc_1: do while (NbCellsInvalid > 0 .and. NbIterReComp < MaxIterPossibleReComp)
        !$OMP SINGLE
        FlagReComp = 1
        NbIterReComp = NbIterReComp + 1
        MaxInvalidCells = max(MaxInvalidCells,NbCellsInvalid)
        !$OMP END SINGLE NOWAIT

        if (not_last_time_step) then
          ! Recompute the time step by recomputing only on corrected cells
          call Recompute_dt_TPCorrec(U_n,t_n,t_n_p1,dt,dt_changed)
        else
          !$OMP SINGLE
          dt_changed = .FALSE.
          !$OMP END SINGLE
        end if

        ! Recompute the sum of flux only on corrected cells
        call RecomputePhiTPCorrec(U_n,t_n,Phi)

        if (dt_changed) then
          ! Recompute the solution on all the cells, as dt changed
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          loop_elt_1_correc: do CellK = 1, pmesh%NbCells
            U1(:,CellK) = U_n(:,CellK) - dt*Phi(:,CellK)
          end do loop_elt_1_correc
          !$OMP END DO
        else
          ! Recompute the solution only on updated cells as dt did not change
          !$OMP DO
          loop_CellsTopUpdate_1_correc: do idxK = 1, NbCellsToUpdate
            CellK = CellsToUpdate(idxK)
            U1(:,CellK) = U_n(:,CellK) - dt*Phi(:,CellK)
          end do loop_CellsTopUpdate_1_correc
          !$OMP END DO
        endif

        call GetCellsToCorrec(U_n,U1,NbIterReComp)
      end do loop_correc_1
      !$OMP SINGLE
      if (NbCellsInvalid > 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Bailing out ..."
        write(StdErr,'(I9,A)') NbCellsInvalid, ' invalid cells,'
        write(StdErr,'(I9,A)') NbCellsToUpdate, ' cells to update,'
        write(StdErr,'(I9,A)') NbIterReComp, ' loop of re-computations done.'
        FLUSH StdErr
        STOP
      endif
      MaxIterReComp = max(MaxIterReComp,NbIterReComp)
      TotIterReComp = TotIterReComp + NbIterReComp
      TotReComp = TotReComp + FlagReComp
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _TP_CORREC_ && _TP_FLUX_ && _HO_MOOD_ */
      !======================================================================

      t2 = t_n + dt

      !======================================================================
#if defined(_HO_MOOD_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      !$OMP END SINGLE
      call InitMOOD_iter(U1)
      !$OMP SINGLE
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _HO_MOOD_ */

#if !defined(_TP_FLUX_)
      call compute_nu_all(U1,t2)
#if defined(_SRC_ENABLE_)
      call compute_delta_all()
#endif /* _SRC_ENABLE_ */
#endif /* _TP_FLUX_ */

      call compute_Phi(U1,t2,Phi)

      !$OMP DO &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      loop_elt_n: do CellK = 1, pmesh%NbCells
        U_n_p1(:,CellK) = 0.5_PR*(U_n(:,CellK) + U1(:,CellK) - dt*Phi(:,CellK))
      end do loop_elt_n
      !$OMP END DO

#if defined(_HO_MOOD_) && !defined(_HO_MOOD_UNLIMITED_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      FlagReComp = 0
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
      FlagReCompTP = 0
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
      NbIterReComp = 0
      !$OMP END SINGLE
      call GetCellsToUpdate(U1,U_n_p1,t_n,dt,NbIterReComp)
      loop_MOOD_n: do while (NBCellsInvalid > 0 .and. NbIterReComp < MaxIterPossibleReComp)
        !$OMP SINGLE
        FlagReComp = 1
        NbIterReComp = NbIterReComp + 1
        MaxInvalidCells = max(MaxInvalidCells,NbCellsInvalid)
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
        test_has_invalidTP_cells_n: if (NbCellsInvalidTP > 0) then
          FlagReCompTP = 1
        end if test_has_invalidTP_cells_n
        MaxInvalidCellsTP = max(MaxInvalidCellsTP,NbCellsInvalidTP)
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
        !$OMP END SINGLE NOWAIT
        call RecomputePhiMOOD(U1,t2,Phi)
        ! Recompute only on decremented cell and their neigbors
        ! as dt does not change
        !$OMP DO
        loop_elt_n_MOOD: do idxK = 1, NbCellsToUpdate
          CellK = CellsToUpdate(idxK)
          U_n_p1(:,CellK) = 0.5_PR*(U_n(:,CellK) + U1(:,CellK) - dt*Phi(:,CellK))
        end do loop_elt_n_MOOD
        !$OMP END DO
        call GetCellsToUpdate(U1,U_n_p1,t_n,dt,NbIterReComp)
      end do loop_MOOD_n
      !$OMP SINGLE
      if (NbCellsInvalid > 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Bailing out ..."
        write(StdErr,'(I9,A)') NbCellsInvalid, ' invalid cells,'
        write(StdErr,'(I9,A)') NbCellsToUpdate, ' cells to update,'
        write(StdErr,'(I9,A)') NbIterReComp, ' loop of re-computations done.'
        FLUSH StdErr
        STOP
      endif
      MaxIterReComp = max(MaxIterReComp,NbIterReComp)
      TotIterReComp = TotIterReComp + NbIterReComp
      TotReComp = TotReComp + FlagReComp
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
      TotReCompTP = TotReCompTP + FlagReCompTP
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _HO_MOOD_ && _HO_MOOD_UNLIMITED_ */

#if defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      FlagReComp = 0
      NbIterReComp = 0
      !$OMP END SINGLE
      call GetCellsToCorrec(U1,U_n_p1,NbIterReComp)
      loop_correc_n: do while (NbCellsInvalid > 0 .and. NbIterReComp < MaxIterPossibleReComp)
        !$OMP SINGLE
        FlagReComp = 1
        NbIterReComp = NbIterReComp + 1
        MaxInvalidCells = max(MaxInvalidCells,NbCellsInvalid)
        !$OMP END SINGLE NOWAIT

        if (not_last_time_step) then
          ! Recompute the time step by recomputing only on corrected cells
          call Recompute_dt_TPCorrec(U1,t_n,t_n_p1,dt,dt_changed)
          !$OMP SINGLE
          if (dt_changed) then
            t2 = t_n + 0.5_PR*dt
          end if
          !$OMP END SINGLE
        else
          !$OMP SINGLE
          dt_changed = .FALSE.
          !$OMP END SINGLE
        end if
        ! Recompute the sum of flux only on corrected cells
        call RecomputePhiTPCorrec(U1,t_n,Phi)

        if (dt_changed) then
          ! Recompute the solution on all the cells, as dt changed
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          loop_elt_n_correc: do CellK = 1, pmesh%NbCells
            U_n_p1(:,CellK) = 0.5_PR*(U_n(:,CellK) + U1(:,CellK) - dt*Phi(:,CellK))
          end do loop_elt_n_correc
          !$OMP END DO
        else
          ! Recompute the solution only on updated cells as dt did not change
          !$OMP DO
          loop_CellsToUpdate_n_correc: do idxK = 1, NbCellsToUpdate
            CellK = CellsToUpdate(idxK)
            U_n_p1(:,CellK) = 0.5_PR*(U_n(:,CellK) + U1(:,CellK) - dt*Phi(:,CellK))
          end do loop_CellsToUpdate_n_correc
          !$OMP END DO
        endif

        call GetCellsToCorrec(U1,U_n_p1,NbIterReComp)
      end do loop_correc_n
      !$OMP SINGLE
      if (NbCellsInvalid > 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Bailing out ..."
        write(StdErr,'(I9,A)') NbCellsInvalid, ' invalid cells,'
        write(StdErr,'(I9,A)') NbCellsToUpdate, ' cells to update,'
        write(StdErr,'(I9,A)') NbIterReComp, ' loop of re-computations done.'
        FLUSH StdErr
        STOP
      endif
      MaxIterReComp = max(MaxIterReComp,NbIterReComp)
      TotIterReComp = TotIterReComp + NbIterReComp
      TotReComp = TotReComp + FlagReComp
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _TP_CORREC_ && _TP_FLUX_ && _HO_MOOD_ */
      !======================================================================

  end subroutine SSPRK22

#elif defined (_TIME_SSPRK33_)
  !--------------------------------------------------------------------------
  !> @brief
  !> Optimal SSPRK3 for nonlinear operators (see (4.2) in \cite GotShuTad_2001 )
  !>
  !> \f[\begin{array}{rcl}
  !>    U^{(1)} &=& U^n - \Delta t \ \Phi(U^n)\\
  !>    U^{(2)} &=& \frac{1}{4} \left( 3 U^n + U^{(1)} - \Delta t \ \Phi(U^{(1)}) \right) \\
  !>    U^{n+1} &=& \frac{1}{3} \left( U^{n} + 2 U^{(2)} - 2 \Delta t \ \Phi(U^{(2)}) \right)
  !> \end{array}\f]
  !>
  !> see also (3.3) is \cite GotShu_1998 , and (2.18,2.19) in \cite ShuOsh_1988
  !>
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[inout]  t_n_p1, dt Next time and time step
  !> @param[inout] Phi Working array
  !> @param[out] U_n_p1 solution at next time step
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine SSPRK33(U_n,t_n,t_n_p1,dt,Phi,U_n_p1)
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n
      real(PR), intent(inout) :: t_n_p1, dt
      real(PR), dimension(:,:), intent(inout) :: Phi
      real(PR), dimension(:,:), intent(out) :: U_n_p1

      real(PR) :: t2, t3
      integer :: CellK
#if defined(_TP_CORREC_) || defined(_HO_MOOD_)
      integer :: idxK
#endif /*_TP_CORREC_ || _HO_MOOD_ */


      !======================================================================
      call compute_Phi(U_n,t_n,Phi)

      !$OMP DO &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      loop_elt_1: do CellK = 1, pmesh%NbCells
        U1(:,CellK) = U_n(:,CellK) - dt*Phi(:,CellK)
      end do loop_elt_1
      !$OMP END DO

#if defined(_HO_MOOD_) && !defined(_HO_MOOD_UNLIMITED_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      FlagReComp = 0
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
      FlagReCompTP = 0
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
      NbIterReComp = 0
      !$OMP END SINGLE
      call GetCellsToUpdate(U_n,U1,t_n,dt,NbIterReComp)
      loop_MOOD_1: do while (NbCellsInvalid > 0 .and. NbIterReComp < MaxIterPossibleReComp)
        !$OMP SINGLE
        FlagReComp = 1
        NbIterReComp = NbIterReComp + 1
        MaxInvalidCells = max(MaxInvalidCells,NbCellsInvalid)
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
        test_has_invalidTP_cells_1: if (NbCellsInvalidTP > 0) then
          FlagReCompTP = 1
        end if test_has_invalidTP_cells_1
        MaxInvalidCellsTP = max(MaxInvalidCellsTP,NbCellsInvalidTP)
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
        !$OMP END SINGLE NOWAIT
        call RecomputePhiMOOD(U_n,t_n,Phi)
        ! Recompute only on decremented cell and their neigbors
        ! as dt does not change
        !$OMP DO
        loop_elt_1_MOOD: do idxK = 1, NbCellsToUpdate
          CellK = CellsToUpdate(idxK)
          U1(:,CellK) = U_n(:,CellK) - dt*Phi(:,CellK)
        end do loop_elt_1_MOOD
        !$OMP END DO
        call GetCellsToUpdate(U_n,U1,t_n,dt,NbIterReComp)
      end do loop_MOOD_1
      !$OMP SINGLE
      if (NbCellsInvalid > 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Bailing out ..."
        write(StdErr,'(I9,A)') NbCellsInvalid, ' invalid cells,'
        write(StdErr,'(I9,A)') NbCellsToUpdate, ' cells to update,'
        write(StdErr,'(I9,A)') NbIterReComp, ' loop of re-computations done.'
        FLUSH StdErr
        STOP
      endif
      MaxIterReComp = max(MaxIterReComp,NbIterReComp)
      TotIterReComp = TotIterReComp + NbIterReComp
      TotReComp = TotReComp + FlagReComp
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
      TotReCompTP = TotReCompTP + FlagReCompTP
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _HO_MOOD_ && _HO_MOOD_UNLIMITED_ */

#if defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      FlagReComp = 0
      NbIterReComp = 0
      !$OMP END SINGLE
      call GetCellsToCorrec(U_n,U1,NbIterReComp)
      loop_correc_1: do while (NbCellsInvalid > 0 .and. NbIterReComp < MaxIterPossibleReComp)
        !$OMP SINGLE
        FlagReComp = 1
        NbIterReComp = NbIterReComp + 1
        MaxInvalidCells = max(MaxInvalidCells,NbCellsInvalid)
        !$OMP END SINGLE NOWAIT

        if (not_last_time_step) then
          ! Recompute the time step by recomputing only on corrected cells
          call Recompute_dt_TPCorrec(U_n,t_n,t_n_p1,dt,dt_changed)
        else
          !$OMP SINGLE
          dt_changed = .FALSE.
          !$OMP END SINGLE
        end if

        ! Recompute the sum of flux only on corrected cells
        call RecomputePhiTPCorrec(U_n,t_n,Phi)

        if (dt_changed) then
          ! Recompute the solution on all the cells, as dt changed
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          loop_elt_1_correc: do CellK = 1, pmesh%NbCells
            U1(:,CellK) = U_n(:,CellK) - dt*Phi(:,CellK)
          end do loop_elt_1_correc
          !$OMP END DO
        else
          ! Recompute the solution only on updated cells as dt did not change
          !$OMP DO
          loop_CellsToUpdate_1_correc: do idxK = 1, NbCellsToUpdate
            CellK = CellsToUpdate(idxK)
            U1(:,CellK) = U_n(:,CellK) - dt*Phi(:,CellK)
          end do loop_CellsToUpdate_1_correc
          !$OMP END DO
        endif

        call GetCellsToCorrec(U_n,U1,NbIterReComp)
      end do loop_correc_1
      !$OMP SINGLE
      if (NbCellsInvalid > 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Bailing out ..."
        write(StdErr,'(I9,A)') NbCellsInvalid, ' invalid cells,'
        write(StdErr,'(I9,A)') NbCellsToUpdate, ' cells to update,'
        write(StdErr,'(I9,A)') NbIterReComp, ' loop of re-computations done.'
        FLUSH StdErr
        STOP
      endif
      MaxIterReComp = max(MaxIterReComp,NbIterReComp)
      TotIterReComp = TotIterReComp + NbIterReComp
      TotReComp = TotReComp + FlagReComp
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _TP_CORREC_ && _TP_FLUX_ && _HO_MOOD_ */
      !======================================================================

      t2 = t_n + dt

      !======================================================================
#if defined(_HO_MOOD_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      !$OMP END SINGLE
      call InitMOOD_iter(U1)
      !$OMP SINGLE
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _HO_MOOD_ */

#if !defined(_TP_FLUX_)
      call compute_nu_all(U1,t2)
#if defined(_SRC_ENABLE_)
      call compute_delta_all()
#endif /* _SRC_ENABLE_ */
#endif /* _TP_FLUX_ */

      call compute_Phi(U1,t2,Phi)

      !$OMP DO &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      loop_elt_2: do CellK = 1, pmesh%NbCells
        U2(:,CellK) = 0.25_PR*(3._PR*U_n(:,CellK) + U1(:,CellK) - dt*Phi(:,CellK))
      end do loop_elt_2
      !$OMP END DO

#if defined(_HO_MOOD_) && !defined(_HO_MOOD_UNLIMITED_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      FlagReComp = 0
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
      FlagReCompTP = 0
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
      NbIterReComp = 0
      !$OMP END SINGLE
      call GetCellsToUpdate(U1,U2,t_n,dt,NbIterReComp)
      loop_MOOD_2: do while (NbCellsInvalid > 0 .and. NbIterReComp < MaxIterPossibleReComp)
        !$OMP SINGLE
        FlagReComp = 1
        NbIterReComp = NbIterReComp + 1
        MaxInvalidCells = max(MaxInvalidCells,NbCellsInvalid)
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
        test_has_invalidTP_cells_2: if (NbCellsInvalidTP > 0) then
          FlagReCompTP = 1
        end if test_has_invalidTP_cells_2
        MaxInvalidCellsTP = max(MaxInvalidCellsTP,NbCellsInvalidTP)
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
        !$OMP END SINGLE NOWAIT
        call RecomputePhiMOOD(U1,t2,Phi)
        ! Recompute only on decremented cell and their neigbors
        ! as dt does not change
        !$OMP DO
        loop_elt_2_MOOD: do idxK = 1, NbCellsToUpdate
          CellK = CellsToUpdate(idxK)
          U2(:,CellK) = 0.25_PR*(3._PR*U_n(:,CellK) + U1(:,CellK) - dt*Phi(:,CellK))
        end do loop_elt_2_MOOD
        !$OMP END DO
        call GetCellsToUpdate(U1,U2,t_n,dt,NbIterReComp)
      end do loop_MOOD_2
      !$OMP SINGLE
      if (NbCellsInvalid > 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Bailing out ..."
        write(StdErr,'(I9,A)') NbCellsInvalid, ' invalid cells,'
        write(StdErr,'(I9,A)') NbCellsToUpdate, ' cells to update,'
        write(StdErr,'(I9,A)') NbIterReComp, ' loop of re-computations done.'
        FLUSH StdErr
        STOP
      endif
      MaxIterReComp = max(MaxIterReComp,NbIterReComp)
      TotIterReComp = TotIterReComp + NbIterReComp
      TotReComp = TotReComp + FlagReComp
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
      TotReCompTP = TotReCompTP + FlagReCompTP
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _HO_MOOD_ && _HO_MOOD_UNLIMITED_ */

#if defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      FlagReComp = 0
      NbIterReComp = 0
      !$OMP END SINGLE
      call GetCellsToCorrec(U1,U2,NbIterReComp)
      loop_correc_2: do while (NbCellsInvalid > 0 .and. NbIterReComp < MaxIterPossibleReComp)
        !$OMP SINGLE
        FlagReComp = 1
        NbIterReComp = NbIterReComp + 1
        MaxInvalidCells = max(MaxInvalidCells,NbCellsInvalid)
        !$OMP END SINGLE NOWAIT

        if (not_last_time_step) then
          ! Recompute the time step by recomputing only on corrected cells
          call Recompute_dt_TPCorrec(U1,t_n,t_n_p1,dt,dt_changed)
          !$OMP SINGLE
          if (dt_changed) then
            t2 = t_n + dt
          end if
          !$OMP END SINGLE
        else
          !$OMP SINGLE
          dt_changed = .FALSE.
          !$OMP END SINGLE
        end if

        ! Recompute the sum of flux only on corrected cells
        call RecomputePhiTPCorrec(U1,t_n,Phi)


        if (dt_changed) then
          ! Recompute the solution on all the cells, as dt changed
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          loop_elt_2_correc: do CellK = 1, pmesh%NbCells
            U2(:,CellK) = 0.25_PR*(3._PR*U_n(:,CellK) + U1(:,CellK) - dt*Phi(:,CellK))
          end do loop_elt_2_correc
          !$OMP END DO
        else
          ! Recompute the solution only on updated cells as dt did not change
          !$OMP DO
          loop_CellsTopUpdate_2_correc: do idxK = 1, NbCellsToUpdate
            CellK = CellsToUpdate(idxK)
            U2(:,CellK) = 0.25_PR*(3._PR*U_n(:,CellK) + U1(:,CellK) - dt*Phi(:,CellK))
          end do loop_CellsTopUpdate_2_correc
          !$OMP END DO
        endif

        call GetCellsToCorrec(U1,U2,NbIterReComp)
      end do loop_correc_2
      !$OMP SINGLE
      if (NbCellsInvalid > 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Bailing out ..."
        write(StdErr,'(I9,A)') NbCellsInvalid, ' invalid cells,'
        write(StdErr,'(I9,A)') NbCellsToUpdate, ' cells to update,'
        write(StdErr,'(I9,A)') NbIterReComp, ' loop of re-computations done.'
        FLUSH StdErr
        STOP
      endif
      MaxIterReComp = max(MaxIterReComp,NbIterReComp)
      TotIterReComp = TotIterReComp + NbIterReComp
      TotReComp = TotReComp + FlagReComp
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _TP_CORREC_ && _TP_FLUX_ && _HO_MOOD_ */
      !======================================================================

      t3 = t_n + 0.5_PR*dt

      !======================================================================
#if defined(_HO_MOOD_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      !$OMP END SINGLE
      call InitMOOD_iter(U2)
      !$OMP SINGLE
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _HO_MOOD_ */

#if !defined(_TP_FLUX_)
      call compute_nu_all(U2,t3)
#if defined(_SRC_ENABLE_)
      call compute_delta_all()
#endif /* _SRC_ENABLE_ */
#endif /* _TP_FLUX_ */

      call compute_Phi(U2,t3,Phi)

      !$OMP DO &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      loop_elt_n: do CellK = 1, pmesh%NbCells
        U_n_p1(:,CellK) = (U_n(:,CellK) + 2._PR*U2(:,CellK) - 2._PR*dt*Phi(:,CellK))/3._PR
      end do loop_elt_n
      !$OMP END DO

#if defined(_HO_MOOD_) && !defined(_HO_MOOD_UNLIMITED_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      FlagReComp = 0
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
      FlagReCompTP = 0
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
      NbIterReComp = 0
      !$OMP END SINGLE
      call GetCellsToUpdate(U2,U_n_p1,t_n,dt,NbIterReComp)
      loop_MOOD_n: do while (NbCellsInvalid > 0 .and. NbIterReComp < MaxIterPossibleReComp)
        !$OMP SINGLE
        FlagReComp = 1
        NbIterReComp = NbIterReComp + 1
        MaxInvalidCells = max(MaxInvalidCells,NbCellsInvalid)
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
        test_has_invalidTP_cells_n: if (NbCellsInvalidTP > 0) then
          FlagReCompTP = 1
        end if test_has_invalidTP_cells_n
        MaxInvalidCellsTP = max(MaxInvalidCellsTP,NbCellsInvalidTP)
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
        !$OMP END SINGLE NOWAIT
        call RecomputePhiMOOD(U2,t3,Phi)
        ! Recompute only on decremented cell and their neigbors
        ! as dt does not change
        !$OMP DO
        loop_elt_n_MOOD: do idxK = 1, NbCellsToUpdate
          CellK = CellsToUpdate(idxK)
          U_n_p1(:,CellK) = (U_n(:,CellK) + 2._PR*U2(:,CellK) - 2._PR*dt*Phi(:,CellK))/3._PR
        end do loop_elt_n_MOOD
        !$OMP END DO
        call GetCellsToUpdate(U2,U_n_p1,t_n,dt,NbIterReComp)
      end do loop_MOOD_n
      !$OMP SINGLE
      if (NbCellsInvalid > 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Bailing out ..."
        write(StdErr,'(I9,A)') NbCellsInvalid, ' invalid cells,'
        write(StdErr,'(I9,A)') NbCellsToUpdate, ' cells to update,'
        write(StdErr,'(I9,A)') NbIterReComp, ' loop of re-computations done.'
        FLUSH StdErr
        STOP
      endif
      MaxIterReComp = max(MaxIterReComp,NbIterReComp)
      TotIterReComp = TotIterReComp + NbIterReComp
      TotReComp = TotReComp + FlagReComp
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
      TotReCompTP = TotReCompTP + FlagReCompTP
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _HO_MOOD_ && _HO_MOOD_UNLIMITED_ */

#if defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      FlagReComp = 0
      NbIterReComp = 0
      !$OMP END SINGLE
      call GetCellsToCorrec(U2,U_n_p1,NbIterReComp)
      loop_correc_n: do while (NbCellsInvalid > 0 .and. NbIterReComp < MaxIterPossibleReComp)
        !$OMP SINGLE
        FlagReComp = 1
        NbIterReComp = NbIterReComp + 1
        MaxInvalidCells = max(MaxInvalidCells,NbCellsInvalid)
        !$OMP END SINGLE NOWAIT

        if (not_last_time_step) then
          ! Recompute the time step by recomputing only on corrected cells
          call Recompute_dt_TPCorrec(U2,t_n,t_n_p1,dt,dt_changed)
          !$OMP SINGLE
          if (dt_changed) then
            t3 = t_n + 0.5_PR*dt
          end if
          !$OMP END SINGLE
        else
          !$OMP SINGLE
          dt_changed = .FALSE.
          !$OMP END SINGLE
        end if

        ! Recompute the sum of flux only on corrected cells
        call RecomputePhiTPCorrec(U2,t_n,Phi)

        if (dt_changed) then
          ! Recompute the solution on all the cells, as dt changed
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          loop_elt_n_correc: do CellK = 1, pmesh%NbCells
            U_n_p1(:,CellK) = (U_n(:,CellK) + 2._PR*U2(:,CellK) - 2._PR*dt*Phi(:,CellK))/3._PR
          end do loop_elt_n_correc
          !$OMP END DO
        else
          ! Recompute the solution only on updated cells as dt did not change
          !$OMP DO
          loop_CellsTopUpdate_n_correc: do idxK = 1, NbCellsToUpdate
            CellK = CellsToUpdate(idxK)
            U_n_p1(:,CellK) = (U_n(:,CellK) + 2._PR*U2(:,CellK) - 2._PR*dt*Phi(:,CellK))/3._PR
          end do loop_CellsTopUpdate_n_correc
          !$OMP END DO
        endif

        call GetCellsToCorrec(U2,U_n_p1,NbIterReComp)
      end do loop_correc_n
      !$OMP SINGLE
      if (NbCellsInvalid > 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Bailing out ..."
        write(StdErr,'(I9,A)') NbCellsInvalid, ' invalid cells,'
        write(StdErr,'(I9,A)') NbCellsToUpdate, ' cells to update,'
        write(StdErr,'(I9,A)') NbIterReComp, ' loop of re-computations done.'
        FLUSH StdErr
        STOP
      endif
      MaxIterReComp = max(MaxIterReComp,NbIterReComp)
      TotIterReComp = TotIterReComp + NbIterReComp
      TotReComp = TotReComp + FlagReComp
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _TP_CORREC_ && _TP_FLUX_ && _HO_MOOD_ */
      !======================================================================

  end subroutine SSPRK33

#elif defined (_TIME_RK3TVD_)
  !--------------------------------------------------------------------------
  !> @brief
  !> RK3-TVD (see (1.8) in \cite Dio_2012 )
  !>
  !> \f[\begin{array}{rcl}
  !>    U^{(1)} &=& U^n - \Delta t \ \Phi(t^n,U^n)\\
  !>    U^{(2)} &=& U^{(1)} - \Delta t \ \Phi(t^n+ \Delta a,t U^{(1)}) \\
  !>    U^{(3)} &=& \frac{3 U^{n} + U^{(2)}}{4} - \Delta t \ \Phi(t^n + \frac{\Delta t}{2}, U^{(2)}) \\
  !>    U^{n+1} &=& \frac{U^{n} + 2 U^{(3)}}{3}
  !> \end{array}\f]
  !>
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[inout]  t_n_p1, dt Next time and time step
  !> @param[inout] Phi Working array
  !> @param[out] U_n_p1 solution at next time step
  !>
  !> @remarks We are using the same array for \f$U^{(1)}\f$ and \f$U^{(1)}\f$: \c U13
  !> @remarks We temporaly save \f$\frac{3 U^{n} + U^{(2)}}{4}\f$ in \c U_n_p1
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine RK3TVD(U_n,t_n,t_n_p1,dt,Phi,U_n_p1)
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n
      real(PR), intent(inout) :: t_n_p1, dt
      real(PR), dimension(:,:), intent(inout) :: Phi
      real(PR), dimension(:,:), intent(out) :: U_n_p1

      real(PR) :: t2, t3
      integer :: CellK
#if defined(_TP_CORREC_) || defined(_HO_MOOD_)
      integer :: idxK
#endif /*_TP_CORREC_ || _HO_MOOD_ */

      !======================================================================
      call compute_Phi(U_n,t_n,Phi)

      !$OMP DO &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      loop_elt_1: do CellK = 1, pmesh%NbCells
        U13(:,CellK) = U_n(:,CellK) - dt*Phi(:,CellK)
      end do loop_elt_1
      !$OMP END DO


#if defined(_HO_MOOD_) && !defined(_HO_MOOD_UNLIMITED_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      FlagReComp = 0
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
      FlagReCompTP = 0
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
      NbIterReComp = 0
      !$OMP END SINGLE
      call GetCellsToUpdate(U_n,U13,t_n,dt,NbIterReComp)
      loop_MOOD_1: do while (NbCellsInvalid > 0 .and. NbIterReComp < MaxIterPossibleReComp)
        !$OMP SINGLE
        FlagReComp = 1
        NbIterReComp = NbIterReComp + 1
        MaxInvalidCells = max(MaxInvalidCells,NbCellsInvalid)
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
        test_has_invalidTP_cells_1: if (NbCellsInvalidTP > 0) then
          FlagReCompTP = 1
        end if test_has_invalidTP_cells_1
        MaxInvalidCellsTP = max(MaxInvalidCellsTP,NbCellsInvalidTP)
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
        !$OMP END SINGLE NOWAIT
        call RecomputePhiMOOD(U_n,t_n,Phi)
        ! Recompute only on decremented cell and their neigbors
        ! as dt does not change
        !$OMP DO
        loop_elt_1_MOOD: do idxK = 1, NbCellsToUpdate
          CellK = CellsToUpdate(idxK)
          U13(:,CellK) = U_n(:,CellK) - dt*Phi(:,CellK)
        end do loop_elt_1_MOOD
        !$OMP END DO
        call GetCellsToUpdate(U_n,U13,t_n,dt,NbIterReComp)
      end do loop_MOOD_1
      !$OMP SINGLE
      if (NbCellsInvalid > 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Bailing out ..."
        write(StdErr,'(I9,A)') NbCellsInvalid, ' invalid cells,'
        write(StdErr,'(I9,A)') NbCellsToUpdate, ' cells to update,'
        write(StdErr,'(I9,A)') NbIterReComp, ' loop of re-computations done.'
        FLUSH StdErr
        STOP
      endif
      MaxIterReComp = max(MaxIterReComp,NbIterReComp)
      TotIterReComp = TotIterReComp + NbIterReComp
      TotReComp = TotReComp + FlagReComp
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
      TotReCompTP = TotReCompTP + FlagReCompTP
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _HO_MOOD_ && _HO_MOOD_UNLIMITED_ */

#if defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      FlagReComp = 0
      NbIterReComp = 0
      !$OMP END SINGLE
      call GetCellsToCorrec(U_n,U13,NbIterReComp)
      loop_correc_1: do while (NbCellsInvalid > 0 .and. NbIterReComp < MaxIterPossibleReComp)
        !$OMP SINGLE
        FlagReComp = 1
        NbIterReComp = NbIterReComp + 1
        MaxInvalidCells = max(MaxInvalidCells,NbCellsInvalid)
        !$OMP END SINGLE NOWAIT

        if (not_last_time_step) then
          ! Recompute the time step by recomputing only on corrected cells
          call Recompute_dt_TPCorrec(U_n,t_n,t_n_p1,dt,dt_changed)
        else
          !$OMP SINGLE
          dt_changed = .FALSE.
          !$OMP END SINGLE
        end if

        ! Recompute the sum of flux only on corrected cells
        call RecomputePhiTPCorrec(U_n,t_n,Phi)

        if (dt_changed) then
          ! Recompute the solution on all the cells, as dt changed
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          loop_elt_1_correc: do CellK = 1, pmesh%NbCells
            U13(:,CellK) = U_n(:,CellK) - dt*Phi(:,CellK)
          end do loop_elt_1_correc
          !$OMP END DO
        else
          ! Recompute the solution only on updated cells as dt did not change
          !$OMP DO
          loop_CellsTopUpdate_1_correc: do idxK = 1, NbCellsToUpdate
            CellK = CellsToUpdate(idxK)
            U13(:,CellK) = U_n(:,CellK) - dt*Phi(:,CellK)
          end do loop_CellsTopUpdate_1_correc
          !$OMP END DO
        endif

        call GetCellsToCorrec(U_n,U13,NbIterReComp)
      end do loop_correc_1
      !$OMP SINGLE
      if (NbCellsInvalid > 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Bailing out ..."
        write(StdErr,'(I9,A)') NbCellsInvalid, ' invalid cells,'
        write(StdErr,'(I9,A)') NbCellsToUpdate, ' cells to update,'
        write(StdErr,'(I9,A)') NbIterReComp, ' loop of re-computations done.'
        FLUSH StdErr
        STOP
      endif
      MaxIterReComp = max(MaxIterReComp,NbIterReComp)
      TotIterReComp = TotIterReComp + NbIterReComp
      TotReComp = TotReComp + FlagReComp
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _TP_CORREC_ && _TP_FLUX_ && _HO_MOOD_ */
      !======================================================================

      t2 = t_n + dt

      !======================================================================
#if defined(_HO_MOOD_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      !$OMP END SINGLE
      call InitMOOD_iter(U13)
      !$OMP SINGLE
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _HO_MOOD_ */

#if !defined(_TP_FLUX_)
      call compute_nu_all(U13,t2)
#if defined(_SRC_ENABLE_)
      call compute_delta_all()
#endif /* _SRC_ENABLE_ */
#endif /* _TP_FLUX_ */

      call compute_Phi(U13,t2,Phi)

      !$OMP DO &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      loop_elt_2: do CellK = 1, pmesh%NbCells
        U2(:,CellK) = U13(:,CellK) - dt*Phi(:,CellK)
      end do loop_elt_2
      !$OMP END DO

#if defined(_HO_MOOD_) && !defined(_HO_MOOD_UNLIMITED_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      FlagReComp = 0
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
      FlagReCompTP = 0
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
      NbIterReComp = 0
      !$OMP END SINGLE
      call GetCellsToUpdate(U13,U2,t_n,dt,NbIterReComp)
      loop_MOOD_2: do while (NbCellsInvalid > 0 .and. NbIterReComp < MaxIterPossibleReComp)
        !$OMP SINGLE
        FlagReComp = 1
        NbIterReComp = NbIterReComp + 1
        MaxInvalidCells = max(MaxInvalidCells,NbCellsInvalid)
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
        test_has_invalidTP_cells_2: if (NbCellsInvalidTP > 0) then
          FlagReCompTP = 1
        end if test_has_invalidTP_cells_2
        MaxInvalidCellsTP = max(MaxInvalidCellsTP,NbCellsInvalidTP)
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
        !$OMP END SINGLE NOWAIT
        call RecomputePhiMOOD(U13,t2,Phi)
        ! Recompute only on decremented cell and their neigbors
        ! as dt does not change
        !$OMP DO
        loop_elt_2_MOOD: do idxK = 1, NbCellsToUpdate
          CellK = CellsToUpdate(idxK)
          U2(:,CellK) = U13(:,CellK) - dt*Phi(:,CellK)
        end do loop_elt_2_MOOD
        !$OMP END DO
        call GetCellsToUpdate(U13,U2,t_n,dt,NbIterReComp)
      end do loop_MOOD_2
      !$OMP SINGLE
      if (NbCellsInvalid > 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Bailing out ..."
        write(StdErr,'(I9,A)') NbCellsInvalid, ' invalid cells,'
        write(StdErr,'(I9,A)') NbCellsToUpdate, ' cells to update,'
        write(StdErr,'(I9,A)') NbIterReComp, ' loop of re-computations done.'
        FLUSH StdErr
        STOP
      endif
      MaxIterReComp = max(MaxIterReComp,NbIterReComp)
      TotIterReComp = TotIterReComp + NbIterReComp
      TotReComp = TotReComp + FlagReComp
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
      TotReCompTP = TotReCompTP + FlagReCompTP
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _HO_MOOD_ && _HO_MOOD_UNLIMITED_ */

#if defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      FlagReComp = 0
      NbIterReComp = 0
      !$OMP END SINGLE
      call GetCellsToCorrec(U13,U2,NbIterReComp)
      loop_correc_2: do while (NbCellsInvalid > 0 .and. NbIterReComp < MaxIterPossibleReComp)
        !$OMP SINGLE
        FlagReComp = 1
        NbIterReComp = NbIterReComp + 1
        MaxInvalidCells = max(MaxInvalidCells,NbCellsInvalid)
        !$OMP END SINGLE NOWAIT

        if (not_last_time_step) then
          ! Recompute the time step by recomputing only on corrected cells
          call Recompute_dt_TPCorrec(U13,t_n,t_n_p1,dt,dt_changed)
          !$OMP SINGLE
          if (dt_changed) then
            t2 = t_n + dt
          end if
          !$OMP END SINGLE
        else
          !$OMP SINGLE
          dt_changed = .FALSE.
          !$OMP END SINGLE
        end if

        ! Recompute the sum of flux only on corrected cells
        call RecomputePhiTPCorrec(U13,t2,Phi)

        if (dt_changed) then
          ! Recompute the solution on all the cells, as dt changed
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          loop_elt_2_correc: do CellK = 1, pmesh%NbCells
            U2(:,CellK) = U13(:,CellK) - dt*Phi(:,CellK)
          end do loop_elt_2_correc
          !$OMP END DO
        else
          ! Recompute the solution only on updated cells as dt did not change
          !$OMP DO
          loop_CellsTopUpdate_2_correc: do idxK = 1, NbCellsToUpdate
            CellK = CellsToUpdate(idxK)
            U2(:,CellK) = U13(:,CellK) - dt*Phi(:,CellK)
          end do loop_CellsTopUpdate_2_correc
          !$OMP END DO
        endif

        call GetCellsToCorrec(U13,U2,NbIterReComp)
      end do loop_correc_2
      !$OMP SINGLE
      if (NbCellsInvalid > 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Bailing out ..."
        write(StdErr,'(I9,A)') NbCellsInvalid, ' invalid cells,'
        write(StdErr,'(I9,A)') NbCellsToUpdate, ' cells to update,'
        write(StdErr,'(I9,A)') NbIterReComp, ' loop of re-computations done.'
        FLUSH StdErr
        STOP
      endif
      MaxIterReComp = max(MaxIterReComp,NbIterReComp)
      TotIterReComp = TotIterReComp + NbIterReComp
      TotReComp = TotReComp + FlagReComp
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _TP_CORREC_ && _TP_FLUX_ && _HO_MOOD_ */
      !======================================================================

      t3 = t_n + 0.5_PR*dt

      !======================================================================
      !$OMP DO &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      loop_elt_3_tmp: do CellK = 1, pmesh%NbCells
        U_n_p1(:,CellK) = 0.25_PR*(3._PR*U_n(:,CellK) + U2(:,CellK))
      end do loop_elt_3_tmp
      !$OMP END DO

#if defined(_HO_MOOD_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      !$OMP END SINGLE
      call InitMOOD_iter(U_n_p1)
      !$OMP SINGLE
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _HO_MOOD_ */


#if !defined(_TP_FLUX_)
      call compute_nu_all(U_n_p1,t3)
#if defined(_SRC_ENABLE_)
      call compute_delta_all()
#endif /* _SRC_ENABLE_ */
#endif /* _TP_FLUX_ */

      call compute_Phi(U_n_p1,t3,Phi)

      !$OMP DO &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      loop_elt_3: do CellK = 1, pmesh%NbCells
        U13(:,CellK) = U_n_p1(:,CellK) - dt*Phi(:,CellK)
      end do loop_elt_3
      !$OMP END DO

#if defined(_HO_MOOD_) && !defined(_HO_MOOD_UNLIMITED_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      FlagReComp = 0
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
      FlagReCompTP = 0
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
      NbIterReComp = 0
      !$OMP END SINGLE
      call GetCellsToUpdate(U_n_p1,U13,t_n,dt,NbIterReComp)
      loop_MOOD_3: do while (NbCellsInvalid > 0 .and. NbIterReComp < MaxIterPossibleReComp)
        !$OMP SINGLE
        FlagReComp = 1
        NbIterReComp = NbIterReComp + 1
        MaxInvalidCells = max(MaxInvalidCells,NbCellsInvalid)
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
        test_has_invalidTP_cells_3: if (NbCellsInvalidTP > 0) then
          FlagReCompTP = 1
        end if test_has_invalidTP_cells_3
        MaxInvalidCellsTP = max(MaxInvalidCellsTP,NbCellsInvalidTP)
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
        !$OMP END SINGLE NOWAIT
        call RecomputePhiMOOD(U_n_p1,t3,Phi)
        ! Recompute only on decremented cell and their neigbors
        ! as dt does not change
        !$OMP DO
        loop_elt_3_MOOD: do idxK = 1, NbCellsToUpdate
          CellK = CellsToUpdate(idxK)
          U13(:,CellK) = U_n_p1(:,CellK) - dt*Phi(:,CellK)
        end do loop_elt_3_MOOD
        !$OMP END DO
        call GetCellsToUpdate(U_n_p1,U13,t_n,dt,NbIterReComp)
      end do loop_MOOD_3
      !$OMP SINGLE
      if (NbCellsInvalid > 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Bailing out ..."
        write(StdErr,'(I9,A)') NbCellsInvalid, ' invalid cells,'
        write(StdErr,'(I9,A)') NbCellsToUpdate, ' cells to update,'
        write(StdErr,'(I9,A)') NbIterReComp, ' loop of re-computations done.'
        FLUSH StdErr
        STOP
      endif
      MaxIterReComp = max(MaxIterReComp,NbIterReComp)
      TotIterReComp = TotIterReComp + NbIterReComp
      TotReComp = TotReComp + FlagReComp
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
      TotReCompTP = TotReCompTP + FlagReCompTP
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _HO_MOOD_ && _HO_MOOD_UNLIMITED_ */

#if defined(_TP_CORREC_) && !defined(_TP_FLUX_) && !defined(_HO_MOOD_)
      !$OMP SINGLE
      TimeCounters%ReComp(1) = GetTime()
      FlagReComp = 0
      NbIterReComp = 0
      !$OMP END SINGLE
      call GetCellsToCorrec(U_n_p1,U13,NbIterReComp)
      loop_correc_3: do while (NbCellsInvalid > 0 .and. NbIterReComp < MaxIterPossibleReComp)
        !$OMP SINGLE
        FlagReComp = 1
        NbIterReComp = NbIterReComp + 1
        MaxInvalidCells = max(MaxInvalidCells,NbCellsInvalid)
        !$OMP END SINGLE NOWAIT

        if (not_last_time_step) then
          ! Recompute the time step by recomputing only on corrected cells
          call Recompute_dt_TPCorrec(U2,t_n,t_n_p1,dt,dt_changed)
          !$OMP SINGLE
          if (dt_changed) then
            t3 = t_n + 0.5_PR*dt
          end if
          !$OMP END SINGLE
        else
          !$OMP SINGLE
          dt_changed = .FALSE.
          !$OMP END SINGLE
        end if

        ! Recompute the sum of flux only on corrected cells
        call RecomputePhiTPCorrec(U2,t3,Phi)

        if (dt_changed) then
          ! Recompute the solution on all the cells, as dt changed
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          loop_elt_3_correc: do CellK = 1, pmesh%NbCells
            U13(:,CellK) = U_n_p1(:,CellK) - dt*Phi(:,CellK)
          end do loop_elt_3_correc
          !$OMP END DO
        else
          ! Recompute the solution only on updated cells as dt did not change
          !$OMP DO
          loop_CellsTopUpdate_3_correc: do idxK = 1, NbCellsToUpdate
            CellK = CellsToUpdate(idxK)
            U13(:,CellK) = U_n_p1(:,CellK) - dt*Phi(:,CellK)
          end do loop_CellsTopUpdate_3_correc
          !$OMP END DO
        endif

        call GetCellsToCorrec(U_n_p1,U13,NbIterReComp)
      end do loop_correc_3
      !$OMP SINGLE
      if (NbCellsInvalid > 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Bailing out ..."
        write(StdErr,'(I9,A)') NbCellsInvalid, ' invalid cells,'
        write(StdErr,'(I9,A)') NbCellsToUpdate, ' cells to update,'
        write(StdErr,'(I9,A)') NbIterReComp, ' loop of re-computations done.'
        FLUSH StdErr
        STOP
      endif
      MaxIterReComp = max(MaxIterReComp,NbIterReComp)
      TotIterReComp = TotIterReComp + NbIterReComp
      TotReComp = TotReComp + FlagReComp
      TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
      !$OMP END SINGLE
#endif /* _TP_CORREC_ && _TP_FLUX_ && _HO_MOOD_ */
      !======================================================================

      !======================================================================
      !$OMP DO &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      loop_elt_n: do CellK = 1, pmesh%NbCells
        U_n_p1(:,CellK) = (U_n(:,CellK) + 2._PR*U13(:,CellK))/3._PR
      end do loop_elt_n
      !$OMP END DO
      !======================================================================

  end subroutine RK3TVD

#endif /* _TIME_SCHEME_ */

end module mod_time_explicit
