!> Performs a time iteration with the right time scheme
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_time_explicit_Butcher

  use mod_numerics, only: PR
  use type_IO, only: StdErr
  use type_mesh, only: pmesh
#if defined(_FC_PAR_OMP_)
  use type_mesh, only: chunk_NbPrimaryCells
#endif /* _FC_PAR_OMP_ */
  use mod_operators, only: compute_Phi
#if defined(_HO_MOOD_)
  use mod_HO_MOOD, only: InitMOOD_iter
  use mod_HO_MOOD_degree, only: GetCellsToUpdate
  use mod_HO_MOOD_recompute, only: RecomputePhiMOOD
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
  use type_recompute, only: NbCellsInvalidTP
  use type_recompute, only: TotReCompTP
  use type_recompute, only: MaxInvalidCellsTP
  use type_recompute, only: FlagReCompTP
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
#endif /* _HO_MOOD_ */
#if defined(_TP_CORREC_) || defined(_HO_MOOD_)
  use type_recompute, only: NbCellsInvalid, NbCellsToUpdate
  use type_recompute, only: CellsToUpdate
  use mod_tools_time, only: GetTime
  use mod_tools_time, only: TimeCounters
  use type_recompute, only: TotReComp, TotIterReComp
  use type_recompute, only: MaxInvalidCells, MaxIterReComp
  use type_recompute, only: MaxIterPossibleReComp
  use type_recompute, only: FlagReComp, NbIterReComp
#endif /* _TP_CORREC_ || _HO_MOOD_ */
#if !defined(_TP_FLUX_)
  use mod_coeff_DLP, only: compute_nu_all
#if defined(_SRC_ENABLE_)
  use mod_delta, only: compute_delta_all
#endif /* _SRC_ENABLE_ */
#endif /* .not. _TP_FLUX_ */

  use mod_Butcher_Tableau, only: ButcherTableau

  IMPLICIT NONE

#if !defined(_FC_PGI_)
  private
  public compute_U_n_p1_Butcher
  public U_But
  public Phi_But
#endif /* !defined(_FC_PGI_) */

  real(PR), dimension(:,:,:), allocatable, save :: U_But
  real(PR), dimension(:,:,:), allocatable, save :: Phi_But

CONTAINS
  !--------------------------------------------------------------------------
  !> @brief
  !> Performs a time iteration using a time scheme defined with a Butcher tableau as in \cite SpiRuu_2002
  !>
  !> \f[U^{n} \Rightarrow U^{n+1}\f]
  !>
  !> @param[in] U_n Solution at current time
  !> @param[in] t_n Current time
  !> @param[in] t_n_p1, dt Next time and time step
  !> @param[out] U_n_p1 solution at next time step
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine compute_U_n_p1_Butcher(U_n,t_n,t_n_p1,dt,U_n_p1)
      real(PR), dimension(:,:), intent(in) :: U_n
      real(PR), intent(in) :: t_n, t_n_p1, dt
      real(PR), dimension(:,:), intent(out) :: U_n_p1
      real(PR) :: t_i
      integer :: CellK
#if defined(_HO_MOOD_)
      integer :: idxK
#endif /* defined(_HO_MOOD_) */
      integer :: i, k

#if defined(_TP_CORREC_) && !defined(_HO_MOOD_)
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Routine not defined..."
      FLUSH StdErr
      STOP
#endif /* _TP_CORREC_ && !_HO_MOOD_ */

#if !(defined(_TIME_FE_) || defined(_TIME_SSPRK11_) || \
defined(_TIME_SSPRK21_) || defined(_TIME_SSPRK22_) || \
defined(_TIME_SSPRK31_) || defined(_TIME_SSPRK32_) || defined(_TIME_SSPRK33_) || \
defined(_TIME_SSPRK42_) || defined(_TIME_SSPRK43_) || \
defined(_TIME_SSPRK52_) || defined(_TIME_SSPRK53_) || defined(_TIME_SSPRK54_) || \
defined(_TIME_SSPRK62_) || defined(_TIME_SSPRK63_) || defined(_TIME_SSPRK64_) || \
defined(_TIME_SSPRK72_) || defined(_TIME_SSPRK73_) || defined(_TIME_SSPRK74_) || \
defined(_TIME_SSPRK82_) || defined(_TIME_SSPRK83_) || defined(_TIME_SSPRK84_))
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Routine not usable with this time scheme: "//_TIME_
      FLUSH StdErr
      STOP
#endif /* _TIME_SCHEME_ */

      i = 0
      !$OMP DO &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      loop_elt_0: do CellK = 1, pmesh%NbCells
        U_But(:,CellK,i) = U_n(:,CellK)
      end do loop_elt_0
      !$OMP END DO

      !======================================================================
      loop_stages: do i = 1, ButcherTableau%NbStages
        t_i = t_n + ButcherTableau%d(i - 1)*dt
        test_first: if (i > 1) then
#if defined(_HO_MOOD_)
          !$OMP SINGLE
          TimeCounters%ReComp(1) = GetTime()
          !$OMP END SINGLE NOWAIT
          call InitMOOD_iter(U_But(:,:,i-1))
          !$OMP SINGLE
          TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
          !$OMP END SINGLE NOWAIT
#endif /* _HO_MOOD_ */

#if !defined(_TP_FLUX_)
          call compute_nu_all(U_But(:,:,i-1),t_i)
#if defined(_SRC_ENABLE_)
          call compute_delta_all()
#endif /* _SRC_ENABLE_ */
#endif /* _TP_FLUX_ */
        endif test_first

        call compute_Phi(U_But(:,:,i-1),t_i,Phi_But(:,:,i-1))

        !$OMP DO &
        !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
        do CellK = 1, pmesh%NbCells
           U_But(:,CellK,i)= 0._PR
        end do
        !$OMP END DO

        do k = 0, i - 1
          !$OMP DO &
          !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
          loop_elt_i: do CellK = 1, pmesh%NbCells
             U_But(:,CellK,i) = U_But(:,CellK,i)&
              + ButcherTableau%alpha(k,i)*U_But(:,CellK,k) - dt*ButcherTableau%beta(k,i)*Phi_But(:,CellK,k)
          end do loop_elt_i
          !$OMP END DO
        end do

#if defined(_HO_MOOD_) && !defined(_HO_MOOD_UNLIMITED_)
        !$OMP SINGLE
        TimeCounters%ReComp(1) = GetTime()
        FlagReComp = 0
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
        FlagReCompTP = 0
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
        NbIterReComp = 0
        !$OMP END SINGLE
        call GetCellsToUpdate(U_But(:,:,i-1),U_But(:,:,i),t_i,dt,NbIterReComp)
        loop_MOOD_i: do while (NbCellsInvalid > 0 .and. NbIterReComp < MaxIterPossibleReComp)
          !$OMP SINGLE
          FlagReComp = 1
          NbIterReComp = NbIterReComp + 1
          MaxInvalidCells = max(MaxInvalidCells,NbCellsInvalid)
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
          test_has_invalidTP_cells_i: if (NbCellsInvalidTP > 0) then
            FlagReCompTP = 1
          end if test_has_invalidTP_cells_i
          MaxInvalidCellsTP = max(MaxInvalidCellsTP,NbCellsInvalidTP)
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
          !$OMP END SINGLE NOWAIT
          call RecomputePhiMOOD(U_But(:,:,i-1),t_i,Phi_But(:,:,i-1))
          ! Recompute only on decremented cell and their neigbors
          ! as dt does not change
          !$OMP DO
          do idxK = 1, NbCellsToUpdate
            CellK = CellsToUpdate(idxK)
            U_But(:,CellK,i)= 0._PR
          end do
          !$OMP END DO
          do k = 0, i - 1
            !$OMP DO
            loop_elt_i_MOOD: do idxK = 1, NbCellsToUpdate
              CellK = CellsToUpdate(idxK)
              U_But(:,CellK,i) = U_But(:,CellK,i)&
                + ButcherTableau%alpha(k,i)*U_But(:,CellK,k) - dt*ButcherTableau%beta(k,i)*Phi_But(:,CellK,k)
            end do loop_elt_i_MOOD
            !$OMP END DO
          end do
          call GetCellsToUpdate(U_But(:,:,i-1),U_But(:,:,i),t_i,dt,NbIterReComp)
        end do loop_MOOD_i
        !$OMP SINGLE
        if (NbCellsInvalid > 0) then
          write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Bailing out ..."
          write(StdErr,'(I9,A)') NbCellsInvalid, ' invalid cells,'
          write(StdErr,'(I9,A)') NbCellsToUpdate, ' cells to update,'
          write(StdErr,'(I9,A)') NbIterReComp, ' loop of re-computations done.'
          FLUSH StdErr
          STOP
        endif
        MaxIterReComp = max(MaxIterReComp,NbIterReComp)
        TotIterReComp = TotIterReComp + NbIterReComp
        TotReComp = TotReComp + FlagReComp
#if defined(_TP_CORREC_) && (_HO_MOOD_MAX_DEGREE_ >= 1)
        TotReCompTP = TotReCompTP + FlagReCompTP
#endif /* _TP_CORREC_ && (_HO_MOOD_MAX_DEGREE_ >= 1) */
        TimeCounters%ReComp(2) = TimeCounters%ReComp(2) + GetTime() - TimeCounters%ReComp(1)
        !$OMP END SINGLE NOWAIT
#endif /* _HO_MOOD_ && !_HO_MOOD_UNLIMITED_ */
      end do loop_stages
      !======================================================================

      i = ButcherTableau%NbStages
      !$OMP DO &
      !$OMP SCHEDULE(STATIC,chunk_NbPrimaryCells)
      loop_elt_n_p1: do CellK = 1, pmesh%NbCells
        U_n_p1(:,CellK) = U_But(:,CellK,i)
      end do loop_elt_n_p1
      !$OMP END DO

  end subroutine compute_U_n_p1_Butcher

end module mod_time_explicit_Butcher
