!> Some generic tools
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_tools

  use mod_numerics, only: PR, eps
  use mod_numerics, only: MaxLen
  use type_IO, only: StdErr
#if defined(_FC_INTEL_) || (defined(_FC_GNU_) && __GNUC__>=5) || defined(_FC_PGI_)
  use, INTRINSIC :: IEEE_ARITHMETIC, only: IEEE_IS_NAN
#endif /* _FC_ */

  IMPLICIT NONE

  private
  public CrossProduct2D
  public RotationMatrix
  public Solve2by2LinearSystem
  public SolvePolynomial2
  public plus1, minus1
  public plus2, minus2
  public factorial
  public BinomialCoeff, BinomialCoeff_Rec
  public x_to_alpha
  public min_NaN, max_NaN
  public LenInt

  public Int2Str
  interface Int2Str
    module procedure Int2Str_free, Int2Str_fix
  end interface

contains
  !--------------------------------------------------------------------------
  !> @brief
  !> Determinant of two vectors in 2D
  !>
  !> @remark
  !> Is the same as the third component of a cross product
  !>
  !> @param[in] u,v two 2D vectors
  !> @return \f$u*v\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function CrossProduct2D(u,v)
    real(PR), dimension(:), intent(in) :: u,v
    real(PR) :: CrossProduct2D

    CrossProduct2D = u(1)*v(2) - u(2)*v(1)
  end function CrossProduct2D

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the rotation matrix of angle \f$\theta\f$
  !>
  !> @param[in] theta Angle of the rotation
  !> @param[out] R Rotation matrix
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine RotationMatrix(theta,R)
      real(PR), intent(in) :: theta
      real(PR), dimension(:,:), intent(out) :: R

      real(PR) :: c, s

      c = cos(theta)
      s = sin(theta)

      R(1,1) =  c
      R(2,1) =  s
      R(1,2) = -s
      R(2,2) =  c

  end subroutine RotationMatrix

  !--------------------------------------------------------------------------
  !> @brief
  !> Add one to n with periodic condition p
  !>
  !> @param[in] n integer
  !> @param[in] p period
  !> @return
  !> * n + 1 if n+1 \f$\leq\f$ p
  !> * 1     else
  !>
  !> @warning the second formulation is only valid for n<=p
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function plus1(n,p)
    integer, intent(in) :: n, p
    integer :: plus1

    plus1 = 1 + mod(n,p)

    !if (n + 1 <= p) then
      !plus1 = n + 1
    !else
      !plus1 = 1
    !end if

  end function plus1

  !--------------------------------------------------------------------------
  !> @brief
  !> Add two to n with periodic condition p
  !>
  !> @param[in] n integer
  !> @param[in] p period
  !> @return
  !> * n + 2 if n+2 \f$\leq\f$ p
  !> * 1     if n+1 \f$\leq\f$ p
  !> * 2     else
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function plus2(n,p)
    integer, intent(in) :: n, p
    integer :: plus2

    if (n + 2 <= p) then
      plus2 = n + 2
    else if (n + 1 <= p) then
      plus2 = 1
    else
      plus2 = 2
    end if

  end function plus2

  !--------------------------------------------------------------------------
  !> @brief
  !> Subtract one to n with periodic condition p
  !>
  !> @param[in] n integer
  !> @param[in] p period
  !> @return
  !> * n - 1 if n-1 \f$\geq\f$ 1
  !> * p else
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function minus1(n,p)
    integer, intent(in) :: n, p
    integer :: minus1

    !minus1 = n*(1/i) + mod(i-1,n)

    if (n - 1 >=1) then
      minus1 = n - 1
    else
      minus1 = p
    end if

  end function minus1

  !--------------------------------------------------------------------------
  !> @brief
  !> Subtract two to n with periodic condition p
  !>
  !> @param[in] n integer
  !> @param[in] p period
  !> @return
  !> * n - 2 if n-2 \f$\geq\f$ 1
  !> * p - 1 if n-1 \f$\geq\f$ 1
  !> * p else
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function minus2(n,p)
    integer, intent(in) :: n, p
    integer :: minus2

    if (n - 2 >=1) then
      minus2 = n - 2
    else if (n - 1 >=1) then
      minus2 = p - 1
    else
      minus2 = p
    end if

  end function minus2

  !--------------------------------------------------------------------------
  !> @brief
  !> Solve a \f$2\f$ by \f$2\f$ linear system \f$A x = b\f$
  !>
  !> @param[in] A, b  Matrix and RHS of the linear system
  !> @param[out] x    Solution of the linear system
  !> @param[out] IERR Error code, 0 on success
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine Solve2by2LinearSystem(A,b,x,IERR)
      real(PR), dimension(:,:), intent(in)  :: A
      real(PR), dimension(:), intent(in)    :: b
      real(PR), dimension(:), intent(out)   :: x
      integer, intent(out)                  :: IERR

      real(PR) :: DetA

      DetA = A(1,1)*A(2,2) - A(2,1)*A(1,2)

      !if (IsZero(DetA)) then
      if (abs(DetA) < eps) then
        IERR = -1
      else
        x(1) = A(2,2)*b(1) - A(1,2)*b(2)
        x(2) = A(1,1)*b(2) - A(2,1)*b(1)
        x = x/DetA
        IERR = 0
      end if

  end subroutine Solve2by2LinearSystem

  !--------------------------------------------------------------------------
  !> @brief
  !> Solve second order polynomial equation:
  !> \f[ a_2*x^2 + a_1*x + a_0 = 0 \f]
  !>
  !> @param[in] a_2, a_1, a_0 Coefficients of the polynom
  !> @param[out] x_m, x_p Two roots of the equations
  !> @param[out] IERR Error code, 0 on success
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  subroutine SolvePolynomial2(a_2,a_1,a_0,x_m,x_p,IERR)
      real(PR), intent(in)  :: a_2, a_1, a_0
      real(PR), intent(out) :: x_m, x_p
      integer, intent(out)  :: IERR

      real(PR) :: delta

      if (abs(a_2) > eps) then
        delta = a_1**2 - 4._PR*a_2*a_0

        if (delta < 0._PR) then
          IERR = -1
        else
          IERR = 0
        end if

        delta = sqrt(delta)
        x_m = 0.5_PR*(-a_1 - delta)/a_2
        x_p = 0.5_PR*(-a_1 + delta)/a_2
      else
        IERR = 0
        ! If a_2 = 0, it's a linear equation
        x_m = -a_0/a_1
        x_p = x_m
      end if

  end subroutine SolvePolynomial2

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the factorial of an integer \f$n\f$.
  !>
  !> @param[in] n An integer
  !> @return \f$n!=1 \times 2 \times \dots \times (n-1) \times n\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function factorial(n)
      integer, intent(in) :: n
      integer  :: factorial

      integer :: i

      factorial = 1
      do i = 2, n
        factorial = i*factorial
      end do
  end function factorial

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the binomial coefficent.
  !>
  !> @param[in] k,n Integers
  !> @return \f$\begin{pmatrix} n \\ k \end{pmatrix} = \frac{n!}{k!(n-k)!}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function BinomialCoeff(k,n)
      integer, intent(in) :: k, n
      integer :: BinomialCoeff

      integer :: i

      !BinomialCoeff = factorial(n)/(factorial(k)*factorial(n-k))

      BinomialCoeff = 1
      do i = n, n - k + 1, -1
        BinomialCoeff = BinomialCoeff*i
      end do
      do i = 1, k
        BinomialCoeff = BinomialCoeff/i
      end do
  end function BinomialCoeff

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the binomial coefficent (recursive version).
  !>
  !> @param[in] k,n Integers
  !> @return \f$\begin{pmatrix} n \\ k \end{pmatrix} = \frac{n!}{k!(n-k)!}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  recursive function BinomialCoeff_Rec(k,n) result(BinomialCoeff_res)
    integer, intent(in) :: k, n
    integer :: BinomialCoeff_res

    if (k == n) then
      BinomialCoeff_res = 1
    else if (k == 0) then
      BinomialCoeff_res = 1
    else if (k == 1) then
      BinomialCoeff_res = n
    else
      BinomialCoeff_res = BinomialCoeff_Rec(k-1,n-1) + BinomialCoeff_Rec(k,n-1)
    end if
  end function BinomialCoeff_Rec

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the vector \f$\boldsymbol{x}\f$ to \f$\alpha\f$
  !>
  !> @param[in] x Vector
  !> @param[in] alpha Multi-index
  !> @return \f$\boldsymbol{x}^{\boldsymbol{\alpha}} = x_1^{\alpha_1} x_2^{\alpha_2} \dots x_n^{\alpha_n}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function x_to_alpha(x,alpha)
      real(PR), dimension(:), intent(in) :: x
      integer, dimension(:), intent(in) :: alpha
      real(PR) :: x_to_alpha

      integer :: row

      if (size(x) == size(alpha)) then
        x_to_alpha = 1._PR
        do row = 1, size(x)
          x_to_alpha = x_to_alpha*x(row)**alpha(row)
        end do
      else
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Arrays do not have the same size"
        FLUSH StdErr
        STOP
      endif
  end function x_to_alpha

  !--------------------------------------------------------------------------
  !> @brief
  !> Minimum function that propagates NaN
  !>
  !> @param[in] x, y Reals to compare
  !> @return \f$\min(x,y)\f$ by taking cares of NaN
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function min_NaN(x,y)
      real(PR), intent(in) :: x, y
      real(PR) :: min_NaN

#if defined(_FC_INTEL_) || (defined(_FC_GNU_) && __GNUC__>=5) || defined(_FC_PGI_)
      if (IEEE_IS_NAN(x)) then
        min_NaN = x
      else if (IEEE_IS_NAN(y)) then
        min_NaN = y
      else
        min_NaN = min(x,y)
      end if
#else
      if (ISNAN(x)) then
        min_NaN = x
      else if (ISNAN(y)) then
        min_NaN = y
      else
        min_NaN = min(x,y)
      end if
#endif /* _FC */

  end function min_NaN

  !--------------------------------------------------------------------------
  !> @brief
  !> Maximum function that propagates NaN
  !>
  !> @param[in] x, y Reals to compare
  !> @return \f$\max(x,y)\f$ by taking cares of NaN
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function max_NaN(x,y)
      real(PR), intent(in) :: x, y
      real(PR) :: max_NaN

#if defined(_FC_INTEL_) || (defined(_FC_GNU_) && __GNUC__>=5) || defined(_FC_PGI_)
      if (IEEE_IS_NAN(x)) then
        max_NaN = x
      else if (IEEE_IS_NAN(y)) then
        max_NaN = y
      else
        max_NaN = max(x,y)
      end if
#else
      if (ISNAN(x)) then
        max_NaN = x
      else if (ISNAN(y)) then
        max_NaN = y
      else
        max_NaN = max(x,y)
      end if
#endif /* _FC */

  end function max_NaN

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the number of digits in a postitve integer
  !>
  !> @remarks See https://stackoverflow.com/questions/1068849/how-do-i-determine-the-number-of-digits-of-an-integer-in-c
  !> and especially https://stackoverflow.com/a/1068937
  !>
  !> @remarks This is a "pure" function in order to be called inside parameter declarations
  !> (for instance see mod_tools::int2str_free)
  !>
  !> @param[in] k Integer
  !> @return Number of digit of k
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  PURE function LenInt(k) result(len)
    integer, intent(in) :: k
    integer :: len

    if (k == 0) then
      len = 1
    else
      len = floor(log10(abs(real(k,PR)))) + 1
    end if

  end function LenInt

  !--------------------------------------------------------------------------
  !> @brief
  !> Convert a positive integer to a string
  !>
  !> @param[in] k Integer to convert
  !> @return k in a string
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function Int2Str_free(k) result(str)
    integer, intent(in) :: k
    character(len=LenInt(k)) :: str

    character(len=MaxLen) :: str_tmp

    write (str_tmp, *) k
    str = trim(adjustl(str_tmp))

  end function Int2Str_free

  !--------------------------------------------------------------------------
  !> @brief
  !> Convert a positive integer to a string by setting the length of the string desired
  !>
  !> If the integer is too short for the length add some leading zeros
  !>
  !> @param[in] k Integer to convert
  !> @param[in] length Length of the desired string
  !> @return k in a string of length \c length
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function Int2Str_fix(k,length) result(str)
    integer, intent(in) :: k, length
    character(len=length) :: str

    character(len=LenInt(length)) :: length_str

    length_str = Int2Str_free(length)

    if (LenInt(k)>length) then
      write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Length too short"
      FLUSH StdErr
      STOP
    endif

    write(str,'(I'//length_str//'.'//length_str//')') k
    str = trim(adjustl(str))

  end function Int2Str_fix

end module mod_tools
