!> Some generic tools for arrays
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_tools_array

  use mod_numerics, only: PR, zero, eps
  use type_IO, only: StdErr

  IMPLICIT NONE

  private
  public IsPresent
  public GetIndex
  public FindIndex
  public norm_2
  public normalize
  public tensor_product
#if defined(_LIB_LAPACK_)
  public ConditioningNumber
#endif /* _LIB_LAPACK_ */
#if defined(_FC_PGI_)
  public NORM2
#endif /* defined(_FC_PGI_) */

contains
  !--------------------------------------------------------------------------
  !> @brief
  !> Test if the integer \c K exists in the array \c v.
  !>
  !> @warning Don't use it on big arrays !
  !>
  !> @param[in] K, v
  !> @return TRUE if \c K is inside \c v.
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function IsPresent(K,v)
    integer, intent(in)               :: K
    integer, dimension(:), intent(in) :: v
    logical :: IsPresent

    integer :: row

    IsPresent = .FALSE.

    !if (minval(abs(v - K),1) == 0) IsPresent = .TRUE.

    do row = 1, size(v)
      if (v(row) == K) then
        IsPresent = .TRUE.
        exit
      end if
    end do
  end function IsPresent

  !--------------------------------------------------------------------------
  !> @brief
  !> Get the index of \c K in the array \c v.
  !>
  !> @warning Don't use it on big arrays !
  !>
  !> @param[in] K, v
  !> @return \c idx such that v(idx) = K, and -1 in case of failure.
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function GetIndex(K,v)
    integer, intent(in)               :: K
    integer, dimension(:), intent(in) :: v
    integer :: GetIndex

    integer :: row

    !GetIndex = minloc(abs(v - K),1)
    !if (v(GetIndex) /= K) GetIndex = -1

    GetIndex = -1
    do row = 1, size(v)
      if (v(row) == K) then
        GetIndex = row
        exit
      end if
    end do

  end function GetIndex

  !--------------------------------------------------------------------------
  !> @brief
  !> Find the first index such that \f$ \text{array}(i) \leq \text{scalar} \leq \text{array}(i+1)\f$
  !>
  !> This routine is needed for the linear interpolation of values.
  !> If \c scalar is not inside the the values of the table return the first
  !> or the last but one index, depending if it is smaller or larger than all
  !> the values in the table.
  !>
  !> @warning Return -1 in case of failure
  !>
  !> @param[in] scalar Value to find in the array
  !> @param[in] array Table of values
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function FindIndex(scalar,array)
    real(PR), intent(in) :: scalar
    real(PR), dimension(:), intent(in) :: array
    integer :: FindIndex

    integer :: n
    logical :: found

    n = size(array)
    found = .FALSE.

    !if (scalar < array(1) .or. scalar > array(n)) then
      !write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> scalar not in the table"
      !write(StdErr,'(2(A,ES23.16))') 'min = ', array(1), ' ; max = ', array(n)
      !write(StdErr,'(A,ES23.16)')'value = ', scalar
      !FLUSH StdErr
      !STOP
    !end if

    if (scalar < array(1)) then
      FindIndex = 1
      found = .TRUE.
    else if (scalar > array(n)) then
      FindIndex = n-1
      found = .TRUE.
    else
      loop_array: do FindIndex = 1, n-1
        if (scalar >= array(FindIndex) .and. scalar <= array(FindIndex+1)) then
          found = .TRUE.
          exit loop_array
        end if
      end do loop_array
    endif

    ! If we cannot find the index, set it to -1 to indicate an error
    if (.not. found) FindIndex = -1

  end function FindIndex

#if defined(_FC_PGI_)
  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the euclidean norm of a vector
  !>
  !> @param[in] x a vector
  !> return euclidean norm of x: \f$\Vert x \Vert_2\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function NORM2(x)
    real(PR), dimension(:), intent(in) :: x
    real(PR) :: NORM2

    NORM2 = sqrt(dot_product(x,x))

  end function NORM2
#endif /* defined(_FC_PGI_) */

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the squared euclidean norm of a vector
  !>
  !> @param[in] x a vector
  !> return the squared euclidean norm of x: \f$\Vert x \Vert_2^2\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function norm_2(x)
    real(PR), dimension(:), intent(in) :: x
    real(PR) :: norm_2

    norm_2 = dot_product(x,x)

  end function norm_2

  !--------------------------------------------------------------------------
  !> @brief
  !> Normalize a vector
  !>
  !> @param[in] x a vector
  !> @return the vector divided by is norm if it is non zero
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function normalize(x)
    real(PR), dimension(:), intent(in) :: x
    real(PR), dimension(size(x)) :: normalize

    real(PR) :: norm

    norm = NORM2(x)

    !if (IsZero(norm)) then
    if (abs(norm) < eps) then
      normalize(:) = zero
    else
      normalize(:) = x(:)/norm
    end if
  end function

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the tensorial product of two vectors
  !>
  !> @param[in] x,y two vectors
  !> @return a \f$m\f$ by \f$n\f$ matrix
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function tensor_product(x,y)
    real(PR), dimension(:), intent(in) :: x, y
    real(PR), dimension(size(x),size(y)) :: tensor_product

    integer :: row, col

    do col = 1, size(y)
      do row = 1, size(x)
        tensor_product(row,col) = x(row)*y(col)
      end do
    end do

  end function tensor_product

#if defined(_LIB_LAPACK_)
  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the conditioning number of a general matrix
  !>
  !> \f[ \kappa(A) = \frac{\sigma_{max}(A)}{\sigma_{min}(A)} \f]
  !>
  !> @param[in] A real rectangular matrix
  !> @return kappa Conditioning number of \f$A\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function ConditioningNumber(A) result(kappa)
    real(PR), dimension(:,:), intent(in) :: A
    real(PR) :: kappa

    real(PR), dimension(:,:), allocatable :: A_tmp
    real(PR), dimension(:), allocatable :: S
    real(PR), dimension(:,:), allocatable :: U, VT
    real(PR), dimension(:), allocatable :: WORK
    integer :: INFO

    integer :: m, n, min_mn

    m = size(A,1)
    n = size(A,2)

    min_mn = min(m,n)

    allocate(A_tmp(m,n))
    allocate(S(min_mn))
    allocate(WORK(5*min_mn))
#if defined(_DEBUG_)
    allocate(U(m,m))
    allocate(VT(n,n))
#endif /* _DEBUG_ */

    ! warning: DGESVD erase the content of the input matrix, a copy is needed
    !A_tmp(:,:) = A(:,:)
    call DCOPY(m*n,A,1,A_tmp,1)

    call DGESVD('N', 'N', m, n, A_tmp, m, S, U, m, VT, n, WORK, 5*min_mn, INFO)
    if (INFO /= 0) then
      write(StdErr,'(2A,I4,A,I4)')__FILE__,':', __LINE__, ' -> DGESVD failed, INFO = ', INFO
      FLUSH StdErr
      STOP
    endif

    kappa = S(1)/S(min_mn)

    deallocate(A_tmp)
    deallocate(S)
    deallocate(WORK)
#if defined(_DEBUG_)
    deallocate(U)
    deallocate(VT)
#endif /* _DEBUG_ */


  end function ConditioningNumber
#endif /* _LIB_LAPACK_ */

end module mod_tools_array
