!> Some tools for linear interpoçlation
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_tools_interp

  use mod_numerics, only: PR
  use type_IO, only: StdErr

  implicit none

  private
  public LinearInterp
  public BiLinearInterp
  public TriLinearInterp

contains
  !--------------------------------------------------------------------------
  !> @brief
  !> Linear interpolation in 1D
  !>
  !> https://en.wikipedia.org/wiki/Linear_interpolation
  !>
  !> @param[in] xA, xB Coordinates of the two known points
  !> @param[in] fA, fB Value of the function at the two points
  !> @param[in] xG  Coordinates of the interpolation point
  !> @return fG Interpolation of the function at G
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function LinearInterp(xA,fA,xB,fB,xG) result(fG)
    real(PR), intent(in) :: xA, xB
    real(PR), intent(in) :: fA, fB
    real(PR), intent(in) :: xG
    real(PR) :: fG

    ! Slope version
    fG = fA + (fB - fA)*(xG - xA)/(xB - xA)
    ! Barycentric coordiantes version
    !fG = fA*(1._PR - (xG - xA)/(xB - xA)) + fB)*(xG - xA)/(xB - xA)
  end function LinearInterp

  !--------------------------------------------------------------------------
  !> @brief
  !> Linear interpolation in 2D
  !>
  !> https://en.wikipedia.org/wiki/Bilinear_interpolation
  !> http://www.scratchapixel.com/old/lessons/3d-advanced-lessons/interpolation/bilinear-interpolation/
  !>
  !> @param[in] x1, x2, y1, y2 Boundaries of the square of vertices (x1,y1), (x2,y1), (x2,y2), (x1,y2)
  !> @param[in] f11, f21, f22, f12 Value of the function at the four points
  !> @param[in] x, y  Coordinates of the interpolation point
  !> @return f Interpolation of the function at (x,y)
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function BiLinearInterp(x1,x2,y1,y2,f11,f21,f22,f12,x,y) result(f)
    real(PR), intent(in) :: x1, x2, y1, y2
    real(PR), intent(in) :: f11, f21, f22, f12
    real(PR), intent(in) :: x, y
    real(PR) :: f

    real(PR) :: fy1, fy2

    ! Double linear interpolation version
    ! Two linear interpolations in the x-direction
    fy1 = LinearInterp(x1,f11,x2,f21,x)
    fy2 = LinearInterp(x1,f12,x2,f22,x)
    ! Then a linear interpolation in the y-direction
    f = LinearInterp(y1,fy1,y2,fy2,y)

    ! Direct version
    !f = 1._PR/((x2-x1)*(y2-y1))*(&
        !f11*(x2-x)*(y2-y) + f21*(x-x1)*(y2-y) +&
        !f12*(x2-x)*(y-y1) + f22*(x-x1)*(y-y1))

  end function BiLinearInterp

  !--------------------------------------------------------------------------
  !> @brief
  !> Linear interpolation in 3D
  !>
  !> https://en.wikipedia.org/wiki/Trilinear_interpolation
  !> http://www.scratchapixel.com/old/lessons/3d-advanced-lessons/interpolation/trilinear-interpolation/
  !> http://paulbourke.net/miscellaneous/interpolation/
  !>
  !> @param[in] x1, x2, y1, y2, z1, z2 Boundaries of the cube of vertices
  !>   (x1,y1,z1), (x2,y1,z1), (x2,y2,z1), (x1,y2,z1),
  !>   (x1,y1,z2), (x2,y1,z2), (x2,y2,z2), (x1,y2,z2)
  !> @param[in] f111, f211, f221, f121, f112, f212, f222, f122
  !>    Value of the function at the eight points
  !> @param[in] x, y, z  Coordinates of the interpolation point
  !> @return f Interpolation of the function at (x,y,z)
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function TriLinearInterp(x1,x2,y1,y2,z1,z2,&
        f111,f211,f221,f121,f112,f212,f222,f122,&
        x,y,z) &
        result(f)
    real(PR), intent(in) :: x1, x2, y1, y2, z1, z2
    real(PR), intent(in) :: f111, f211, f221, f121, f112, f212, f222, f122
    real(PR), intent(in) :: x, y, z
    real(PR) :: f

    real(PR) :: fz1, fz2

    ! Double bilinear interpolation version
    ! One bilinear interpolation in the (x,y,z1) plane
    fz1 = BiLinearInterp(x1,x2,y1,y2,f111,f211,f221,f121,x,y)
    ! A second  bilinear interpolation in the (x,y,z2) plane
    fz2 = BiLinearInterp(x1,x2,y1,y2,f112,f212,f222,f122,x,y)
    ! Then a linear interpolation in the z-direction
    f = LinearInterp(z1,fz1,z2,fz2,z)

    ! Direct version
    !f = ...

  end function TriLinearInterp
end module mod_tools_interp
