!> Some generic tools in relation with the mesh
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_tools_mesh

  use mod_numerics, only: PR, pi, eps
  use type_IO, only: StdErr
  use mod_tools, only: CrossProduct2D, plus1, Solve2by2LinearSystem
  use type_mesh, only: pmesh, nodes, dim_
#if defined(_FC_PGI_)
  use mod_tools_array, only: NORM2
#endif /* defined(_FC_PGI_) */


  IMPLICIT NONE

  private
  public TriangleArea, compute_tangent
  public Circumcenter
#if defined(_MESH_D2P_)
  public compute_normal
#endif /* _MESH_ */
  public CDG, IsOut
  public ProjectionPointToLine
  public is_inside_bounding_box, is_inside_triangle
  public compute_bar_coord_triangle
  public wrapper_compute_bar_coord_stared_polygon
  public compute_bar_coord_stared_polygon
  public compute_bar_coord_stared_polygon_full

contains

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the area of a triangle with the coordinates of the vertices
  !>
  !> @param[in] xA,xB,xC Coordinates of the vertices
  !>
  !> @remark \f$ |T| = \frac{1}{2} \big| (x_A - x_C) (y_B - y_A) - (x_A - x_B) (y_C - y_A) \big| \f$
  !>
  !> @author Rodolphe Turpault
  !--------------------------------------------------------------------!
  function TriangleArea(xA,xB,xC)
      real(PR), dimension(:), intent(in) :: xA, xB, xC
      real(PR)                           :: TriangleArea

      TriangleArea = 0.5_PR*abs((xB(1)-xA(1))*(xC(2)-xA(2))-(xB(2)-xA(2))*(xC(1)-xA(1)))

  end function TriangleArea

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the coordinates of the circumcenter
  !>
  !> @param[in] CellK Index of the cell
  !> @return Coordinates of the circumcenter
  !>
  !> @remark See https://en.wikipedia.org/wiki/Circumscribed_circle#Circumcenter_coordinates
  !> and http://mathworld.wolfram.com/BarycentricCoordinates.html
  !> @warning Only defined for a triangle
  !>
  !> @author Rodolphe Turpault
  !--------------------------------------------------------------------!
  function Circumcenter(CellK)
      integer, intent(in) :: CellK
      real(PR), dimension(dim_) :: Circumcenter

      real(PR), dimension(dim_) :: xA, xB, xC
      real(PR) :: a, b, c
      real(PR) :: w_A, w_B, w_C

      if (pmesh%elt(CellK)%NbInterfaces == 3) then
        c = pmesh%elt(CellK)%le(1)
        a = pmesh%elt(CellK)%le(2)
        b = pmesh%elt(CellK)%le(3)

        xA(:) = nodes(pmesh%elt(CellK)%vertices_glob(1))%x(:)
        xB(:) = nodes(pmesh%elt(CellK)%vertices_glob(2))%x(:)
        xC(:) = nodes(pmesh%elt(CellK)%vertices_glob(3))%x(:)

        w_A = a**2*(b**2 + c**2 - a**2)
        w_B = b**2*(c**2 + a**2 - b**2)
        w_C = c**2*(a**2 + b**2 - c**2)

        Circumcenter(:) = (w_A*xA(:) + w_B*xB(:) + w_C*xC(:))/(w_A + w_B + w_C)
      else
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Function not defined for general polygons"
        FLUSH StdErr
        STOP
      endif

  end function Circumcenter

#if defined(_MESH_D2P_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the unitary outgoing normal vector to an interface
  !>
  !> @param[in] xA,xB Coordinates of the vertices of the interface
  !> @param[in] xK Coordinates of an other point
  !> @result Coordinates of the normal such that KA.n > 0
  !>
  !> @warning Use it only on the diamond mesh
  !>
  !> @author Rodolphe Turpault
  !--------------------------------------------------------------------!
  function compute_normal(xA,xB,xK) result(n)
      real(PR), dimension(:), intent(in) :: xA, xB, xK
      real(PR), dimension(dim_)          :: n

      real(PR) :: scal

      ! The normal is orthogonal to the tangent to the interfaces
      n(1) = xB(2) - xA(2)
      n(2) = xA(1) - xB(1)
      ! scal = KA.n
      scal = (xA(1) - xK(1))*n(1) + (xA(2) - xK(2))*n(2)
      if (scal < 0._PR) then
        n=-n
      end if

      n(:) = n(:)/NORM2(n)

  end function compute_normal
#endif /* _MESH_ */

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the tangent vector \f$\vec{t}\f$ to have a direct basis with \f$\vec{n}\f$
  !>
  !> @param[in] normal Coordinates of the normal to the interface
  !> @return Coordinates of the tangent vector to the interface
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function compute_tangent(normal) result(tangent)
      real(PR), dimension(:), intent(in) :: normal
      real(PR), dimension(dim_)             :: tangent

      ! Apply a rotation of +pi/2 in order to have the tangent
      ! from the normal vector
      tangent(1) = - normal(2)
      tangent(2) =   normal(1)

  end function compute_tangent

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the coordinates of the gravity center of a cell
  !>
  !> @param[in] CellK Cell index
  !> @return The coordinates of the gravity center
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function CDG(CellK)
    integer, intent(in) :: CellK
    real(PR), dimension(dim_) :: CDG

    integer :: i_K

    CDG(:) = 0._PR
    do i_K = 1, pmesh%elt(CellK)%NbInterfaces
      CDG(:) = CDG(:) + nodes(pmesh%elt(CellK)%vertices_glob(i_K))%x(:)
    end do

    CDG(:) = CDG(:)/real(pmesh%elt(CellK)%NbInterfaces,PR)

  end function CDG

  !!--------------------------------------------------------------------!
  !!> @brief
  !!> Compute the coordinates of the intersection of the diagonal
  !!> of a quadrangle
  !!>
  !!> @param[in] xA, xB, xC, xD Coordinates of the four vertices
  !!> @return The coordinates of points of the intersection
  !!>
  !!> @author Florian Blachère
  !!--------------------------------------------------------------------!
  !function IntersectionDiagonalQuad(xA,xB,xC,xD) result(x)
      !real(PR), dimension(dim_), intent(in) :: xA, xB, xC, xD
      !real(PR), dimension(dim_) :: x

      !real(PR), dimension(dim_,dim_) :: Matrix
      !real(PR), dimension(dim_) :: RHS
      !integer :: IERR

      !Matrix(1,1) = xC(2) - xA(2)
      !Matrix(2,1) = xD(2) - xB(2)
      !Matrix(1,2) = xA(1) - xC(1)
      !Matrix(2,2) = xB(1) - xD(1)

      !RHS(1) = xA(1)*xC(2) - xA(2)*xC(1)
      !RHS(2) = xB(1)*xD(2) - xB(2)*xD(1)

      !call Solve2by2LinearSystem(Matrix,RHS,x,IERR)

      !if (IERR /=0) then
        !write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Trouble to find the intersection"
        !FLUSH StdErr
        !STOP
      !end if

  !end function IntersectionDiagonalQuad

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the othogonal projection of a point on a line
  !>
  !> @param[in] xA, xB Two points on a line
  !> @param[in] xC Point to project
  !> @return The coordinates of the projection point
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function ProjectionPointToLine(xA,xB,xC) result(xP)
      real(PR), dimension(:), intent(in) :: xA, xB, xC
      real(PR), dimension(dim_) :: xP

      real(PR), dimension(dim_,dim_) :: Matrix
      real(PR), dimension(dim_) :: RHS
      integer :: IERR

      if (dim_ /= 2) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Routine not defined"
        FLUSH StdErr
        STOP
      end if

      Matrix(1,1) = xB(2) - xA(2)
      Matrix(2,1) = xA(1) - xB(1)
      Matrix(1,2) = Matrix(2,1)
      Matrix(2,2) = - Matrix(1,1)

      RHS(1) = - (xB(1)*xA(2) - xA(1)*xB(2))
      RHS(2) = - (xC(1)*(xB(1) - xA(1)) + xC(2)*(xB(2) - xA(2)))

      call Solve2by2LinearSystem(Matrix,RHS,xP,IERR)

      if (IERR /= 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Trouble to find the projection"
        FLUSH StdErr
        STOP
      end if

  end function ProjectionPointToLine

  !--------------------------------------------------------------------!
  !> @brief
  !> Test if a point is outside the computational domain
  !>
  !> @param[in] x Coordinates of the point
  !> @return True is point is outside, false if it is inside
  !>
  !> @warning only true for a rectangular domain
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function IsOut(x)
    real(PR), dimension(:), intent(in) :: x
    logical :: isout

    integer :: i

    isout = .FALSE.

    do i = 1, dim_
      if (x(i) < pmesh%x_min(i) .or. x(i) > pmesh%x_max(i)) then
        isout = .TRUE.
        exit
      end if
    end do

  end function IsOut

  !--------------------------------------------------------------------!
  !> @brief
  !> Test if the point M is inside the triangle (A,B,C)
  !>
  !> @remark
  !> Using the method of the barycentric coordinates, if the three barycentric
  !> coordinates are positive: the point is inside the triangle.<BR>
  !> See http://totologic.blogspot.fr/2014/01/accurate-point-in-triangle-test.html
  !>
  !> @param[in] M, A, B, C Coordinates of the points
  !> @return True is point is inside, false if it is outside
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  function is_inside_triangle(M,A,B,C)
    real(PR), dimension(:), intent(in) :: M, A, B, C
    logical :: is_inside_triangle

    real(PR), dimension(3) :: coord_bar

    call compute_bar_coord_triangle(M, A, B, C, coord_bar)

    if (minval(coord_bar) < 0._PR) then
      is_inside_triangle = .FALSE.
    else
      is_inside_triangle = .TRUE.
    end if

  end function is_inside_triangle

  !--------------------------------------------------------------------!
  !> @brief
  !> Test if a point M is inside a bounding box
  !>
  !> @remark
  !> The bounding box is the rectangle surrounding the polygon
  !>
  !> @param[in] x Coordinates of the point to test
  !> @param[in] v Coordinates of the points constituting the polygon
  !> @return True is point is inside, false if it is outside
  !>
  !> @author Florian Blachère
  !>!--------------------------------------------------------------------!
  function is_inside_bounding_box(x,v)
    real(PR), dimension(:), intent(in) :: x
    real(PR), dimension(:,:), intent(in) :: v
    logical :: is_inside_bounding_box

    real(PR) :: coord_min, coord_max
    integer :: i

    is_inside_bounding_box = .TRUE.

    do i = 1, dim_
      coord_min = minval(v(:,i))
      coord_max = maxval(v(:,i))
      if (x(i) < coord_min .or. x(i) > coord_max) then
        is_inside_bounding_box = .FALSE.
        exit
      end if
    end do

  end function is_inside_bounding_box

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the barycentric coordinates of a point regarding to a triangle
  !>
  !> @remark
  !> \f[\left\{\begin{array}{ccccccc}
  !>  G & = & w_A A & + & w_B B & + & w_C C \\
  !>  1 & = & w_A   & + & w_B   & + & w_C  \\
  !>  \end{array}\right. \f]
  !>
  !> @param[in] G Coordinates of the barycentre
  !> @param[in] A, B, C Coordinates of the points of the triangle
  !> @param[out] coord The three barycentric coordinates
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine compute_bar_coord_triangle(G,A,B,C,coord)
    real(PR), dimension(:), intent(in) :: G, A, B, C
    real(PR), dimension(:), intent(out) :: coord

    real(PR) :: den

    if (size(G) /= 2) then
      write(StdErr,'(2A,I4,A,I1)')__FILE__,':', __LINE__, " -> Routine not defined for higer dimension, dim=", size(G)
      FLUSH StdErr
      STOP
    end if

    den = (B(2) - C(2))*(A(1) - C(1)) + (C(1) - B(1))*(A(2) - C(2))

    if (abs(den) > eps) then
      coord(1) = ((B(2) - C(2))*(G(1) - C(1)) + (C(1) - B(1))*(G(2) - C(2)))/den
      coord(2) = ((C(2) - A(2))*(G(1) - C(1)) + (A(1) - C(1))*(G(2) - C(2)))/den
      coord(3) = 1 - (coord(2) + coord(1))
    else
      coord(:) = - 1._PR
    end if


    ! FIX: To avoid the case coord(3) = -epsilon, which leads to say that G is outside the triangle
    !if (IsZero(coord(3))) coord(3) = 0._PR
    !if (abs(coord(3)) < eps) coord(3) = 0._PR

  end subroutine compute_bar_coord_triangle

  !--------------------------------------------------------------------!
  !> @brief
  !> Wrapper for the computation of the barycentric coordinates
  !> of a point regarding to starred polygon
  !>
  !> @remark
  !> This wrapper first test if the point is in a bounding box
  !> with is_inside_bounding_box() and then compute the barycentric coordinates
  !> with compute_bar_coord_stared_polygon()
  !>
  !> @param[in] xM Coordinates of the barycentre
  !> @param[in] n Number of vertices in the starred polygon
  !> @param[in] v Coordinates of the points of the starred polygon
  !> @param[out] coord The barycentric coordinates
  !> @param[out] ok If true the point is inside the kernel of the starred polygon
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine wrapper_compute_bar_coord_stared_polygon(xM,n,v,coord,ok)
    real(PR), dimension(:), intent(in) :: xM
    integer, intent(in) :: n
    real(PR), dimension(:,:), intent(in) :: v
    real(PR), dimension(n), intent(out) :: coord
    logical, intent(out) :: ok

    ok = is_inside_bounding_box(xM,v)

    if (ok) then
      call compute_bar_coord_stared_polygon(xM,n,v,coord)

      if (minval(coord) < 0._PR) then
        ok = .FALSE.
      else
        ok = .TRUE.
      end if

    else
      coord(:) = -1._PR
    end if

  end subroutine wrapper_compute_bar_coord_stared_polygon

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the barycentric coordinates of a point regarding to a starred polygon \cite Flo_2003 .
  !>
  !> @param[in] xM Coordinates of the barycentre
  !> @param[in] n Number of vertices in the starred polygon
  !> @param[in] v Coordinates of the points of the starred polygon
  !> @param[out] w The barycentric coordinates
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine compute_bar_coord_stared_polygon(xM,n,v,w)
    real(PR), dimension(:), intent(in) :: xM
    integer, intent(in) :: n
    real(PR), dimension(:,:), intent(in) :: v
    real(PR), dimension(:), intent(out) :: w

    real(PR), dimension(dim_,n) :: s
    real(PR), dimension(n) :: r, sin_alpha, A
    real(PR) :: sin_beta_i, B_i
    real(PR) :: sum_w
    integer :: i

    do i=1,n
      s(:,i) = v(:,i) - xM(:)
      r(i) = NORM2(s(:,i))
      if (abs(r(i)) < eps) then
        w(:) = -1._PR
        return
      endif
    end do

    do i=1, n-1
      sin_alpha(i) = CrossProduct2D(s(:,i),s(:,i+1))/(r(i)*r(i+1))
      A(i) = r(i)*r(i+1)*0.5_PR*sin_alpha(i)
      if (abs(A(i)) < eps) then
        w(:) = -1._PR
        return
      endif
    end do

    i = n
    sin_alpha(i) = CrossProduct2D(s(:,i),s(:,1))/(r(i)*r(1))
    A(i) = r(i)*r(1)*0.5_PR*sin_alpha(i)
    if (abs(A(i)) < eps) then
      w(:) = -1._PR
      return
    endif

    i = 1
    sin_beta_i = CrossProduct2D(s(:,n),s(:,i+1))/(r(n)*r(i+1))
    B_i = r(n)*r(i+1)*0.5_PR*sin_beta_i
    w(i) = (r(n)*A(i) - r(i)*B_i + r(i+1)*A(n))/(A(n)*A(i))
    sum_w = w(i)

    do i=2, n-1
      sin_beta_i = CrossProduct2D(s(:,i-1),s(:,i+1))/(r(i-1)*r(i+1))
      B_i = r(i-1)*r(i+1)*0.5_PR*sin_beta_i
      w(i) = (r(i-1)*A(i) - r(i)*B_i + r(i+1)*A(i-1))/(A(i-1)*A(i))
      sum_w = sum_w + w(i)
    end do

    i = n
    sin_beta_i = CrossProduct2D(s(:,i-1),s(:,1))/(r(i-1)*r(1))
    B_i = r(i-1)*r(1)*0.5_PR*sin_beta_i
    w(i) = (r(i-1)*A(i) - r(i)*B_i + r(1)*A(i-1))/(A(i-1)*A(i))
    sum_w = sum_w + w(i)

    if (abs(sum_w) > eps) then
      w(:) = w(:)/sum_w
    else
      w(:) = -1._PR
      return
    endif

  end subroutine compute_bar_coord_stared_polygon

  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the barycentric coordinates of a point regarding to a starred polygon \cite Flo_2003
  !> (full definition version, which is defined everywhere)
  !>
  !> @param[in] xM Coordinates of the barycentre
  !> @param[in] n Number of vertices in the starred polygon
  !> @param[in] v Coordinates of the points of the starred polygon
  !> @param[out] w The barycentric coordinates
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine compute_bar_coord_stared_polygon_full(xM,n,v,w)
    real(PR), dimension(:), intent(in) :: xM
    integer, intent(in) :: n
    real(PR), dimension(:,:), intent(in) :: v
    real(PR), dimension(:), intent(out) :: w

    real(PR), dimension(dim_,n) :: d
    real(PR), dimension(n) :: r
    real(PR) :: sum_w
    integer :: i, j, jp1

    do i=1,n
      d(:,i) = v(:,i) - xM(:)
      r(i) = NORM2(d(:,i))
    end do

    i = 1
    w(i) = sqrt(r(n)*r(i+1) - dot_product(d(:,n),d(:,i+1)))
    do j=2, n-1
      w(i) = w(i) * sqrt(r(j)*r(j+1) + dot_product(d(:,j),d(:,j+1)))
    end do
    sum_w = w(i)

    do i=2, n-1
      w(i) = sqrt(r(i-1)*r(i+1) - dot_product(d(:,i-1),d(:,i+1)))
      do j=1, n
        if (j /= (i-1) .and. j /= i) then
          jp1 = plus1(j,n)
          w(i) = w(i) * sqrt(r(j)*r(jp1) + dot_product(d(:,j),d(:,jp1)))
        end if
      end do
      sum_w = sum_w + w(i)
    end do

    i = n
    w(i) = sqrt(r(i-1)*r(1) - dot_product(d(:,i-1),d(:,1)))
    do j=1, n-2
      w(i) = w(i) * sqrt(r(j)*r(j+1) + dot_product(d(:,j),d(:,j+1)))
    end do
    sum_w = sum_w + w(i)

    w(:)= w(:)/sum_w

  end subroutine compute_bar_coord_stared_polygon_full

end module mod_tools_mesh
