!> Tools for OpenMP
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_tools_omp
#if defined(_FC_PAR_OMP_)

  use mod_numerics, only: MaxLen
  use type_IO, only: StdOut, StdErr
  use type_mesh, only: pmesh
  use type_mesh, only: chunk_NbPrimaryCells, chunk_IntInterfaces, chunk_BndInterfaces
  use OMP_LIB, only: OMP_IN_PARALLEL
  use OMP_LIB, only: OMP_GET_NUM_THREADS, OMP_GET_MAX_THREADS, OMP_GET_NUM_PROCS
  use OMP_LIB, only: OMP_GET_DYNAMIC, OMP_GET_NESTED
  use OMP_LIB, only: OMP_GET_SCHEDULE, OMP_SCHED_KIND
#if (_OPENMP >= 201307) /* OpenMP 4.0 */
  use OMP_LIB, only: OMP_GET_PROC_BIND
#endif /* _OPENMP */


  implicit none

  private
  public compute_chunk
  public DisplayOMP_status

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Compute the chunks regarding to the different sizes
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine compute_chunk()

      if (OMP_IN_PARALLEL()) then
        !$OMP SINGLE
        chunk_NbPrimaryCells = pmesh%NbCells/OMP_GET_NUM_THREADS()
        chunk_IntInterfaces = pmesh%NbIntInterfaces/OMP_GET_NUM_THREADS()
        chunk_BndInterfaces = pmesh%NbBndInterfaces/OMP_GET_NUM_THREADS()
        !$OMP END SINGLE NOWAIT
      else
        !$OMP PARALLEL &
        !$OMP DEFAULT(NONE) &
        !$OMP SHARED(pmesh) &
        !$OMP SHARED(chunk_NbPrimaryCells, chunk_IntInterfaces, chunk_BndInterfaces)
        !$OMP SINGLE
        chunk_NbPrimaryCells = pmesh%NbCells/OMP_GET_NUM_THREADS()
        chunk_IntInterfaces = pmesh%NbIntInterfaces/OMP_GET_NUM_THREADS()
        chunk_BndInterfaces = pmesh%NbBndInterfaces/OMP_GET_NUM_THREADS()
        !$OMP END SINGLE NOWAIT
        !$OMP END PARALLEL
      end if

  end subroutine compute_chunk

  !--------------------------------------------------------------------!
  !> @brief
  !> Display OpenMP status
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine DisplayOMP_status()
      integer, parameter :: len_OMP_var = 17
      integer, parameter :: NbOMP_var = 11
      character(len=len_OMP_var), dimension(NbOMP_var), parameter :: OMP_var = [&
          ' OMP_DYNAMIC     ', &
          ' OMP_NESTED      ', &
          ' OMP_NUM_THREADS ', &
          ' OMP_PLACES      ', &
          ' OMP_PROC_BIND   ', &
          'GOMP_CPU_AFFINITY', &
          ' KMP_AFFINITY    ', &
          ' OMP_SCHEDULE    ', &
          ' OMP_STACKSIZE   ', &
          'GOMP_STACKSIZE   ', &
          ' KMP_STACKSIZE   ' &
      ]
      integer(OMP_SCHED_KIND) :: schedule
      character(len=MaxLen) :: str
      integer :: len_arg, IERR, idx
      integer :: modifier

      !$OMP PARALLEL &
      !$OMP DEFAULT(NONE) &
      !$OMP PRIVATE(schedule, modifier)
      !$OMP SINGLE
      write(StdOut,'(A,I2,A)')      "With          : ", OMP_GET_NUM_THREADS(), " thread(s)"
      write(StdOut,'(A,I2,A,I2,A)') "over          : ", OMP_GET_MAX_THREADS(), " thread(s) &
          &and ", OMP_GET_NUM_PROCS(), " proc(s)"
      write(StdOut,'(A,L,A,L)')     "Dynamic       : ", OMP_GET_DYNAMIC(), ", Nested : ", OMP_GET_NESTED()
#if (_OPENMP >= 201307)
      write(StdOut,'(A,I1)')        "Binding       : ", OMP_GET_PROC_BIND()
#endif /* _OPENMP */
      call OMP_GET_SCHEDULE(schedule,modifier)
      write(StdOut,'(A,I1)')        "Schedule      : ", schedule
      !$OMP END SINGLE NOWAIT
      !$OMP END PARALLEL


#if defined(_LIB_LAPACK_)
#if defined(_LIB_LAPACK_OPENBLAS_)
      !::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::!
      call get_environment_variable('OPENBLAS_NUM_THREADS', str, len_arg, IERR, .TRUE.)
      select case (IERR)
      case (1)
        write(StdOut,'(A)') 'OPENBLAS_NUM_THREADS does not exist'
      case (0)
        !if (len_arg > 0) then
        write(StdOut,'(A,A)')"OPENBLAS_NUM_THREADS = ", trim(str)
        !end if
      case (- 1)
        write(StdOut,'(A,I8,A,I8)') ' length = ', len_arg, ' truncated to', MaxLen
        write(StdErr,'(A)')"======================================================="
        FLUSH StdErr
        STOP
      case default !(>= 2)
        write(StdOut,'(A,I1)') 'get_environment_variable failed: IERR = ', IERR
        write(StdErr,'(A)')"======================================================="
        FLUSH StdErr
        STOP
      end select
      !::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::!
#elif defined(_LIB_LAPACK_MKL_)
      !::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::!
      call get_environment_variable('MKL_NUM_THREADS', str, len_arg, IERR, .TRUE.)
      select case (IERR)
      case (1)
        write(StdOut,'(A)') ' MKL_NUM_THREADS  does not exist'
      case (0)
        !if (len_arg > 0) then
        write(StdOut,'(A,A)')" MKL_NUM_THREADS  = ", trim(str)
        !end if
      case (- 1)
        write(StdOut,'(A,I8,A,I8)') ' length = ', len_arg, ' truncated to', MaxLen
        write(StdErr,'(A)')"======================================================="
        FLUSH StdErr
        STOP
      case default !(>= 2)
        write(StdOut,'(A,I1)') 'get_environment_variable failed: IERR = ', IERR
        write(StdErr,'(A)')"======================================================="
        FLUSH StdErr
        STOP
      end select
      !::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::!
#endif /* _LIB_LAPACK_ */
#endif /* _LIB_LAPACK_ */

      do idx = 1, NbOMP_var
        call get_environment_variable(trim(adjustl(OMP_var(idx))), str, len_arg, IERR, .TRUE.)
        select case (IERR)
        case (1)
          write(StdOut,'(A)') OMP_var(idx)//' does not exist'
        case (0)
          !if (len_arg > 0) then
          write(StdOut,'(A,A)') OMP_var(idx)//" = ", trim(str)
          !end if
        case (- 1)
          write(StdOut,'(A,I8,A,I8)') ' length = ', len_arg, ' truncated to', MaxLen
          write(StdErr,'(A)')"======================================================="
          FLUSH StdErr
          STOP
        case default !(>= 2)
          write(StdOut,'(A,I1)') 'get_environment_variable failed: IERR = ', IERR
          write(StdErr,'(A)')"======================================================="
          FLUSH StdErr
          STOP
        end select
      end do

      write(StdOut,'(A)')"......................................................."

  end subroutine DisplayOMP_status

#endif /* _FC_PAR_OMP_ */
end module mod_tools_omp
