!> Some subroutines for statistical computations
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_tools_stat

  use mod_numerics, only: PR, one

  IMPLICIT NONE

  !> Compute the mean value of an array
  !>
  !> \f[\bar{x} = \frac{1}{n}\sum_{i=1}^{n} x_i \f]
  interface mean
    module procedure mean_r1, mean_r2
  end interface

  !> Compute the standard deviation of an array
  !>
  !> \f[\sigma = \sqrt{\frac{1}{n}\sum_{i=1}^{n} x_i^2 - \bar{x}} \f]
  interface StdDev
    module procedure StdDev_r1, StdDev_r2
  end interface

  contains

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the mean of an array of rank 1
  !>
  !> @param[in] x Array
  !> return Mean of the array: \f$\bar{x}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function mean_r1(x)
    real(PR), dimension(:), intent(in) :: x
    real(PR) :: mean_r1

    mean_r1 = sum(x)/real(size(x),PR)

  end function mean_r1

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the mean of an array of rank 2
  !>
  !> @param[in] x Array
  !> return Mean of the array: \f$\bar{x}\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function mean_r2(x)
    real(PR), dimension(:,:), intent(in) :: x
    real(PR) :: mean_r2

    mean_r2 = sum(x)/real(size(x),PR)

  end function mean_r2

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the standard deviation of an array or rank 1
  !>
  !> @param[in] x Array
  !> return Standard deviation of the array: \f$\sigma\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function StdDev_r1(x)
    real(PR), dimension(:), intent(in) :: x
    real(PR) :: StdDev_r1

    StdDev_r1 = sqrt(one/real(size(x),PR)*sum(x**2) - mean(x)**2)

  end function StdDev_r1

  !--------------------------------------------------------------------------
  !> @brief
  !> Compute the standard deviation of an array or rank 2
  !>
  !> @param[in] x Array
  !> return Standard deviation of the array: \f$\sigma\f$
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------------
  function StdDev_r2(x)
    real(PR), dimension(:,:), intent(in) :: x
    real(PR) :: StdDev_r2

    StdDev_r2 = sqrt(one/real(size(x),PR)*sum(x**2) - mean(x)**2)

  end function StdDev_r2
end module mod_tools_stat
