!> Tools to interact with the system
!>
!> @author Florian Blachère
!>
!> @copyright © 2016 CNRS,
!> @copyright © 2016 Université de Nantes.
!> @copyright This program is free software: you can redistribute it and/or modify
!> it under the terms of the GNU General Public License as published by
!> the Free Software Foundation, either version 3 of the License, or
!> (at your option) any later version.
!>
!> @copyright This program is distributed in the hope that it will be useful,
!> but WITHOUT ANY WARRANTY; without even the implied warranty of
!> MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!> GNU General Public License for more details.
!>
!> @copyright You should have received a copy of the GNU General Public License
!> along with this program.  If not, see <http://www.gnu.org/licenses/>.
module mod_tools_sys

  use type_IO, only: StdOut, StdErr
#if defined(_FC_INTEL_)
  use IFPORT, only: access, GETPID, GETCWD, SYSTEM, HOSTNM
#elif defined(_FC_GNU_)
  use iso_fortran_env, only: compiler_version, compiler_options
#endif /* _FC_ */
  use mod_numerics, only: MaxLen
  use mod_physics_var, only: test_case, Output_Directory
  use mod_read, only: RestartFile, Restarting
#if defined(_FC_PAR_OMP_)
  use mod_tools_omp, only: DisplayOMP_status
#endif /* _FC_PAR_OMP_ */
  use mod_tools_pbs, only: DisplayPBS_status
  use type_mesh, only: basename_mesh_file

  implicit none

  private
  public Initialize
#if !defined(_FC_PGI_)
  public DisplayMemoryUsage
#endif /* !defined(_FC_PGI_) */

contains
  !--------------------------------------------------------------------!
  !> @brief
  !> Get and check the arguments
  !>
  !> @param[in] DATE, TIME Date and time of the last compilation
  !> @param[in] SVNVER, SVNTAG, GITVER, GITTAG Revision and tag of the source code
  !> @param[out] mesh_file Full path to the mesh file
  !> @param[out] info Informations about the run
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine Initialize(DATE,TIME,SVNVER,SVNTAG,GITTAG,GITVER,mesh_file,info)
      character(len=*), intent(in)  :: DATE, TIME
      character(len=*), intent(in)  :: SVNVER, SVNTAG, GITVER, GITTAG
      character(len=*), intent(out) :: mesh_file
      character(len=*), intent(out) :: info

      character(len=MaxLen) :: str
      integer               :: pid_int
      integer               :: len_arg
      integer               :: IERR
#if defined(_FC_GNU_)
      integer               :: status
      character(len=MaxLen) :: MSG
#endif /* _FC_ */
      integer               :: trim_left, trim_right
      integer, dimension(8) :: date_run

      write(StdOut,'(A)')"======================================================="
      write(StdOut,'(A)')"                        AdmAP2D"
      write(StdOut,'(A)')"-------------------------------------------------------"
      write(StdOut,'(A)')"Copyright © 2016 CNRS,"
      write(StdOut,'(A)')"Copyright © 2016 Université de Nantes."
      write(StdOut,'(A)')" "
      write(StdOut,'(A)')"This program is free software: you can redistribute it and/or modify"
      write(StdOut,'(A)')"it under the terms of the GNU General Public License as published by"
      write(StdOut,'(A)')"the Free Software Foundation, either version 3 of the License, or"
      write(StdOut,'(A)')"(at your option) any later version."
      write(StdOut,'(A)')" "
      write(StdOut,'(A)')"This program is distributed in the hope that it will be useful,"
      write(StdOut,'(A)')"but WITHOUT ANY WARRANTY; without even the implied warranty of"
      write(StdOut,'(A)')"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the"
      write(StdOut,'(A)')"GNU General Public License for more details."
      write(StdOut,'(A)')" "
      write(StdOut,'(A)')"You should have received a copy of the GNU General Public License"
      write(StdOut,'(A)')"along with this program.  If not, see <http://www.gnu.org/licenses/>."
      write(StdOut,'(A)')"-------------------------------------------------------"
#if !defined(_OMPP_)
      write(StdOut,'(2A)')"Compiled on: ", _MAKE_HOST_
      write(StdOut,'(2A)')"Architecture: ", _MAKE_ARCH_
      write(StdOut,'(4A)')"By: ", _MAKE_USER_, ", with: ", _FC_
#if defined(_FC_INTEL_)
      write(StdOut,'(A,I4,A,I8)')"Compiler version: ", __INTEL_COMPILER, ', build on: ', __INTEL_COMPILER_BUILD_DATE
#elif defined(_FC_GNU_)
      write(StdOut,'(2A)')"Compiler version: ", compiler_version()
      !write(StdOut,'(2A)')"Compiler version: ",__VERSION__
      !write(StdOut,'(A,1I1,A,1I1,A,1I1)')"Compiler version: ", __GNUC__, '.', __GNUC_MINOR__, '.', __GNUC_PATCHLEVEL__
#endif /* _FC_ */
#if defined(_FC_PAR_OMP_)
      write(StdOut,'(A,I6)')"OpenMP specification: ", _OPENMP
#endif /* _FC_PAR_OMP_ */
      write(StdOut,'(4A)')"Svn branch/tag: ", SVNTAG, ', revision: ', SVNVER
      write(StdOut,'(4A)')"Git branch/tag: ", GITTAG, ', revision: ', GITVER
      write(StdOut,'(4A)')"Date of the latest compilation: ", DATE,' ', TIME
#if defined(_LIB_LAPACK_)
      write(StdOut,'(A)')"......................................................."
      write(StdOut,'(2A)')"LAPACK used: ", _LIB_LAPACK_
#endif /* _LIB_LAPACK_ */
      write(StdOut,'(A)')"......................................................."
      write(StdOut,'(2A)')"FFLAGS  : ", _FFLAGS_
      write(StdOut,'(2A)')"INCLUDE : ", _INCDIR_
      write(StdOut,'(2A)')"LDPATH  : ", _LDPATH_
      write(StdOut,'(2A)')"LDFLAGS : ", _LDFLAGS_
#if defined(_FC_GNU_)
      write(StdOut,'(2A)')"Options : ", compiler_options()
#endif /* _FC_ */
      write(StdOut,'(A)')"......................................................."
#endif /* .not. _OMPPP_ */
#if !defined(_FC_PGI_)
      pid_int = GETPID()
      write(str,'(I9)') pid_int
      write(StdOut,'(2A)')"PID of the run: ", trim(adjustl(str))
      IERR = HOSTNM(str)
      if (IERR == 0) then
        write(StdOut,'(2A)') "HOST          : ", trim(adjustl(str))
      else
        write(StdErr,'(A)')'Trouble to find the hostname'
        write(StdErr,'(A)')"======================================================="
        FLUSH StdErr
        STOP
      end if
      IERR = GETCWD(str)
      if (IERR == 0) then
        write(StdOut,'(2A)') "PWD           : ", trim(adjustl(str))
      else
        write(StdErr,'(A)')'Trouble to define the current working directory'
        write(StdErr,'(A)')"======================================================="
        FLUSH StdErr
        STOP
      end if
#endif /* !defined(_FC_PGI_) */

      ! Get the name of the program
      call get_command_argument(0, str, len_arg, IERR)
      if (IERR == 0) then
        write(StdOut,'(2A)') "Program run   : ", trim(str)
      endif
      call  date_and_time(VALUES=date_run)
      write(StdOut,'(A,I2.2,A,I2.2,A,I2.2,A,I4.4,A,I2.2,A,I2.2)') &
          "At            : ", date_run(5),':',date_run(6),':',date_run(7), &
          ", on: ", date_run(1),'/',date_run(2),'/',date_run(3)

      write(StdOut,'(A)')"......................................................."
#if defined(_FC_PAR_OMP_)
      call DisplayOMP_status()
#endif /* _FC_PAR_OMP_ */

      call DisplayPBS_status()

      ! Get and check the mesh file
      call get_command_argument(1, mesh_file, len_arg, IERR)
      if (IERR > 0) then
        write(StdErr,'(A)')'You have to give a mesh file in the first argument'
        write(StdErr,'(A)')"======================================================="
        FLUSH StdErr
        STOP
      elseif (IERR == -1) then
        write(StdErr,'(A,I8)')'The name of the file is too long, maximum authorized: ',MaxLen
        write(StdErr,'(A)')"======================================================="
        FLUSH StdErr
        STOP
#if !defined(_FC_PGI_)
      elseif (IERR == 0) then
        if( access(mesh_file, 'r') /=0 ) then
          write(StdErr,'(3A)') "The file ", "'"//trim(mesh_file)//"'", " does not exist or is not readable"
          write(StdErr,'(A)')"======================================================="
          FLUSH StdErr
          STOP
        endif
#endif /* !defined(_FC_PGI_) */
      endif

      trim_left  = index(mesh_file,'/',BACK=.TRUE.)    + 1
      trim_right = index(mesh_file,'.msh',BACK=.TRUE.) - 1
      basename_mesh_file = mesh_file(trim_left:trim_right)

      ! Get and check the test case
      call get_command_argument(2, test_case, len_arg, IERR)
      if (IERR > 0) then
        write(StdErr,'(A)')'You have to give a test case in the second argument'
        write(StdErr,'(A)')"======================================================="
        FLUSH StdErr
        STOP
      elseif (IERR == -1) then
        write(StdErr,'(A,I8)')'The name of the test case is too long, maximum authorized: ',MaxLen
        write(StdErr,'(A)')"======================================================="
        FLUSH StdErr
        STOP
      endif

      ! Get and check if there is a third argument (file to read for a restart)
      call get_command_argument(3, RestartFile, len_arg, IERR)
      if (IERR == 0) then
        write(StdOut,'(A,A)') 'File given for the restart: ', trim(RestartFile)
        Restarting = .TRUE.
      elseif (IERR > 0 .and. len_arg == 0) then
        write(StdOut,'(A)') 'No restart file given'
        Restarting = .FALSE.
      elseif (IERR == -1) then
        write(StdErr,'(A,I8)')'The name of the restart file is too long, maximum authorized: ',MaxLen
        write(StdErr,'(A)')"======================================================="
        FLUSH StdErr
        STOP
      endif

#if !defined(_OMPP_)
      info = _INFO_//'-'//trim(basename_mesh_file)//'-'//trim(test_case)
#else
      info = 'ompp-'//trim(basename_mesh_file)//'-'//trim(test_case)
#endif

      ! Create the directory for the outputs if it does not exist
#if !defined(_OMPP_)
      Output_Directory = _IO_OUTDIR_//'/'//trim(info)//'/.' ! Trick to test if the folder is writable
#else
      Output_Directory = './ompp-'//trim(info)//'/.' ! Trick to test if the folder is writable
#endif

#if !defined(_FC_PGI_)
      if( access(Output_Directory, 'w') /=0 ) then
#if defined(_FC_INTEL_)
        IERR = SYSTEM('mkdir -p '//Output_Directory)
        if (IERR /= 0) then
          write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Error while creating the output directory."
          write(StdErr,'(2(A))') "Output directory: ", trim(Output_Directory)
          FLUSH StdErr
          STOP
        end if
#elif defined(_FC_GNU_)
        CALL EXECUTE_COMMAND_LINE('mkdir -p '//Output_Directory, &
            WAIT=.TRUE., EXITSTAT=IERR, CMDSTAT=status, CMDMSG=msg)
        if (IERR /= 0 .or. status /=0) then
          write(StdErr,'(2A,I4,A,A)')__FILE__,':', __LINE__, " -> Error while creating the output directory: ", trim(msg)
          write(StdErr,'(2(A))') "Output directory: ", trim(Output_Directory)
          FLUSH StdErr
          STOP
        end if
#endif /* _FC_ */
      endif
#endif /* !defined(_FC_PGI_) */
#if !defined(_OMPP_)
      Output_Directory = _IO_OUTDIR_//'/'//trim(info)
#else
      Output_Directory = './ompp-'//trim(info)
#endif
      write(StdOut,'(2(A))') "Output directory: ", trim(Output_Directory)

      write(StdOut,'(A)')"======================================================="
      FLUSH StdOut

  end subroutine Initialize


#if !defined(_FC_PGI_)
  !--------------------------------------------------------------------!
  !> @brief
  !> Display current memory usage
  !>
  !> @author Florian Blachère
  !--------------------------------------------------------------------!
  subroutine DisplayMemoryUsage()
      integer, parameter    :: mem_id = 10
      character(len=MaxLen) :: str
      integer               :: pid_int
      integer               :: IERR
#if defined(_FC_GNU_)
      integer               :: status
      character(len=MaxLen) :: MSG
#endif /* _FC_ */

      ! Get the PID of the running process
      pid_int = GETPID()
      write(str,'(I9)') pid_int

      ! Write the memory usage in a temporary file
#if defined(_FC_INTEL_)
      IERR = SYSTEM('echo '//trim(str)//"| xargs -I {} cat /proc/{}/status 2>/dev/null | grep VmRSS | cut -d: -f2 | tr -d '[[:space:]]' > "//trim(Output_Directory)//'/memory.tmp')
      if (IERR /= 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Error getting memmory usage"
        FLUSH StdErr
        STOP
      end if
#elif defined(_FC_GNU_)
      CALL EXECUTE_COMMAND_LINE('echo '//trim(str)//"| xargs -I {} cat /proc/{}/status 2>/dev/null | grep VmRSS | cut -d: -f2 | tr -d '[[:space:]]' > "//trim(Output_Directory)//'/memory.tmp', &
          WAIT=.TRUE., EXITSTAT=IERR, CMDSTAT=status, CMDMSG=msg)
      if (IERR /= 0 .or. status /=0) then
        write(StdErr,'(2A,I4,A,A)')__FILE__,':', __LINE__, " -> Error getting memmory usage: ", trim(msg)
        FLUSH StdErr
        STOP
      end if

#endif /* _FC_ */

      ! Read this file
      open(unit=mem_id,file=trim(Output_Directory)//'/memory.tmp',action='read')
      read(mem_id,*) str
      close(mem_id)

      ! Delete the temporary file
#if defined(_FC_INTEL_)
      IERR = SYSTEM('rm -f '//trim(Output_Directory)//'/memory.tmp')
      if (IERR /= 0) then
        write(StdErr,'(2A,I4,A)')__FILE__,':', __LINE__, " -> Error while deleting memory.tmp"
        FLUSH StdErr
        STOP
      end if
#elif defined(_FC_GNU_)
      CALL EXECUTE_COMMAND_LINE('rm -f '//trim(Output_Directory)//'/memory.tmp', &
          WAIT=.TRUE., EXITSTAT=IERR, CMDSTAT=status, CMDMSG=msg)
      if (IERR /= 0 .or. status /=0) then
        write(StdErr,'(2A,I4,A,A)')__FILE__,':', __LINE__, " -> Error while deleting memory.tmp: ", trim(msg)
        FLUSH StdErr
        STOP
      end if
#endif /* _FC_ */

      ! Write the result
      write(StdOut,'(A,A)') 'Memory usage             : ', trim(str)
  end subroutine DisplayMemoryUsage
#endif /* !defined(_FC_PGI_) */

end module mod_tools_sys
